Set Implicit Arguments.
Require Import List.
Require Import Permutation.
Require Import MyTactics.
Require Import Resources.
Require Import Syntax.

(* The theory of adoption and abandon. *)

(* The predicate [abandon_all] is used in the typing rule for memory
   locations, [JBlock], so it plays quite a central role. Its definition relies
   on two auxiliary predicates, namely [abandon] and [no_adoptees]. *)

(* The definitions of [abandon], [no_adoptees], and [abandon_all] are made
   somewhat subtle by the fact that we wish to use them also in the case where
   the adopter [x] is a duplicable object and there are no adoptees.  This
   allows us to have just one typing rule, [JBlock], for duplicable and
   exclusive objects. One could probably simplify things here at the cost of
   having two separate typing rules. *)

(* ---------------------------------------------------------------------------- *)
(* ---------------------------------------------------------------------------- *)

(* We begin with a few lemmas that concern resources. *)

(* If [R] maps [x] to no permission, then [x] is not in the domain of [R]. *)

Lemma domain_none:
  forall R x,
  maps R x (@DX_NONE _ _) ->
  domain R x ->
  False.
Proof.
  (*  Recall that the notation [domain] is defined in [MSAClosedAllocationMaps]. *)
  unfold domain. intros. unpack. maps_injective. subst. simpl in *. tauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [R] maps [x] to no permission, and if [R] owns [y] and says [y.adopter == x],
   then we have a contradiction: the heap fragment [R] is not closed. *)

Lemma owns_adoptee_implies_knows_adopter:
  forall R x y tagy vsy,
  heap_maps R x (@DX_NONE _ _) ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  False.
Proof.
  destruct R as (hR, lR).
  destruct hR as [ R hR ].
  unfold SCAM_maps. simpl.
  introv hx hy.
  (* Because [R] is closed and [R] says [y.adopter == x], the address [x]
     must be in the domain of [R]. *)
  forwards: hR hy.
    simpl. eauto.
  (* This contradicts the hypothesis that [x] is not in the domain of [R]. *)
  eauto using domain_none.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [R1] owns [x] and if the conjunction [R1 * R2] says [y.adopter == x], then
   [R1] alone says [y.adopter == x]. *)

Lemma owns_adopter_implies_owns_adoptee:
  forall R1 R2 R x y tagy vsy b,
  heap_maps R1 x (DX_EXCLUSIVE b) ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  consistent R ->
  star R1 R2 = R ->
  heap_maps R1 y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)).
Proof.
  introv hx hy hc. intros; subst.
  (* The conjunction [R1 * R2] maps [y] to the permission [DX_EXCLUSIVE ...].
     Hence, either [R1] or [R2] has exclusive permission for [y]. *)
  maps_split_exclusive hy.
  (* If it is [R1], then we are done. *)
  assumption.
  (* If it is [R2], then we reach the interesting case of this proof. We
     must argue that this case is impossible. *)
  false.
  (* [R1] maps [x] to an exclusive permission, so [R2] must map it to no
     permission. *)
  forwards: maps_exclusive_star; [ eexact hx | eauto | ].
  (* The result follows. *)
  eauto using owns_adoptee_implies_knows_adopter.
Qed.

(* ---------------------------------------------------------------------------- *)
(* ---------------------------------------------------------------------------- *)

(* The predicate [abandon R x y R'] means that, in the instrumented heap [R],
   the object [x] is the adopter of the object [y], and if [x] abandons [y],
   then we obtain the instrumented heap [R']. Here, [x] and [y] are memory
   locations. *)

(* Although this predicate is called [abandon], it really describes both
   adoption (in the state [R]) and abandon (in the state [R']). *)

Inductive abandon: resource -> nat -> nat -> resource -> Prop :=
| Abandon:
    forall R x y R' tagy vsy tagx vsx,
    (* In the heap [R], [x.adopter] is [null]. This premise (together with
       the following one) forces [x] and [y] to be distinct objects. It
       also forces adopters to be exclusive objects. *)
    heap_maps R x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
    (* In the heap [R], [y.adopter] is [x]. *)
    heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
    (* Setting [y.adopter] to [null] yields the heap [R']. *)
    heap_upd R y Exclusive (Orphan tagy vsy) = R' ->
    abandon R x y R'.

Ltac select_abandon action :=
  match goal with h: abandon _ _ _ _ |- _ => action h end.
Ltac abandon_inversion :=
  gji select_abandon.
    
(* Inductive abandon : resource -> nat -> nat -> resource -> Prop := *)
(* | Abandon: *)
(*     forall R x y R' tagy vsy tagx vsx, *)
(*     (* In the heap [R], [x.adopter] is [null]. This premise (together with *)
(*        the following one) forces [x] and [y] to be distinct objects. It *)
(*        also forces adopters to be exclusive objects. *) *)
(*     CAM_maps R x (DX_EXCLUSIVE (Block tagx Null vsx)) -> *)
(*     (* In the heap [R], [y.adopter] is [x]. *) *)
(*     CAM_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) -> *)
(*     (* Setting [y.adopter] to [null] yields the heap [R']. *) *)
(*     upd R y Exclusive (Orphan tagy vsy) = R' -> *)
(*     abandon R x y R'. *)

(* Ltac select_abandon action := *)
(*   match goal with h: abandon _ _ _ _ |- _ => action h end. *)
(* Ltac abandon_inversion := *)
(*   gji select_abandon. *)

(* ---------------------------------------------------------------------------- *)

(* A repetition of the definition. *)

(* This lemma is convenient when we have an [abandon] fact and we already have
   names, [tagy] and [vsy], for the tag and fields of the object [y]. *)

Lemma abandon_as_update:
  forall R x y R' tagy vsy,
  abandon R x y R' ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  R' = heap_upd R y Exclusive (Orphan tagy vsy).
Proof.
  intros. abandon_inversion.
  SCAM_maps_injective. reflexivity.
Qed.

(* The definition implies that [y.adopter] is null in [R']. *)

Lemma abandon_yields_null_adopter:
  forall R x y R',
  abandon R x y R' ->
  forall tag adopter vs,
  heap_maps R' y (DX_EXCLUSIVE (Block tag adopter vs)) ->
  adopter = Null.
Proof.
  intros. abandon_inversion. heap_maps_update_inversion.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [x] has an adoptee [y], then [x] and [y] differ. *)

Lemma no_self_adoption:
  forall R x y R',
  abandon R x y R' ->
  x <> y.
Proof.
  (* The definition of [abandon] implies [x.adopter == null] and
     [y.adopter != null]. *)
  intros. abandon_inversion. intro. subst. SCAM_maps_injective. congruence.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Abandon is the opposite of adoption. *)

Lemma abandon_adopt:
  forall R y x tagx vsx tagy vsy req,
  heap_maps R x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy Null vsy)) ->
  abandon (adopt R y x tagy vsy req) x y R.
Proof.

intros.
econstructor.
  unpack; eapply (SCAM_maps_update_elsewhere heap_sub_resource); eauto.
  eapply (SCAM_maps_update_there heap_sub_resource); eauto.
  rewrite <- (@ip_inverse _ _ heap_project heap_inject heap_sub_resource R) at 2.
  unfold heap_upd, adopt.
  unfold SCAM_maps, SCAM_update.

  rewrite inject_erase.
  eapply injectA_equality.

  erewrite (@CAM_update_equality
             _ _ _ _
             (heap_project (heap_inject _ _))
             _ _ _ _ _ 
             (@pi_inverse _ _ heap_project heap_inject heap_sub_resource _ _)).

  unfold CAM_update.
  came.

  decompose_pairs.
came.
  rewrite update_overwrite.
  eapply update_identity. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [R'] owns [z] after [x] abandons [y], then this was true as well before the
   abandon took place. *)

Lemma abandon_owns_reverse:
  forall R x y R',
  abandon R x y R' ->
  forall z,
  owns R' z ->
  owns R z.
Proof.
  dependent_destruction. intros. unpack. subst.
  heap_maps_update_inversion; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Abandon preserves consistency. *)

Lemma consistent_abandon:
  forall R x y R',
  abandon R x y R' ->
  consistent R ->
  consistent R'.
Proof.
  intros. abandon_inversion. eauto using heap_consistent_upd.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Abandon commutes with [rio]. Intuitively, this means that actions by other
   threads cannot take our adoptees away from us while we are asleep. *)

Lemma rio_abandon:
  forall R1 x y R1',
  abandon R1 x y R1' ->
  forall R2,
  rio R1 R2 ->
  exists R2',
  abandon R2 x y R2' /\
  rio R1' R2'.
Proof.
  intros. abandon_inversion. xplit.
  (* The [abandon] subgoal. *)
  econstructor; eauto using (SCAM_maps_stable heap_project_morphism), @stable_exclusive.
  (* The [rio] subgoal. *)
  eapply (SCAM_rio_update_update heap_project_morphism heap_inject_bimorphism); eauto using @rio_reflexive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Abandon is affine. Intuitively, this means that receiving some extra state
   that we do not need does not disturb the relation between us and our
   adoptees. *)

(* This property exploits the fact that we have exclusive ownership of [y]. *)

Lemma star_abandon_1:
  forall R1 x y R1',
  abandon R1 x y R1' ->
  forall R2,
  consistent (star R1 R2) ->
  abandon (star R1 R2) x y (star R1' R2).
Proof.
  intros. abandon_inversion.
  econstructor.
    SCAM_maps_maximal heap_project_morphism.
    SCAM_maps_maximal heap_project_morphism.
    symmetry. eauto using star_upd_exclusive_left.
Qed.

Lemma star_abandon_2:
  forall R1 x y R1',
  abandon R1 x y R1' ->
  forall R2,
  consistent (star R1 R2) ->
  consistent (star R1' R2).
Proof.
  intros. abandon_inversion.
  erewrite star_upd_exclusive_left; eauto using heap_consistent_upd.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Conversely, if a piece of state is taken away from us and concerns neither
   the adopter [x] nor the adoptee [y], then the relation between adopter and
   adoptee is not disturbed. *)

Lemma star_abandon_reverse:
  forall R x y R',
  abandon R x y R' ->
  forall R'1 R'2,
  (* The state after adoption is a conjunction [R'1 * R'2]. *)
  R' = star R'1 R'2 ->
  consistent R ->
  forall tagx vsx tagy vsy,
  (* [R'2] owns both [x] and [y]. *)
  heap_maps R'2 x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
  heap_maps R'2 y (DX_EXCLUSIVE (Block tagy Null vsy)) ->
  (* As a result, [R'1] is completely independent and can be hoisted out
     of the [abandon] predicate. *)
  exists R2,
  abandon R2 x y R'2 /\
  R = star R'1 R2.
Proof.
  introv ha ? hc hx hy. subst.
  (* We would like to claim that an appropriate [R2] is obtained by
     starting with [R'2] and letting [x] adopt [y]. We first need to
     check that this adoption operation is permitted. Here is the
     proof obligation. *)
  assert (req : x <> y /\ owns R'2 x).
    solve [ eauto using no_self_adoption ].
  (* We will also meet a proof obligation that [R'] is consistent. *)
  forwards hc': consistent_abandon; [ eauto | eauto | ].
  (* Destruct the hypothesis. *)
  abandon_inversion.
  (* Provide the witness for [R2]. *)
  exists (adopt R'2 y x tagy vsy req); xplit.
  (* Sub-goal 1. *)
  solve [ eauto using abandon_adopt ].
  (* Sub-goal 2. *)
  (* Hoist [adopt] out of [star]. *)
  erewrite star_adopt_right_LR with (hc := hc') by eassumption.
  (* The goal is almost within reach. But first, we have two variables
     that represent the tag of [y], and two variables that represent
     the fields of [y]. We must argue that they are in fact equal. *)
  match goal with
  h: heap_upd ?R _ _ (Orphan ?tagy1 ?vsy1) = ?R' |- ?R = adopt ?R' ?y _ ?tagy2 ?vsy2 _ =>
    assert (heap_maps R' y (DX_EXCLUSIVE (Block tagy1 Null vsy1))); [ rewrite <- h |
    assert (heap_maps R' y (DX_EXCLUSIVE (Block tagy2 Null vsy2))) ]
  end.
    (* Proof of the first assertion. *)
    eapply (SCAM_maps_update_there heap_sub_resource); eauto.
    (* Proof of the second assertion. *)
    SCAM_maps_maximal heap_project_morphism.
  (* As a consequence of these assertions, we obtain the desired equality. *)
  SCAM_maps_injective.
  (* The result follows. *)
  eapply (SCAM_update_reverse heap_sub_resource); eassumption.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The access rights that we have for the adopter [x] are the same in [R]
   and [R'], that is, before and after [x] abandons [y]. *)

Lemma abandon_maps:
  forall R x y R',
  abandon R x y R' ->
  forall p,
  heap_maps R' x p ->
  heap_maps R x p.
Proof.
  intros. abandon_inversion. heap_maps_update_inversion.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [x] has at least one adoptee [y], then [x] cannot be duplicable. *)

Lemma adopter_is_exclusive:
  forall R x y R',
  abandon R x y R' ->
  forall p,
  heap_maps R' x p ->
  duplicable p ->
  False.
Proof.
  introv ha hm hd.
  (* [x] is mapped to [p] by [R'], hence also by [R]. *)
  forwards: abandon_maps; [ eauto | eauto | ].
  (* But [abandon] implies that [x] is mapped by [R] to some exclusive
     permission... *)
  dependent destruction ha. subst R'.
  (* ... which contradicts the fact that [p] is duplicable. *)
  SCAM_maps_injective.
  eapply @not_duplicable_exclusive with (A := orphan) (B := block).
  eassumption.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Abandon commutes with an update of [x], provided [x] remains exclusive
   and [x.adopter] remains [null]. That is, updating the tag or fields of
   the object at [x] does not affect its relationship with its adoptees. *)

Lemma abandon_update:
  forall R x y R',
  abandon R x y R' ->
  forall o,
  abandon (heap_upd R x Exclusive o) x y (heap_upd R' x Exclusive o).
Proof.
  intros.
  forwards: no_self_adoption. eauto.
  abandon_inversion. destruct o. econstructor.
  eapply (SCAM_maps_update_there heap_sub_resource); eauto.
  eapply (SCAM_maps_update_elsewhere heap_sub_resource); eauto.

  unfold heap_upd.
  unfold SCAM_update.
  unfold SCAM_maps in *.
  repeat rewrite inject_erase. 
  eapply injectA_equality.
  came.
  erewrite (@pi_inverse _ _ _ _ heap_sub_resource).
  erewrite (@pi_inverse _ _ _ _ heap_sub_resource).
  came.
  destruct (heap_project R).
  came.
  eapply update_update; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Two successive abandon operations commute. Intuitively, this means that the
   order of the adoptees does not matter. *)

Lemma abandon_abandon:
  forall R x y1 R1 y2 R2,
  abandon R x y1 R1 ->
  abandon R1 x y2 R2 ->
  exists R1',
  abandon R x y2 R1' /\
  abandon R1' x y1 R2.
Proof.
  introv h1 h2.
  (* Decompose the hypotheses. *)
  unpack (no_self_adoption h2).
  abandon_inversion.
  heap_maps_update_inversion.
  (* Build the conclusion. *)
  xplit.
  econstructor; eauto.
  econstructor.
    eapply (SCAM_maps_update_elsewhere heap_sub_resource); eauto.
    eapply (SCAM_maps_update_elsewhere heap_sub_resource); eauto.

  unfold heap_upd.
  unfold SCAM_update.
  unfold SCAM_maps in *.
  repeat rewrite inject_erase. 
  eapply injectA_equality.
  came.
  erewrite (@pi_inverse _ _ _ _ heap_sub_resource).
  erewrite (@pi_inverse _ _ _ _ heap_sub_resource).
  came.
  destruct (heap_project R).
  came.
  eapply update_update; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)
(* ---------------------------------------------------------------------------- *)

(* Absence of adoptees. *)

(* The predicate [no_adoptees R x] means that [R] has some permission over [x]
   and if [x] is exclusive then [x] has no adoptees, i.e. [x] cannot abandon
   any more objects. *)

Inductive no_adoptees : resource -> nat -> Prop :=
| NoAdoptees:
    forall R x,
    (* [x] is in the domain of [R]. *)
    (exists p, heap_maps R x p /\ present p) ->
    (* [x] cannot abandon one more object. *)
    (
      forall y tagx tagy vsx vsy,
      heap_maps R x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
      heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
      False
    ) ->
    no_adoptees R x.

(* ---------------------------------------------------------------------------- *)

(* This property is preserved by [rio]. Intuitively, this means that, while we
   are asleep, actions by other threads cannot break the fact that we have no
   adoptees. *)

Lemma no_adoptees_rio:
  forall R1 x,
  no_adoptees R1 x ->
  forall R2,
  rio R1 R2 ->
  no_adoptees R2 x.
Proof.
  destruct 1 as [ R1 x hdx hnoy ]. introv hrio.
  econstructor.
  (* Sub-goal 1. *)
  (* [x] is in the domain of [R1], so it is also in the domain of [R2]. *)
  destruct hdx as [ p [ hmx ? ]].
  unpack (SCAM_maps_rio heap_project_morphism hmx hrio).
  eauto using @present_rio.
  (* Sub-goal 2. *)
  introv hmx hmy.
  (* [R2] has exclusive permissions for [x] and [y], so [R1] must have the
     same permissions. *)
  unpack (maps_rio_exclusive_reverse hmx hrio).
  unpack (maps_rio_exclusive_reverse hmy hrio).
  (* We obtain a contradiction, since [x] has no adoptees in [R1]. *)
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This property is preserved by [star]. Intuitively, this means that receiving
   a bit of extra state that we do not need cannot break the fact that we have
   no adoptees. *)

Lemma no_adoptees_star:
  forall R1 x,
  no_adoptees R1 x ->
  forall R2,
  consistent (star R1 R2) ->
  no_adoptees (star R1 R2) x.
Proof.
  destruct 1 as [ R1 x [ p [ hmxp ? ]] hnoy ]. introv hc.
  econstructor. 
  (* Sub-goal 1. *)
  unpack (SCAM_maps_maximal_left heap_project_morphism hmxp hc).
  eauto using present_maximal.
  (* Sub-goal 2. *)
  introv hmx hmy.
  (* [x] is in the domain of [R1] and exclusive in [R1 * R2], so it must be
     exclusive in [R1], and absent in [R2]. *)
  maps_split_exclusive hmx.
  (* The conjunction [R1 * R2] owns [y], so either [R1] or [R2] owns [y]. *)
  maps_split_exclusive hmy.
  (* If [R1] owns [y], we are done. *)
  solve [ eauto ].
  (* If [R2] owns [y], then the fact that [R2] is closed implies that [x] is
     the domain of [R2], which contradicts the fact that [x] is absent in [R2]. *)
  eauto using owns_adoptee_implies_knows_adopter.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Conversely, if a piece of state is taken away from us and does not concern [x],
   then the fact that [x] has no adoptees is not violated. *)

Lemma no_adoptees_star_reverse:
  forall R x,
  no_adoptees R x ->
  consistent R ->
  forall R1 R2,
  R = star R1 R2 ->
  forall tagx vsx,
  heap_maps R2 x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
  no_adoptees R2 x.
Proof.
  dependent_destruction. intros. subst. econstructor.
  xplit. eauto. simpl; tauto.
  eauto using (SCAM_maps_maximal_right heap_project_morphism), @maximal_exclusive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This property is preserved by an update. *)

(* More precisely, if [x] has no adoptees in [R], and if [x] is exclusive,
   then an update to the tag and/or fields of [x] preserves the fact that
   [x] has no adoptees. *)

Lemma no_adoptees_update:
  forall R x,
  no_adoptees R x ->
  forall tagx vx,
  heap_maps R x (DX_EXCLUSIVE (Block tagx Null vx)) ->
  forall o,
  no_adoptees (heap_upd R x Exclusive o) x.
Proof.
  destruct 1 as [ R1 x _ hnoy ]. destruct o. econstructor.
  (* Sub-goal 1. *)
  (* [x] remains present after the update. *)
  xplit. eapply (SCAM_maps_update_there heap_sub_resource); eauto. simpl; tauto.
  (* Sub-goal 2. *)
  introv _ hmy.
  (* The locations [x] and [y] must be distinct, because [x] has just
     been updated with a null [adopter] pointer, while [y.adopter] is
     [x]. *)
  destruct (@SCAM_maps_update_inversion _ _ _ _ _ heap_project heap_inject _ _ _ _ _ _ _ _ hmy); unpack; [
    subst
  | match goal with h: _ = _ |- _ => inversion h end
  ].
  (* The result follows. If no object [y] could be an adopter of [x]
     prior to the update, then the same holds after the update. *)
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This property is trivially true of duplicable objects. *)

(* This follows from the careful manner in which the definition of [no_adoptees]
   is stated. For [x] to have an adoptee, [x] is required to be exclusive, so the
   fact that duplicable objects have no adoptees is a tautology. *)

Lemma duplicable_objects_have_no_adoptees:
  forall R x o,
  heap_maps R x (DX_DUPLICABLE o) ->
  no_adoptees R x.
Proof.
  introv hm. econstructor; intros.
  (* Sub-goal 1. *)
  xplit. eauto. simpl; eauto.
  (* Sub-goal 2. *)
  SCAM_maps_injective. congruence.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This property is true of newly allocated objects. *)

Lemma new_objects_have_no_adoptees:
  forall R x m o req,
  consistent R ->
  x = heap_limit R ->
  no_adoptees (heap_allocate R (mo2dx m o) req) x.
Proof.
  intros. econstructor.
  (* Sub-goal 1. *)
  xplit.
  eapply (SCAM_maps_allocate_there heap_project_morphism heap_sub_resource); eauto.
  eapply present_mo2dx.
  (* Sub-goal 2. *)
  introv hmx hmy.
  (* The locations [x] and [y] must be distinct, because [x.adopter]
     is null, while [y.adopter] is [x]. As a result, the object [y]
     is unaffected by the allocation. *)
  destruct (@SCAM_maps_allocate_inversion _ _ _ _ _ heap_project heap_inject _ _ _ _ _ _ hmy); [
    clear hmx hmy; subst
  | destruct m, o; simpl in *; congruence ].
  (* Now, we find that, prior to allocation, [y.adopter] was already
     equal to [x], which was not an allocated location. This contradicts
     the fact that [R] is closed. *)
  destruct R as (hR, lR);
  destruct hR as [ R hR ]; simpl in *.
  unfold SCAM_maps, SCAM_limit in *. came.
  forwards: hR; [ eauto | simpl; eauto | ].
  unfold domain in *. unpack.
  forwards: maps_limit; [ eauto | ].
  omega.
Qed.

(* ---------------------------------------------------------------------------- *)
(* ---------------------------------------------------------------------------- *)

(* Iterated abandon. *)

(* The predicate [abandon_all R x ys R'] means that, in the instrumented heap
   [R], the object [x] currently is the adopter of the objects [ys], and if it
   abandons them, then we obtain the instrumented heap [R']. (This is just the
   reflexive, transitive closure of [abandon].) Furthermore, this predicate
   guarantees that the list [ys] is complete, that is, in [R'], the object [x]
   no longer has any adoptee. *)

Inductive abandon_all : resource -> nat -> list nat -> resource -> Prop :=
| AbandonAllZero:
    forall R x,
    no_adoptees R x ->
    abandon_all R x nil R
| AbandonAllSucc:
    forall R x y R' ys R'',
    abandon R x y R' ->
    abandon_all R' x ys R'' ->
    abandon_all R x (y :: ys) R''.

Hint Constructors abandon_all : abandon_all.

Ltac select_abandon_all action :=
  match goal with h: abandon_all _ _ _ _ |- _ => action h end.
Ltac abandon_all_inversion :=
  gji select_abandon_all.

(* ---------------------------------------------------------------------------- *)

(* The properties of [abandon] extend to [abandon_all] in a trivial manner. *)

Lemma consistent_abandon_all:
  forall R x ys R',
  abandon_all R x ys R' ->
  consistent R ->
  consistent R'.
Proof.
  induction 1; eauto using consistent_abandon.
Qed.

Ltac consistent_abandon_all :=
  repeat match goal with h: abandon_all _ _ _ _ |- _ =>
    forwards: consistent_abandon_all; [ eexact h | eauto 6 with consistent_split | revert h ]
  end; intros.

Hint Resolve consistent_abandon_all : rio_star_hints.
  (* Used in JMonotonicity, case [JBlock]. *)

(* ---------------------------------------------------------------------------- *)

Lemma abandon_all_owns_reverse:
  forall R x ys R',
  abandon_all R x ys R' ->
  forall z,
  owns R' z ->
  owns R z.
Proof.
  induction 1; eauto using abandon_owns_reverse.
Qed.

(* ---------------------------------------------------------------------------- *)

Lemma rio_abandon_all:
  forall R1 x ys R1',
  abandon_all R1 x ys R1' ->
  forall R2,
  rio R1 R2 ->
  exists R2',
  abandon_all R2 x ys R2' /\
  rio R1' R2'.
Proof.
  induction 1; intros.
  (* AbandonAllZero *)
  eauto using no_adoptees_rio with abandon_all.
  (* AbandonAllSucc *)
  forwards: rio_abandon; [ eauto | eauto | unpack ].
  forwards: IHabandon_all; [ eauto | unpack ].
  eauto with abandon_all.
Qed.

Ltac rio_abandon_all :=
  match goal with
  | h1: abandon_all ?R1 _ _ _, h2: rio ?R1 _ |- _ =>
      forwards: (>> rio_abandon_all h1 h2);
      unpack
  end.

(* ---------------------------------------------------------------------------- *)

Lemma star_abandon_all_1:
  forall R1 x ys R1',
  abandon_all R1 x ys R1' ->
  forall R2,
  consistent (star R1 R2) ->
  abandon_all (star R1 R2) x ys (star R1' R2).
Proof.
  induction 1; eauto using star_abandon_1, star_abandon_2, no_adoptees_star
                      with abandon_all.
Qed.

Lemma star_abandon_all_2:
  forall R1 x ys R1',
  abandon_all R1 x ys R1' ->
  forall R2,
  consistent (star R1 R2) ->
  consistent (star R1' R2).
Proof.
  induction 1; eauto using star_abandon_2.
Qed.

Lemma star_abandon_all_1_symmetric:
  forall R2 x ys R2',
  abandon_all R2 x ys R2' ->
  forall R1,
  consistent (star R1 R2) ->
  abandon_all (star R1 R2) x ys (star R1 R2').
Proof.
  introv ? hc.
  rewrite (@star_commutative _ _ R1) in hc.
  rewrite (@star_commutative _ _ R1) at 1.
  rewrite (@star_commutative _ _ R1) at -2. (* what the f??? *)
  eauto using star_abandon_all_1.
Qed.

Lemma star_abandon_all_2_symmetric:
  forall R2 x ys R2',
  abandon_all R2 x ys R2' ->
  forall R1,
  consistent (star R1 R2) ->
  consistent (star R1 R2').
Proof.
  introv ? hc.
  rewrite (@star_commutative _ _ R1) in hc.
  rewrite (@star_commutative _ _ R1).
  eauto using star_abandon_all_2.
Qed.

Hint Resolve star_abandon_all_1 star_abandon_all_2
star_abandon_all_1_symmetric star_abandon_all_2_symmetric :
star_abandon_all.

(* ---------------------------------------------------------------------------- *)

Lemma abandon_all_update:
  forall R x ys R',
  abandon_all R x ys R' ->
  forall tagx vx,
  heap_maps R' x (DX_EXCLUSIVE (Block tagx Null vx)) ->
  forall o,
  abandon_all (heap_upd R x Exclusive o) x ys (heap_upd R' x Exclusive o).
Proof.
  induction 1; intros; econstructor; eauto using abandon_update, no_adoptees_update.
Qed.

(* ---------------------------------------------------------------------------- *)

Lemma adopters_are_exclusive:
  forall R x ys R',
  abandon_all R x ys R' ->
  forall p,
  heap_maps R' x p ->
  duplicable p ->
  ys = nil.
Proof.
  induction 1; introv ? hd.
  (* Zero. *)
  eauto.
  (* Step. *)
  false.
  (* Use the induction hypothesis to deduce that [ys] is [nil]. *)
  forwards: IHabandon_all; [ eauto | eauto | subst ys; clear IHabandon_all ].
  (* Draw the consequences. *)
  revert hd. (* this hypothesis blocks [abandon_all_inversion] *)
  abandon_all_inversion.
  (* Conclude. *)
  eauto using adopter_is_exclusive.
Qed.

(* ---------------------------------------------------------------------------- *)

Lemma abandon_all_maps:
  forall R x ys R',
  abandon_all R x ys R' ->
  forall p,
  heap_maps R' x p ->
  heap_maps R x p.
Proof.
  induction 1; intros; eauto using abandon_maps.
Qed.

Ltac abandon_all_maps :=
  repeat match goal with h: abandon_all _ _ _ _ |- _ =>
    forwards: abandon_all_maps; [ eexact h | SCAM_maps_maximal heap_project_morphism | revert h ]
  end; intros.

(* ---------------------------------------------------------------------------- *)

Lemma star_abandon_all_reverse:
  forall R x ys R',
  abandon_all R x ys R' ->
  forall R'1 R'2,
  (* The state after adoption is a conjunction [R'1 * R'2]. *)
  R' = star R'1 R'2 ->
  consistent R ->
  (* [R'2] owns [x] as well as every adoptee [y]. *)
  forall tagx vsx,
  heap_maps R'2 x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
  Forall (fun y =>
    exists tagy vsy,
    heap_maps R'2 y (DX_EXCLUSIVE (Block tagy Null vsy))
  ) ys ->
  (* As a result, [R'1] is completely independent and can be hoisted out
     of the [abandon_all] predicate. *)
  exists R2,
  abandon_all R2 x ys R'2 /\
  R = star R'1 R2.
Proof.
  induction 1; intros; subst; destruction_Forall; unpack.
  (* Base case. *)
  solve [ eauto using no_adoptees_star_reverse with abandon_all ].
  (* Step. *)
  (* Apply the induction hypothesis to hoist [R'1] out of the tail
     of the list of adoptees. *)
  forwards [ R2b [ ha ? ]]: IHabandon_all; eauto using consistent_abandon.
  subst. clear IHabandon_all hdd.
  (* Now, we would like to apply the lemma [star_abandon_reverse], so as
     to hoist [R'1] out of the adoption relation between [x] and [y]. In
     order to do that, though, the hypothesis that [R'2] owns [y] must be
     hoisted up through the tail of the list, to find that [R2b] owns [y]. *)
  forwards ho: abandon_all_owns_reverse. eexact ha. eauto.
  (* We have just proved that [R2b] owns [y], but we must go a little
     further and prove that, in [R2b], [y.adopter] is null. Let [b]
     be the block at address [y]. *)
  destruct ho as [ b ? ].
  (* Then, [R'1 * R2b] also maps [y] to [b]. *)
  assert (ho : heap_maps (star R'1 R2b) y (DX_EXCLUSIVE b)).
    eauto using (SCAM_maps_maximal_right heap_project_morphism), @maximal_exclusive, consistent_abandon.
  (* But [R'1 * R2b] is the state obtained by abandoning [y], so in this
     state, [y.adopter] must be [null]. *)
  destruct b. forwards: abandon_yields_null_adopter; [ eauto | eauto | ]. subst.
  (* We have now proved that [R2b] says [y.adopter] is null. We can apply
     the lemma [star_abandon_reverse]. *)
  forwards: star_abandon_reverse; eauto using abandon_all_maps. unpack. subst.
  (* The result follows. *)
  eauto with abandon_all.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A key property of the predicate [abandon_all] is that the list [ys] is a list
   of all the current adoptees of [x]. The following lemma states this fact: if
   [y.adopter] is [x], then [y] must be a member of the list [ys]. *)

Lemma no_lost_sheep:
  forall R x ys R' y tagx tagy vsx vsy,
  abandon_all R x ys R' ->
  heap_maps R x (DX_EXCLUSIVE (Block tagx Null vsx)) ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  In y ys.
Proof.
  induction 1; simpl; intros.
  (* Base case. *)
  (* Here, the fact that [y] cannot possibly point to [x] follows
     directly from the definition of [no_adoptees]. *)
  match goal with h: no_adoptees _ _ |- _ => inversion h; subst end.
  solve [ eauto ].
  (* Step case. *)
  (* The object [z] is known to be an adoptee of [x]. If [z] and [y]
     are equal, then the result is immediate. If [z] and [y] differ,
     then we must apply the induction hypothesis, so the goal boils
     down to proving that our two [CAM_maps] hypotheses remain true
     after [x] abandons [z]. *)
  match goal with h: abandon _ _ ?z _ |- _ =>
    inversion h; subst;
    destruct (eq_nat_dec z y); [ eauto | right ]
  end.
  eapply IHabandon_all;
  eapply (SCAM_maps_update_elsewhere heap_sub_resource);
  eauto using no_self_adoption.
Qed.

(* ---------------------------------------------------------------------------- *)

(* As part of the [take] operation, once we have proven that some object [y] is a
   member of the list [ys], we are interested in moving [y] to the front of the
   list, so as to be able to then easily take it out of the list. The following
   lemma states that [y] can indeed be brought to the front. *)

Lemma the_last_will_be_the_first:
  forall R x ys R',
  abandon_all R x ys R' ->
  forall y,
  In y ys ->
  exists Ry zs,
  Permutation ys (y :: zs) /\
  abandon R x y Ry /\
  abandon_all Ry x zs R'.
Proof.
  induction 1; simpl; introv h; [ tauto | destruct h ].
  (* If we have reached [y], the result is immediate. *)
  subst. eauto.
  (* If we are looking at some other element [z], then we must apply
     the induction hypothesis to bring [y] in front of the list tail,
     then permute [y] and [z]. *)
  forwards: IHabandon_all; [ eauto | unpack; subst ].
  forwards: abandon_abandon; [ | eauto | unpack ]. eauto.
  xplit; eauto using perm_trans, perm_swap with abandon_all.
Qed.

