Set Implicit Arguments.
Require Import Relation_Operators.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WkInversion.
Require Import Resources.
Require Import Row.
Require Import Abandon.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JHeap.
Require Import JCanonicalization.
Require Import JDecomposition.
Require Import JSubstitution.
Require Import JDrop.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import Reduction.
Require Import SubjectReduction.
Require Import Progress.
(* Require Import CSubjectReduction. *)
(* Require Import CProgress. *)
(* Require Import TypeSoundness. *)
Require Import Harmless.
Require Import Inhabited.

(* A well-typed concurrent program allows no data race. *)

(* A predicate for terms that are ready to access a location. *)

Inductive access_mode :=
  | Read
  | Write
.

Inductive access: access_mode -> term -> nat -> Prop :=
 
  | AccRead:
      forall l field,
      access Read (TRead (TLoc l) field) l

  | AccWrite:
      forall l field v,
      access Write (TWrite (TLoc l) field v) l

  | AccWriteTag:
      forall l tag,
      access Write (TWriteTag (TLoc l) tag) l

  | AccSwitch:
      forall l ts t,
      access Read (TSwitch (TLoc l) (fill_row ts t)) l

  | AccGive:
      forall ly lx,
      access Write (TGive (TLoc ly) (TLoc lx)) ly

  | AccContext:
      forall am E t l,
      eval_context E ->
      access am t l ->
      access am (fill E t) l

  | AccThread:
      forall am t l,
      access am t l ->
      access am (TThread t) l

  | AccParLeft:
      forall am t1 t2 l,
      access am t1 l ->
      access am (TPar t1 t2) l

  | AccParRight:
      forall am t1 t2 l,
      access am t2 l ->
      access am (TPar t1 t2) l

.

(* A value cannot access a memory location. *)

Lemma access_non_value:
  forall am t l K,
  access am t l ->
  wk K t VALUE ->
  False.
Proof.
  induction 1; intros; try open_eval_context; wk_inversion.
Qed.


(* There is a data_race in a concurrent program whenever one can find two
   distinct threads that access the same location [l], where at least one
   of the accesses is a write. *)

Inductive data_race: term -> Prop :=

  | DataRace:
      forall am1 am2 t1 t2 l,
      access am1 t1 l ->
      access am2 t2 l ->
      (am1 = Write \/ am2 = Write) ->
      data_race (TPar t1 t2)

  | DRLeft:
      forall t1 t2,
      data_race t1 ->
      data_race (TPar t1 t2)

  | DRRight:
      forall t1 t2,
      data_race t2 ->
      data_race (TPar t1 t2)

.

(* [JcTop] judgments are stable by extension of the resources. *)

Lemma jctop_split_right:
  forall R2 K v T,
  jctop R2 K v T ->
  forall R R1,
  star R1 R2 = R ->
  consistent R ->
  jctop R K v T.
Proof.
  introv hj hs.
  rewrite star_commutative in hs.
  eauto using jpi_drop_right.
Qed.

(* A location that is accessed by a well-typed program contains
   a block. If moreover the access is a [write] access, then
   the block is exclusive. *)

Lemma seq_typed_access:
  forall n Rh K P t T,
  jt n Rh K P t T ->
  wk zilch t PROG ->
  forall am l,
  access am t l ->
  forall R Rp,
  jpi Rp K P ->
  star Rh Rp = R ->
  consistent R ->
  exists m tag Ts U,
  (am = Write -> m = Exclusive) /\
  jctop R K (TLoc l) (TyBlock m tag Ts U).
Proof.
  induction 1; introv hwk hacc hjpi hs hc; subst.
  (* Singleton. *)
  false; eauto using access_non_value.
  (* Frame. *)
  eauto using jps_sound, JpsDisappearsRight with consistent_split.
  (* Cut. *)
  eauto using reconfigure07 with j.
  (* Abs. *)
  inversion hwk. false; eauto using access_non_value.
  (* SubLeft. *)
  eauto using jps_sound with consistent_split.
  (* SubRight. *)
  eauto.
  (* Application. *)
  inversion hacc; open_eval_context. 
  match goal with
    | h: TApp _ _ = TApp _ _ |- _ => inversion h; subst; clear h
  end.
  inversion hwk. wk_inversion. 
  eapply IHj; eauto using jps_sound, JpsDisappearsLeft with consistent_split.
  (* Block. *)
  inversion hacc; open_eval_context; discriminate.
  (* Fold. *)
  eauto.
  (* Dynamic. *)
  inversion hacc; open_eval_context; discriminate.
  (* New. *)
  inversion hacc; open_eval_context; discriminate.
  (* Read. *)
  classification.
  inversion hacc; subst.
  do 4 eexists; xplit; intros; try discriminate; eauto using jctop_split_right.
  open_eval_context; discriminate.
  (* Write. *)
  inversion hacc; subst.
  do 4 eexists; xplit;
  eauto using jps_sound, JpsDisappearsRight, jpi_drop_right, @star_commutative.
  open_eval_context; discriminate.
  (* WriteTag. *)
  classification.
  inversion hacc; subst.
  do 4 eexists; xplit; eauto using jctop_split_right.
  open_eval_context; discriminate.
  (* Switch. *)
  inversion hacc; subst.
  forwards: decomposition_TyCon;
    eauto using jps_sound, JpsDisappearsLeft with consistent_split.
  unpack.
  do 4 eexists; xplit; intros; try discriminate; eauto using jctop_split_right.
  open_eval_context; discriminate.
  (* Give. *)
  inversion hacc; subst.
  forwards: decomposition_adoptee;
    eauto using jps_sound, JpsDisappearsRight with consistent_split.
  unpack.
  do 4 eexists; xplit; eauto using jctop_split_right.
  open_eval_context; discriminate.
  (* Take. *)
  inversion hacc; open_eval_context; discriminate.
  (* Fail. *)
  inversion hacc; open_eval_context; discriminate.
  (* Spawn. *)
  inversion hacc; open_eval_context; discriminate.
  (* Lck. *)
  inversion hacc; open_eval_context; discriminate.
  (* Locked. *)
  inversion hacc; open_eval_context; discriminate.
  (* NewLck. *)
  inversion hacc; open_eval_context; discriminate.
  (* Acquire. *)
  inversion hacc; open_eval_context; discriminate.
  (* Release. *)
  inversion hacc; open_eval_context; discriminate.
  (* ExistsElim. *)
  inversion hacc; open_eval_context; discriminate.
  (* ForallIntro. *)
  jpi_inversion.
  assert (st: shift 0 t=t).
   eapply closed_lift_invariant; eauto using wk_closed, length_empty.
  forwards ih: IHj; try rewrite st; eauto.
  named_unpack ih.

  forwards: all_substitutable_kinds_inhabited; [ eauto | unpack ].
  do 4 eexists; xplit; auto.
  eapply j_substitution; eauto using wk_empty.
  simpl_subst_all; eauto.

Grab Existential Variables.
auto. auto. auto. auto. auto.

Qed.

Lemma typed_access:
  forall am R t l,
  access am t l ->
  jcp R t ->
  consistent R ->
  exists K m tag Ts U,
  (am = Write -> m = Exclusive) /\
  jctop R K (TLoc l) (TyBlock m tag Ts U).
Proof.
  introv hacc.
  revert R.
  induction hacc; introv hjcp hc; inversion hjcp; subst.
  (* Eval context. *)
  open_eval_context. discriminate.
  (* Eval context. *)
  open_eval_context. discriminate.
  (* Thread. *)
  make_jt_hypothesis hjt.
  forwards: seq_typed_access;
    eauto using JpiEmpty, @consistent_by_equality, @core_right_unit.
  (* ParLeft. *)
  eauto using jcp_drop_right.
  (* ParRight. *)
  eauto using jcp_drop_left.
Qed.

(* If a concurrent program is well-typed, then it contains no data-race. *)

Lemma data_race_freedom:
  forall h t,
  jconf (Cf h t) ->
  data_race t ->
  False.
(* The following tactic inverts an access hypothesis using
   the previous lemmas. *)
Ltac access_inversion i1 i2 :=
  match goal with
    | h: access _ _ _ |- _ =>
        (* The first step reveals the block that is accessed. *)
        forwards i1: typed_access h;
        [ eauto
        | eauto using jh_consistent with consistent_split
        | named_unpack i1 ];
        (* The second step decomposes the block. *)
        forwards i2: decomposition_TyBlock i1;
        [ eauto using jh_consistent with consistent_split
        | named_unpack i2 ];
        clear h
  end.
Proof.
  introv hjconf hdr.
  inversion hjconf as [ h' t' R Ravl Rlck hwkcp hs hjh hjpi hjcp ]; subst.
  clear hjconf.
  (* Induction on the data race. *)
  induction hdr.
  (* Interesting case: two parallel accesses to the same location [l]. *)
  jcp_inversion.
  access_inversion j2 m2.
  access_inversion j1 m1.
  subst.
  (* The inversions above introduce two heap fragments [R1] and [R2] in
     which the location [l] appears. We show that the combination 
     [star R1 R2] of these two fragments is consistent. *)
  match goal with
    | haa1: abandon_all _ ?l _ (star _ ?R1),
      haa2: abandon_all _ ?l _ (star _ ?R2) |- _ =>
        assert (consistent (star R1 R2));
        [ eapply consistent_split_right;
            rewrite <- reconfigure11;
            eapply star_abandon_all_2; eauto;
            eapply star_abandon_all_2_symmetric; eauto;
            eapply consistent_split_right;
            rewrite <- reconfigure11;
            eauto using jh_consistent with consistent_split
        | auto ]
  end.
  (* Case on which of the accesses is a write. *)
  match goal with 
    | h: _ \/ _ |- _ => destruct h
  end;
  (* Deduce that the corresponding block is exclusive in one of the
     heap fragments [R1] or [R2]. *)
  match goal with
    | h1: ?e, h2: ?e -> _ |- _ => forwards: h2 h1; subst; simpl in *
  end;
  (* Contradiction with the existence of the block in the other
     fragment. *)
  match goal with 
    | m: mode |- _ =>
        destruct m; simpl in *;
        [ forwards: maps_exclusive_star;
            eauto using @consistent_by_equality, @star_commutative;
            SCAM_maps_injective; discriminate
        | forwards: exclusive_exclusive_star m1 m2; eauto ]
  end.

  (* Iteration. *)
  jcp_inversion.
  wk_inversion.
  eapply IHhdr;
    eauto using JConf, jcp_star_left with wk.

  jcp_inversion.
  wk_inversion.
  eapply IHhdr;
    eauto using JConf, jcp_star_right with wk.

Qed.

