Set Implicit Arguments.
Require Import MyTactics.
Require Import MyLogic.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Row.
Require Import Intro.
Require Import Dup.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JDrop.
Require Import JClassification.
Require Import JDecomposition.

(* Focus and defocus lemmas. *)

(* ------------------------------------------------------------------------- *)

(* This auxiliary lemma explains why the permission subsumption rule
   [JpsDecomposeBlock] is sound. It is proved by induction on the number of
   components that we have to skip before finding the one that interests
   us. *)

Lemma JpsDecomposeBlockSound:
  forall f vs,
  row_context f vs ->
  forall R K v T Ts,
  row_context f Ts ->
  jpi R K (zip (fill_row vs v) (fill_row Ts T)) ->
  exists R1 R2,
  jpi R1 K (zip (fill_row vs v) (fill_row Ts (TySingleton v))) /\
  jctop R2 K v T /\
  star R1 R2 = R.
Proof.
  induction 1; inversion 1; subst; simpl_subst_goal; simpl; intro hjpi;
  dependent destruction hjpi; subst;
  repeat match goal with h: jpi _ _ _ |- _ =>
    rewrite subst_lift in h
  end;
  repeat rewrite subst_lift.
  (* Zero *)
  solve [ eauto 9 using JcTopSingleton, @star_commutative, @core_left_unit with j wk ].
  (* Succ *)
  forwards: IHrow_context; [ eauto | eauto | unpack; subst ].
  solve [ eauto 8 using @star_associative_symmetric with j ].
Qed.

(* ------------------------------------------------------------------------- *)

(* This auxiliary lemma explains why the permission subsumption rule
   [JpsRecomposeBlock] is sound. *)

Lemma JpsRecomposeBlockSound:
  forall f vs,
  row_context f vs ->
  forall R1 R2 K v w T Ts,
  row_context f Ts ->
  jpi R1 K (zip (fill_row vs v) (fill_row Ts (TySingleton w))) ->
  jctop R2 K w T ->
  consistent (star R1 R2) ->
  jpi (star R1 R2) K (zip (fill_row vs v) (fill_row Ts T)).
Proof.
  induction 1; inversion 1; subst; simpl_subst_goal; simpl; introv hjpi ? ?;
  dependent destruction hjpi; subst;
  repeat match goal with h: jpi _ _ _ |- _ =>
    rewrite subst_lift in h
  end;
  repeat rewrite subst_lift.
  (* Zero *)
  decomposition. (* [v] and [w] must be equal. *)
  solve [ eauto using jpi_drop_right, reconfigure04 with j ].
  (* Succ *)
  solve [ eauto using consistent_12p3_23, @star_associative with j ].
Qed.

(* ------------------------------------------------------------------------- *)

(* This auxiliary lemma isolates the remark made in the above proof, i.e.
   the hypotheses of the above lemma imply [v = w]. This can be viewed as
   a special case of the lemma [row_context_read]. *)

Lemma DecomposedBlockEquality:
  forall f vs,
  row_context f vs ->
  forall R K v w Ts,
  row_context f Ts ->
  jpi R K (zip (fill_row vs v) (fill_row Ts (TySingleton w))) ->
  v = w.
Proof.
  induction 1; inversion 1; subst; simpl_subst_goal; simpl; introv hjpi;
  dependent destruction hjpi; subst;
  repeat match goal with h: jpi _ _ _ |- _ =>
    rewrite subst_lift in h
  end;
  repeat rewrite subst_lift.
  (* Zero *)
  decomposition. reflexivity.
  (* Succ *)
  eauto.
Qed.

Ltac decomposed_block_equality :=
  match goal with
  h1: row_context ?f ?vs,
  h2: row_context ?f ?Ts,
  h3: jpi _ _ (zip (fill_row ?vs ?v) (fill_row ?Ts _)) |- _ =>
    forwards: DecomposedBlockEquality h1 h2 h3; subst v
  end.

(* ------------------------------------------------------------------------- *)

(* This auxiliary lemma explains why the permission subsumption rule
   [JpsUnifyBlock] is sound. *)

Lemma JpsUnifyBlockSound:
  forall R1 R2 K v m1 m2 tag1 tag2 Ts1 Ts2 field v1 v2 U1 U2,
  row_context field Ts1 ->
  row_context field Ts2 ->
  jctop R1 K v (TyBlock m1 tag1 (fill_row Ts1 (TySingleton v1)) U1) ->
  jctop R2 K v (TyBlock m2 tag2 (fill_row Ts2 (TySingleton v2)) U2) ->
  consistent (star R1 R2) ->
  v1 = v2.
Proof.
  intros.
  classification.
  do 2 decomposition_TyBlock_no_adoptees; decomposition.
  do 2 decomposed_block_equality.
  match goal with h1: heap_maps _ _ _, h2: heap_maps _ _ _ |- _ =>
    forwards: heap_maps_mtagvs2dx_injective h1 h2
  end.
  eauto using consistent_1234_24.
  unpack.
  row_context_equality_inversion.
  reflexivity.
Qed.

(* ------------------------------------------------------------------------- *)

(* This auxiliary lemma, which is closely related to the above lemma, explains
   why the permission subsumption rule [JpsDirectedUnfold] is sound. *)

Lemma JpsDirectedUnfoldSound:
  forall R1 R2 K v m1 m2 tag1 tag2 Ts1 Ts2 U1 U2,
  jctop R1 K v (TyBlock m1 tag1 Ts1 U1) ->
  jctop R2 K v (TyBlock m2 tag2 Ts2 U2) ->
  consistent (star R1 R2) ->
  tag1 = tag2.
Proof.
  intros.
  classification.
  do 2 decomposition_TyBlock_no_adoptees; intros.
  match goal with h1: heap_maps _ _ _, h2: heap_maps _ _ _ |- _ =>
    forwards: heap_maps_mtagvs2dx_injective h2 h1
  end.
  eauto using consistent_1234_24.
  unpack.
  assumption.
Qed.

