Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.

(* ---------------------------------------------------------------------------- *)

(* A version of [JpsReflexive] that does not require immediate syntactic
   equality. *)

Lemma JpsReflexiveEq:
  forall n K P Q,
  P = Q ->
  jps n K P Q.
Proof.
  intros. subst. eapply JpsReflexive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The permission [x @ (=y | y @ T)] can be transformed to [x @ T]. *)

Lemma JpsSingletonElimination:
  forall K x y T,
  wk K T TYPE ->
  jps 1 K
    (TyAnchored x (TyMixStar (TySingleton y) (TyAnchored y T)))
    (TyAnchored x T).
Proof.
  intros.
  jps_left JpsMixStarElim.
  eapply JpsEqualsForEqualsReverseDrop with (P := (TyAnchored (TVar 0) (shift 0 T)));
    eauto using wk_weakening with wk;
    simpl_subst_goal; eauto with f_equal subst_lift.
Qed.

(* Hence, the type [=v | v @ T] is a subtype of [T]. *)

Lemma SubSingletonElimination:
  forall K v T,
  wk K T TYPE ->
  sub 2 K
    (TyMixStar (TySingleton v) (TyAnchored v T))
    T.
Proof.
  (* Unfold the definition of [sub] in terms of [jps]. *)
  econstructor.
  simpl_lift_goal.
  (* The result is then immediate. *)
  eauto using JpsSingletonElimination, wk_weakening.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Some of the rules that define [jps] can be re-formulated in terms of [sub].
   Here is an example. *)

Lemma SubCoArrow:
  forall m n K T1 T2 U1 U2,
  sub m K U1 T1 ->
  sub m K T2 U2 ->
  wk K U1 TYPE ->
  sub (S n) K
    (TyArrow T1 T2)
    (TyArrow U1 U2).
Proof.
  intros.
  eapply Sub.
  simpl_lift_goal.
  eapply JpsCoArrow.
    eapply j_weakening; eauto.
    eapply j_weakening; eauto.
    eauto using wk_weakening.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The type constructor [TyAnchored] is covariant. This is trivial: it is
   a direct consequence of the manner in which [sub] is defined in terms
   of [jps]. *)

Lemma JpsCoAnchored:
  forall n K T U,
  sub (S n) K T U ->
  forall v,
  wk K v VALUE ->
  jps n K
    (TyAnchored v T)
    (TyAnchored v U).
Proof.
  intros. sub_inversion.
  eapply j_substitution; eauto;
    simpl_subst_goal; eauto with f_equal subst_lift substitutable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The constructor [TyMixStar], which represents a conjunction of a type and a
   permission, is covariant. *)

Lemma SubCoMixStar:
  forall n K T U P Q,
  sub n K T U ->
  jps n K P Q ->
  sub (S (S (S (S n)))) K (TyMixStar T P) (TyMixStar U Q).
Proof.
  econstructor. simpl_lift_goal.
  (* The goal is to transform [x @ (T | P)] into [x @ (U | Q)]. *)
  jps_left JpsMixStarElim.
  jps_left JpsCoStar.
    sub_inversion. eapply jps_index_monotonic; eauto with omega.
    eapply j_weakening; eauto.
  eapply JpsMixStarIntro.
Qed.

(* ---------------------------------------------------------------------------- *)

(* [TyBottom] is the least type. *)

Lemma SubBottom:
  forall n K U,
  wk K U TYPE ->
  sub (S n) K TyBottom U.
Proof.
  econstructor. simpl_lift_goal.
  eauto 6 using wk_weakening with j wk.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Universal types can be eliminated. *)

Lemma SubForallElim:
  forall n K U kappa T,
  wk K U kappa ->
  sub (S n) K (TyForall kappa T) (subst U 0 T).
Proof.
  econstructor.
  simpl_lift_goal.
  econstructor.
  eauto with lift_subst.
  eauto using wk_weakening.
Qed.

(* Universal types are covariant. *)

Lemma SubCoForall:
  forall n K kappa T U,
  sub n (insert 0 kappa K) T U ->
  sub (S (S (S n))) K (TyForall kappa T) (TyForall kappa U).
Proof.
  intros. sub_inversion. econstructor. simpl_lift_goal. rewrite plus_0_r.
  jps_left JpsForallType2Perm; [ simpl_lift_goal; rewrite plus_0_r; reflexivity | ].
  jps_right JpsForallPerm2Type; [ | simpl_lift_goal; rewrite plus_0_r; reflexivity].
  eapply JpsCoForall.
  (* We are almost there, but we must exchange the indices 0 and 1. Tricky.
     I am doing this as an exercise, although probably making this subtyping
     axiom primitive would be much simpler than attempting to derive it in
     this way. Could the DeBruijn library be extended to facilitate this kind
     of exchange? *)
  (* Introduce a fresh name at index 2. *)
  forwards h: (get1 j_weakening); eauto.
  instantiate (1 := 1 + 1) in h.
  simpl_lift_all.
  (* Replace 0 with the new name at 2, i.e. technically, substitute TVar 1 for 0. *)
  forwards i: (get1 j_substitution).
    eauto.
    rewrite insert_insert. f_equal. omega.
    eauto.
    eauto.
    instantiate (1 := TVar 1). eapply WkTVar. eauto with lookup_insert.
    eauto with substitutable.
  simpl_subst_all.
  (* Now, check the goal. *)
  equates 1. equates 2. equates 3. eassumption.
  eauto with insert_insert.
  rewrite <- lift_lift by omega. rewrite subst_lift. reflexivity.
  rewrite <- lift_lift by omega. rewrite subst_lift. reflexivity.
Qed.

