Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Row.
Require Import Intro.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import DerivedSubtyping.
Require Import Harmless.

(* The following rules are meant to be used for type-checking source programs.
   They are not used in the type soundness proof. For this reason, instead of
   trying to allow an arbitrary resource [R], we specialized the rules and work
   with [void]. *)

(* ---------------------------------------------------------------------------- *)

(* A lemma that helps adjust the resource that appears in a typing judgement. *)

Lemma JSameResource:
  forall f n R1 R2 K P t T,
  j f n R1 K P t T ->
  R2 = R1 ->
  j f n R2 K P t T.
Proof.
  intros. subst. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A variant of [JFrame] that frames everything. *)

Lemma JFrameAll:
  forall n K Q v T,
  jt n void K TyEmpty v T ->
  jt (S (S n)) void K Q v (TyMixStar T Q).
Proof.
  eauto using (@JSubLeft 0), JpsEmptyAppearsLeft, JFrame.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A derived typing rule for values. *)

(* The native typing rule for values, [JSingleton], asserts that a value [v]
   has the singleton type [=v]. This judgement does not require any permission
   as part of its pre-condition. Here, we develop a variant of this rule,
   which states that, if we have the permission [v @ T], then the value [v]
   has type [T]. *)

Lemma JValue:
  forall K v T,
  wk K v VALUE ->
  wk K T TYPE ->
  jt 3 void K (TyAnchored v T) v T.
Proof.
  intros.
  (* We build the derivation from the bottom up. *)
  eapply JSubRight; [ | eapply SubSingletonElimination; eauto ].
  eapply JFrameAll.
  eapply JSingleton; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Derived typing rules for sequencing. *)

(* In this version, all of the permissions that are initially available,
   represented by [P], are sent into the left-hand side [t]. Thus, the
   only permission that is available to the right-hand side [u] is the
   permission [x @ T]. *)

(* K |- { P } t : T
   K; x: VALUE ||- { x @ T } u : U
   x # T, U
   -------------------------------------------
   K |- { P } let x = t in u : U            *)

Lemma JLet:
  forall m n K P T U t u,
  jt n void K P t T ->
  jt m void
    (insert 0 VALUE K)
    (TyAnchored (TVar 0) (shift 0 T))
    u
    (shift 0 U)
  ->
  wk K T TYPE ->
  jt (S (S n)) void K P (TLet t u) U.
Proof.
  intros.
  eapply JCut; [ | | eapply duplicable_void ].
  (* The application [TLet t u] is well-typed under
     the pre-condition [(TAbs u) @ (TyArrow T U)]. *)
  solve [ eapply JApplication; eauto with wk ].
  (* The permission [(TAbs u) @ (TyArrow T U)] is justified. *)
  eapply JpiAtomicEmpty; [ eauto with wk | ].
  eapply JFunction; [ eauto | eauto with dup | ].
  rewrite core_void.
  unfold intro.
  eauto using (@JSubLeft 0), JpsDisappearsLeft.
Qed.

(* In this version, we have built in an application of the frame rule. The
   permissions that are initially available are split as [P * Q]. The
   permission [P] is sent into the left-hand side [t], whereas [Q] is not made
   available to [t] and remains available for consumption by [u]. *)

(* K |- { P } t : T
   K; x: VALUE ||- { Q * x @ T } u : U
   x # Q, T, U
   -----------------------------------------------
   K |- { P * Q } let x = t in u : U            *)

Lemma JLetFrame:
  forall m n K P Q T U t u,
  jt n void K P t T ->
  jt m void
    (insert 0 VALUE K)
    (TyStar (shift 0 Q) (TyAnchored (TVar 0) (shift 0 T)))
    u
    (shift 0 U) 
  ->
  wk K T TYPE ->
  wk K Q PERM ->
  jt (S (S (S n))) void K (TyStar P Q) (TLet t u) U.
Proof.
  intros.
  eapply JCut; [ | | eapply duplicable_void ].
  (* [t] has type [T] under the pre-condition [P]. We frame [Q] around
     this judgement, so [t] has type [T | Q] under [P * Q]. *)
  eapply JApplication; [ eauto with wk | ].
  solve [ eapply JFrame; eauto ].
  (* The permission [(TAbs u) @ (TyArrow (T | Q) U)] is justified by [core R]. *)
  eapply JpiAtomicEmpty; [ eauto with wk | ].
  eapply JFunction; [ eauto with wk | eauto with dup | simpl_lift_goal ].
  rewrite core_void.
  eapply (@JSubLeft 2); [ eassumption | unfold intro; simpl_lift_goal ].
  jps_left JpsDisappearsLeft.
  jps_left JpsMixStarElim.
  eauto with j.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A derived typing rule for applications. *)

(* The rule states that if the function [t] has type [U -> T] under the
   precondition [P] and if the argument [u] has type [U] under the
   precondition [Q] then the application has type [T] under the combined
   precondition [P * Q]. *)

(* The application is not written under the form [t u]. Indeed, this form
   is not considered well-kinded unless [t] is a value. Furthermore, even
   if we were to require that [t] be a value, that would not be sufficient:
   we would have to also require a canonical derivation for [t] under an
   empty precondition, a very strong requirement. Thus, the present
   formulation seems more satisfactory. It reflects the fact that the
   type system requires A-normal form. *)

Lemma JNormalApp:
  forall m n K P Q t u T U,
  jt n void K P t (TyArrow U T) ->
  jt m void K Q u U ->
  wk K P PERM ->
  wk K Q PERM ->
  jt (S (S (S n))) void K (TyStar P Q) (TLet t (TApp (TVar 0) (shift 0 u))) T.
Proof.
  intros.
  eapply JLetFrame.
    eauto.
    eapply (@JSubLeft 0).
      eapply JApplication.
        eauto with wk.
        eapply j_weakening; eauto.
      simpl_lift_goal. eapply JpsStarCommutative.
    eauto with wk.
    eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A variant of [JForallIntro]. Instead of allowing an arbitrary [P] in the
   premise and introducing [forall x.P] in the conclusion, we require that
   [P] do not mention [x], and we use [P] in the premise and conclusion. *)

Lemma JForallIntroVariant:
  forall n R K P t kappa T,
  wk K t PROG ->
  harmless t ->
  jt n R (insert 0 kappa K) (shift 0 P) (shift 0 t) T ->
  substitutable kappa ->
  jt (S (S n)) R K P t (TyForall kappa T).
Proof.
  intros.
  (* Apply JForallIntro. *)
  forwards: JForallIntro; eauto.
  (* We now have a vacuous quantification [forall x.P], where [x]
     does not occur in [P]. Use permission subsumption to argue
     that this is equivalent to [P]. *)
  eauto using (@JSubLeft 1), JpsVacuousForallIntro.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The term [TUnit] has type [TyUnit] in any context. *)

Lemma JUnitMix:
  forall n R K P,
  jt (S (S n)) R K P TUnit (TyMixStar TyUnit P).
Proof.
  intros.
  eapply JCut.
  eapply JFrame; eapply JNew; simpl; eauto with wk row_length.
  eauto with j.
  eauto using @core_right_unit.
Qed.

Lemma JUnit:
  forall n R K,
  jt n R K TyEmpty TUnit TyUnit.
Proof.
  intros.
  eapply JNew; simpl; eauto with wk row_length.
Qed.