Set Implicit Arguments.
Require Import List.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Signature.

(* ---------------------------------------------------------------------------- *)

(* A subset of types and permissions are duplicable. *)

(* For the moment, this judgement is not parameterized with an environment.
   Hence, a permission variable, or a type variable, cannot be considered
   duplicable. This shortcoming will be addressed in the future. *)

(* One must not confuse the syntactic judgement that is defined here, which
   concerns permissions and types, with the semantic notion of [duplicable]
   that is defined in [MSA], which concerns resources. *)

(* The rule [DupTyCon] states that the algebraic data type [TyCon tc] is
   duplicable if [tc] is declared as duplicable *and* every branch in the
   definition of [tc] is composed of duplicable fields. This rule is the
   reason why the predicate [dup] must be co-inductively defined. An
   inductive definition of [dup] would be sound but would prevent a
   recursive type, such as [list], from being considered duplicable. *)

CoInductive dup : term -> Prop :=
| DupBottom:
    dup TyBottom
| DupForall:
    forall kappa P,
    dup P ->
    dup (TyForall kappa P)
| DupExists:
    forall P,
    dup P ->
    dup (TyExists P)
| DupSingleton:
    forall v,
    dup (TySingleton v)
| DupDynamic:
    dup TyDynamic
| DupLck:
    forall P,
    dup (TyLck P)
| DupArrow:
    forall T U,
    dup (TyArrow T U)
| DupMixStar:
    forall T P,
    dup T ->
    dup P ->
    dup (TyMixStar T P)
| DupBlock:
    forall tag Ts U,
    dup Ts ->
    dup (TyBlock Duplicable tag Ts U)
| DupTyCon:
    forall tc,
    tc_mode tc = Duplicable ->
    Forall dup (tc_branches tc) ->
    dup (TyCon tc)
| DupAnchored:
    forall v T,
    dup T ->
    dup (TyAnchored v T)
| DupEmpty:
    dup TyEmpty
| DupStar:
    forall P Q,
    dup P ->
    dup Q ->
    dup (TyStar P Q)
| DupFalse:
    dup TyFalse
| DupNil:
    dup RowNil
| DupCons:
    forall T Ts,
    dup T ->
    dup Ts ->
    dup (RowCons T Ts)
.

Hint Constructors dup : dup.

Ltac select_dup action :=
  match goal with h: dup _ |- _ => action h end.
Ltac dup_inversion :=
  gji select_dup.

(* ---------------------------------------------------------------------------- *)

(* The judgement [dup] is stable by lifting. *)

Lemma dup_weakening:
  forall T,
  dup T ->
  forall x,
  dup (shift x T).
Proof.
  cofix. inversion 1; intros; simpl_lift_goal; eauto with dup.
Qed.

Hint Resolve dup_weakening : j_weakening.

(* ---------------------------------------------------------------------------- *)

(* The judgement [dup] is stable by substitution. *)

Lemma dup_substitution:
  forall T,
  dup T ->
  forall w x,
  dup (subst w x T).
Proof.
  cofix. inversion 1; intros; simpl_subst_goal; eauto with dup.
Qed.

Hint Resolve dup_substitution : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* The judgement [pcore P Q] means that the permission [Q] is *some* duplicable
   fragment of the permission [P]. *)

(* For the moment, it does not play an important role. It only serves in the
   statement of the auxiliary lemma [j_duplicable_preliminary]. *)

Inductive pcore : term -> term -> Prop :=

(* If [P] is a conjunction, then each of the conjuncts can be treated
   independently. *)

| PCoreStar:
    forall P1 P2 Q1 Q2,
    pcore P1 Q1 ->
    pcore P2 Q2 ->
    pcore (TyStar P1 P2) (TyStar Q1 Q2)

(* One might think that we need the following rule: if [P] is an existential
   permission, then we take a duplicable fragment of its body, and preserve
   the existential quantifier. Actually, we are able to work without it. The
   need for it does not arise in the proof of the lemma
   [j_duplicable_preliminary]. *)

(* We do need such a rule for universal quantifiers. *)

| PCoreForall:
    forall kappa P Q,
    pcore P Q ->
    pcore (TyForall kappa P) (TyForall kappa Q)

(* Any permission [P] can be discarded, regardless of whether it is or is not
   duplicable. This conservative rule makes the relation [pcore]
   non-deterministic. It is required because a variable of kind [PERM] that is
   not known to be duplicable may later be instantiated with a duplicable
   permission. *)

| PCoreEmpty:
    forall P,
    pcore P TyEmpty

(* If [P] is known to be duplicable, then it can be kept. *)

| PCoreDuplicable:
    forall P,
    dup P ->
    pcore P P

.

Hint Constructors pcore : pcore.

(* The tactic [pcore_inversion] destructs a [pcore _ _] hypothesis,
   even if this introduces multiple sub-cases. *)

Ltac pcore_inversion :=
  match goal with
  | h: pcore _ _ |- _ =>
      inversion h; subst; clear h
  end.

(* ---------------------------------------------------------------------------- *)

(* The relation [pcore] is preserved by substitution. *)

Lemma pcore_subst:
  forall P Q,
  pcore P Q ->
  forall v x,
  pcore (subst v x P) (subst v x Q).
Proof.
  induction 1; intros; simpl_subst_goal; econstructor; eauto using dup_substitution.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A duplicable fragment of the empty permission must be the empty permission. *)

Lemma pcore_empty_inversion:
  forall Q,
  pcore TyEmpty Q ->
  Q = TyEmpty.
Proof.
  inversion 1; subst; eauto.
Qed.

(* A duplicable fragment of a conjunction [P1 * P2] must be either empty or a
   conjunction of two fragments of [P1] and [P2]. *)

Lemma pcore_star_inversion:
  forall P1 P2 Q,
  pcore (TyStar P1 P2) Q ->
  Q = TyEmpty \/
  exists Q1 Q2, Q = TyStar Q1 Q2 /\ pcore P1 Q1 /\ pcore P2 Q2.
Proof.
  inversion 1; subst; dup_inversion; eauto 8 with pcore.
Qed.

