Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Reduction.
Require Import Signature.
Require Import Exclusive.
Require Import Adopts.
Require Import Typing.
Require Import Top.
Require Import JDecomposition.

(* ---------------------------------------------------------------------------- *)

(* If the memory location [x] has type [TyBlock Exclusive _ _ _] in [R], then
   [R] owns [x], and [x.adopter] must be [null]. This serves as a preliminary
   statement for the next lemma, which is more general. *)

Lemma exclusive_block_type_implies_maps_exclusive:
  forall R K x tag Ts U,
  jctop R K (TLoc x) (TyBlock Exclusive tag Ts U) ->
  consistent R ->
  exists tag vs,
  heap_maps R x (DX_EXCLUSIVE (Block tag Null vs)).
Proof.
  intros. dca. do 2 eexists. SCAM_maps_maximal heap_project_morphism.
Qed.

(* If the memory location [x] has exclusive type in [R], then [R] owns [x],
   and [x.adopter] must be [null]. *)

Lemma exclusive_type_implies_maps_exclusive:
  forall R K x T,
  jctop R K (TLoc x) T ->
  exclusive T ->
  consistent R ->
  exists tag vs,
  heap_maps R x (DX_EXCLUSIVE (Block tag Null vs)).
Proof.
  introv ? hex ?.
  inversion hex; subst; decomposition_TyCon;
  repeat match goal with h: _ = Exclusive |- _ => rewrite h in *; clear h end;
  eauto using exclusive_block_type_implies_maps_exclusive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If the memory locations [x] and [y] have exclusive type, then they must be
   distinct locations. *)

Lemma exclusive_implies_distinct_locations:
  forall R1 R2 K x y T U,
  jctop R2 K (TLoc x) T ->
  jctop R1 K (TLoc y) U ->
  exclusive T ->
  exclusive U ->
  consistent (star R1 R2) ->
  x <> y.
Proof.
  introv hx hy hT hU hc.
  forwards: (exclusive_type_implies_maps_exclusive hx hT);
    [ eauto with consistent_split | unpack ].
  forwards: (exclusive_type_implies_maps_exclusive hy hU);
    [ eauto with consistent_split | unpack ].
  eapply exclusive_exclusive_star; eauto using @consistent_star_commutative.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [x] has an exclusive block type and claims to adopt objects of type [U],
   and if [y] separately has type [U], then, after [x] adopts [y], the block
   [x] remains well-typed, with the same type. This is the key reason why
   [give] is sound. *)

(* The premise [CAM_maps R1 y (DX_EXCLUSIVE _)] may seem redundant, because
   the hypotheses [jctop R1 K (TLoc y) U] and [exclusive U] imply that [R1]
   must map [y] to *some* exclusive block. However, we use this premise to
   get access to the tag and fields of the object at [y], [tagy] and [vsy],
   and we use them in the conclusion. *)

(* TEMPORARY as an experiment, try defining a version of [adopt] that does
   not require the arguments [tag] and [vs]? *)

Lemma give_TyBlock:
  forall R1 R2 K x y tagx Ts tagy vsy U req,
  jctop R2 K (TLoc x) (TyBlock Exclusive tagx Ts U) ->
  jctop R1 K (TLoc y) U ->
  consistent (star R1 R2) ->
  heap_maps R1 y (DX_EXCLUSIVE (Block tagy Null vsy)) ->
  jctop
    (adopt (star R1 R2) y x tagy vsy req)
    K
    (TLoc x)
    (TyBlock Exclusive tagx Ts U).
Proof.
  (* The following is a hint for [CAM_maps_maximal]. *)
  generalize consistent_1p23_13. intro.
  (* Decompose our hypotheses. Only the judgement about [x] is
     decomposed, because it exhibits a [TyBlock] type. *)
  introv hx hy hc hm.
  revert req. (* avoid unpacking [req] *)
  dca.
  (* We have a combination of three resources, where [R1] concerns the
     location [y], [R2a] concerns the fields of [x], and [R3] concerns the
     location [x] per se and its current adoptees. We will need to justify
     that [R2a] is not concerned by [adopt], or in other words, we will need
     to hoist it out of [adopt]. Here is a proof that this makes sense. *)
  match type of hc with consistent (star ?R1 (star ?R2a ?R2b)) =>
    assert (obligation : x <> y /\ exists p, heap_maps (star R1 R2b) x (DX_EXCLUSIVE p)); [
      unpack; split; [ eauto | eexists; SCAM_maps_maximal heap_project_morphism ]
    | ]
  end.
  (* Now, construct a new judgement about the memory location [x]. *)
  eapply JcTopLoc.
    (* Sub-goal 1. *)
    (* Reconfigure in preparation for applying [star_adopt_right]. *)
    reconfigure06_under_adopt.
    rewrite reconfigure06 in hc.
    (* We can now apply [star_adopt_right]. *)
    eapply star_adopt_right with (req1 := obligation); [ eauto | SCAM_maps_maximal heap_project_morphism ].
    (* Sub-goal 2. *)
    eassumption.
    (* Sub-goal 3. *)
    eauto.
    (* Sub-goal 4. *)
    (* [x] now has more one adoptee, namely [y]. *)
    eapply AbandonAllSucc.
      eapply abandon_adopt; SCAM_maps_maximal heap_project_morphism.
      eauto with star_abandon_all.
    (* Sub-goal 5. *)
    eapply star_associative_symmetric.
    (* Sub-goal 6. *)
    econstructor; eauto.
    (* Sub-goal 7. *)
    eauto.
    (* Sub-goal 8. *)
    eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The previous lemma is generalized to the case where [x] has an arbitrary
   exclusive type [T] and the types [T] and [U] are in the relation [adopts]. *)

Lemma give_general:
  forall R1 R2 K x y T U,
  adopts T U ->
  jctop R2 K (TLoc x) T ->
  jctop R1 K (TLoc y) U ->
  consistent (star R1 R2) ->
  forall tagy vsy,
  heap_maps R1 y (DX_EXCLUSIVE (Block tagy Null vsy)) ->
  forall req,
  jctop
    (adopt (star R1 R2) y x tagy vsy req)
    K
    (TLoc x)
    T.
Proof.
  introv ha hx hy hc hm. intros. dependent destruction ha; subst.
  (* Case: [T] is [TyCon tc]. *)
  forwards: decomposition_TyCon; [ eauto | unpack ].
  (* Do not clear [hx]; it will be used again. *)
  match goal with h: _ = Exclusive |- _ => rewrite h in * end.
  eapply JcTopFold; [ | eauto | eauto | eauto ].
  eapply give_TyBlock; eauto.
  (* Case: [T] is [TyBlock _ _ _ _]. *)
  eapply give_TyBlock; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The above version of the lemma requires the user to construct the proof
   obligation [req]. We now show that this proof obligation in fact follows
   from the first four hypotheses of the lemma. *)

Lemma give_req:
  forall R1 R2 K x y T U,
  adopts T U ->
  jctop R2 K (TLoc x) T ->
  jctop R1 K (TLoc y) U ->
  consistent (star R1 R2) ->
  x <> y /\ exists p, heap_maps (star R1 R2) x (DX_EXCLUSIVE p).
Proof.
  introv ? hx ? ?. split.
  (* Sub-goal 1. *)
  (* [x] and [y] must be distinct because they have exclusive types. *)
  eauto using exclusive_implies_distinct_locations, adopters_are_exclusive, adoptees_are_exclusive.
  (* Sub-goal 2. *)
  (* [R1] owns [x] because [x] has an exclusive type. Thus, [star R1 R2]
     owns [x] as well. *)
  forwards: (exclusive_type_implies_maps_exclusive hx).
    eauto using adopters_are_exclusive.
    eauto with consistent_split.
  unpack.
  eexists.
  SCAM_maps_maximal heap_project_morphism.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This allows us to offer a version of the lemma where the proof obligation
   is fulfilled by us (as opposed to by the user of the lemma). *)

Lemma give:
  forall R1 R2 K x y T U,
  forall ha : adopts T U,
  forall hx : jctop R2 K (TLoc x) T,
  forall hy : jctop R1 K (TLoc y) U,
  forall hc : consistent (star R1 R2),
  forall tagy vsy,
  heap_maps R1 y (DX_EXCLUSIVE (Block tagy Null vsy)) ->
  jctop
    (adopt (star R1 R2) y x tagy vsy (give_req ha hx hy hc))
    K
    (TLoc x)
    T.
Proof.
  eauto using give_general.
Qed.

