Set Implicit Arguments.
Require Import MyTactics.
Require Import Syntax.
Require Import Reduction.
Require Import DeBruijn.
Require Import WkInversion.
Require Import Row.
Require Import Environments.

(* ------------------------------------------------------------------------- *)

(* The simplest form of the value restriction states that [JForallIntro] is
   restricted to values. We wish to be slightly more permissive. In
   particular, we wish to allow a memory allocation expression to receive a
   polymorphic type, otherwise we would not be able to construct a polymorphic
   empty list. Thus, we define a subcategory of harmless terms. This
   subcategory must be stable by substitution, preserved by reduction, and
   must have the property that a harmless term cannot allocate hidden state
   (i.e., cannot create a new lock). *)

Inductive harmless: term -> Prop :=

  | HlTVar:
      forall x,
      harmless (TVar x)

  | HlTAbs:
      forall t,
      harmless (TAbs t)

  (* A function application is in principle never harmless. As a special
     case, [TLet u t], which is sugar for [TApp (TAbs t) u], can be
     considered harmless if [u] and [t] are harmless. *)

  | HlTLet:
      forall u t,
      harmless u ->
      harmless t ->
      harmless (TLet u t)

  | HlTLoc:
      forall l,
      harmless (TLoc l)

  (* Memory allocation is harmless, even if this is mutable memory. *)

  | HlTNew:
      forall tag vs,
      harmless (TNew tag vs)

  | HlTRead:
      forall v field,
      harmless (TRead v field)

  | HlTWrite:
      forall v1 field v2,
      harmless (TWrite v1 field v2)

  | HlTWriteTag:
      forall v tag,
      harmless (TWriteTag v tag)

  | HlTSwitch:
      forall v ts,
      harmless ts ->
      harmless (TSwitch v ts)

  | HlTGive:
      forall v1 v2,
      harmless (TGive v1 v2)

  | HlTTake:
      forall v1 v2,
      harmless (TTake v1 v2)

  | HlTFail:
      harmless TFail

  (* [TSpawn] is considered harmful. Perhaps we could consider it harmless,
     but that would force us to add a universal introduction rule at the
     level of soups. It is probably not worth the trouble. *)

  | HlTLck:
      forall l,
      harmless (TLck l)

  (* [TNewLock] is considered harmful. This is essential, and is the reason
     for the existence of the [harmless] predicate in the first place. If
     it was considered harmless, then the invariant of the newly-allocated
     lock could mention a type variable introduced by [JForallIntro], and
     this would break subject reduction. *)

  | HlTAcquire:
      forall v,
      harmless (TAcquire v)

  | HlTRelease:
      forall v,
      harmless (TRelease v)

  | HlRowNil:
      harmless RowNil

  | HlRowCons:
      forall t ts,
      harmless t ->
      harmless ts ->
      harmless (RowCons t ts)

.

Hint Constructors harmless: harmless.

(* ------------------------------------------------------------------------- *)

(* Values are harmless. *)

Lemma harmless_value:
  forall K t,
  wk K t VALUE ->
  harmless t.
Proof.
  dependent_induction; eauto with harmless.
Qed.

(* ------------------------------------------------------------------------- *)

(* Harmlessness is preserved by weakening. *)

Lemma harmless_lift:
  forall t,
  harmless t ->
  forall k l,
  harmless (lift k l t).
Proof.
  induction 1; intros; simpl_lift_all; auto with harmless.
Qed.

(* Harmlessness is preserved by inverse weakening. *)

Lemma harmless_lift_inverse:
  forall t',
  harmless t' ->
  forall k l t,
  lift k l t = t' ->
  harmless t.
Proof.
  induction 1; intros; subst;
  repeat match goal with h: lift _ _ ?t = _ |- _ =>
    destruct t;
    try solve [ discriminate | eauto with harmless ];
    simpl_lift_in h;
    injections;
    try solve [ eauto with harmless ]
  end.
Qed.

(* ------------------------------------------------------------------------- *)

(* Harmlessness is preserved by substitution of a harmless term for
   a variable. *)

Lemma harmless_subst_harmless:
  forall t,
  harmless t ->
  forall u k,
  harmless u ->
  harmless (subst u k t).
Proof.
  induction 1; intros; simpl_subst_goal;
  eauto using harmless_lift with harmless.
  (* TVar *)
  unfold subst_idx. dblib_by_cases; eauto with harmless.
Qed.

(* Harmlessness is also preserved by substitution at a logical kind.
   Indeed, such a substitution, applied to a term, is the identity. *)

(* The two situations where harmlessness is preserved by substitution
   are summarized as follows. *)

Lemma harmless_subst:
  forall K' t2 kappa2,
  harmless t2 ->
  wk K' t2 kappa2 ->
  ~ logic kappa2 ->
  forall t1 x K kappa1,
  insert x kappa1 K = K' ->
  wk K t1 kappa1 ->
  substitutable kappa1 ->
  harmless (subst t1 x t2).
Proof.
  (* By unfolding [substitutable], we get two cases. *)
  introv ? ? ? ? [ ? | ? ].
  (* In this case, the result of [subst t1 x t2] is just [t2], up to
     an inverse shift, so the goal follows from the hypothesis. *)
  eauto using harmless_lift_inverse, wk_substitution_logic.
  (* In this case, [kappa1] is [VALUE], so we [t1] is harmless and
     we use the lemma [harmless_subst_harmless]. *)
  subst. eauto using harmless_subst_harmless, harmless_value.
Qed.

(* ------------------------------------------------------------------------- *)

(* [TSpawn] is harmful. *)

Lemma harmful_spawn:
  forall v1 v2,
  harmless (TSpawn v1 v2) ->
  False.
Proof.
  introv hl; inversion hl.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a row of terms is harmless, then every term in this row is harmless. *)

Lemma harmless_fill_row_inverse:
  forall f ts,
  row_context f ts ->
  forall t,
  harmless (fill_row ts t) ->
  harmless t.
Proof.
  induction 1; introv h; simpl_subst_all; dependent destruction h; auto.
Qed.

(* ------------------------------------------------------------------------- *)

(* Harmlessness is preserved by reduction. *)

Lemma harmless_red:
  forall t1 t2 h1 h2,
  red (Cf h1 t1) (Cf h2 t2) ->
  wkh h1 ->
  harmless t1 ->
  harmless t2.
Proof.
  introv hred ? hl. dependent induction hred;
  try solve [
    try open_eval_context;
    dependent destruction hl;
    eauto using harmless_subst_harmless, harmless_fill_row_inverse with harmless
  ].
  (* RedRead. *)
  wk_inversion. eauto using harmless_value.
Qed.
