Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Dup.
Require Import Exclusive.
Require Import Adopts.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JctopInduction.

(* Classification. *)

(* When we have a canonical type derivation about a value, which exhibits a
   known type constructor, what can we say of the value? *)

(* ------------------------------------------------------------------------- *)

(* Bottom. *)

(* A value cannot have type bottom. *)

Lemma classification_TyBottom:
  forall R K v,
  jctop R K v TyBottom ->
  False.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* Functions. *)

(* If [v] has a function type, then [v] must be a function. *)

Lemma classification_TyArrow:
  forall R K v T U,
  jctop R K v (TyArrow T U) ->
  exists t,
  v = TAbs t.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has a memory block type, then [v] must be a memory location. *)

Lemma classification_TyBlock:
  forall R K v m tag Ts U,
  jctop R K v (TyBlock m tag Ts U) ->
  exists l,
  v = TLoc l.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has an algebraic data type, then [v] must be a memory location. *)

Lemma classification_TyCon:
  forall R K v tc,
  jctop R K v (TyCon tc) ->
  exists l,
  v = TLoc l.
Proof.
  jctop_induction; eauto using classification_TyBlock with j.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has type [dynamic], then [v] must be a memory location. *)

Lemma classification_TyDynamic:
  forall R K v,
  jctop R K v TyDynamic ->
  exists l,
  v = TLoc l.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has a lock type, then [v] must be a lock memory location. *)

Lemma classification_TyLck:
  forall R K v P,
  jctop R K v (TyLck P) ->
  exists l,
  v = TLck l.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has an exclusive type, then [v] must be a memory location. *)

Lemma classification_exclusive:
  forall R K v T,
  jctop R K v T ->
  exclusive T ->
  exists l,
  v = TLoc l.
Proof.
  inversion 2; subst; eauto using classification_TyBlock, classification_TyCon.
Qed.

Lemma classification_adopter:
  forall R K v T U,
  jctop R K v T ->
  adopts T U ->
  exists l,
  v = TLoc l.
Proof.
  eauto using adopters_are_exclusive, classification_exclusive.
Qed.

Lemma classification_adoptee:
  forall R K v T U,
  jctop R K v U ->
  adopts T U ->
  exists l,
  v = TLoc l.
Proof.
  eauto using adoptees_are_exclusive, classification_exclusive.
Qed.

(* ------------------------------------------------------------------------- *)

(* The following tactic applies one of the above classification lemmas. *)

Ltac classification :=
  repeat match goal with
  | h: jctop _ _ _ TyBottom |- _ =>
      generalize (classification_TyBottom h); intro;
      tauto;
      revert h
  | h: jctop _ _ _ (TyArrow _ _) |- _ =>
      generalize (classification_TyArrow h); intro;
      unpack; subst; wk_inversion;
      revert h
  | h: jctop _ _ _ (TyBlock _ _ _ _) |- _ =>
      generalize (classification_TyBlock h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ (TyCon _) |- _ =>
      generalize (classification_TyCon h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ TyDynamic |- _ =>
      generalize (classification_TyDynamic h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ (TyLck _) |- _ =>
      generalize (classification_TyLck h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ ?T, he: exclusive ?T |- _ =>
      generalize (classification_exclusive h he); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ ?T, ha: adopts ?T _ |- _ =>
      generalize (classification_adopter h ha); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ ?U, ha: adopts _ ?U |- _ =>
      generalize (classification_adoptee h ha); intro;
      unpack; subst;
      revert h
  end;
  intros.

