Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Signature.
Require Import Dup.
Require Import Exclusive.
Require Import Adopts.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JWeakening.
Require Import JDuplication.
Require Import JctopInduction.

(* Decomposing a canonical typing judgement [jctop]. *)

(* ------------------------------------------------------------------------- *)

(* If [v1] has type [=v2], then [v1] and [v2] are equal. *)

Lemma decomposition_TySingleton:
  forall R K v1 v2,
  jctop R K v1 (TySingleton v2) ->
  v1 = v2.
Proof.
  jctop_induction; eauto 3 with j.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a function [\x.t] has type [T -> U] under the resource [R], then the
   function body [t] has type [U] under the resource [core R] and under the
   precondition [x @ T]. *)

Lemma decomposition_TyArrow:
  forall R K t T U,
  jctop R K (TAbs t) (TyArrow T U) ->
  consistent R ->
  exists n,
  jt n
    (core R)
    (insert 0 VALUE K)
    (TyAnchored (TVar 0) (shift 0 T))
    t
    (shift 0 U).
Proof.
  jctop_induction.
  (* JCut *)
  (* Just apply the induction hypothesis. *)
  forwards: IHh; eauto using @consistent_by_equality with j wk aac.
  match goal with h: exists n, jt n ?R1 _ _ _ _ |- exists n, jt n ?R2 _ _ _ _ =>
    assert (eq : R1 = R2); [ aac_reflexivity | rewrite <- eq ]
  end.
  assumption.
  (* JFunction *)
  clear IHh. eexists.
  (* We use [JCut] in order to hide the fact that the body of
     the function relies on the permission [P]. This suggests that,
     instead of defining [JFunction] to allow a duplicable permission [P]
     to be captured by the function, perhaps we could have adopted a
     stricter version of [JFunction] that does not allow this, and recovered
     the lenient version by combining [JFunction] and [JCut]. In the
     end, it does not make much difference, I suppose. The present
     lemma effectively gives us inversion for the strict variant. *)
  eapply JCut.
    eassumption.
    eapply j_weakening; [ | eauto | eauto ].
      eapply j_duplicable; eauto with consistent_split.
    rewrite star_commutative. eauto using @star_core.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has type [T | Q] under [R], then under the same resource,
   separately, [v] has type [T] and the permission [Q] is justified. *)

Lemma decomposition_TyMixStar:
  forall R K v T Q,
  jctop R K v (TyMixStar T Q) ->
  exists R1 R2,
  jctop R1 K v T /\
  jpi R2 K Q /\
  R = star R1 R2.
Proof.
  jctop_induction; jpi_inversion.
  (* JFrame *)
  solve [ xplit; eauto with j aac ].
  (* JCut *)
  forwards: IHh; eauto with j. unpack.
  xplit; eauto using congruence00.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a memory location [l] has type [TyBlock m tag Ts U] under [R], then,
   well, you see what I mean. *)

Lemma decomposition_TyBlock:
  forall R K l m tag Ts U,
  jctop R K (TLoc l) (TyBlock m tag Ts U) ->
  consistent R ->
  exists R1 R2 R2' R2'a R2'b vs ys,
  star R1 R2 = R /\
  wk zilch vs (ROW VALUE) /\
  jpi R1 K (zip vs Ts) /\
  abandon_all R2 l ys R2' /\
  star R2'a R2'b = R2' /\
  jpi R2'a K (zap ys U) /\
  wk K U TYPE /\
  heap_maps R2'b l (mtagvs2dx m tag vs).
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto using @consistent_by_equality with j wk aac. unpack. subst.
  solve [ eauto 16 using congruence01 ]. (* waoh! *)
  (* JBlock *)
  (* Hmm. We are repeating the argument in [JDrop.v]. Inelegant. *)
  xplit;
    eauto using @star_associative,
                (SCAM_maps_maximal_left heap_project_morphism), maximal_mo2dx,
                consistent_12p3_23
           with star_abandon_all.
Qed.

(* In some cases, we do not care at all about the adoptees. Then, we can
   use a simplified decomposition lemma. *)

Lemma decomposition_TyBlock_no_adoptees:
  forall R K l m tag Ts U,
  jctop R K (TLoc l) (TyBlock m tag Ts U) ->
  consistent R ->
  exists R1 R2 vs,
  star R1 R2 = R /\
  wk zilch vs (ROW VALUE) /\
  jpi R1 K (zip vs Ts) /\
  heap_maps R2 l (mtagvs2dx m tag vs).
Proof.
  intros.
  forwards: decomposition_TyBlock; eauto. unpack. subst.
  xplit; eauto.
  consistent_abandon_all.
  abandon_all_maps.
  SCAM_maps_maximal heap_project_morphism.
Qed.

Ltac decomposition_TyBlock_no_adoptees :=
  match goal with
  h: jctop ?R _ (TLoc _) (TyBlock _ _ _ _) |- _ =>
    forwards: decomposition_TyBlock_no_adoptees h; [
      eauto with consistent_split
    | unpack; subst R;
      revert h (* move [h] out of the way *)
    ]
  end.

(* ------------------------------------------------------------------------- *)

(* If a memory location [l] has type [TyCon tc], then for some [tag] and [Ts]
   associated with [tc], this memory location has type [TyBlock (tc_mode tc)
   tag Ts (tc_adopts tc)]. *)

Lemma decomposition_TyCon:
  forall R K l tc,
  jctop R K (TLoc l) (TyCon tc) ->
  exists tag Ts,
  tc_tag tc tag Ts /\
  jctop R K (TLoc l) (TyBlock (tc_mode tc) tag Ts (tc_adopts tc)).
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto with j wk. unpack. subst.
  solve [ xplit; eauto using JcTopSameResource with aac ].
  (* JFold *)
  xplit; eauto with j wk.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a memory location [l] has type [dynamic], then [l] must be a valid (i.e.,
   allocated) memory location. *)

Lemma decomposition_TyDynamic:
  forall R K l,
  jctop R K (TLoc l) TyDynamic ->
  consistent R ->
  l < heap_limit R.
Proof.
  jctop_induction.
  (* JCut *)
  forwards: IHh; eauto using @consistent_by_equality with j wk aac. rewrite reconfigure07. eassumption.
  (* JDynamic *)
  rewrite (SCAM_limit_split_left heap_project_morphism); eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a memory location [l] has type [TyLck P], then the lock resource maps [l]
   to the permission [P]. *)

Lemma decomposition_TyLck:
  forall R K l P,
  jctop R K (TLck l) (TyLck P) ->
  consistent R ->
  exists lp,
  lck_maps R l lp /\ is_locked_permission lp P.
Proof.
  jctop_induction.
  (* JCut. *)
  forwards: IHh; eauto using @consistent_by_equality with j wk aac. rewrite reconfigure07. eassumption.
  (*  *)
  eauto using lck_resource_maps_star.
Qed. 

(* If a memory location [l] has type [TyLocked], then the lock resource maps
   [l] to the permission [P] with the [HELD] flag. *)

Lemma decomposition_TyLocked:
  forall R K l,
  jctop R K (TLck l) (TyLocked) ->
  consistent R ->
  exists P,
  lck_maps R l (p2hl P).
Proof.
  jctop_induction.
  (* JCut. *)
  forwards: IHh; eauto using @consistent_by_equality with j wk aac. rewrite reconfigure07. eassumption.
  (*  *)
  forwards: lck_resource_maps_star. eauto. right; eauto. eauto. eauto. unpack.
  forwards: (SCAM_maps_star_inversion lck_project_morphism H1).
  unpack.
  SCAM_maps_injective.
  exists P.
  assert (star (p2hl P) x1 = p2hl P).
    inversion H3; eauto.
    false.
    forwards: (split_lock H0).
    unpack.
    discriminate.
  rewrite H0 in *. auto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a memory location admits an exclusive type [T], then it admits a type of
   the form [TyBlock Exclusive tag Ts U]. *)

Lemma decomposition_exclusive:
  forall R K l T,
  jctop R K (TLoc l) T ->
  exclusive T ->
  exists m tag Ts U,
  jctop R K (TLoc l) (TyBlock m tag Ts U) /\
  m = Exclusive.
Proof.
  inversion 2; intros; subst.
  forwards: decomposition_TyCon; [ eauto | unpack ]. xplit; eauto.
  xplit; eauto.
Qed.

Lemma decomposition_adopter:
  forall R K l T U,
  jctop R K (TLoc l) T ->
  adopts T U ->
  exists m tag Ts,
  jctop R K (TLoc l) (TyBlock m tag Ts U) /\ (* note: same type [U] *)
  m = Exclusive.
Proof.
  inversion 2; intros; subst.
  forwards: decomposition_TyCon; [ eauto | unpack ]. xplit; eauto.
  xplit; eauto.
Qed.

Lemma decomposition_adoptee:
  forall R K l T U,
  jctop R K (TLoc l) U ->
  adopts T U ->
  exists m tag Ts U,
  jctop R K (TLoc l) (TyBlock m tag Ts U) /\
  m = Exclusive.
Proof.
  eauto using adoptees_are_exclusive, decomposition_exclusive.
Qed.

(* ------------------------------------------------------------------------- *)

(* If the permission [vs @ Ts[f:=T]] is satisfiable, then [vs] must of the
   form [ws[f:=w]]. *)

Lemma zip_subst_inverse:
  forall f Ts,
  row_context f Ts ->
  forall R K vs T,
  jpi R K (zip vs (fill_row Ts T)) ->
  exists w ws,
  row_context f ws /\
  vs = fill_row ws w.
Proof.
  induction 1; introv h; subst; simpl_subst_all;
  destruct vs; simpl in h; dependent destruction h; subst.
  (* Zero. *)
  xplit.
  econstructor; eauto.
  simpl_subst_goal. f_equal; eauto with subst_lift.
  (* Succ. *)
  forwards: IHrow_context. eauto. unpack.
  xplit.
  econstructor; eauto.
  simpl_subst_goal. f_equal; eauto with subst_lift.
Qed.

(* ------------------------------------------------------------------------- *)

(* The tactic [decomposition] applies the decomposition lemmas as much
   as possible. It also decomposes [jpi] hypotheses, when they involve
   existentials and stars, so as to get access to the [jctop]
   judgements that might be hidden in them. *)

Ltac select_decomposition action :=
  match goal with
  | h: jctop _ _ ?v1 (TySingleton ?v2) |- _ =>
      (* [h] implies that [v1] and [v2] coincide *)
      generalize (decomposition_TySingleton h);
      (* [h] also implies that this value is well-kinded *)
      generalize (jctop_value h);
      (* we must clear [h] in order to avoid looping *)
      clear h; intros ? ?; try subst v1
  | h: jctop _ _ (TAbs _) (TyArrow _ _) |- _ =>
      forwards: (decomposition_TyArrow h); [
        eauto with consistent_split
      | clear h; unpack
      ]
  | h: jctop _ _ _ (TyMixStar _ _) |- _ =>
      generalize (decomposition_TyMixStar h); clear h;
      intros ?; unpack; try subst
  | h: jctop _ _ (TLoc _) (TyBlock _ _ _ _) |- _ =>
      forwards: (decomposition_TyBlock h); [
        eauto with consistent_split
      | clear h; unpack; try subst
      ]
  | h: jctop _ _ (TLoc _) TyDynamic |- _ =>
      forwards: (decomposition_TyDynamic h); [
        eauto with consistent_split
      | clear h
      ]
  | h: jctop _ _ (TLck _) (TyLck _) |- _ =>
      forwards: (decomposition_TyLck h); [
        eauto 6 with consistent_split
      | clear h; unpack; try subst
      ]
  | h: jctop _ _ (TLck _) TyLocked |- _ =>
      forwards: (decomposition_TyLocked h); [
        eauto 6 with consistent_split
      | clear h; unpack; try subst
      ]
  | h1: row_context _ ?Ts,
    h2: jpi _ _ (zip ?vs (subst _ 0 ?Ts)) |- _ =>
      match vs with
      | subst _ 0 _ =>
          revert h2
      | _ =>
          forwards: zip_subst_inverse; [ eexact h1 | eexact h2 | unpack; subst; revert h2 ]
      end
  | h: jpi _ _ (TyStar _ _) |- _ =>
      action h
  | h: jpi _ _ (TyExists _) |- _ =>
      action h
  (* We are careful not to destruct judgements of the form
     [jpi _ _ (TyAnchored _ _)], since these are [jctop] judgements. *)
  end.

Ltac decomposition :=
  gji select_decomposition.

(* The tactic [decomposition_TyCon] applies the lemma by the same name. *)

Ltac decomposition_TyCon :=
  repeat match goal with
  | h: jctop _ _ (TLoc _) (TyCon _) |- _ =>
      forwards: (decomposition_TyCon h); [
        clear h; unpack; try subst
      ]
  end.

Ltac decomposition_TyLck :=
  repeat match goal with
  | h: jctop _ _ (TLoc _) (TyLck _) |- _ =>
      forwards: (decomposition_TyLck h); [
        clear h; unpack; try subst
      ]
  end.

(* A shortcut. *)

Ltac dca :=
  decomposition;
  consistent_abandon_all;
  abandon_all_maps.

