Set Implicit Arguments.
Require Import MyTactics.
Require Import MyArithmetic.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import Intro.
Require Import Dup.
Require Import Typing.
Require Import JBasic.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are affine with respect to [R]: the more resources
   are available, the better. *)

(* This property is needed in order to prove that [empty] is a unit for the
   conjunction of permissions: indeed, in order to convert [P * empty] to [P],
   we must abandon the resource that is owned by the [empty] conjunct. This
   resource is not necessarily trivial, because we have an affine
   interpretation of permissions. *)

(* It so happens that this property can be proved for [j] first, without
   mentioning [jpi]. Indeed, in the rules that mention [jpi], the current
   resource is split, and it suffices to send all of the ``excess resource''
   into the side that does not mention [jpi]. *)

(* This lemma is named [j_drop_right] because, when read from bottom to top,
   it can be used to drop the right-hand side of a [star]. *)

Lemma j_jbs_drop_right:
(
  forall f n R1 K P t T,
  j f n R1 K P t T ->
  forall R2 R,
  star R1 R2 = R ->
  consistent R ->
  j f n R K P t T
) /\ (
  forall f n R1 K P tag Tss ts T,
  jbs f n R1 K P tag Tss ts T ->
  forall R2 R,
  star R1 R2 = R ->
  consistent R ->
  jbs f n R K P tag Tss ts T
).
Proof.
  apply j_jbs_mutind; intros; subst;
  (* Clear the original type derivation, which can mislead [eauto]
     when attempting to construct a new type derivation. *)
  repeat match goal with
  | h: j _ _ _ _ _ _ _ |- _ => clear h
         end;
  try match goal with
  | h: is_locked_permission _ _ |- _ => forwards: lck_resource_maps_star; eauto; unpack
  end;
  (* The lemma [star_associative] has the effect of sending the
     excess resource into the right-hand side of a split of the
     form [star R1 R2 = R]. *)
  try econstructor (solve [
    eauto using @star_core, @star_associative,
                @consistent_core, consistent_12p3_23,
                @SCAM_maps_maximal_left, maximal_mo2dx, maximal_p2hl,
                @SCAM_limit_split_left,
                heap_project_morphism, lck_project_morphism
           with aac star_abandon_all lt_lt wk
  ]).

(* econstructor. auto. *)
(* Problem here: [L_LOCK] is not maximal, and we might have
     [lck_maps (star R R2) l (p2hl)]
   instead of
     [lck_maps (star R R2) l (p2l)]. *)

(* forwards hmaps: lck_resource_maps_star; eauto. *)
(* unpack; eauto with j. *)
Qed.

Definition j_drop_right := get1 j_jbs_drop_right.

(* The symmetric version can also be useful. *)

Lemma j_drop_left:
  forall f n R2 K P t T,
  j f n R2 K P t T ->
  forall R1 R,
  star R1 R2 = R ->
  consistent R ->
  j f n R K P t T.
Proof.
  introv ? ?; subst. rewrite star_commutative. eauto using @j_drop_right.
Qed.

(* We now prove that [jpi] is affine as well. *)

Lemma jpi_drop_right:
  forall R1 K P,
  jpi R1 K P ->
  forall R2 R,
  star R1 R2 = R ->
  consistent R ->
  jpi R K P.
Proof.
  induction 1; intros; subst; try solve [ econstructor; eauto 2 using j_drop_right ].
  (* JpiStar *)
  (* Send the surplus [R2] arbitrarily into one side. *)
  econstructor; [ | | eapply star_associative ]; eauto using consistent_12p3_23.
  (* JpiAtomic *)
  econstructor; [ | | | eapply star_associative ]; eauto using consistent_12p3_23.
Qed.

(* And now, [jcp] is affine. *)

(* Apparently, a variant of this lemma in file [Progress.v] does not
   require the consistency hypothesis. *)

Lemma jcp_drop_right:
  forall R1 R2 t,
  jcp R1 t ->
  consistent (star R1 R2) ->
  jcp (star R1 R2) t.
Proof.
  induction 1.
  eauto using j_drop_right with jcp.
  subst.
  repeat rewrite <- star_associative.
  eauto with jcp consistent_split.
Qed.

Lemma jcp_drop_left:
  forall R1 R2 t,
  jcp R2 t ->
  consistent (star R1 R2) ->
  jcp (star R1 R2) t.
Proof.
  introv ?.
  rewrite star_commutative.
  eauto using jcp_drop_right.
Qed.
