Set Implicit Arguments.
Require Import MyList.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Signature.
Require Import Dup.
Require Import Typing.
Require Import JBasic.
Require Import DerivedTyping.
Require Import Top.

(* ---------------------------------------------------------------------------- *)

(* If [Ts] is duplicable, then the permission [vs @ Ts] is duplicable. *)

Lemma dup_zip:
  forall Ts,
  dup Ts ->
  forall vs,
  dup (zip vs Ts).
Proof.
  induction Ts; inversion 1; destruct vs; simpl; eauto with dup.
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [jpi R K P] holds and [P] is duplicable, then [jpi (core R) K P] holds
   as well. In other words, duplicable permissions are justified by duplicable
   resources. *)

(* The above statement is not directly amenable to an inductive proof. In the
   case [JpiAtomic], for instance, the hidden permission is not necessarily
   duplicable, so the induction hypothesis cannot be applied. We generalize
   this statement as follows: if [jpi R K P] holds, then for *every* duplicable
   fragment [Q] of [P], the judgement [jpi (core R) K Q] holds as well. This is
   the reason why we need to define the relation [pcore]. *)

(* In order to prove this property of [jpi], we must state a related property
   about the typing judgement [j]. Fortunately, this property needs only
   concern values (as opposed to terms) and needs only concern canonical
   typing derivations (as opposed to arbitrary derivations). We prove that, if
   a canonical typing derivation concludes that a value [v] has duplicable
   type [T] under [R] and [P], then the same is true under [core R] and under
   *some* duplicable fragment [Q] of [P]. *)

Lemma j_duplicable_preliminary:
(
  forall f n R1 K P t T,
  j f n R1 K P t T ->
  f = Canonical ->
  wk K P PERM ->
  wk K t VALUE ->
  consistent R1 ->
  dup T ->
  forall R2,
  core R1 = R2 ->
  exists m Q,
  jc m R2 K Q t T /\
  pcore P Q
) /\ (
  forall R1 K P,
  jpi R1 K P ->
  consistent R1 ->
  forall R2,
  core R1 = R2 ->
  forall Q,
  pcore P Q ->
  jpi R2 K Q
).
(* A tactic for recognizing and invoking the induction hypothesis. *)
Ltac j_duplication_ih :=
  match goal with ih: Canonical = Canonical -> _ |- _ =>
    forwards: ih; eauto with wk dup consistent_split; unpack
  end.
Proof.
  apply j_jpi_mutind; intros; subst; try discriminate; simpl;
  dup_inversion; wk_inversion;
  (* Clear the induction hypothesis when it is clearly not applicable. *)
  try match goal with ih: Unrestricted = Canonical -> _ |- _ => clear ih end.

  (* JSingleton *)
  solve [ eauto using (JSingleton 0) with pcore ].

  (* JFrame *)
  j_duplication_ih.
  (* The framed out permission must be duplicable, so it is a duplicable
     fragment of itself. We are able to reconstruct an instance of the
     frame rule. *)
  solve [ eauto 8 using JFrame with pcore ].

  (* JCut *)
  j_duplication_ih.
  (* The induction hypothesis yields the existence of *some* duplicable
     fragment [Q] of the conjunction [P1 * P2]. By analysis of the definition
     of [pcore], this must be either [TyEmpty] or a conjunction of two
     fragments of [P1] and [P2]. *)
  forwards [ | ]: pcore_star_inversion; [ eauto | | ];
  intros; instantiate; unpack; subst.
  (* In the former case, the result follows directly, whereas in the latter
     case, it follows from an application of [JCut]. *)
  solve [ eauto 6 using JSameResource, @core_star_right with pcore ].
  solve [ eauto 8 using JCut, @star_core with consistent_split ].

  (* JFunction *)
  (* The captured permission must be duplicable, so it is its own
     core. Furthermore, the function body is type-checked under
     [core R], which is the same as [core (core R)]. Thus, we are
     able to reconstruct an instance of [JFunction]. *)
  solve [ eauto 8 using JFunction, JSameResource, @core_idempotent with pcore ].

  (* JBlock *)
  consistent_abandon_all.
  (* We are looking at a duplicable object, so there are no adoptees. *)
  forwards: adopters_are_exclusive; [
    eauto
  | SCAM_maps_maximal heap_project_morphism
  | eapply duplicable_duplicable | ].
  subst.
  abandon_all_inversion.
  simpl in *.
  xplit.
  eapply (JBlock 0);
    eauto using @star_core, @SCAM_maps_core, dup_zip,
                @duplicable_implies_self_core_symmetric, @duplicable_duplicable,
                duplicable_objects_have_no_adoptees,
                heap_project_morphism
           with abandon_all consistent_split pcore SCAM_maps_maximal_heap.
  eapply PCoreEmpty.

  (* JFold *)
  (* A duplicable algebraic data type unfolds to a duplicable block type,
     so everything works out as desired. *)
  match goal with h: _ = Duplicable |- _ => rewrite h in * end.
  j_duplication_ih.
    econstructor. eapply is_nth_Forall; eauto.
  solve [ eauto using JFold ].

  (* JDynamic *)
  solve [ eauto 7 using (JDynamic 0), @SCAM_limit_core with lt_lt pcore typeclass_instances ].

  (* JLck *)
  solve [ eauto 7 using (JLck 0), @SCAM_maps_core, is_locked_permission_core
                   with pcore typeclass_instances ].

  (* JForallIntro *)
  j_duplication_ih. eauto using wk_weakening.
  solve [ xplit; eauto with j pcore ].

  (* JpiEmpty *)
  (* A fragment of the empty permission must be itself empty. *)
  forwards: pcore_empty_inversion; eauto. subst.
  solve [ eauto using JpiEmpty ].

  (* JpiStar *)
  (* A fragment of [P1 * P2] must be either empty or a conjunction of
     fragments of [P1] and [P2]. In the former case, the result is
     immediate. In the latter case, it follows from the induction
     hypotheses. *)
  forwards [ | ]: pcore_star_inversion; [ eauto | | ];
  intros; instantiate; unpack; subst.
  solve [ eauto using JpiEmpty ].
  solve [ eauto 7 using @star_core with j consistent_split ].

  (* JpiForall *)
  (* A fragment of a universal permission must be either a universal
     permission or empty or a (duplicable) universal permission.
     In either case, the result is easily obtained. *)
  pcore_inversion; dup_inversion;
  solve [ eauto with j pcore ].
 
  (* JpiExists *)
  (* A fragment of an existential permission must be either empty or
     a (duplicable) existential permission. In the former case, the
     result is immediate. In the latter case, it follows from the
     induction hypothesis.*)
  pcore_inversion; dup_inversion.
  solve [ eauto using JpiEmpty ].
  solve [ eauto using JpiExists, pcore_subst with pcore ].

  (* JpiAtomic *)
  (* A fragment of the permission [v @ T] must be either empty or
     [v @ T] itself, where [T] is duplicable. In the former case, the
     result is immediate. In the latter case, it follows from the
     induction hypothesis.*)
  pcore_inversion; dup_inversion.
  solve [ eauto using JpiEmpty ].
  j_duplication_ih.
  solve [ eauto using @star_core with j consistent_split ].

Qed.

(* ---------------------------------------------------------------------------- *)

(* This is the sought-after corollary: duplicable permissions are justified by
   duplicable resources. *)

Lemma j_duplicable:
  forall R K P,
  jpi R K P ->
  consistent R ->
  dup P ->
  jpi (core R) K P.
Proof.
  intros. eapply j_duplicable_preliminary; eauto with pcore.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The same property holds of [jctop] derivations. *)

Lemma jctop_duplicable:
  forall R K v T,
  jctop R K v T ->
  consistent R ->
  dup T ->
  jctop (core R) K v T.
Proof.
  dependent_destruction; intros; subst.
  forwards:  get1 j_duplicable_preliminary; eauto with consistent_split wk. unpack.
  forwards: last2 j_duplicable_preliminary; eauto with consistent_split.
  econstructor; eauto using @star_core.
Qed.

