Set Implicit Arguments.
Require Import MyTactics.
Require Import Syntax.
Require Import Row.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import MyList.
Require Import SubsumptionSoundness.
Require Import JSubstitution.
Require Import Environments.
Require Import Harmless.
Require Import Inhabited.
Require Import MSASAMOperations.
Require Import DeBruijn.

(* Typing judgements for heaps and closed configurations. *)

(* These judgements are used in the type soundness proof. They encode the global
   invariant that the type system guarantees. *)

(* ---------------------------------------------------------------------------- *)

(* The judgement [jh h R] means that the uninstrumented heap [h] agrees with
   the resource [R]. A resource is an instrumented heap, so this judgement
   basically means that [h] is a projection of [R]. *)

Notation heap_agrees :=
  (@SCAM_agrees _ _ concrete_heap_project _ _ _ _ _ heap_project heap_agree).
Notation lck_agrees :=
  (@SCAM_agrees _ _ concrete_lck_project  _ _ _ _ _ lck_project  lck_agree ).

Definition jh (h: concrete_resource) (R: resource): Prop :=
     heap_agrees h R /\ lck_agrees h R.

Ltac lck_agrees_maps_left_to_right :=
  match goal with
    | hjh: jh ?h _, hmaps: concrete_lck_maps ?h _ _ |- _ =>
        inversion hjh as [ _ hagrees ];
        unpack (SCAM_agrees_maps_left_to_right hagrees hmaps);
        clear hagrees
  end.

Ltac is_locked_permission_inversion_aux h :=
  inversion h as [ hp | hp ]; inversion hp; subst; clear h hp.

Ltac is_locked_permission_inversion_step :=
  match goal with
    | h: is_locked_permission (L_LOCK _) _ |- _ =>
        is_locked_permission_inversion_aux h
    | h: is_locked_permission (L_HELD_LOCK _) _ |- _ =>
        is_locked_permission_inversion_aux h
    | h: is_locked_permission L_ERROR _ |- _ =>
        is_locked_permission_inversion_aux h
    | h: is_locked_permission ?l _ |- _ =>
        destruct l
  end.

Ltac is_locked_permission_inversion :=
  repeat is_locked_permission_inversion_step.

Ltac lck_agree_inversion_aux h :=
  inversion h as [ ? ? h_agree | ? ? h_agree ]; inversion h_agree; subst; clear h h_agree.

Ltac lck_agree_inversion_step :=
  match goal with
    | h: lck_agree TakenLock _ |- _ =>
        lck_agree_inversion_aux h
    | h: lck_agree FreeLock _ |- _ =>
        lck_agree_inversion_aux h
    | h: lck_agree ?l _ |- _ =>
        destruct l
  end.

Ltac lck_agree_inversion :=
  repeat lck_agree_inversion_step.

Ltac lck_maps_lock_split_aux :=
  match goal with
    | hmaps: lck_maps (star _ _) _ (L_LOCK _) |- _ =>
        let hmsi := fresh in
          forwards hmsi: (SCAM_maps_star_inversion lck_project_morphism hmaps);
        named_unpack hmsi;
        forwards: (split_lock hmsi);
        unpack; subst;
        clear hmaps hmsi
  end.

Ltac lck_maps_lock_split :=
  repeat lck_maps_lock_split_aux;
  SCAM_maps_injective; subst;
  try discriminate.

Ltac lck_maps_held_lock_split_aux :=
  match goal with
    | hmaps: lck_maps (star _ _) _ (L_HELD_LOCK _) |- _ =>
        let hmsi := fresh in
          forwards hmsi: (SCAM_maps_star_inversion lck_project_morphism hmaps);
        named_unpack hmsi;
        destruct (split_held_lock hmsi);
        unpack; subst;
        clear hmaps hmsi
  end.

Ltac lck_maps_held_lock_split :=
  repeat (lck_maps_held_lock_split_aux; lck_maps_lock_split);
  repeat is_locked_permission_inversion.


(* ---------------------------------------------------------------------------- *)

(* Properties of the judgement [jh]. *)

(* Recall that [jh h R] means that the heap [h] and the instrumented heap [R]
   agree. *)

(* The property [jh h R] happens to imply that [R] is consistent. This is
   not essential, but allows omitting a few consistency hypotheses here
   and there. *)

Lemma jh_consistent:
  forall h R,
  jh h R ->
  consistent R.
Proof.
  inversion 1.
  decompose_pairs; consistent_product.
  split.
  replace c with (heap_project (c, c0)). eapply SCAM_agrees_consistent;
    [ eauto using @dx_permission_agrees_consistent | eapply H0 ].
    auto.
  replace c0 with (lck_project (c, c0)). eapply SCAM_agrees_consistent;
    [ eauto using @l_permission_agrees_consistent | eapply H1 ].
    auto.
Qed.

(* The empty heap and the empty resource agree. *)

Lemma jh_empty:
  jh concrete_empty void.
Proof.
  split; unfold SCAM_agrees; simpl; eapply agrees_empty.
Qed.

(* Allocating a new address in the heap requires allocating a new address in
   the instrumented heap as well. There must be agreement between the value
   and the permission. *)

Lemma jh_heap_allocate:
  forall h R m tag vs,
  jh h R ->
  jh (concrete_heap_allocate h (Block tag Null vs)) (heap_alloc R m (Orphan tag vs)).
Proof.
  inversion 1.
  split.
  eapply SCAM_agrees_allocate; eauto using agree_mo2dx.
  decompose_pairs; eassumption.
Qed.

Lemma jh_lck_allocate:
  forall h R P,
  jh h R ->
  wk zilch P PERM ->
  jh (concrete_lck_allocate h FreeLock) (lck_allocate R (p2l P)).
Proof.
  inversion 1.
  split.
  decompose_pairs; eassumption.
  eapply SCAM_agrees_allocate;
    eauto using agree_p2l, wk_closed, length_empty.
Qed.

(* Updating the heap with a new block requires updating the instrumented heap
   as well. There must be agreement between the block and the permission. *)

Lemma jh_heap_update:
  forall h R l m tag vs,
  jh h R ->
  jh (concrete_heap_update h l (Block tag Null vs)) (heap_upd R l m (Orphan tag vs)).
Proof.
  inversion 1.
  split.
  eapply SCAM_agrees_update; eauto using agree_mo2dx.
  decompose_pairs; eassumption.
Qed.


(* Without modifying the heap, it is possible to update the instrumented
   heap. There must be agreement between the pre-existing block and the new
   permission. *)

(* Unused. *)

Lemma jh_freeze:
  forall h R l m tag vs,
  jh h R ->
  concrete_heap_maps h l (Block tag Null vs) ->
  jh h (heap_upd R l m (Orphan tag vs)).
Proof.
  inversion 1.
  introv hmaps.
  erewrite <- (SAM_update_identity concrete_heap_sub_resource hmaps).
  eauto using jh_heap_update.
Qed.

(* Having [x] adopt [y] in the physical heap requires adjusting the instrumented
   heap as well. *)

Lemma jh_adopt:
  forall h R y x tag vs req,
  jh h R ->
  jh (concrete_heap_update h y (Block tag (NonNull x) vs)) (adopt R y x tag vs req).
Proof.
  inversion 1.
  split.
  eapply SCAM_agrees_update.
    eauto.
    econstructor; eauto.
  decompose_pairs; eassumption.
Qed.

(* If the heap and instrumented heap agree, then they have the same allocation
   limit. *)

Lemma jh_heap_limit:
  forall h R1 l,
  jh h R1 ->
  heap_limit R1 = l ->
  concrete_heap_limit h = l.
Proof.
  inversion 1; intros; subst.
  eapply SCAM_agrees_limit; eauto.
Qed.

Lemma jh_lck_limit:
  forall h R1 l,
  jh h R1 ->
  lck_limit R1 = l ->
  concrete_lck_limit h = l.
Proof.
  inversion 1; intros; subst.
  eapply SCAM_agrees_limit; eauto.
Qed.

Lemma jh_lck_update_Free:
  forall h R l P,
  jh h R ->
  closed 0 P ->
  jh (concrete_lck_update h l FreeLock) (lck_update R l (p2l P)).
Proof.
  inversion 1.
  split.
  decompose_pairs. eapply H0.
  eapply SCAM_agrees_update. auto. econstructor. econstructor. auto.
Qed.

Lemma jh_lck_update_Taken:
  forall h R l P,
  jh h R ->
  closed 0 P ->
  jh (concrete_lck_update h l TakenLock) (lck_update R l (p2hl P)).
Proof.
  inversion 1.
  split.
  decompose_pairs. eapply H0.
  eapply SCAM_agrees_update. auto. econstructor. econstructor. auto.
Qed.

(* If the heap and instrumented heap agree, and if the instrumented heap
   maps [l] to a permission that agrees with the block [b], then the heap
   maps [l] to [b]. *)

Lemma jh_heap_maps_right_to_left:
  forall h R l p b,
  jh h R ->
  heap_maps R l p ->
  heap_agree b p ->
  concrete_heap_maps h l b.
Proof.
  introv h1 h2 h3.
  inversion h1 as [ h11 _ ].
  unpack (SCAM_agrees_maps_right_to_left h11 h2).
  equates 1. eauto.
  eapply heap_permission_determines_block; eauto.
Qed.

(* Lemma jh_lck_maps_right_to_left: *)
(*   forall h R l p, *)
(*   jh h R -> *)
(*   lck_maps R l p -> *)
(*   exists b, *)
(*   concrete_lck_maps h l b. *)
(* Proof. *)
(*   introv h1 h2. *)
(*   dependent destruction h. *)
(*   destruct h1 as (h11, h12). *)
(*   unpack (SCAM_agrees_maps_right_to_left h12 h2). *)
(*   eauto. *)
(* Qed. *)

(* If the heap and instrumented heap agree, if the heap maps [l] to [b], and
   if the instrumented heap maps [l] to a permission that mentions [o], then
   [b] and [o] must agree. *)

Lemma jh_maps_maps:
  forall h R l m b o,
  jh h R ->
  concrete_heap_maps h l b ->
  heap_maps R l (mo2dx m o) ->
  block_orphan_agree b o.
Proof.
  introv h1 h2 h3.
  inversion h1 as [ h11 _ ].
  unpack (SCAM_agrees_maps_maps h11 h2 h3).
  eauto using agree_mo2dx_implies_block_orphan_agree.
Qed.

(* The tactic [jh_maps_maps] helps apply the lemma [jh_maps_maps]
   in a situation where its hypotheses are available. *)

Ltac jh_maps_maps_aux b o :=
  assert (block_orphan_agree b o); [
    (* Proof of the assertion [block_orphan_agree b o]. *)
    eapply jh_maps_maps; [
      eauto
    | eauto
    | SCAM_maps_maximal heap_project_morphism
    ]
  | (* Continuation of the assertion. *)
    (* Destruct the agreement assertion. *)
    block_orphan_agree_inversion;
    (* If the equality between the fields has the form [subst _ 0 vs1 =
       subst _ 0 vs2] and if we can find two hypotheses of the form
       [row_context f vs1] and [row_context f vs2], then we can further
       decompose this equality. *)
    try row_context_equality_inversion
  ].

Ltac jh_maps_maps :=
  match goal with
  | h1: concrete_heap_maps _ ?l ?b, h2: heap_maps _ ?l (mo2dx _ ?o) |- _ =>
      jh_maps_maps_aux b o
  | h1: concrete_heap_maps _ ?l ?b, h2: heap_maps _ ?l (DX_EXCLUSIVE (Block ?tag Null ?vs)) |- _ =>
      jh_maps_maps_aux b (Orphan tag vs)
  end.

(* If the heap [h] and instrumented heap [R] agree, if [l] is below the allocation
   limit (hence a valid memory location), then there must exist a memory block [b]
   and a permission [p] such that the heap [h] maps [l] to the block [b] and the
   instrumented heap [R] maps [l] to the permission [p] and [b] and [p] are in
   agreement. *)

Lemma jh_heap_limit_maps_maps:
  forall h R,
  jh h R ->
  consistent R ->
  forall l,
  l < heap_limit R ->
  exists b p,
  concrete_heap_maps h l b /\
  heap_maps R l p /\
  heap_agree b p.
Proof.
  introv hj hc hl.
  inversion hj as [ hjj _ ].
  (* Because [l] is below the allocation limit, [R] must map [l] to
     some permission [p]. *)
  generalize (SCAM_limit_maps heap_project_morphism R hl hc); intros [ p hp ].
  (* Because [h] and [R] agree, [h] must map [l] to some memory block
     that agrees with the description given by [p]. *)
  unpack (SCAM_agrees_maps_right_to_left hjj hp).
  (* We are done. *)
  solve [ eauto ].
Qed.

(* Guarded permission. *)

Definition lck_selection (l: lock) (lp: lck_permission) : term :=
  match l with
    | TakenLock => TyEmpty
    | FreeLock => 
        match lp with
          | L_ERROR => TyEmpty
          | L_LOCK P => P
          (* The last case is not meaningful in a well-typed heap.
             However, it is convenient for [guarded_permission_drop_right]. *)
          | L_HELD_LOCK P => P
        end
  end.

Definition guarded_permission_list (h: concrete_resource)
                                   (R: resource): list term :=
  let lh := concrete_lck_project h in
  let (lR, _) := lck_project R in
  match lh, lR with
    | MapUndefined, _
    | _, MapUndefined  => nil
    | Map mh, Map mR   => zip lck_selection mh mR
  end.

Definition guarded_permission (h: concrete_resource) (R: resource) : term :=
  fold_left TyStar (guarded_permission_list h R) TyEmpty.

Lemma jpi_closed:
  forall R K kappa P,
  jpi R (insert 0 kappa K) P ->
  closed 0 P ->
  substitutable kappa ->
  jpi R K P.
Proof.
  intros.
  forwards: all_substitutable_kinds_inhabited; [ eauto | unpack ].
  eapply j_substitution; eauto using closed_subst_invariant, wk_empty.
Qed.

Ltac map_undefined_inversion :=
  repeat match goal with
           | h: context[MapUndefined _] |- _ => inversion h; clear h
         end.

Ltac lck_agrees_inversion :=
  match goal with
    | h: lck_agrees _ _ |- _ => inversion h; subst; simpl in *; subst
  end.

Ltac map_heap_cleanup :=
  repeat match goal with
           | h: jh _ _ |- _ => clear h
           | h: heap_agrees _ _ |- _ => clear h
           | h: lck_agrees _ _ |- _ => clear h
           | c: MSAClosedAllocationMaps.closed _ |- _ => clear c
           | c: concrete_heap |- _ => clear c
           | l: list heap_permission |- _ => clear l
           | x: raw_heap_resource |- _ => clear x
         end.

Ltac forall2_inversion :=
  match goal with
    | h: Forall2 _ _ _ |- _ => inversion h; subst; clear h
  end.

Lemma jh_guarded_permission_list_closed:
  forall h R,
  jh h R ->
  Forall (closed 0) (guarded_permission_list h R).
Proof.
  inversion 1.
  decompose_pairs. unfold guarded_permission_list. came. dam.
  map_undefined_inversion.
  map_undefined_inversion.
  map_undefined_inversion.
  lck_agrees_inversion; map_heap_cleanup.
  induction_lists; eauto.
  econstructor.
    forall2_inversion.
      lck_agree_inversion; simpl; [ eauto | econstructor ].
    forall2_inversion.
      eauto.
Qed.

Lemma TyStar_closed:
  forall P1 P2,
  closed 0 P1 ->
  closed 0 P2 ->
  closed 0 (TyStar P1 P2).
Proof.
  introv hc1 hc2. unfold closed in *; simpl_lift_all.
  rewrite hc1. rewrite hc2. auto.
Qed.

Lemma fold_left_forall:
  forall (A: Type) (l: list A) (Pi: A) (f: A -> A -> A) (P: A -> Prop),
  P Pi ->
  Forall P l ->
  (forall p1 p2, P p1 -> P p2 -> P (f p1 p2)) ->
  P (fold_left f l Pi).
Proof.
  introv ? hforall ?. generalize dependent Pi. induction l.
    auto.
    dependent destruction hforall. eauto.
Qed.

Lemma fold_left_TyStar_closed:
  forall l Pi,
  Forall (closed 0) l ->
  closed 0 Pi ->
  closed 0 (fold_left TyStar l Pi).
Proof.
  intros. eapply fold_left_forall; eauto using TyStar_closed.
Qed.
  
Lemma jh_guarded_permission_closed:
  forall h R,
  jh h R ->
  closed 0 (guarded_permission h R).
Proof.
  inversion 1.
  forwards: jh_guarded_permission_list_closed; eauto.
  unfold guarded_permission.
  unfold guarded_permission_list in *.
  decompose_pairs. came. dam.
  map_undefined_inversion.
  map_undefined_inversion.
  map_undefined_inversion.
  map_heap_cleanup.
  match goal with
    | c: concrete_lck |- _ => destruct c
  end.
    unfold closed; simpl; simpl_lift_goal; auto.
    induction_lists.
      econstructor.
      econstructor.
      econstructor.
      eapply fold_left_TyStar_closed.
        eauto.
        unfold closed; simpl_lift_goal; erewrite closed_lift_invariant by eauto; auto.
Qed.

(* Lemma fold_left_TyStar_jps_acc: *)
(*   forall l n K P Q, *)
(*   jps n K P Q -> *)
(*   jps (List.length l + n) K (fold_left TyStar l P) *)
(*                        (fold_left TyStar l Q). *)
(* Proof. *)
(*   intros l n K P Q hjps; revert l. *)
(*   apply rev_ind. *)
(*     simpl. auto. *)
(*     intros. *)
(*     repeat rewrite fold_left_app. *)
(*     simpl. *)
(*     replace (List.length (l ++ x :: nil) + n) with (S (List.length l + n)). *)
(*     eauto with j. *)
(*     rewrite app_length. *)
(*     simpl. *)
(*     omega. *)
(* Qed. *)

Lemma fold_left_TyStar_jps_acc:
  forall l n K P Q,
  jps n K P Q ->
  exists k,
  jps k K (fold_left TyStar l P)
          (fold_left TyStar l Q).
Proof.
  intros l n K P Q hjps; revert l.
  apply rev_ind; [ eauto | ].
    intros; unpack.
    repeat rewrite fold_left_app.
    simpl; eauto with j.
Qed.

Lemma jps_star_associative_symmetric:
  forall n K P1 P2 P3,
  jps (S (S (S n))) K (TyStar (TyStar P1 P2) P3) (TyStar P1 (TyStar P2 P3)).
Proof.
  intros.
  eauto with j.
Qed.

Lemma fold_left_TyStar_right:
  forall l K P Q,
  exists n,
  jps n K (fold_left TyStar l (TyStar P Q))
          (TyStar Q (fold_left TyStar l P)).
Proof.
  intros l K P Q; revert l.
  apply rev_ind; [ exists 0; simpl; eauto with j | ].
    intros; unpack.
    repeat rewrite fold_left_app; simpl.
    eexists.
    eapply JpsTransitive; [ | eapply jps_star_associative_symmetric ].
    eauto with j.
Qed.

Lemma fold_left_TyStar_right_inverse:
  forall l K P Q,
  exists n,
  jps n K (TyStar Q (fold_left TyStar l P))
          (fold_left TyStar l (TyStar P Q)).
Proof.
  intros l K P Q; revert l.
  apply rev_ind; [ exists 0; simpl; eauto with j | ].
    intros; unpack.
    repeat rewrite fold_left_app; simpl.
    eauto with j.
Qed.

Lemma fold_left_TyStar_is_nth_P:
  forall n l P K Pi,
  is_nth n l P ->
  exists k,
  jps k K (fold_left TyStar l Pi)
          (TyStar P (fold_left TyStar (set_nth n l TyEmpty) Pi)).
Proof.
  introv hisnth; revert Pi; induction hisnth; simpl; [ | eauto ].
    intros.
    forwards h1: fold_left_TyStar_jps_acc.
      eapply JpsCoStar; [ eapply (JpsEmptyAppearsRight 0) | eapply JpsReflexive ].
    forwards h2: fold_left_TyStar_right.
    destruct h1 as (x1, h1).
    destruct h2 as (x2, h2).
    exists (S (x1 + x2)).
    eapply JpsTransitive.
      eapply jps_index_monotonic; [ eapply h1 | eauto with arith ].
      eapply jps_index_monotonic; [ eapply h2 | eauto with arith ].
Qed.

Lemma fold_left_TyStar_is_nth_empty:
  forall n l P K Pi,
  is_nth n l TyEmpty ->
  exists k,
  jps k K (TyStar P (fold_left TyStar l Pi))
          (fold_left TyStar (set_nth n l P) Pi).
Proof.
  intro n; induction n.
    introv hisnth; inversion hisnth; subst; simpl.
    forwards h1: fold_left_TyStar_jps_acc.
      eapply (JpsDisappearsRight 0).
    forwards h2: fold_left_TyStar_right_inverse.
    destruct h1 as (x1, h1).
    destruct h2 as (x2, h2).
    exists (S ((S x1) + x2)).
    eapply JpsTransitive.
      eapply jps_index_monotonic; [ eapply JpsCoStar | eauto with arith ].
        eapply JpsReflexive.
        eapply h1.
      eapply jps_index_monotonic; [ eapply h2 | eauto with arith ].
  intro l; destruct l; simpl; introv hisnth; inversion hisnth; subst.
    eauto.
Qed.

Lemma guarded_permission_heap_invariant:
  forall hh1 hh2 lh hR1 hR2 lR,
  guarded_permission (hh1, lh) (hR1, lR) = guarded_permission (hh2, lh) (hR2, lR).
Proof.
  unfold guarded_permission. simpl. auto.
Qed.

Lemma guarded_permission_heap_inject_invariant:
  forall h R hh hR,
  guarded_permission h R = guarded_permission (concrete_heap_inject hh h) 
                                              (heap_inject hR R).
Proof.
  decompose_pairs. simpl. eauto using guarded_permission_heap_invariant.
Qed.

Lemma guarded_permission_concrete_heap_allocate:
  forall h b R,
  guarded_permission (concrete_heap_allocate h b) R = guarded_permission h R.
Proof.
  decompose_pairs. eauto using guarded_permission_heap_invariant.
Qed.

Lemma guarded_permission_heap_alloc:
  forall h R m o,
  guarded_permission h (heap_alloc R m o) = guarded_permission h R.
Proof.
  decompose_pairs. eauto using guarded_permission_heap_invariant.
Qed.

Lemma guarded_permission_new_lck:
  forall h R P,
  jh h R ->
  guarded_permission (concrete_lck_allocate h FreeLock) (lck_allocate R (p2l P))
  = TyStar (guarded_permission h R) P.
Proof.
  inversion 1.
  decompose_pairs.
  unfold guarded_permission, guarded_permission_list.
  unfold SCAM_allocate, SAM_allocate.
  repeat rewrite pi_inverse.
  came. dam.
  map_undefined_inversion.
  map_undefined_inversion.
  map_undefined_inversion.
  dependent destruction c2; simpl.
    map_undefined_inversion.
    rewrite zip_app.
      rewrite fold_left_app; simpl; auto.
      forwards: jh_lck_limit; eauto.
      auto.
Qed.
(*
Lemma guarded_permission_new_lck:
  forall h R p,
  jh h R ->
  guarded_permission (concrete_lck_allocate h TakenLock) (lck_allocate R p)
  = TyStar (guarded_permission h R) TyEmpty.
Proof.
  inversion 1.
  decompose_pairs.
  unfold guarded_permission, guarded_permission_list.
  unfold SCAM_allocate, SAM_allocate.
  repeat rewrite pi_inverse.
  came. dam.
  map_undefined_inversion.
  map_undefined_inversion.
  map_undefined_inversion.
  dependent destruction c2; simpl.
    map_undefined_inversion.
    rewrite zip_app.
      rewrite fold_left_app; simpl; auto.
      forwards: jh_lck_limit; eauto.
      auto.
Qed.
*)

Inductive lck_permission_compat: lck_permission -> lck_permission -> Prop :=

| Lck_Error_compat:
    lck_permission_compat L_ERROR L_ERROR

| Lck_Perm_compat:
    forall lp1 lp2 P,
    is_locked_permission lp1 P ->
    is_locked_permission lp2 P ->
    lck_permission_compat lp1 lp2.

Lemma lck_selection_compat:
  forall l lp1 lp2,
  lck_permission_compat lp1 lp2 ->
  lck_selection l lp1 = lck_selection l lp2.
Proof.
  inversion 1; is_locked_permission_inversion; destruct l; simpl; auto.
Qed.

Inductive lck_compat: resource -> resource -> Prop :=
| Lck_compat:
    forall R1 R2,
    lck_limit R1 = lck_limit R2 ->
    (forall l lp1 lp2,
     lck_maps R1 l lp1 ->
     lck_maps R2 l lp2 ->
     lck_permission_compat lp1 lp2) ->
    lck_compat R1 R2.
    
Lemma guarded_permission_compat:
  forall h R1 R2,
  lck_compat R1 R2 ->
  guarded_permission h R1 = guarded_permission h R2.
Proof.
  dependent_destruction.
  unfold guarded_permission; f_equal.
  unfold guarded_permission_list, SCAM_maps, SCAM_limit in *.
  decompose_pairs.
  came.
  map_heap_cleanup.
  destruct c4; [ auto | ].
  dam; induction_lists; auto.
    eapply lck_selection_compat. eapply H0.
      econstructor. econstructor.
      econstructor. econstructor.
    eapply IHxs; eauto.
      introv hm1 hm2;
        inversion hm1; inversion hm2; subst.
      eapply H0.
        econstructor. econstructor. eauto.
        econstructor. econstructor. eauto.
Qed.

Lemma star_consistent_compat:
  forall lp1 lp2,
  consistent (star lp1 lp2) ->
  lck_permission_compat (star lp1 lp2) lp1.
Proof.
  intros.
  forwards: consistent_lp_inversion; eauto. unpack.
  econstructor; eauto using is_locked_permission_split_left.
Qed.

Lemma drop_right_lck_compat:
  forall R R1 R2,
  star R1 R2 = R ->
  consistent (star R1 R2) ->
  lck_compat R R1.
Proof.
  introv hs hc. subst.
  econstructor.
  eapply SCAM_limit_split_left; eauto with typeclass_instances.
  introv hmaps12 hmaps1.
    forwards: (SCAM_maps_star_inversion lck_project_morphism hmaps12). unpack.
    SCAM_maps_injective.
    forwards: (SCAM_maps_consistent lck_project_morphism (star R1 R2)); eauto.
    eauto using star_consistent_compat.
Qed.  

Lemma guarded_permission_drop_right:
  forall h R R1 R2,
  consistent R ->
  star R1 R2 = R ->
  guarded_permission h R = guarded_permission h R1.
Proof.
  intros; subst.
  eauto using guarded_permission_compat, drop_right_lck_compat.
Qed.

Lemma guarded_permission_list_set_empty:
  forall h l R lp,
  guarded_permission_list (concrete_lck_update h l TakenLock) (lck_update R l lp)
  = set_nth l (guarded_permission_list h R) TyEmpty.
Proof.
  decompose_pairs.
  unfold guarded_permission_list, SAM_update.
  came.
  map_heap_cleanup.
  generalize dependent x; generalize dependent c2; induction l.
    intro c2; destruct c2; intro x; destruct x; simpl; auto.
      induction_lists; auto.
    intro c2; destruct c2; intro x; destruct x; simpl; auto.
      induction_lists; auto.
      forwards: (IHl (Map xs) (Map ys)). auto.
Qed.

Lemma guarded_permission_list_set_P:
  forall h l R P,
  jh h R ->
  guarded_permission_list (concrete_lck_update h l FreeLock) (lck_update R l (p2l P))
  = set_nth l (guarded_permission_list h R) P.
Proof.
  inversion 1. decompose_pairs.
  unfold guarded_permission_list, SAM_update.
  came.
  map_heap_cleanup.
  generalize dependent x; generalize dependent c2; induction l.
    intro c2; destruct c2; intro x; destruct x; simpl; auto.
      induction_lists; auto.
    intro c2; destruct c2; intro x; destruct x; simpl; auto.
      induction_lists; auto.
      forwards: (IHl (Map xs) (Map ys)). auto.
Qed.

Lemma guarded_permission_acquire_jps:
  forall h l R K P,
  jh h R ->
  concrete_lck_maps h l FreeLock ->
  lck_maps R l (p2l P) ->
  exists n,
  jps n K (guarded_permission h R)
          (TyStar P (guarded_permission (concrete_lck_update h l (TakenLock))
                                        (lck_update R l (p2hl P)))).
Proof.
  inversion 1.
  introv hcmaps hmaps.
  unfold guarded_permission.
  rewrite guarded_permission_list_set_empty.
  eapply fold_left_TyStar_is_nth_P.
  unfold SCAM_maps in hmaps. unfold CAM_maps in hmaps. unfold SAM_maps in hcmaps.
  unfold guarded_permission_list.
  decompose_pairs. came. dam.
  map_undefined_inversion.
  map_undefined_inversion.
  map_undefined_inversion.
  destruct c2.
    map_undefined_inversion.
    inversion hmaps; inversion hcmaps; subst.
    eapply is_nth_zip_eq; eauto.
Qed.

Lemma guarded_permission_release_jps:
  forall h l R K P,
  jh h R ->
  concrete_lck_maps h l TakenLock ->
  lck_maps R l (p2hl P) ->
  exists n,
  jps n K (TyStar P (guarded_permission h R))
          (guarded_permission (concrete_lck_update h l FreeLock) (lck_update R l (p2l P))).
Proof.
  inversion 1.
  introv hcmaps hmaps.
  unfold guarded_permission.
  erewrite guarded_permission_list_set_P by eauto.
  eapply fold_left_TyStar_is_nth_empty.
  unfold SCAM_maps in hmaps. unfold CAM_maps in hmaps. unfold SAM_maps in hcmaps.
  unfold guarded_permission_list.
  decompose_pairs. came. dam.
  map_undefined_inversion.
  map_undefined_inversion.
  map_undefined_inversion.
  destruct c2.
    map_undefined_inversion.
    inversion hmaps; inversion hcmaps; subst.
    eapply is_nth_zip_eq; eauto.
Qed.

Lemma guarded_permission_acquire:
  forall h R Rlck Ravl l P,
  jh h R ->
  concrete_lck_maps h l FreeLock ->
  lck_maps R l (p2l P) ->
  star Rlck Ravl = R ->
  jpi Rlck zilch (guarded_permission h R) ->
  exists Rlckp Rlckr,
  star Rlckp Rlckr = Rlck /\
  jpi Rlckp zilch P /\
  jpi Rlckr zilch (guarded_permission (concrete_lck_update h l TakenLock) (lck_update R l (p2hl P))).
Proof.
  introv hjh hcm hm hs hg.
  forwards hjps: guarded_permission_acquire_jps; eauto. named_unpack hjps.
  forwards hjpi: jps_sound;
    [ eapply hjps
    | eapply hg
    | subst; eauto using jh_consistent with consistent_split
    | ].
  inversion hjpi; subst.
  eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The judgement [jconf c] means that the configuration [c] is well-typed. *)

(* [jconf] is the global invariant that holds throughout execution. It requires
   that the heap [h] and the term [t] be well-formed (i.e., [wkc (Cf h t)])
   and be well-typed under a common resource [R]. *)

(* Requiring the syntactic pre-condition to be [TyEmpty] is not restrictive
   thanks to the presence of the rule [JCut], which can be used at
   the bottom of the derivation of [jt _ _ TyEmpty t T]. *)

Inductive jconf : configuration -> Prop :=
| JConf:
    forall h t R Ravl Rlck,
    wkcp (Cf h t) ->
    star Ravl Rlck = R ->
    jh h R ->
    jpi Rlck zilch (guarded_permission h Ravl) ->
    jcp Ravl t ->
    jconf (Cf h t).

