Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Adopts.
Require Import Resources.
Require Import Row.
Require Import Intro.
Require Import Dup.
Require Import Signature.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JWeakening.
Require Import Harmless.

(* ---------------------------------------------------------------------------- *)

(* A typing judgement that uses an empty environment and/or the global signature
   of algebraic data type definitions must produce a closed type. Hence, this
   judgement must be stable under substitution. This argument is used when
   reasoning about the premises of some typing rules, such as [JBlock], [JFold],
   etc. *)

Ltac j_substitution_closed_tac := 
  intros;
  (* Argue that the type is closed. *)
  repeat erewrite closed_subst_invariant by
    eauto using wk_closed, length_empty, wk_zip_inverse with wk omega;
  (* The result follows. *)
  eauto.

Ltac j_substitution_closed_tac_in h :=
  intros;
  (* Argue that the type is closed. *)
  repeat erewrite closed_subst_invariant in h by
    eauto using wk_closed, length_empty, wk_zip_inverse with wk omega;
  (* The result follows. *)
  eauto.

Ltac j_substitution_closed_tac_hyp :=
  match goal with
  | h: context[fill _ _] |- _ => j_substitution_closed_tac_in h
  end.

Lemma jpi_zip_substitution:
  forall R K vs Ts w x,
  jpi R K (zip (subst w x vs) Ts) ->
  wk zilch vs (ROW VALUE) ->
  jpi R K (zip vs Ts).
Proof.
  intros. replace vs with (subst w x vs). assumption. j_substitution_closed_tac.
Qed.

Lemma wk_zilch_substitution:
  forall T kappa,
  wk zilch T kappa ->
  forall w x,
  wk zilch (subst w x T) kappa.
Proof.
  j_substitution_closed_tac.
Qed.

Lemma tc_tag_substitution:
  forall tc tag Ts w x,
  tc_tag tc tag Ts ->
  tc_tag tc tag (subst w x Ts).
Proof.
  j_substitution_closed_tac.
Qed.

Lemma tc_is_record_substitution:
  forall tc Ts w x,
  tc_is_record tc Ts ->
  tc_is_record tc (subst w x Ts).
Proof.
  j_substitution_closed_tac.
Qed.

Lemma tc_adopts_substitution:
  forall tc w x,
  tc_adopts tc = subst w x (tc_adopts tc).
Proof.
  j_substitution_closed_tac.
Qed.

Lemma switch_precondition_substitution:
  forall tag Ts U w x tc,
  wk zilch Ts (ROW TYPE) ->
  wk zilch U TYPE ->
  TyBlock (tc_mode tc) tag (subst w x Ts) (subst w x U) = TyBlock (tc_mode tc) tag Ts U.
Proof.
  j_substitution_closed_tac.
Qed.

Hint Resolve jpi_zip_substitution wk_zilch_substitution
tc_tag_substitution tc_is_record_substitution tc_adopts_substitution
switch_precondition_substitution : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* The judgement [adopts] is stable by substitution. *)

Lemma adopts_substitution:
  forall T U,
  adopts T U ->
  forall w x,
  adopts (subst w x T) (subst w x U).
Proof.
  induction 1; intros; subst; simpl_subst_goal;
  econstructor; eauto with j_substitution.
Qed.

Hint Resolve adopts_substitution : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* This undecipherable lemma appears as a leaf in the proof of the substitution
   lemma below, in the case of [JpsDecomposeBlock]. We prove it separately,
   because we do not want to allow [eauto] to search too far down. *)

Lemma undecipherable:
  forall x (T Ts : term),
  subst (TySingleton (TVar 0)) 0 (shift 1 (subst (shift 0 T) (1 + x) Ts)) =
  subst (shift 0 T) (1 + x) (subst (TySingleton (TVar 0)) 0 (shift 1 Ts)).
Proof.
  intros.
  subst_subst_hint.
  f_equal.
  lift_subst_hint.
  f_equal.
  eapply lift_one_lift_zero.
Qed.

(* This variant of [subst_subst] appears as a leaf in the proof of the
   substitution lemma below, in the case of [JpsRecomposeBlock]. We prove it
   separately, because we do not want to allow [eauto] to search too far
   down. *)

Lemma fishy:
  forall x (w T Ts : term),
  subst (TySingleton (subst T x w)) 0 (subst (shift 0 T) (1 + x) Ts) =
  subst T x (subst (TySingleton w) 0 Ts).
Proof.
  (* The reason why we can't just apply [subst_subst] is that
     [TySingleton] and [subst] must somehow commute. Besides,
     there is a problem with [rewrite] which we work around by
     using [symmetry] first. *)
  intros. symmetry. rewrite subst_subst by omega. reflexivity.
Qed.

Hint Resolve undecipherable fishy : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* A variant of the lemma [zip_substitution]. The equality is reversed (this
   is not really important, but it happens to be how the lemma is used) and
   the pathological cases where [zip vs Ts] evaluates to [TyFalse] are avoided
   by requiring a [jpi] hypothesis. This lemma is used in the case [JBlock] of
   the main substitution lemma below. *)

Lemma zip_substitution_variant:
  forall vs Ts R K w x,
  jpi R K (zip vs Ts) ->
  subst w x (zip vs Ts) = zip (subst w x vs) (subst w x Ts).
Proof.
  induction vs; simpl; try solve [ intros; jpi_inversion ];
  intro Ts; case_eq Ts; try solve [ intros; jpi_inversion ];
  intros; subst Ts; simpl_subst_goal; simpl.
  (* RowNil *)
  reflexivity.
  (* RowCons *)
  jpi_inversion. f_equal. eauto.
Qed.

Hint Resolve zip_substitution_variant : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are stable by substitution. *)

Lemma j_substitution:
(
  forall n K P Q,
  jps n K P Q ->
  forall w x kappa K' P' Q',
  insert x kappa K' = K ->
  subst w x P = P' ->
  subst w x Q = Q' ->
  wk K' w kappa ->
  substitutable kappa ->
  jps n K' P' Q'
) /\ (
  forall n K T1 T2,
  sub n K T1 T2 ->
  forall w x kappa K',
  insert x kappa K' = K ->
  wk K' w kappa ->
  substitutable kappa ->
  sub n K' (subst w x T1) (subst w x T2)
) /\ (
  forall n K Ts1 Ts2,
  row_sub n K Ts1 Ts2 ->
  forall w x kappa K',
  insert x kappa K' = K ->
  wk K' w kappa ->
  substitutable kappa ->
  row_sub n K' (subst w x Ts1) (subst w x Ts2)
) /\ (
  forall f n R K P t T,
  j f n R K P t T ->
  forall w x kappa K' P' t' T',
  insert x kappa K' = K ->
  subst w x P = P' ->
  subst w x t = t' ->
  subst w x T = T' ->
  wk K' w kappa ->
  substitutable kappa ->
  j f n R K' P' t' T'
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  wkbs Tss ->
  forall w x kappa K' P' ts' T',
  insert x kappa K' = K ->
  (
    forall tag Ts,
    wk zilch Ts (ROW TYPE) ->
    subst w x (P tag Ts) = P' tag Ts
  ) ->
  subst w x ts = ts' ->
  subst w x T = T' ->    
  wk K' w kappa ->
  substitutable kappa ->
  jbs f n R K' P' tag Tss ts' T'
) /\ (
  forall R K P,
  jpi R K P ->
  forall w x kappa K' P',
  insert x kappa K' = K ->
  subst w x P = P' ->
  wk K' w kappa ->
  substitutable kappa ->
  jpi R K' P'
).
Local Hint Extern 1
  (subst _ _ (TyAnchored _ _)  = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (subst _ _ (TyStar _ _)      = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (subst _ _ (TyBlock _ _ _ _) = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (subst _ _ (TyArrow _ _)     = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (TyAnchored _ _ = TyAnchored _ _)  => (f_equal).
Local Hint Extern 1
  (is_locked_permission _ (subst _ _ _)) => j_substitution_closed_tac.
Proof.
  apply the_great_mutind; intros; subst; simpl_subst_goal;
  try econstructor (solve [
    eauto 7 using harmless_lift, harmless_subst
            with insert_insert insert_concat lift_subst subst_subst j_substitution
  ]).
Qed.

(* The following tactic applies the substitution lemma (in forward mode)
   to a typing judgement [j] that appears among the hypotheses. *)

Ltac substitution :=
  let ht := fresh "ht" in
  forwards ht: (get4 j_substitution); eauto with substitutable;
  instantiate; repeat rewrite subst_lift in ht.

