Set Implicit Arguments.
Require Import MyTactics.
Require Import MyList.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Adopts.
Require Import Resources.
Require Import Row.
Require Import Intro.
Require Import Dup.
Require Import Signature.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import Harmless.

(* ---------------------------------------------------------------------------- *)

(* A typing judgement that uses an empty environment and/or the global signature
   of algebraic data type definitions must produce a closed type. Hence, this
   judgement must be stable under lifting. This argument is used when reasoning
   about the premises of some typing rules, such as [JBlock], [JFold], etc. *)

Ltac j_weakening_closed_tac :=
  intros;
  (* Argue that the type is closed. *)
  repeat erewrite closed_lift_invariant by
    eauto using wk_closed, length_empty, wk_zip_inverse with wk omega;
  (* The result follows. *)
  eauto.

Lemma jpi_zip_weakening:
  forall R K vs Ts x,
  jpi R K (zip (shift x vs) Ts) ->
  wk zilch vs (ROW VALUE) ->
  jpi R K (zip vs Ts).
Proof.
  intros. replace vs with (shift x vs). assumption. j_weakening_closed_tac.
Qed.

Lemma wk_zilch_weakening:
  forall T kappa,
  wk zilch T kappa ->
  forall x,
  wk zilch (shift x T) kappa.
Proof.
  j_weakening_closed_tac.
Qed.

Lemma tc_tag_weakening:
  forall tc tag Ts x,
  tc_tag tc tag Ts ->
  tc_tag tc tag (shift x Ts).
Proof.
  j_weakening_closed_tac.
Qed.

Lemma tc_is_record_weakening:
  forall tc Ts x,
  tc_is_record tc Ts ->
  tc_is_record tc (shift x Ts).
Proof.
  j_weakening_closed_tac.
Qed.

Lemma tc_adopts_weakening:
  forall tc x,
  tc_adopts tc = shift x (tc_adopts tc).
Proof.
  j_weakening_closed_tac.
Qed.

Lemma switch_precondition_weakening:
  forall tag Ts U x tc,
  wk zilch Ts (ROW TYPE) ->
  wk zilch U TYPE ->
  TyBlock (tc_mode tc) tag (shift x Ts) (shift x U) = TyBlock (tc_mode tc) tag Ts U.
Proof.
  j_weakening_closed_tac.
Qed.

Hint Resolve jpi_zip_weakening wk_zilch_weakening tc_tag_weakening
tc_is_record_weakening tc_adopts_weakening
switch_precondition_weakening : j_weakening.

(* ---------------------------------------------------------------------------- *)

(* The judgement [adopts] is stable by lifting. *)

Lemma adopts_weakening:
  forall T U,
  adopts T U ->
  forall x,
  adopts (shift x T) (shift x U).
Proof.
  induction 1; intros; subst; simpl_lift_goal;
  econstructor; eauto with j_weakening.
Qed.

Hint Resolve adopts_weakening : j_weakening.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are stable by lifting. *)

Lemma j_weakening:
(
  forall n K P1 P2,
  jps n K P1 P2 ->
  forall x okappa K' P1' P2',
  raw_insert x okappa K = K' ->
  shift x P1 = P1' ->
  shift x P2 = P2' ->
  jps n K' P1' P2'
) /\ (
  forall n K T1 T2,
  sub n K T1 T2 ->
  forall x okappa K',
  raw_insert x okappa K = K' ->
  sub n K' (shift x T1) (shift x T2)
) /\ (
  forall n K Ts1 Ts2,
  row_sub n K Ts1 Ts2 ->
  forall x okappa K',
  raw_insert x okappa K = K' ->
  row_sub n K' (shift x Ts1) (shift x Ts2)
) /\ (
  forall f n R K P t T,
  j f n R K P t T ->
  forall x okappa K' P' t' T',
  raw_insert x okappa K = K' ->
  shift x P = P' ->
  shift x t = t' ->
  shift x T = T' ->
  j f n R K' P' t' T'
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  wkbs Tss ->
  forall x okappa K' P' ts' T',
  raw_insert x okappa K = K' ->
  (
    forall tag Ts,
    wk zilch Ts (ROW TYPE) ->
    shift x (P tag Ts) = P' tag Ts
  ) ->
  shift x ts = ts' ->
  shift x T = T' ->    
  jbs f n R K' P' tag Tss ts' T'
) /\ (
  forall R K P,
  jpi R K P ->
  forall x okappa K' P',
  raw_insert x okappa K = K' ->
  shift x P = P' ->
  jpi R K' P'
).
Local Hint Extern 1
  (shift _ (TyAnchored _ _)  = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (shift _ (TyStar _ _)      = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (shift _ (TyBlock _ _ _ _) = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (shift _ (TyArrow _ _)     = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (TyAnchored _ _ = TyAnchored _ _) => f_equal.
Local Hint Extern 1
  (subst _ 0 _ = subst _ 0 _) => f_equal.
(* Local Hint Extern 1 *)
(*   (lck_maps _ _ (p2l (shift _ _))) => j_weakening_closed_tac. *)
(* Local Hint Extern 1 *)
(*   (lck_maps _ _ (p2hl (shift _ _))) => j_weakening_closed_tac. *)
Local Hint Extern 1
  (is_locked_permission _ (shift _ _)) => j_weakening_closed_tac.
Proof.
  apply the_great_mutind; intros; subst; simpl_lift_goal;
  try econstructor (solve [
    eauto using harmless_lift with lookup_insert insert_insert insert_concat lift_lift lift_subst j_weakening
  ]).
Qed.

Lemma jpi_zilch:
  forall R K P,
  jpi R zilch P ->
  jpi R K P.
Proof.
  introv hjz.
  induction K.
    auto.
    eapply (last6 j_weakening).
      eauto.
      instantiate (1 := a); instantiate (1 := 0). eauto.
      forwards: (last6 j_wk) hjz.
      erewrite closed_lift_invariant; eauto.
      eapply wk_closed; eauto using length_empty.
Qed.

