Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.

(* ------------------------------------------------------------------------- *)

(* The following axioms characterize a monotonic separation algebra. *)

Class MSA (resource : Type) := {

  (* Resources are combined using [star]. *)

  star: resource -> resource -> resource;

  (* [star] is commutative and associative. *)

  star_commutative:
    forall R1 R2,
    star R1 R2 = star R2 R1;

  star_associative:
    forall R1 R2 R3,
    star R1 (star R2 R3) = star (star R1 R2) R3;

  (* Because we require [star] to be a total function, we must allow for the
     possibility that some resources are inconsistent. An inconsistent
     resource is semantically equivalent to [false]. There may exist several
     syntactically distinct inconsistent resources. *)

  consistent: resource -> Prop;

  (* If a combination is consistent, then its parts are consistent as well. *)

  consistent_split_left:
    forall R1 R2,
    consistent (star R1 R2) ->
    consistent R1;

  (* Every resource [R] has a duplicable core. *)

  core: resource -> resource;

  (* The resource [core R] acts as a unit element for [R] with respect to
     [star]. That is, it is compatible with [R], and adds nothing to it. *)

  core_right_unit:
    forall R,
    star R (core R) = R;

  (* Two compatible resources have a common core. *)

  compatible_implies_common_core_left:
    forall R1 R2 R,
    star R1 R2 = R ->
    consistent R ->
    core R1 = core R;

  (* A duplicable resource is its own core. *)

  duplicable_implies_self_core:
    forall R,
    star R R = R ->
    R = core R;

  (* Every core is duplicable. *)

  cores_are_duplicable:
    forall R,
    star (core R) (core R) = core R;

  (* If a combination is consistent and duplicable, then its parts are
     duplicable as well. *)

  duplicable_split_left:
    forall R1 R2 R,
    star R1 R2 = R ->
    consistent R ->
    star R R = R ->
    star R1 R1 = R1;

  (* The information relation represents how the currently executing thread is
     allowed to modify some other (currently inactive) thread's resource.
     That is, the information relation describes the changes that can be
     imposed to one thread by some other thread of execution. *)

  rio: resource -> resource -> Prop;

  rio_reflexive:
    forall R,
    rio R R;

  (* Note that, although the relation [rio] is usually transitive, this property
     is not technically required anywhere, so we do not require it. *)

  (* [rio] preserves consistency. A consistent resource cannot be forced to
     evolve to an inconsistent one. *)

  (* Technically, one could work without this property, because the axiom
     [rio_star] splits such a move into two smaller moves whose destinations
     are also guaranteed to be consistent. In effect, if we did not require
     [rio] to preserve consistency, we would have to explicitly work within
     the intersection of [rio] and consistency. Life should be slightly
     simpler for our clients if we prove consistency up front. *)

  rio_consistent:
    forall R1 R2,
    consistent R1 ->
    rio R1 R2 ->
    consistent R2;

  (* The information relation is compatible with separating conjunction in
     the following sense: if we have a split and if the outside world requires
     that the sum grow, then each summand is able to grow to reflect this.
     Intuitively, the new duplicable information is simply sent down into
     both sides of the split. *)

  rio_star:
    forall R1a R1b R2,
    rio (star R1a R1b) R2 ->
    consistent (star R1a R1b) ->
    exists R2a R2b,
    star R2a R2b = R2 /\ rio R1a R2a /\ rio R1b R2b;

  rio_core:
    forall R1 R2,
    rio R1 R2 ->
    rio (core R1) (core R2)

}.

(* ------------------------------------------------------------------------- *)

(* An element [R] is duplicable if and only if it combines with itself to
   produce itself. *)

Notation duplicable R :=
  (star R R = R).

(* An element [R1] is maximal if and only if every consistent combination of
   it with some resource [R2] yields [R1] again. *)

Definition maximal `{MSA A} (R1 : A) :=
  forall R2 R, star R1 R2 = R -> consistent R -> R1 = R.

(* An element [R1] is stable if and only if it cannot evolve under [rio]. *)

Definition stable `{MSA A} (R1 : A) :=
  forall R2, rio R1 R2 -> R1 = R2.

(* ------------------------------------------------------------------------- *)

(* The following properties are valid in all monotonic separation algebras. *)

Section Properties.

Context `{MSA A}.

(* The symmetric counterpart of the axiom [star_associative]. *)

Lemma star_associative_symmetric:
  forall R1 R2 R3,
  star (star R1 R2) R3 = star R1 (star R2 R3).
Proof.
  symmetry. apply star_associative.
Qed.

(* The symmetric counterpart of the axiom [consistent_split_left]. *)

Lemma consistent_split_right:
  forall R1 R2,
  consistent (star R1 R2) ->
  consistent R2.
Proof.
  intros. eapply consistent_split_left. rewrite star_commutative. eassumption.
Qed.

(* A tautology that is useful when reasoning modulo AC is required. *)

Lemma consistent_by_equality:
  forall R1 R2,
  consistent R1 ->
  R1 = R2 ->
  consistent R2.
Proof.
  congruence.
Qed.

(* A special case of the above lemma. *)

Lemma consistent_star_commutative:
  forall R1 R2,
  consistent (star R1 R2) ->
  consistent (star R2 R1).
Proof.
  intros. rewrite star_commutative. assumption.
Qed.

(* A version of [consistent_split_left] that is useful when reasoning
   modulo AC is required. *)

Lemma consistent_by_subset:
  forall R1 R2 R3,
  consistent R1 ->
  R1 = star R2 R3 ->
  consistent R2.
Proof.
  eauto using consistent_split_left, consistent_by_equality.
Qed.

(* The symmetric counterpart of the axiom [core_right_unit]. *)

Lemma core_left_unit:
  forall R,
  star (core R) R = R.
Proof.
  intros. rewrite star_commutative. eapply core_right_unit.
Qed.

(* Variants of [core_left_unit] and [core_right_unit]. *)

Lemma core_right_unit_variant:
  forall R1 R2,
  core R1 = R2 ->
  star R1 R2 = R1.
Proof.
  intros; subst. eapply core_right_unit.
Qed.

Lemma core_left_unit_variant:
  forall R1 R2,
  core R1 = R2 ->
  star R2 R1 = R1.
Proof.
  intros; subst. eapply core_left_unit.
Qed.

(* The symmetric counterpart of the axiom [duplicable_split_left]. *)

Lemma duplicable_split_right:
  forall R1 R2 R,
  star R1 R2 = R ->
  consistent R ->
  duplicable R ->
  duplicable R2.
Proof.
  intros. eapply duplicable_split_left.
    rewrite star_commutative. eassumption.
    eassumption.
    eassumption.
Qed.

(* The symmetric counterparts of the axiom [compatible_implies_common_core_left]. *)

Lemma compatible_implies_common_core_right:
  forall R1 R2 R,
  star R1 R2 = R ->
  consistent R ->
  core R2 = core R.
Proof.
  intros. eapply compatible_implies_common_core_left.
    rewrite star_commutative. eassumption.
    eassumption.
Qed.

Lemma compatible_implies_common_core_symmetric:
  forall R1 R2 R,
  star R1 R2 = R ->
  consistent R ->
  core R1 = core R2.
Proof.
  intros.
  transitivity (core R). 
  eauto using compatible_implies_common_core_left.
  symmetry. eauto using compatible_implies_common_core_right.
Qed.

Lemma compatible_implies_common_core_both:
  forall R1 R2,
  consistent (star R1 R2) ->
  core R1 = core (star R1 R2) /\ core R2 = core (star R1 R2).
Proof.
  eauto using compatible_implies_common_core_left, compatible_implies_common_core_right.
Qed.

(* The symmetric counterpart of the axiom [duplicable_implies_self_core]. *)

Lemma duplicable_implies_self_core_symmetric:
  forall R,
  star R R = R ->
  core R = R.
Proof.
  symmetry. eauto using @duplicable_implies_self_core.
Qed.

(* Other ways of proving two cores equal. *)

Lemma simplify_core_core_left_left:
  forall R1 R2 R,
  core R1 = core R ->
  consistent (star R1 R2) ->
  core (star R1 R2) = core R.
Proof.
  intros.
  erewrite <- @compatible_implies_common_core_left by eauto.
  assumption.
Qed.

Lemma simplify_core_core_left_right:
  forall R1 R2 R,
  core R2 = core R ->
  consistent (star R1 R2) ->
  core (star R1 R2) = core R.
Proof.
  intros.
  rewrite @star_commutative.
  eapply simplify_core_core_left_left; eauto.
  rewrite @star_commutative.
  eauto.
Qed.

(* There are several ways to define what it means for a resource [R] to be
   duplicable. One is to state that two copies of [R] amount to one, that is,
   [star R R = R] holds. One other is to state that [R] is equal to its
   core. The two definitions are equivalent. One implication is an axiom
   ([duplicable_implies_self_core]) while the converse implication is derived
   below ([core_implies_duplicable]). A third way would be to state that [R]
   is duplicable if it is a core, that is, if there exists some resource [R0]
   such that [R = core R0]. Again, this definition is equivalent to the first
   one. This is another axiom ([cores_are_duplicable]). *)

Lemma core_implies_duplicable:
  forall R,
  R = core R ->
  duplicable R.
Proof.
  introv h. rewrite h at 2. apply core_right_unit.
Qed.

(* Core is idempotent. *)

Lemma core_idempotent:
  forall R,
  consistent R ->
  core (core R) = core R.
Proof.
  eauto using compatible_implies_common_core_left, core_left_unit.
Qed.

Lemma core_idempotent_symmetric:
  forall R,
  consistent R ->
  core R = core (core R).
Proof.
  symmetry. eauto using core_idempotent.
Qed.

(* A consistent resource has a consistent core. *)

Lemma consistent_core:
  forall R,
  consistent R ->
  consistent (core R).
Proof.
  intros.
  eapply consistent_split_left.
  rewrite core_left_unit.
  assumption.
Qed.

(* The core of a consistent combination is the combination of the cores. *)

Lemma core_star:
  forall R1 R2,
  consistent (star R1 R2) ->
  core (star R1 R2) = star (core R1) (core R2).
Proof.
  intros.
  rewrite <- cores_are_duplicable.
  symmetry. f_equal.
  eauto using compatible_implies_common_core_left.
  eauto using compatible_implies_common_core_right.
Qed.

Lemma star_core: (* symmetric version of the previous lemma *)
  forall R1 R2,
  consistent (star R1 R2) ->
  star (core R1) (core R2) = core (star R1 R2).
Proof.
  intros. symmetry. eauto using core_star.
Qed.

Lemma star_core_right: (* another formulation *)
  forall R1 R2,
  consistent (star R1 R2) ->
  star (core R1) (core R2) = core R2.
Proof.
  intros.
  rewrite star_core by eauto.
  symmetry. eapply compatible_implies_common_core_right; eauto.
Qed.

Lemma star_core_left: (* symmetric version of the previous lemma *)
  forall R1 R2,
  consistent (star R1 R2) ->
  star (core R1) (core R2) = core R1.
Proof.
  intros.
  rewrite star_core by eauto.
  symmetry. eapply compatible_implies_common_core_left; eauto.
Qed.

Lemma core_star_right: (* yet another formulation *)
  forall R1 R2,
  consistent (star R1 R2) ->
  core (star R1 R2) = core R2.
Proof.
  intros.
  rewrite core_star by eauto.
  eapply star_core_right. eauto.
Qed.

Lemma core_star_left: (* symmetric version of the previous lemma *)
  forall R1 R2,
  consistent (star R1 R2) ->
  core (star R1 R2) = core R1.
Proof.
  intros.
  rewrite core_star by eauto.
  eapply star_core_left. eauto.
Qed.

Lemma star_core_3:
  forall R1 R2 R3,
  consistent (star R1 (star R2 R3)) ->
  star (core R1) (star (core R2) (core R3)) = core (star R1 (star R2 R3)).
Proof.
  intros.
  rewrite star_core by eauto using consistent_split_right.
  eapply star_core. eauto.
Qed.

(* The following code is commented out, for the moment, so as to avoid
   an unnecessary dependency over [AAC]. *)

(* BEGIN COMMENTED OUT

Require Import AAC.

Instance sc : Commutative eq star.
Proof. unfold Commutative. eauto using @star_commutative. Qed.
Instance sa: Associative eq star.
Proof. unfold Associative. eauto using @star_associative. Qed.

Goal
  forall R1 R2 R,
  duplicable R ->
  star (star R1 R) (star R2 R) = (star (star R1 R2) R).
Proof.
  introv h. aac_rewrite h. aac_reflexivity.
Qed.

END COMMENTED OUT *)

End Properties.

Hint Resolve @consistent_split_left @consistent_split_right : consistent_split.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps exploit the property [rio_star]. It should work
   iteratively when there are nested stars. The database [rio_star_hints] is
   used to prove the consistency subgoal. *)

Hint Resolve @consistent_split_left @consistent_split_right : rio_star_hints.

Ltac rio_star :=
  repeat match goal with
  h: rio (star ?R1a ?R1b) ?R2 |- _ =>
    forwards: (rio_star _ _ _ h); [
      eauto with rio_star_hints
    | unpack; subst;
      (* Move [h] away, so we look for another instance of [rio_star]. *)
      revert h
    ]
  end;
  (* Re-introduce the hypotheses that were moved away. *)
  intros.

