Set Implicit Arguments.
Generalizable All Variables.
Require Import MyList.
Require Import MyTactics.
Require Import MSA.
Require Import MSAAllocationMaps.

(* The MSA of allocation maps has been defined in [MSAAllocationMaps]. We now
   define several operations on allocation maps, and prove a large number of
   properties of these operations. *)

Section Operations.

Variable A : Type.

Context `{MSA A}.

(* ------------------------------------------------------------------------- *)

(* The empty allocation map. *)

Definition empty_map :=
  Map (@nil A).

Lemma consistent_empty_map:
  consistent empty_map.
Proof.
  econstructor.
Qed.

Lemma duplicable_empty_map:
  duplicable empty_map.
Proof.
  econstructor.
Qed.

(* ------------------------------------------------------------------------- *)

(* A predicate over allocation maps: [everywhere]. *)

(* [everywhere P R] means that [R] is defined and maps every allocated address
   to a value that satisfies [P]. *)

Inductive everywhere (P : A -> Prop) : aMap A -> Prop :=
| Everywhere:
    forall m,
    Forall P m ->
    everywhere P (Map m).

Lemma everywhere_empty:
  forall P,
  everywhere P empty_map.
Proof.
  econstructor. econstructor.
Qed.

(* ------------------------------------------------------------------------- *)

(* A predicate over allocation maps: access. *)

(* The predicate [maps R r a] means that [R] is defined and maps the allocated
   address [r] to the value [a]. *)

Inductive maps : aMap A -> nat -> A -> Prop :=
| Maps:
    forall m r a,
    is_nth r m a ->
    maps (Map m) r a.

Hint Constructors maps : maps.

(* ------------------------------------------------------------------------- *)

(* A collection of lemmas that involve the above properties. *)

(* The empty map is empty. *)

Lemma maps_empty:
  forall r a,
  maps empty_map r a ->
  False.
Proof.
  unfold empty_map. dependent_destruction. my_list_cleanup.
Qed.

(* [maps] is injective. *)

Lemma maps_injective:
  forall R r a1 a2,
  maps R r a1 ->
  maps R r a2 ->
  a1 = a2.
Proof.
  inversion 1; inversion 1; eauto using is_nth_injective.
Qed.

(* If [R] satisfies [P] everywhere, and if [maps R r a] holds, then [P a] must
   hold. *)

Lemma everywhere_maps:
  forall (P : A -> Prop) R r a,
  everywhere P R ->
  maps R r a ->
  P a.
Proof.
  inversion 1; inversion 1; subst. eapply is_nth_Forall; eauto.
Qed.

(* A consistent resource maps every address to a consistent value. *)

Lemma maps_consistent:
  forall R r a,
  maps R r a ->
  consistent R ->
  consistent a.
Proof.
  inversion 1; dam. eapply is_nth_Forall; eauto.
Qed.

(* If [R] maps [r] to [a], then [core R] maps [r] to [core a]. *)

Lemma maps_core:
  forall R r a b,
  maps R r a ->
  core a = b ->
  maps (core R) r b.
Proof.
  inversion 1; dam. econstructor; eauto using is_nth_map.
Qed.

(* If [core R] maps [r] to [b], then [R] maps [r] to some [a] whose core is
   [b]. *)

Lemma maps_core_reverse:
  forall R r b,
  maps (core R) r b ->
  exists a,
  maps R r a /\ core a = b.
Proof.
  dependent_destruction; dam.
  forwards: is_nth_map_reverse. eauto. unpack.
  eauto with maps.
Qed.

(* A simple fact that can be occasionally useful. *)

Lemma maps_star_commutative:
  forall R1 R2 r a,
  maps (star R1 R2) r a ->
  maps (star R2 R1) r a.
Proof.
  intros. rewrite star_commutative. assumption.
Qed.

(* If [R1] maps [l] to [a1], [R2] maps [l] to [a2], and if [R1] and [R2] are
   compatible, then [R1 * R2] maps [l] to [a1 * a2]. This is a direct
   consequence of the definition of [star]. *)

Lemma maps_star_construction:
  forall R1 R2 l a1 a2,
  maps R1 l a1 ->
  maps R2 l a2 ->
  consistent (star R1 R2) ->
  maps (star R1 R2) l (star a1 a2).
Proof.
  do 2 dependent_destruction; dam; econstructor. eauto using is_nth_zip.
Qed.

(* This weaker variant does not require the consistency of [R1 * R2].
   It only requires [R1 * R2] to be defined at one arbitrary location. *)

Lemma maps_star_construction_weak:
  forall R1 R2 l a1 a2,
  maps R1 l a1 ->
  maps R2 l a2 ->
  forall m b,
  maps (star R1 R2) m b ->
  maps (star R1 R2) l (star a1 a2).
Proof.
  do 3 dependent_destruction; dam; econstructor. eauto using is_nth_zip.
Qed.

(* If [R1] maps [l] to [a1], and if [R1] and [R2] are compatible,
   then [R2] maps [l] to [a2], which is compatible with [a1]. *)

Lemma maps_star:
  forall R1 R2 l a1,
  maps R1 l a1 ->
  consistent (star R1 R2) ->
  exists a2,
  maps R2 l a2 /\ consistent (star a1 a2).
Proof.
  dependent_destruction. dam.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eexists. split.
    eauto with maps.
    eapply is_nth_Forall; eauto using is_nth_zip.
Qed.

(* This weaker variant does not require the consistency of [R1 * R2].
   It only requires [R1 * R2] to be defined at one arbitrary location.
   As a result, it does not guarantee the consistency of [a1 * a2]. *)

Lemma maps_star_weak:
  forall R1 R2 l a1,
  maps R1 l a1 ->
  forall m b,
  maps (star R1 R2) m b ->
  exists a2,
  maps R2 l a2.
Proof.
  do 2 dependent_destruction; dam.
  match goal with h: is_nth _ _ _ |- _ => clear h end.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eauto with maps.
Qed.

(* If [star R1 R2] maps [l] to [a], then [a] can be split into [a1],
   which is contributed by [R1], and [a2], which is contributed by
   [R2]. *)

Lemma maps_star_inversion:
  forall R1 R2 l a,
  maps (star R1 R2) l a ->
  exists a1 a2,
  maps R1 l a1 /\ maps R2 l a2 /\ star a1 a2 = a.
Proof.
  dependent_destruction; dam.
  forwards: length_is_nth.
    erewrite <- zip_length_right by eauto. eauto using is_nth_length.
  forwards: length_is_nth.
    erewrite <- zip_length_left by eauto. eauto using is_nth_length.
  unpack.
  repeat econstructor; eauto using is_nth_injective, is_nth_zip. (* cool *)
Qed.

(* If the address [r] exists in [R1] and this resource evolves along [rio] to
   [R2], then [r] still exists and its value has evolved along [rio], one
   level down. *)

Lemma maps_rio:
  forall R1 R2 r a1,
  maps R1 r a1 ->
  rio R1 R2 ->
  exists a2,
  maps R2 r a2 /\ rio a1 a2.
Proof.
  dependent_destruction; dam.
  forwards: growth_is_nth. eauto. eauto. unpack.
  eauto with maps.
Qed.

(* Conversely, if [R1] evolves along [rio] to [R2] and if [r] exists in [R2],
   then either it existed previously and its value has evolved along [rio],
   one level down, or [r] is a newly allocated address and its value is
   consistent and duplicable. *)

Lemma maps_rio_reverse:
  forall R1 R2 r a2,
  maps R2 r a2 ->
  rio R1 R2 ->
  (exists a1, maps R1 r a1 /\ rio a1 a2) \/
  (consistent a2 /\ duplicable a2).
Proof.
  dependent_destruction; dam.
  forwards [ ? | ? ]: growth_is_nth_reverse; [ eauto | eauto | unpack | unpack ];
  eauto with maps.
Qed.

(* If [R1] maps [l] to a maximal element [a], then [star R1 R2] must
   map [l] to the same element, provided it is consistent. *)

(* The tactic [maps_maximal], defined further on, helps use these two lemmas. *)

Lemma maps_maximal_left:
  forall R1 R2 l a,
  maps R1 l a ->
  consistent (star R1 R2) ->
  maximal a ->
  maps (star R1 R2) l a.
Proof.
  unfold maximal. dependent_destruction; introv ? hmax; dam. econstructor.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eapply is_nth_zip_eq; eauto.
  eapply hmax.
    eauto.
    eapply is_nth_Forall; eauto using is_nth_zip.
Qed.

Lemma maps_maximal_right:
  forall R1 R2 l a,
  maps R2 l a ->
  consistent (star R1 R2) ->
  maximal a ->
  maps (star R1 R2) l a.
Proof.
  intros.
  rewrite star_commutative.
  eapply maps_maximal_left; eauto.
  rewrite star_commutative.
  eauto.
Qed.

(* If [R1] maps [l] to a stable element [a], then and if [R1] evolves
   to [R2] via [rio], then [R2] maps [l] to [a]. *)

Lemma maps_stable:
  forall R1 R2 l a,
  maps R1 l a ->
  rio R1 R2 ->
  stable a ->
  maps R2 l a.
Proof.
  unfold stable. dependent_destruction; introv ? hstable; dam. econstructor.
  forwards: growth_is_nth; [ eauto | eauto | unpack ].
  erewrite hstable; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* An operation on allocation maps: allocation. *)

(* The allocation map [allocate R a] extends [R] with a newly allocated
   address, which is mapped to [a]. *)

Definition allocate (R : aMap A) (a : A) :=
  match R with
  | MapUndefined =>
      @MapUndefined _
  | Map m =>
      Map (m ++ a :: nil)
  end.

(* Allocation preserves consistency. *)

Lemma consistent_allocate:
  forall R a,
  consistent R ->
  consistent a ->
  consistent (allocate R a).
Proof.
  dam. eauto using Forall_app.
Qed.

(* [star] commutes with [allocate]. Each side of the split receives a partial
   permission for the new address. *)

Lemma star_allocate:
  forall R1 R2 a1 a2,
  star (allocate R1 a1) (allocate R2 a2) = allocate (star R1 R2) (star a1 a2).
Proof.
  dam; try solve [ do 2 rewrite app_length in *; simpl in *; false; omega ].
  eapply zip_app; eauto.
Qed.

(* A particular case, which shows that [allocate R a] can be decomposed as a
   [star]-combination of a suitably-extended [R] and a suitably-extended [a]. *)

Lemma star_allocate_core:
  forall R a,
  star (allocate R (core a)) (allocate (core R) a) = allocate R a.
Proof.
  intros.
  rewrite <- (core_right_unit R) at 3.
  rewrite <- (core_left_unit a) at 3.
  eapply star_allocate.
Qed.

(* A particular case, where the resource [a] is split into [a] and [core a].
   Roughly speaking, all of [a] is sent into one side. *)

Lemma star_allocate_core_left_unit:
  forall R1 R2 a,
  star (allocate R1 (core a)) (allocate R2 a) = allocate (star R1 R2) a.
Proof.
  intros.
  rewrite <- (core_left_unit a) at 3.
  eapply star_allocate.
Qed.

Lemma star_allocate_core_right_unit:
  forall R1 R2 a,
  star (allocate R1 a) (allocate R2 (core a)) = allocate (star R1 R2) a.
Proof.
  intros.
  rewrite <- (core_right_unit a) at 3.
  eapply star_allocate.
Qed.

(* The allocation of a new address, with a consistent and duplicable
   permission, is permitted by [rio]. *)

Lemma rio_allocate:
  forall R a,
  consistent a ->
  duplicable a ->
  rio R (allocate R a).
Proof.
  dam; eauto using growth_app, rio_reflexive.
Qed.

(* Allocation preserves [everywhere P] if the new value satisfies [P]. *)

Lemma everywhere_allocate:
  forall (P : A -> Prop) R a,
  everywhere P R ->
  P a ->
  everywhere P (allocate R a).
Proof.
  dependent_destruction; intros; econstructor; eauto using Forall_app.
Qed.

(* If [r] points to [a], then this fact remains true after allocating
   a new address. *)

Lemma maps_allocate:
  forall R r a b,
  maps R r a ->
  maps (allocate R b) r a.
Proof.
  dependent_destruction; econstructor. eauto using is_nth_app_near.
Qed.

(* If, in the heap [allocate R a], [r] points to [b], then either this
   fact was true in [R] already, or [b] is [a]. *)

Lemma maps_allocate_inversion:
  forall R r a b,
  maps (allocate R a) r b ->
  maps R r b \/ a = b.
Proof.
  destruct R; dependent_destruction.
  forwards [ | ]: is_nth_inversion; eauto with maps.
Qed.

(* ------------------------------------------------------------------------- *)

(* An operation on allocation maps: the next available memory location. *)

(* For convenience, we make this a total function. *)

Definition limit (R : aMap A) :=
  match R with
  | MapUndefined =>
      0
  | Map m =>
      length m
  end.

Lemma limit_split_left:
  forall R1 R2,
  consistent (star R1 R2) ->
  limit (star R1 R2) = limit R1.
Proof.
  dam. rewrite zip_length_left; eauto.
Qed.

Lemma limit_split_right:
  forall R1 R2,
  consistent (star R1 R2) ->
  limit (star R1 R2) = limit R2.
Proof.
  dam. rewrite zip_length_right; eauto.
Qed.

Lemma limit_core:
  forall R,
  limit (core R) = limit R.
Proof.
  dam. eapply map_length.
Qed.

Lemma simplify_limit_core_right:
  forall R l,
  l = limit R ->
  l = limit (core R).
Proof.
  intros. rewrite limit_core. assumption.
Qed.

Lemma same_core_same_limit:
  forall R1 R2,
  core R1 = core R2 ->
  limit R1 = limit R2.
Proof.
  intros.
  rewrite <- (limit_core R1).
  rewrite <- (limit_core R2).
  congruence.
Qed.

Lemma simplify_limit_limit_left_left:
  forall R1 R2 R,
  limit R1 = limit R ->
  consistent (star R1 R2) ->
  limit (star R1 R2) = limit R.
Proof.
  dam; my_list_cleanup; try tauto.
  rewrite zip_length_left; eauto.
Qed.

Lemma simplify_limit_limit_left_right:
  forall R1 R2 R,
  limit R2 = limit R ->
  consistent (star R1 R2) ->
  limit (star R1 R2) = limit R.
Proof.
  dam; my_list_cleanup; try tauto.
  rewrite zip_length_right; eauto.
Qed.

Lemma maps_allocate_there:
  forall R r a,
  consistent R ->
  r = limit R ->
  maps (allocate R a) r a.
Proof.
  dam. eauto using is_nth_app_far with maps is_nth.
Qed.

Lemma maps_allocate_elsewhere:
  forall R l a b,
  maps R l b ->
  maps (allocate R a) l b.
Proof.
  dependent_destruction. eauto using is_nth_app_near with maps.
Qed.

Lemma limit_rio:
  forall R1 R2,
  rio R1 R2 ->
  limit R1 <= limit R2.
Proof.
  dam; eauto using growth_length.
Qed.

Lemma limit_maps:
  forall R l,
  l < limit R ->
  consistent R ->
  exists a,
  maps R l a.
Proof.
  dam.
  forwards: length_is_nth. eauto. unpack. eauto with maps.
Qed.

Lemma maps_limit:
  forall R l a,
  maps R l a ->
  l < limit R.
Proof.
  dependent_destruction. eauto using is_nth_length.
Qed.

(* ------------------------------------------------------------------------- *)

(* An operation on allocation maps: update. *)

(* The allocation map [update R s a] coincides with [R], except at [s], which
   is now mapped to [a]. In principle, [s] should be an already-allocated
   address. *)

Definition update (R : aMap A) (s : nat) (a : A) :=
  match R with
  | MapUndefined =>
      @MapUndefined _
  | Map m =>
      Map (set_nth s m a)
  end.

(* Update preserves consistency. *)

Lemma consistent_update:
  forall R l a,
  consistent R ->
  consistent a ->
  consistent (update R l a).
Proof.
  dam. eauto using Forall_set_nth.
Qed.

(* A characterization of [update] in terms of [maps]. *)

Lemma maps_update_there:
  forall R r s a1 a2,
  maps R r a1 ->
  r = s ->
  maps (update R s a2) r a2.
Proof.
  dependent_destruction; econstructor; subst.
  eauto using is_nth_set_nth_there.
Qed.

Lemma maps_update_elsewhere:
  forall R r s a1 a2,
  maps R r a1 ->
  r <> s ->
  maps (update R s a2) r a1.
Proof.
  dependent_destruction; econstructor.
  eauto using is_nth_set_nth_elsewhere.
Qed.

Lemma maps_update_inversion:
  forall R r s a b,
  maps (update R r a) s b ->
  (r <> s /\ maps R s b) \/ (r = s /\ a = b).
Proof.
  destruct R; dependent_destruction.
  forwards [ | ]: is_nth_set_nth_inversion; [ eauto | left | right ]; unpack;
  eauto with maps.
Qed.

(* When the new value is the previous value, [update] is the identity. *)

Lemma update_identity:
  forall R s a,
  maps R s a ->
  update R s a = R.
Proof.
  dependent_destruction. f_equal. eauto using set_nth_identity.
Qed.

(* [star] commutes with [update]. Each side of the split receives a partial
   permission for the updated address. *)

Lemma star_update:
  forall R1 R2 s a1 a2 a,
  star a1 a2 = a ->
  star (update R1 s a1) (update R2 s a2) = update (star R1 R2) s a.
Proof.
  dam; try solve [ do 2 rewrite length_set_nth in *; simpl in *; false; omega ].
  eapply zip_set_nth.
Qed.

(* As a particular case of [star_update], if [R2] already maps [s] to the core
   of the new value [a], then [R2] does not change. *)

Lemma star_update_core_left:
  forall R1 R2 s a,
  maps R2 s (core a) ->
  star (update R1 s a) R2 = update (star R1 R2) s a.
Proof.
  introv hmaps.
  rewrite <- (update_identity hmaps) at 1.
  eapply star_update.
  eapply core_right_unit.
Qed.

Lemma star_update_core_right:
  forall R1 R2 s a,
  maps R1 s (core a) ->
  star R1 (update R2 s a) = update (star R1 R2) s a.
Proof.
  introv hmaps.
  rewrite <- (update_identity hmaps) at 1.
  eapply star_update.
  eapply core_left_unit.
Qed.

(* Update commutes with allocation. *)

Lemma update_allocate:
  forall R l a1,
  maps R l a1 ->
  forall b a2,
  update (allocate R b) l a2 = allocate (update R l a2) b.
Proof.
  dependent_destruction. dam.
  eauto using set_nth_app_left.
Qed.

(* The update of an existing address is permitted by [rio], provided
   the permission found at this address is updated in a manner that
   respects [rio], one level down. *)

Lemma rio_update:
  forall R s a1 a2,
  maps R s a1 ->
  rio a1 a2 ->
  rio R (update R s a2).
Proof.
  dependent_destruction. intros.
  eauto using growth_set_nth, rio_reflexive.
Qed.

(* Update at an existing address commutes with [rio]. *)

Lemma rio_update_update:
  forall R1 R2 s a a1 a2,
  maps R1 s a ->
  rio R1 R2 ->
  rio a1 a2 ->
  rio (update R1 s a1) (update R2 s a2).
Proof.
  dependent_destruction. dam.
  eauto using growth_set_nth_set_nth.
Qed.

(* Update preserves [everywhere P] if the new value satisfies [P]. *)

Lemma everywhere_update:
  forall (P : A -> Prop) R l a,
  everywhere P R ->
  P a ->
  everywhere P (update R l a).
Proof.
  dependent_destruction; econstructor. eauto using Forall_set_nth.
Qed.

(* Update commutes with itself. *)

Lemma update_update:
  forall R r1 a1 b1 r2 a2 b2,
  maps R r1 a1 ->
  maps R r2 a2 ->
  r1 <> r2 ->
  update (update R r1 b1) r2 b2 = update (update R r2 b2) r1 b1.
Proof.
  dependent_destruction. dependent_destruction. dam.
  eauto using set_nth_set_nth.
Qed.

Lemma update_overwrite:
  forall R r b1 b2,
  update (update R r b1) r b2 = update R r b2.
Proof.
  dam. eauto using set_nth_overwrite.
Qed.

Lemma update_reverse:
  forall R R' s a b,
  maps R s a ->
  update R s b = R' ->
  R = update R' s a.
Proof.
  intros; subst. symmetry. 
  rewrite update_overwrite.
  eauto using update_identity.
Qed.

(* ------------------------------------------------------------------------- *)

End Operations.

Section Agreement.

Context `{agrees : A -> B -> Prop}.

(* ------------------------------------------------------------------------- *)

(* Two maps agree if and only if they agree on the allocation limit and they
   agree pointwise at every allocated location. *)

Inductive aMap_agrees : aMap A -> aMap B -> Prop :=
| AMapAgrees:
    forall m1 m2,
    Forall2 agrees m1 m2 ->
    aMap_agrees (Map m1) (Map m2).

(* Some immediate consequences of the definition. *)

Lemma agrees_consistent:
  forall `{MSA B},
  (forall a b, agrees a b -> consistent b) ->
  forall R1 R2,
  aMap_agrees R1 R2 ->
  consistent R2.
Proof.
  inversion 2; subst.
  unfold consistent; simpl.
  eauto using Forall2_Forall_modus_ponens.
Qed.

Lemma agrees_maps_left_to_right:
  forall R1 R2 l a,
  aMap_agrees R1 R2 ->
  maps R1 l a ->
  exists b,
  maps R2 l b /\
  agrees a b.
Proof.
  do 2 dependent_destruction.
  forwards: Forall2_is_nth_left_to_right; [ eauto | eauto | unpack ].
  xplit. econstructor; eauto. eauto.
Qed.

Lemma agrees_maps_right_to_left:
  forall R1 R2 l b,
  aMap_agrees R1 R2 ->
  maps R2 l b ->
  exists a,
  maps R1 l a /\
  agrees a b.
Proof.
  do 2 dependent_destruction.
  forwards: Forall2_is_nth_right_to_left; [ eauto | eauto | unpack ].
  xplit. econstructor; eauto. eauto.
Qed.

Lemma agrees_maps_maps:
  forall R1 R2 l a b,
  aMap_agrees R1 R2 ->
  maps R1 l a ->
  maps R2 l b ->
  agrees a b.
Proof.
  do 3 dependent_destruction. eapply Forall2_is_nth; eauto.
Qed.

(* Two empty maps agree. *)

Lemma agrees_empty:
  aMap_agrees (empty_map A) (empty_map B).
Proof.
  repeat econstructor.
Qed.

(* Agreement implies equality of limits. *)

Lemma agrees_limit:
  forall R1 R2,
  aMap_agrees R1 R2 ->
  limit R1 = limit R2.
Proof.
  dependent_destruction. eauto using Forall2_length.
Qed.

(* Allocation preserves agreement. *)

Lemma agrees_allocate:
  forall R1 R2 a b,
  aMap_agrees R1 R2 ->
  agrees a b ->
  aMap_agrees (allocate R1 a) (allocate R2 b).
Proof.
  dependent_destruction; intros; econstructor.
  eapply Forall2_app; eauto. (* part of the standard library! *)
Qed.

(* Update preserves agreement. *)

Lemma agrees_update:
  forall R1 R2 l a b,
  aMap_agrees R1 R2 ->
  agrees a b ->
  aMap_agrees (update R1 l a) (update R2 l b).
Proof.
  dependent_destruction; intros; econstructor.
  eauto using Forall2_set_nth.
Qed.

End Agreement.

Implicit Arguments aMap_agrees [ A B ].

(* ------------------------------------------------------------------------- *)

(* The following tactic helps use the [maps_maximal_*] lemmas in a situation
   where we have a hypothesis of the form [maps R _ a] and a goal of the form
   [maps (..... R .....) _ a]. The idea is to repeatedly apply one of the
   lemmas until we get to a situation where the goal matches the hypothesis.
   The hints in [maps_maximal_hints] are supposed to be able to prove that [a]
   is maximal. The hints in [maps_maximal_reflexivity_hints] are supposed to
   be able to prove that the permissions [a] and [b] are equal. *)

Create HintDb maps_maximal_hints.
Create HintDb maps_maximal_reflexivity_hints.

Ltac maps_maximal :=
  instantiate;
  match goal with
  | h: maps ?R ?l ?a |- maps (star ?R1 ?R2) ?l ?b =>
    (* Force [a] and [b] to be the same element. If this fails, perhaps
       another hypothesis [h] will be tried; there might be several [maps]
       facts at hand. *)
    assert (a = b); [ solve [ eauto with maps_maximal_reflexivity_hints ] |
      (* Accumulate consistency hypotheses as we go down, in order to avoid
         a quadratic overhead. *)
      assert (consistent (star R1 R2)); [ solve [ eauto with consistent_split ] | ];
      (* Determine whether we should go left or right. *)
      match R1 with
      | context[R] =>
          (* [R] occurs inside [R1]: go left. *)
          eapply @maps_maximal_left; [ | eassumption | eauto with maps_maximal_hints ];
          maps_maximal
      | _ =>
          match R2 with
          | context[R] =>
              (* [R] occurs inside [R2]: go right. *)
              eapply @maps_maximal_right; [ | eassumption | eauto with maps_maximal_hints ];
              maps_maximal
          | _ =>
              idtac "maps_maximal: I don't know whether to go left or right."
          end
      end
    ]
  | _ =>
    eassumption (* we are done going down and can hopefully solve the goal *)
  end.

(* ---------------------------------------------------------------------------- *)

(* Tactics to help apply certain lemmas in forward mode. *)

Ltac maps_injective :=
  match goal with h1: maps ?R ?x _, h2: maps ?R ?x _ |- _ =>
    generalize (maps_injective h1 h2); intro
  end.

(* ------------------------------------------------------------------------- *)

(* The notions defined in this module should be opaque. I hope. *)

(* Global Opaque empty_map allocate limit update. *)

(* And the definition of the MSA of allocation maps should now be
   irrelevant. *)

(* Global Opaque msa_map. *)

