Set Implicit Arguments.
Generalizable All Variables.
Require Import MyList.
Require Import MyTactics.
Require Import MSA.

(* ------------------------------------------------------------------------- *)

(* An allocation map is a map of addresses to values, with support for dynamic
   allocation of fresh addresses. *)

(* An address is implemented as a natural integer. An allocation map is
   implemented as a finite list of values, which we interpret as a finite
   function of addresses to values. We refer to the length of this list
   as the ``allocation limit''. *)

(* Because the combination of two allocation maps is undefined when they
   disagree on the allocation limit, we also introduce an undefined map. *)

Inductive aMap A :=
| MapUndefined: aMap A
| Map: list A -> aMap A.

(* ------------------------------------------------------------------------- *)

(* Allocation maps over a monotonic separation algebra form a monotonic
   separation algebra. *)

(* The allocation limit must be the same on both sides of [star]. This is
   required so that when one side decides to allocate a fresh address, the new
   address really is globally fresh. Thus, in the definition of [star], we
   produce an inconsistent result unless both sides agree on the allocation
   limit. *)

(* When the allocation limits agree, the two maps are combined pointwise. *)

Definition aMap_star `{MSA A} R1 R2 :=
  match R1, R2 return _ with
  | MapUndefined, _
  | _, MapUndefined =>
      @MapUndefined _
  | Map m1, Map m2 =>
      if eq_nat_dec (length m1) (length m2) then
        Map (zip star m1 m2)
      else
        @MapUndefined _
  end.

(* The undefined map is of course inconsistent. A defined map is consistent if
   and only if every entry is consistent. *)

Definition aMap_consistent `{MSA A} R :=
  match R return _ with
  | MapUndefined =>
      False
  | Map m =>
      Forall consistent m
  end.

(* The function [core] preserves the allocation limit, and is defined
   pointwise. *)

Definition aMap_core `{MSA A} R :=
  match R return _ with
  | MapUndefined =>
      @MapUndefined _
  | Map m =>
      Map (map core m)
  end.

(* The relation [rio] allows new memory addresses to be allocated. At existing
   addresses, the old and new values must be related by the relation [rio]
   (one level down). At new addresses, the new value can be arbitrarily
   chosen, provided it is consistent and duplicable. The consistency
   requirement allows us to establish the property [rio_consistent]. The
   duplicability requirement means that a passive thread cannot be ``given''
   non-duplicable permissions. It is used in the theory of closed allocation
   maps. *)

Definition aMap_rio `{MSA A} (R1 R2 : aMap A) : Prop :=
  match R1, R2 return _ with
  | MapUndefined, MapUndefined =>
      True
  | Map m1, Map m2 =>
      growth rio (fun a => consistent a /\ duplicable a) m1 m2
  | _, _ =>
      False
  end.

(* ------------------------------------------------------------------------- *)

(* We now gather the above definitions in an instance declaration, and
   prepare to meet the proof obligations that ensue. *)

Program Instance msa_map `{MSA A} : MSA (aMap A) := {
  star := aMap_star;
  consistent := aMap_consistent;
  core := aMap_core;
  rio := aMap_rio
}.

(* ------------------------------------------------------------------------- *)

(* The tactic [dam] -- for ``destruct allocation maps'' -- does the bulk of
   the preparatory work. *)

Ltac dam :=
  unfold aMap_star, aMap_consistent, aMap_core, aMap_rio in *;
  intros;
  repeat match goal with R: aMap _ |- _ => destruct R end;
  simpl in *;
  dblib_by_cases;
  simpl in *;
  injections;
  try solve [ congruence | tauto ];
  try subst;
  try match goal with |- Map _ = Map _ => f_equal end.

(* ------------------------------------------------------------------------- *)

(* These ad hoc tactics select certain hypotheses and perform induction or
   destruction on them. *)

Ltac select_Forall :=
  match goal with h: Forall _ _ |- _ => revert h end.

Ltac select_growth :=
  match goal with h: growth _ _ _ _ |- _ => revert h end.

Ltac destruction_Forall :=
  select_Forall; dependent_destruction.

Ltac induction_growth :=
  select_growth; dependent_induction.

(* ------------------------------------------------------------------------- *)

(* These ad hoc lemmas are used below. *)

Lemma Forall_map_core_1:
  forall `{MSA A} (xs : list A),
  Forall consistent xs ->
  Forall (fun x => consistent x /\ duplicable x) (map core xs).
Proof.
  intros. eapply Forall_map; eauto using @consistent_core, @cores_are_duplicable.
Qed.

Lemma Forall_map_core_2:
  forall `{MSA A} (xs : list A),
  Forall (fun x => consistent x /\ duplicable x) xs ->
  Forall (fun x => consistent x /\ duplicable x) (map core xs).
Proof.
  intros. eapply Forall_map.
    eauto.
    simpl. intuition eauto using @consistent_core, @cores_are_duplicable.
Qed.

(* ------------------------------------------------------------------------- *)

(* This lemma corresponds to the proof obligation for [rio_star]. I state it
   explicitly because I wish to control its statement precisely. (Earlier, I
   was able to prove it on the fly, without stating a separate lemma, using
   [dependent induction]; this worked in Coq 8.3 but for some reason does not
   produce the desired induction hypothesis in Coq 8.4.) *)

Lemma aMap_rio_star:
  forall `{MSA A} (R1 R2 : list A),
  growth rio (fun a => consistent a /\ duplicable a) R1 R2 ->
  forall R1a R1b,
  zip star R1a R1b = R1 ->
  length R1a = length R1b ->
  Forall consistent R1 ->
  exists R2a R2b,
  length R2a = length R2b /\
  zip star R2a R2b = R2 /\
  growth rio (fun a => consistent a /\ duplicable a) R1a R2a /\
  growth rio (fun a => consistent a /\ duplicable a) R1b R2b.
Proof.
  induction 1; intros; my_list_cleanup.
  (* Case: GrowthZero *)
  (* The list [ys] represents the newly allocated memory locations.
     We distribute them in an arbitrary manner into the two sides.
     The lemma [core_right_unit] offers one way of doing so. *)
  match goal with ys: list _ |- _ =>
    exists ys (map core ys)
  end.
  repeat split.
    (* Obligation 1. *)
    symmetry. eapply map_length.
    (* Obligation 2. *)
    eauto using zip_map_right, @core_right_unit.
    (* Obligation 3. *)
    econstructor; eauto.
    (* Obligation 4. *)
    econstructor; eauto using Forall_map_core_2.
  (* Case: GrowthSucc *)
  (* Apply [rio_star] (one level down) for the head of the list. *)
  rio_star.
  (* Apply the induction hypothesis for the tail. *)
  forwards [ R2a [ R2b ? ]]: IHgrowth; [ eauto | eauto | eauto | clear IHgrowth; unpack ].
  (* Exhibit our witnesses and conclude. *)
  match goal with h: rio (star _ _) (star ?a ?b) |- _ =>
    exists (a :: R2a) (b :: R2b)
  end.
  repeat split; simpl; solve [ congruence | econstructor; eauto ].
Qed.

(* ------------------------------------------------------------------------- *)

(* Here come the proof obligations for the above instance declaration. *)

Next Obligation. (* star_commutative *)
  dam. eauto 2 using zip_commutative, @star_commutative.
Qed.

Next Obligation. (* star_associative *)
  dam; eauto 2 using zip_associative, @star_associative;
  (* The remaining cases are impossible. *)
  rewrite zip_length_right in * by eauto; false.
Qed.

Next Obligation. (* consistent_split_left *)
  dam. induction_lists; econstructor; eauto using @consistent_split_left.
Qed.

Next Obligation. (* core_right_unit *)
  dam; eauto 2 using zip_map_right, @core_right_unit;
  (* The remaining cases are impossible. *)
  rewrite map_length in *; false.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  dam. induction_lists; eauto 2 using @compatible_implies_common_core_left.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  dam. induction_lists; eauto 2 using @duplicable_implies_self_core.
Qed.

Next Obligation. (* cores_are_duplicable *)
  dam. induction_lists; eauto 2 using @cores_are_duplicable.
Qed.

Next Obligation. (* duplicable_split_left *)
  dam. induction_lists; eauto 2 using @duplicable_split_left.
Qed.

Next Obligation. (* rio_reflexive *)
  dam; eauto using growth_reflexive, @rio_reflexive.
Qed.

Next Obligation. (* rio_consistent *)
  dam. induction_growth; destruction_Forall; unpack; econstructor; eauto using @rio_consistent.
  (* A little bit of left-over manual work. *)
  eapply Forall_covariant. eauto. simpl. tauto.
Qed.

Next Obligation. (* rio_star *)
  dam.
  forwards [ m2a [ m2b ? ]]: aMap_rio_star; [ eauto | eauto | eauto | eauto | unpack ].
  exists (Map m2a) (Map m2b).
  dblib_by_cases.
  repeat split; congruence.
Qed.

Next Obligation. (* rio_core *)
  dam. induction_growth; econstructor; eauto using Forall_map_core_2, @rio_core.
Qed.

