Unset Implicit Arguments. (* intentional! *)
Generalizable All Variables.
Require Import MyProofIrrelevance.
Require Import MyTactics.
Require Import MSA.
Require Import MSAAllocationMaps.
Require Import MSAAllocationMapOperations.

(* Let [A] be a monotonic separation algebra. *)

(* Thus, [aMap A], the set of the allocation maps over [A], is also a
   monotonic separation algebra. Let us think of them as heap fragments. *)

(* In general, a heap fragment is not closed: that is, it may contain pointers
   that point outside of itself, also known as dangling pointers. *)

(* Nevertheless, it is sometimes interesting to restrict one's attention to
   heap fragments that have no dangling edges, for a certain notion of
   ``edge''.  Not every pointer needs be considered an edge; one is free to
   define a notion of ``edge'' that takes into account only certain pointers,
   such as the ``adopter'' pointers of Mezzo. *)
   
(* In this file, we prove that, under certain assumptions, closed heap
   fragments form a monotonic separation algebra. *)

(* ---------------------------------------------------------------------------- *)

(* For the sake of modularity, we formulate the required assumptions in an
   abstract manner. We need two predicates, [present] and [pointsto], which
   respectively characterize which addresses are in the domain of a heap
   fragment and which edges leave these addresses. Each of these predicates
   must satisfy certain properties, which we group in a type class. *)

(* ---------------------------------------------------------------------------- *)

Class Present A `{MSA A} := {

  (* The predicate [present] allows distinguishing a subset of ``present''
     elements of [A]. If [R] is an allocation map, we think of the addresses
     that are mapped to some ``present'' element as the ``domain'' of the heap
     fragment [R]. *)

  present : A -> Prop;

  (* If [x] is in the domain of [R1] or in the domain of [R2], then [x] must
     be in the domain of [R1 * R2]. *)

  (* Note that [x] could be in the domain of both [R1] and [R2], in which case
     the conjunction [R1 * R2] would be inconsistent. We do not wish to add
     the hypothesis [consistent (star a1 a2)] to the axiom below, because this
     would force us in turn to add a consistency hypothesis to the statement
     of the lemma [closed_star], and [star] would no longer be a total
     function over closed allocation maps. So, we omit the consistency
     hypothesis. In practice, this means that the permission [DX_ERROR] must
     be considered ``present''. *)

  present_star_construction:
    forall a1 a2,
    (present a1 \/ present a2) ->
    present (star a1 a2);

  (* In order to prove that [rio] preserves closedness, we assume that no
     pointer can be created or removed in a thread's local state while this
     thread is passive. As far as [present] is concerned, this is expressed by
     the following requirement. *)

  (* An address cannot disappear while we are asleep. *)

  present_rio:
    forall a b,
    rio a b ->
    present a ->
    present b

}.

(* ---------------------------------------------------------------------------- *)

Class Pointsto A `{MSA A} := {

  (* The predicate [pointsto] defines a relation between elements of [A] and
     addresses. We think of the assertion [pointsto a y] as the existence of a
     pointer from [a] to [y].  Thus, if the address [x] is mapped to some
     element [a] that points to the address [y], then there exists an edge in
     the heap from [x] to [y]. *)

  pointsto : A -> nat -> Prop;

  (* If there is an edge from [x] to [y] in [R1 * R2], then there must be such
     an edge either in [R1] or in [R2]. *)

  pointsto_star_inversion:
    forall a1 a2 y,
    pointsto (star a1 a2) y ->
    pointsto a1 y \/ pointsto a2 y;

  (* In order to prove that [rio] preserves closedness, we assume that no
     pointer can be created or removed in a thread's local state while this
     thread is passive. As far as [pointsto] is concerned, this is expressed
     by the following two requirements. *)

  (* An edge cannot appear at an existing address while we are asleep. *)

  pointsto_rio_reverse:
    forall a b y,
    rio a b ->
    pointsto b y ->
    pointsto a y;
    
  (* An edge cannot appear at a new address while we are asleep. *)

  pointsto_is_exclusive:
    forall a y,
    duplicable a ->
    pointsto a y ->
    False

}.

(* ---------------------------------------------------------------------------- *)

(* We now define closedness and prove that it is preserved by [star], [core],
   and [rio]. *)

Section Closedness.

  Context A `{M : MSA A, @Present A M, @Pointsto A M}.

  (* We may now define what it means for a heap fragment [R] to be closed.
     For every address [x] that is in the domain of [R], if there is an edge
     from [x] to some address [y], then [y] must be in the domain of [R].
     In other words, there are no dangling pointers. *)

  Definition domain R y :=
    exists b,
    maps R y b /\ present b.

  Definition closed R :=
    forall x a,
    maps R x a ->
    forall y,
    pointsto a y ->
    domain R y.

  (* Closedness is preserved by conjunction. *)

  (* Intuitively speaking, this is obvious: if there are no dangling pointers
     in [R1] or in [R2], then there are no dangling pointers in [R1 * R2]. *)

  Lemma closed_star:
    forall R1 R2,
    closed R1 ->
    closed R2 ->
    closed (star R1 R2).
  Proof.
    (* Assume there is an edge from [x] to [y] in [R1 * R2]. *)
    unfold closed; intros.
    (* Then, there must be such an edge in [R1] or in [R2]. *)
    forwards: maps_star_inversion; [ eauto | unpack; subst ].
    forwards [ ? | ? ]: pointsto_star_inversion; [ eauto | | ];
    (* The two sub-cases are symmetric. We treat them at once. *)
    (* Because [Ri] is closed, [y] must be in the domain of [Ri]. *)
    match goal with hclosed: forall x a, maps _ x a -> _ |- _ =>
      forwards: hclosed; [ eassumption | eassumption | unfold domain in *; unpack ]
    end;
    (* Thus, [y] must be in the domain of [R1 * R2]. *)
    (forwards: maps_star_weak; [ eauto | eauto using maps_star_commutative | unpack ]);
    solve [ eauto 7 using maps_star_construction_weak, present_star_construction ].
  Qed.

  (* In order to prove that [core] preserves closedness, it seems that we need
     to require either that [core] throws away all edges (in which case it may
     choose to keep or throw away the vertices, as it wishes) or that [core]
     preserves all vertices (in which case it may choose to keep or throw away
     the edges, as it wishes). We choose the first requirement, which
     corresponds to our intended application, where [core] throws away all
     vertices and all edges (the vertices are the exclusive objects and the
     edges are given by the adopter fields). We have already made a stronger
     requirement anyway: [pointsto_is_exclusive] implies that [core] throws
     away all edges. *)

  Lemma not_pointsto_core:
    forall a y,
    pointsto (core a) y ->
    False.
  Proof.
    intros.
    eapply pointsto_is_exclusive; [ | eassumption ].
    eapply cores_are_duplicable.
  Qed.

  (* Closedness is preserved by [core]. *)

  Lemma closed_core:
    forall R,
    closed R ->
    closed (core R).
  Proof.
    (* Assume there is an edge from [x] to [y] in [core R]. *)
    unfold closed; introv hclosed; intros.
    (* This is impossible! *)
    forwards: maps_core_reverse; [ eauto | unpack; subst ].
    false. eauto using not_pointsto_core.
  Qed.

  (* Closedness is preserved by [rio]. *)

  Lemma closed_rio:
    forall R1 R2,
    closed R1 ->
    rio R1 R2 ->
    closed R2.
  Proof.
    (* Assume there is an edge from [x] to [y] in [R2]. *)
    unfold closed; introv hclosed; intros.
    (* Either [x] was already present in [R1], or it is a newly allocated address. *)
    forwards h: maps_rio_reverse; [ eauto | eauto | ].
    destruct h; unpack.
    (* Sub-case: [x] was already present. *)
    (* Then, the edge from [x] to [y] was already present in [R1]. *)
    forwards: pointsto_rio_reverse; [ eauto | eauto | ].
    (* Because [R1] is closed, [y] was present in [R1]. *)
    forwards: hclosed; [ eauto | eauto | unfold domain in *; unpack ].
    (* Thus, [y] is still present in [R2]. *)
    forwards: maps_rio; [ eauto | eauto | unpack ].
    solve [ eauto using present_rio ].
    (* Sub_case: [x] is new. *)
    (* By definition of [rio], the permission associated with [x] in [R2]
       must be duplicable. Thus, there cannot be an edge from [x] to [y]
       in [R2]. This case is impossible. *)
    false.
    eapply pointsto_is_exclusive; eauto.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* We continue this section with a definition of closed allocation maps. *)

  (* One might think that it is possible to keep the existing type of allocation
     maps and define a new MSA structure for it, where the closedness property
     is built into the definition of the predicate [consistent]. This does not
     work. The MSA axioms require that consistency be preserved by arbitrary
     splits, yet closedness is obviously not preserved by arbitrary splits. *)

  (* Instead, we build closedness into a new type of heap fragments, using a
     subset type. Because closedness is preserved by union, we are able to prove
     that [star] is a total function. We are also able to prove that [core] is a
     total function. We define the relation [rio] as the restriction of the
     original [rio] to closed heap fragments. We prove that these definitions
     give rise to a monotonic separation algebra. *)
  
  (* Here is the type of closed allocation maps. *)

  Definition ClosedAMap :=
    { R : aMap A | closed R }.

  (* Closed allocation maps form a monotonic separation algebra. *)

  Definition ClosedAMap_star (R1 R2 : ClosedAMap) : ClosedAMap :=
    let (R1, hc1) := R1 in
    let (R2, hc2) := R2 in
    exist _ (star R1 R2) (closed_star _ _ hc1 hc2).

  Definition ClosedAMap_consistent (R : ClosedAMap) :=
    let (R, _) := R in
    consistent R.

  Definition ClosedAMap_core (R : ClosedAMap) : ClosedAMap :=
    let (R, hc) := R in
    exist _ (core R) (closed_core _ hc).

  Definition ClosedAMap_rio (R1 R2 : ClosedAMap) :=
    let (R1, _) := R1 in
    let (R2, _) := R2 in
    rio R1 R2.

  Global Program Instance msa_ClosedAMap : MSA ClosedAMap := {
    star := ClosedAMap_star;
    consistent := ClosedAMap_consistent;
    core := ClosedAMap_core;
    rio := ClosedAMap_rio
  }.

(* ---------------------------------------------------------------------------- *)

  (* Here are the proof obligations. *)

  Ltac dcam :=
    intros;
    repeat match goal with R: ClosedAMap |- _ => destruct R end;
    unfold ClosedAMap_star, ClosedAMap_consistent, ClosedAMap_core, ClosedAMap_rio in *;
    exist_exist_hypotheses;
    exist_exist_goal.

  Next Obligation.
    dcam. eauto using @star_commutative.
  Qed.
  
  Next Obligation.
    dcam. eauto using @star_associative.
  Qed.

  Next Obligation.
    dcam. eauto using @consistent_split_left.
  Qed.

  Next Obligation.
    dcam. eauto using @core_right_unit.
  Qed.

  Next Obligation.
    dcam. eauto using @compatible_implies_common_core_left.
  Qed.

  Next Obligation.
    dcam. eauto using @duplicable_implies_self_core.
  Qed.

  Next Obligation.
    dcam. eauto using @cores_are_duplicable.
  Qed.

  Next Obligation.
    dcam. eauto using @duplicable_split_left.
  Qed.

  Next Obligation.
    dcam. eauto using @rio_reflexive.
  Qed.

  Next Obligation.
    dcam. eauto using @rio_consistent.
  Qed.

  Next Obligation. (* rio_star *)
    dcam.
    (* Apply [rio_star] to the underlying allocation maps. *)
    rio_star.
    (* The new maps [R2a] and [R2b] are not a priori known to be closed. *)
    (*    Yet, because they evolve out of a closed map along [rio], they *)
    (*    must in fact be closed. *)
    match goal with h: rio _ (aMap_star ?R2a ?R2b) |- _ =>
      assert (cR2a: closed R2a); [ eauto using closed_rio |
      assert (cR2b: closed R2b); [ eauto using closed_rio |
    ]] end.
    (* This allows us to designate these new maps as the desired witnesses. *)
    exists (exist _ _ cR2a).
    exists (exist _ _ cR2b).
    xplit; exist_exist_goal; eauto.
  Qed.

  Next Obligation.
    dcam. eauto using @rio_core.
  Qed.

End Closedness.

Implicit Arguments domain [ A M H ].
Implicit Arguments closed [ A M H H0 ].

