Set Implicit Arguments.
Generalizable All Variables.
Require Import Relation_Definitions.
Require Import EquivDec.
Require Import MyTactics.
Require Import MSA.

Section Definitions.

Context `{EqDec A eq}.

Inductive l_permission :=
| L_LOCK: A -> l_permission
| L_HELD_LOCK: A -> l_permission
| L_ERROR: l_permission.

Obligation Tactic := idtac.

Global Program Instance msa_l_permission : MSA l_permission := {

  star p1 p2 :=
    match p1, p2 return _ with
    | L_ERROR, _
    | _, L_ERROR
    | L_HELD_LOCK _, L_HELD_LOCK _  =>
        L_ERROR
    | L_LOCK a1, L_LOCK a2 =>
        if equiv_dec a1 a2 then L_LOCK a1 else L_ERROR
    | L_HELD_LOCK a1, L_LOCK a2 =>
        if equiv_dec a1 a2 then L_HELD_LOCK a1 else L_ERROR
    | L_LOCK a1, L_HELD_LOCK a2 =>
        if equiv_dec a1 a2 then L_HELD_LOCK a1 else L_ERROR
    end;

  consistent p :=
    match p return _ with
    | L_ERROR =>
        False
    | _ =>
        True
    end;

  core p :=
    match p return _ with
    | L_ERROR =>
        L_ERROR
    | L_LOCK a =>
        L_LOCK a
    | L_HELD_LOCK a =>
        L_LOCK a
    end;

  rio p1 p2 := p1 = p2

}.

Ltac msa_l :=
  intros;
  repeat match goal with R: l_permission |- _ => destruct R end;
  simpl in *;
  equiv_dec_by_cases.

Next Obligation.
  msa_l.
Qed.

Next Obligation.
  msa_l.
Qed.

Next Obligation.
  msa_l.
Qed.

Next Obligation.
  msa_l.
Qed.

Next Obligation.
  msa_l; congruence.
Qed.

Next Obligation.
  msa_l; congruence.
Qed.

Next Obligation.
  msa_l.
Qed.

Next Obligation.
  msa_l; congruence.
Qed.

Next Obligation.
  msa_l.
Qed.

Next Obligation.
  msa_l; congruence.
Qed.

Next Obligation.
  intros.
  case_eq R1a; case_eq R1b; case_eq R2; intros;
    try solve [ exists R1a R1b; subst; equiv_dec_by_cases; eauto ].
Qed.

Next Obligation.
  msa_l; congruence.
Qed.

Lemma duplicable_lock:
  forall a,
  duplicable (L_LOCK a).
Proof.
  msa_l.
Qed.

Lemma not_duplicable_held_lock:
  forall a,
  ~ duplicable (L_HELD_LOCK a).
Proof.
  msa_l; congruence.
Qed.

Lemma consistent_lock:
  forall a,
  consistent (L_LOCK a).
Proof.
  msa_l.
Qed.

Lemma consistent_held_lock:
  forall a,
  consistent (L_HELD_LOCK a).
Proof.
  msa_l.
Qed.

Lemma maximal_held_lock:
  forall a,
  maximal (L_HELD_LOCK a).
Proof.
  unfold maximal. intros. msa_l; congruence.
Qed.

Lemma stable_lock:
  forall a,
  stable (L_LOCK a).
Proof.
  unfold stable. intros. msa_l.
Qed.

Lemma stable_held_lock:
  forall a,
  stable (L_HELD_LOCK a).
Proof.
  unfold stable. intros. msa_l.
Qed.

Lemma maximal_lock_error:
  maximal L_ERROR.
Proof.
  unfold maximal. msa_l.
Qed.

Lemma split_lock:
  forall { p1 p2 a },
  star p1 p2 = L_LOCK a ->
  p1 = L_LOCK a /\ p2 = L_LOCK a.
Proof.
  msa_l; congruence.
Qed.

Lemma split_held_lock:
  forall { p1 p2 a },
  star p1 p2 = L_HELD_LOCK a ->
  (p1 = L_HELD_LOCK a /\ p2 = L_LOCK a) \/
  (p1 = L_LOCK a /\ p2 = L_HELD_LOCK a).
Proof.
  msa_l; try congruence;
  match goal with
    | h: L_HELD_LOCK _ = L_HELD_LOCK _ |- _ =>
        inversion h; subst; auto
  end.
Qed.

Lemma consistent_star_held_lock:
  forall { b p },
  consistent (star (L_HELD_LOCK b) p) ->
  p = L_LOCK b.
Proof.
  msa_l.
Qed.

Lemma weak_maximal_lock:
  forall { b p },
  consistent (star (L_LOCK b) p) ->
  p = L_LOCK b \/ p = L_HELD_LOCK b.
Proof.
  msa_l.
Qed.

Lemma consistent_l_permission_inversion:
  forall p,
  consistent p ->
  exists b,
  p = L_LOCK b \/ p = L_HELD_LOCK b.
Proof.
  msa_l.
  eexists; left; eauto.
  eexists; right; eauto.
Qed.

End Definitions.

Implicit Arguments l_permission [ ].

Section Agreement.

Context `{A: Type, C: Type}.
Context `{l_agrees : C -> A -> Prop, hl_agrees : C -> A -> Prop}.
Context `{H : EqDec A eq}.

Inductive l_permission_agrees : C -> l_permission A -> Prop :=
| LPermissionAgreesLock:
    forall c a,
    l_agrees c a ->
    l_permission_agrees c (L_LOCK a)
| LPermissionAgreesHeldLock:
    forall c a,
    hl_agrees c a ->
    l_permission_agrees c (L_HELD_LOCK a).

Lemma l_permission_agrees_consistent:
  forall c p,
  l_permission_agrees c p ->
  consistent p.
Proof.
  inversion 1; subst; simpl; tauto.
Qed.

End Agreement.

Implicit Arguments l_permission_agrees [ A C ].