Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MSA.
Require Import MSAAllocationMaps.
Require Import MSAClosedAllocationMaps.

(* ------------------------------------------------------------------------- *)

(* The following axioms characterize a morphism of monotonic separation 
   algebras. *)

Class MSAMorphism `{MSA A, MSA B} (morph: A -> B) := {

  (* The morphism has to preserve the basic MSA operations. *)

  morph_star:
    forall R1 R2,
    morph (star R1 R2) = star (morph R1) (morph R2);

  morph_consistent:
    forall R,
    consistent R -> consistent (morph R);
  
  morph_core:
    forall R,
    morph (core R) = core (morph R);

  morph_rio:
    forall R1 R2,
    rio R1 R2 -> rio (morph R1) (morph R2)

  (* (* Some additional axioms. *) *)

  (* morph_maximal: *)
  (*   forall R, *)
  (*   maximal R -> maximal (morph R); *)

  (* morph_stable: *)
  (*   forall R, *)
  (*   stable R -> stable (morph R) *)

}.

Section MSAMorphismProperties.

Context `{MSAMorphism A B morph}.

Lemma morph_duplicable:
  forall R,
  duplicable R -> duplicable (morph R).
Proof.
  introv hd.
  rewrite <- morph_star.
  rewrite hd.
  reflexivity.
Qed.

End MSAMorphismProperties.


Class MSABiMorphism `{MSA A, MSA B} (bimorph: A -> B -> B) := {

  (* The bi-morphism has to preserve the basic MSA operations. *)

  bimorph_star:
    forall RA1 RA2 RB1 RB2,
    bimorph (star RA1 RA2) (star RB1 RB2)
    = star (bimorph RA1 RB1) (bimorph RA2 RB2);

  bimorph_consistent:
    forall RA RB,
    consistent RA ->
    consistent RB ->
    consistent (bimorph RA RB);
  
  bimorph_core:
    forall RA RB,
    bimorph (core RA) (core RB) = core (bimorph RA RB);

  bimorph_rio:
    forall RA1 RA2 RB1 RB2,
    rio RA1 RA2 ->
    rio RB1 RB2 ->
    rio (bimorph RA1 RB1) (bimorph RA2 RB2)

  (* (* Some additional axioms. *) *)

  (* bimorph_maximal: *)
  (*   forall RA RB, *)
  (*   maximal RA -> *)
  (*   maximal RB -> *)
  (*   maximal (bimorph RA RB); *)

  (* bimorph_stable: *)
  (*   forall RA RB, *)
  (*   stable RA -> *)
  (*   stable RB -> *)
  (*   stable (bimorph RA RB) *)

}.

Section MSABiMorphismProperties.

Context `{MSABiMorphism A B bimorph}.

Lemma bimorph_duplicable:
  forall RA RB,
  duplicable RA ->
  duplicable RB ->
  duplicable (bimorph RA RB).
Proof.
  introv hda hdb.
  rewrite <- bimorph_star.
  rewrite hda; rewrite hdb.
  reflexivity.
Qed.

End MSABiMorphismProperties.


(* Class MSASub `{MSA A, MSA B} *)
(*               (project: B -> A) *)
(*               (inject: A -> B -> B) *)
(*               (MSAP: MSAMorphism project) *)
(*               (MSAI: MSABiMorphism inject) *)
Class SubStructure `{A: Type, B: Type}
                    (project: B -> A)
                    (inject : A -> B -> B)
:= {

  pi_inverse:
    forall RA RB,
    project (inject RA RB) = RA;

  ip_inverse:
    forall RB,
    inject (project RB) RB = RB;

  inject_injective:
    forall  RA1 RA2 RB1 RB2,
    inject RA1 RB1 = inject RA2 RB2 ->
    RA1 = RA2;

  inject_erase:
    forall RA1 RA2 RB,
    inject RA2 (inject RA1 RB) = inject RA2 RB

}.

Section MSASubProperties.

Context `{SubStructure A B project inject}.
(* Context `{MSASub A B project inject}. *)

Lemma injectA_equality:
  forall RA1 RA2 RB,
  RA1 = RA2 ->
  inject RA1 RB = inject RA2 RB.
Proof.
  intros.
  subst.
  reflexivity.
Qed.

Lemma injectB_equality:
  forall RA RB1 RB2,
  RB1 = RB2 ->
  inject RA RB1 = inject RA RB2.
Proof.
  intros.
  subst.
  reflexivity.
Qed.

Lemma inject_idempotent:
  forall RA RB,
  inject RA (inject RA RB) = inject RA RB.
Proof.
  eauto using @inject_erase.
Qed.

Context (msa_a: MSA A)
        (msa_b: MSA B)
        (project_morphism: MSAMorphism project)
        (inject_bimorphism: MSABiMorphism inject).

Lemma project_maximal:
  forall RB,
  consistent RB ->
  maximal RB ->
  maximal (project RB).
Proof.
  unfold maximal.
  introv hcb hmb.
  intros RA2 RA hs hcr.
  forwards hb: (hmb (inject RA2 (core RB)) (inject RA RB)).
    rewrite <- (ip_inverse RB) at 1.
    rewrite <- bimorph_star.
    rewrite core_right_unit.
    subst; reflexivity.
  eauto using @bimorph_consistent.
  rewrite hb.
  eapply pi_inverse.
Qed.

Lemma project_stable:
  forall RB,
  stable RB ->
  stable (project RB).
Proof.
  unfold stable.
  introv hsb.
  intros RA2 hr.
  forwards hb: (hsb (inject RA2 RB)).
    rewrite <- (ip_inverse RB) at 1.
    eauto using @bimorph_rio, @rio_reflexive.
  rewrite hb.
  eapply pi_inverse.
Qed.

End MSASubProperties.


(* Section MSASubLift. *)

(* Context `{MSASub A B project inject}. *)

(* Definition lift_function_1 (f: A -> A): *)
(*   B -> B := *)
(*   fun (x: B) => inject (f (project x)) x. *)

(* Definition lift_function_2 `(f: A -> C -> A): *)
(*   B -> C -> B := *)
(*   fun (x: B) (c: C) => inject (f (project x) c) x. *)

(* Definition lift_function_3 `(f: A -> C -> D -> A): *)
(*   B -> C -> D -> B := *)
(*   fun (x: B) (c: C) (d: D) => inject (f (project x) c d) x. *)

(* Definition lift_function_3bis `(f: A -> forall c: C, ((forall c: C, D) -> A)): *)
(*   B -> forall c: C, ((forall c: C, D) -> B) := *)
(*   fun (x: B) (c: C) (d: C -> D)  => inject (f (project x) c d) x. *)

(* Check lift_function_3bis. *)
(* Check @lift_function_3. *)

(* Definition lift_function_4 `(f: A -> C -> D -> E -> A): *)
(*   B -> C -> D -> E -> B := *)
(*   fun (x: B) (c: C) (d: D) (e: E) => inject (f (project x) c d e) x. *)

(* End MSASubLift. *)

(* Section SubCAM. *)

(* Context `{MSA A}. *)
(* Context (pt: Pointsto A).   *)
(* Context (pr: Present A). *)

(* (* Goal MSA (aMap A). *) *)
(* (* eauto with typeclass_instances. *) *)
(* (* Qed. *) *)

(* (* Goal MSA (ClosedAMap A). *) *)
(* (* eauto with typeclass_instances. *) *)
(* (* Qed. *) *)

(* Definition content_project (R: ClosedAMap A) := *)
(*   let (R, hc) := R in R. *)

(* Program Instance content_project_morphism: MSAMorphism content_project. *)

(* Next Obligation. *)
(*   destruct R1, R2. auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   destruct R. auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   destruct R. auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   destruct R1, R2. auto. *)
(* Qed. *)

(* Check exist. *)

(* Program I *)

(* End SubCAM. *)

(* Section SubProduct. *)

(* Context `{A, B}. *)

(* Definition left_projection (c: A*B) := *)
(*   let (a, b) := c in a *)
(* . *)

(* Program Instance left_projection_morphism `{MSA A, MSA B}: *)
(*   MSAMorphism left_projection. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl. *)
(*   reflexivity. *)
(* Qed. *)

(* Next Obligation.  *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl in *. *)
(*   unpack. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl in *. *)
(*   unpack. *)
(*   auto. *)
(* Qed. *)


(* Definition left_injection (ia: A) (c: A*B) := *)
(*   let (a, b) := c in (ia, b) *)
(* . *)

(* Program Instance left_injection_bimorphism `{MSA A, MSA B}: *)
(*   MSABiMorphism left_injection. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl in *. *)
(*   unpack. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl in *. *)
(*   unpack. *)
(*   auto. *)
(* Qed. *)

(* Program Instance left_sub `{MSA A, MSA B}: *)
(*   MSASub left_projection_morphism left_injection_bimorphism. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   introv _ _ heq. *)
(*   decompose_pairs. *)
(*   simpl in *. *)
(*   inversion heq. *)
(*   auto. *)
(* Qed. *)

(* Next Obligation. *)
(*   intros. *)
(*   decompose_pairs. *)
(*   simpl in *. *)
(*   auto. *)
(* Qed. *)

(* End SubProduct. *)