Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MSA.
Require Import MSAMorphism.

(* ------------------------------------------------------------------------- *)

(* The product of two monotonic separation algebras forms a monotonic
   separation algebra. *)

Obligation Tactic := idtac.

Program Instance msa_pair `{MSA A, MSA B} : MSA (A * B) := {

  star R1 R2 :=
    let (RA1, RB1) := R1 in
    let (RA2, RB2) := R2 in
    (star RA1 RA2, star RB1 RB2);
  consistent R :=
    let (RA, RB) := R in
    consistent RA /\ consistent RB;
  core R :=
    let (RA, RB) := R in
    (core RA, core RB);
  rio R1 R2 :=
    let (RA1, RB1) := R1 in
    let (RA2, RB2) := R2 in
    rio RA1 RA2 /\ rio RB1 RB2

}.

Ltac decompose_pairs :=
  intros;
  repeat match goal with R: (_ * _)%type |- _ => destruct R end;
  repeat match goal with h: (_, _) = (_, _) |- _ => injection h; clear h; intros; try subst end;
  repeat match goal with h: _ /\ _ |- _ => destruct h end.

Ltac star_product :=
  match goal with
    | |- context[star (?a1, ?b1) (?a2, ?b2)] =>
        replace (star (a1, b1) (a2, b2)) with (star a1 a2, star b1 b2);
        [ | eauto ]
  end.

Ltac star_product_context :=
  match goal with
    | h: context[star (?a1, ?b1) (?a2, ?b2)] |- _ =>
        replace (star (a1, b1) (a2, b2)) with (star a1 a2, star b1 b2) in h;
        [ | eauto ]
  end.

Ltac consistent_product :=
  match goal with
    | |- context[consistent (?a, ?b)] =>
        replace (consistent (a, b)) with (consistent a /\ consistent b);
        [ | eauto ]
  end.

Ltac consistent_product_context :=
  match goal with
    | h: context[consistent (?a, ?b)] |- _ =>
        replace (consistent (a, b)) with (consistent a /\ consistent b) in h;
        [ | eauto ]
  end.

Ltac core_product :=
  match goal with
    | |- context[core (?a, ?b)] =>
        replace (core (a, b)) with (core a, core b);
        [ | eauto ]
  end.

Ltac core_product_context :=
  match goal with
    | h: context[core (?a, ?b)] |- _ =>
        replace (core (a, b)) with (core a, core b) in h;
        [ | eauto ]
  end.

Ltac rio_product :=
  match goal with
    | |- context[rio (?a1, ?b1) (?a2, ?b2)] =>
        replace (rio (a1, b1) (a2, b2)) with (rio a1 a2 /\ rio b1 b2);
        [ | eauto ]
  end.

Ltac rio_product_context :=
  match goal with
    | h: context[rio (?a1, ?b1) (?a2, ?b2)] |- _ =>
        replace (rio (a1, b1) (a2, b2)) with (rio a1 a2 /\ rio b1 b2) in h;
        [ | eauto ]
  end.

Next Obligation. (* star_commutative *)
  decompose_pairs. f_equal; eapply @star_commutative.
Qed.

Next Obligation. (* star_associative *)
  decompose_pairs. f_equal; eapply @star_associative.
Qed.

Next Obligation. (* consistent_split_left *)
  decompose_pairs. eauto using @consistent_split_left.
Qed.

Next Obligation. (* core_right_unit *)
  decompose_pairs. f_equal; eapply @core_right_unit.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  decompose_pairs. f_equal; eauto using @compatible_implies_common_core_left.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  decompose_pairs. f_equal; eauto using @duplicable_implies_self_core.
Qed.

Next Obligation. (* cores_are_duplicable *)
  decompose_pairs. f_equal; eauto using @cores_are_duplicable.
Qed.

Next Obligation. (* duplicable_split_left *)
  decompose_pairs. f_equal; eauto using @duplicable_split_left.
Qed.

Next Obligation. (* rio_reflexive *)
  decompose_pairs. eauto using @rio_reflexive.
Qed.

Next Obligation. (* rio_consistent *)
  decompose_pairs. eauto using @rio_consistent.
Qed.

Next Obligation. (* rio_star *)
  decompose_pairs.
  forwards [ RA2a [ RA2b ? ]]: @rio_star A; eauto.
  forwards [ RB2a [ RB2b ? ]]: @rio_star B; eauto.
  exists (RA2a, RB2a) (RA2b, RB2b).
  unpack. subst. tauto.
Qed.

Next Obligation. (* rio_core *)
  decompose_pairs. eauto using @rio_core.
Qed.

Lemma duplicable_product:
  forall `{MSA A, MSA B} (a: A) (b: B),
  duplicable a -> duplicable b -> duplicable (a, b).
Proof.
  introv da db.
  simpl.
  rewrite da; rewrite db.
  reflexivity.
Qed.

Lemma consistent_product:
  forall `{MSA A, MSA B} (a: A) (b: B),
  consistent a -> consistent b -> consistent (a, b).
Proof.
  introv da db.
  split; auto.
Qed.

Lemma consistent_product_drop_1:
  forall `{MSA A, MSA B} (a: A) (b: B),
  consistent (a, b) -> consistent a.
Proof.
  introv dab.
  consistent_product_context.
  unpack. auto.
Qed.

Lemma consistent_product_drop_2:
  forall `{MSA A, MSA B} (a: A) (b: B),
  consistent (a, b) -> consistent b.
Proof.
  introv dab.
  consistent_product_context.
  unpack. auto.
Qed.

Section SubProduct.

Context `{A: Type, B: Type}.

Definition left_projection (c: A*B) :=
  let (a, b) := c in a
.

Program Instance left_projection_morphism `{MSA A, MSA B}:
  MSAMorphism left_projection.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  reflexivity.
Qed.

Next Obligation. 
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.


Definition left_injection (ia: A) (c: A*B) :=
  let (a, b) := c in (ia, b)
.

Program Instance left_injection_bimorphism `{MSA A, MSA B}:
  MSABiMorphism left_injection.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.

Program Instance left_sub:
  SubStructure left_projection left_injection.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  introv heq.
  decompose_pairs.
  simpl in *.
  inversion heq.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  auto.
Qed.

Definition right_projection (c: A*B) :=
  let (a, b) := c in b
.

Program Instance right_projection_morphism `{MSA A, MSA B}:
  MSAMorphism right_projection.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  reflexivity.
Qed.

Next Obligation. 
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.


Definition right_injection (ib: B) (c: A*B) :=
  let (a, b) := c in (a, ib)
.

Program Instance right_injection_bimorphism `{MSA A, MSA B}:
  MSABiMorphism right_injection.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  unpack.
  auto.
Qed.

Program Instance right_sub:
  SubStructure right_projection right_injection.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl.
  auto.
Qed.

Next Obligation.
  introv heq.
  decompose_pairs.
  simpl in *.
  inversion heq.
  auto.
Qed.

Next Obligation.
  intros.
  decompose_pairs.
  simpl in *.
  auto.
Qed.

End SubProduct.