Set Implicit Arguments.
Require Import MyArithmetic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Reduction.
Require Import Exclusive.
Require Import Adopts.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JSubstitution.
Require Import JClassification.
Require Import JDecomposition.
Require Import SubsumptionSoundness.
Require Import JHeap.
Require Import JbsInversion.
Require Import WkInversion.

(* ---------------------------------------------------------------------------- *)

(* Progress. *)

(* A term is an answer iff it is a value or [fail]. *)

Inductive answer : term -> Prop :=
| AnswerValue:
    forall v,
    wk zilch v VALUE ->
    answer v
| AnswerFail:
    answer TFail.

(*

With the concurrent reduction semantics, a spawner is not a normal form
anymore. Remove it from the acceptable terms.

However, this notion is useful for the preliminary result.

Third thought: have a preliminary result for sequential programs, and includes
spawn in the final definition of an acceptable configuration.

*)

(* A term is a spawner iff it is a spawn operation in the focus of a
   deep evaluation context. *)

Inductive spawner : term -> Prop :=
| Spawner:
    forall D v1 v2,
    wk zilch v1 VALUE ->
    wk zilch v2 VALUE ->
    deep_eval_context D ->
    spawner (fill D (TSpawn v1 v2)).

(* A spawner in an evaluation context is still a spawner. *)

Lemma spawner_in_hole:
  forall E t,
  eval_context E ->
  spawner t ->
  spawner (fill E t).
Proof.
  introv he hs.
  inversion hs; subst.
  rewrite <- fill_nested_context by eauto using shallow_is_deep.
  eapply Spawner; eauto using DeepEvalContextInd.
Qed.

Lemma let_spawner:
  forall t x,
  wk (singleton VALUE) x PROG ->
  spawner t ->
  spawner (TLet t x).
Proof.
  introv hwk hsp.
  replace (TLet t x) with (fill (TLet (TVar 0) x) t).
    eauto using spawner_in_hole, tlet_eval_context.
    eauto using fill_eval_context, tlet_eval_context.
Qed.

(* A term [t] waits under a resource [h] if [t] tries to acquire/release a lock
   that is not in the expected state. *)

Inductive waits : term -> concrete_resource -> Prop :=
| WaitAcquire:
    forall D h l,
    deep_eval_context D ->
    concrete_lck_maps h l TakenLock ->
    waits (fill D (TAcquire (TLck l))) h.

Lemma waits_in_hole:
  forall E t h,
  eval_context E ->
  waits t h ->
  waits (fill E t) h.
Proof.
  introv he hw.
  inversion hw; subst.
  rewrite <- fill_nested_context by eauto using shallow_is_deep.
  eapply WaitAcquire; eauto using DeepEvalContextInd.
Qed.

(* A configuration is acceptable iff its term component is an answer or it
   is able to reduce. *)

Inductive seq_acceptable : configuration -> Prop :=
| SeqAcceptableAnswer:
    forall h a,
    answer a ->
    seq_acceptable (Cf h a)
| SeqAcceptableRed:
    forall c1 c2,
    red c1 c2 ->
    seq_acceptable c1
| SeqAcceptableSpawner:
    forall h s,
    spawner s ->
    seq_acceptable (Cf h s)
| SeqAcceptableWaits:
    forall h t,
    waits t h ->
    seq_acceptable (Cf h t)
.

Inductive answait: configuration -> Prop :=
| AnswaitValue:
    forall h v,
    wk zilch v VALUE ->
    answait (Cf h v)
| AnswaitFail:
    forall h,
    answait (Cf h TFail)
| AnswaitWaits:
    forall h t,
    waits t h ->
    answait (Cf h t)
| AnswaitThread:
    forall h t,
    answait (Cf h t) ->
    answait (Cf h (TThread t))
| AnswaitPar:
    forall h t1 t2,
    answait (Cf h t1) ->
    answait (Cf h t2) ->
    answait (Cf h (TPar t1 t2))
.

Inductive acceptable: configuration -> Prop :=
| AcceptableAnswait:
    forall c,
    answait c ->
    acceptable c
| AcceptableRed:
    forall c1 c2,
    red c1 c2 ->
    acceptable c1
.
(* TEMPORARY this definition of [acceptable] is not good: a configuration
   is judged acceptable as soon as at least one thread is able to make
   progress. This does not rule out the fact that other threads may crash.
   I am fixing this in reconstruction/. -fpottier *)

(* ---------------------------------------------------------------------------- *)

(* The pointer [y.adopter] is either equal to [x] or different from [x]. Thus,
   an abandon operation either succeeds or fails. *)

Lemma abandon_succeeds_or_fails:
  forall (p : pointer) x,
  p = NonNull x \/ p <> NonNull x.
Proof.
  intros. destruct p as [ | ya ].
    right. congruence.
    case (eq_nat_dec ya x); intro; subst.
      left. congruence.
      right. congruence.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Every well-typed configuration is acceptable. *)

(* As usual, because of the presence of [JExistsElim], the proof must be
   by induction over the height of the type derivation. *)

(* Because of the presence of [JForallIntro], which is not restricted
   to values, the statement must allow for a nonempty environment [K].
   Nevertheless, the term [t] itself is closed; this is stated
   separately. An alternative approach would be to keep a statement of
   progress under an empty environment, and in [JForallIntro], to use
   substitution in order to replace the type variable with an arbitrary
   type. This would require exploiting the fact that every kind is
   inhabited. I avoid this route because I do not wish to rely on this
   fact. *)

Lemma progress_preliminary:
  forall n R1t1 K P t T,
  jt n R1t1 K P t T -> (* [t] is well-typed *)
  wk zilch t PROG ->   (* [t] is closed *)
  forall R1t2 R1t,
  jpi R1t2 K P ->
  star R1t1 R1t2 = R1t ->
  forall R1f R1 h,
  jh h R1 ->
  star R1t R1f = R1 ->
  seq_acceptable (Cf h t).
(* The following tactic is common to the cases that deal with memory
   blocks. *)
Ltac progress_block_conclude :=
  eapply SeqAcceptableRed;
  econstructor;
  eauto using jh_heap_maps_right_to_left, agree_mo2dx, agree_Exclusive with SCAM_maps_maximal_heap.
Ltac progress_block :=
  classification;
  dca;
  progress_block_conclude.
Proof.
  intro n. induction_wf ih: lt_wf n.
  introv h hwk; dependent destruction h; intros; try subst; decomposition;
  try solve [
    (* If the term at hand is obviously a value, we are done. *)
    eapply SeqAcceptableAnswer; eapply AnswerValue;
      eauto 6 using reconfigure02, wk_empty_reverse with wk
  ];
  (* Introduce the fact that [R1] is consistent. *)
  (forwards: jh_consistent; [ eassumption | ]).
  (* JFrame *)
  solve [ eauto using reconfigure01 ].
  (* JCut *)
  solve [ eauto using reconfigure03 with j ].
  (* JSubLeft *)
  solve [ eauto 6 using jps_sound with consistent_split ].
  (* JSubRight *)
  (* Just discard this instance of [JSubRight] and work with what's under it. *)
  solve [ eauto ].
  (* JApplication *)
  (* We have a canonical type derivation for the term [t] on the left-hand
     side of the application. Thus, [t] must be a lambda-abstraction. *)
  classification.
  (* Now, either the term [u] on the right-hand side of the application
     is a value, in which case the whole application is a beta-redex;
     or [u] reduces, in which case the whole application reduces; or
     [u] is [fail], in which case the whole application reduces. *)
  (* Last case: [u] may be a [spawn] operation, which makes an acceptable
     configuration. *)
  forwards ha: ih; eauto using reconfigure02, wk_TApp_inversion.
  dependent destruction ha;
  try match goal with h: answer _ |- _ => inversion h; subst end;
  repeat match goal with c: configuration |- _ => destruct c end.
  (* The second and third cases could be simplified by taking into account
     the very simple form of the evaluation contexts. *)
     (* [u] is a value. *)
     eapply SeqAcceptableRed.
     solve [ eauto with red wk ].
     (* [u] is [fail]. *)
     eapply SeqAcceptableRed.
     erewrite <- fill_eval_context;
     solve [ eauto using tlet_eval_context,
                         wk_TAbs_inversion, wk_TApp_inversion_v with red ].
     (* [u] reduces *)
     eapply SeqAcceptableRed.
     erewrite <- fill_eval_context;
     solve [ eauto using tlet_eval_context,
                         wk_TAbs_inversion, wk_TApp_inversion_v with red ].
     (* [u] spawns *)
     eapply SeqAcceptableSpawner.
     erewrite <- fill_eval_context;
     solve [ eauto using spawner_in_hole, tlet_eval_context,
                         wk_TAbs_inversion, wk_TApp_inversion_v with red ].
     (* [u] waits *)
     eapply SeqAcceptableWaits.
     erewrite <- fill_eval_context;
     solve [ eauto using waits_in_hole, tlet_eval_context,
                         wk_TAbs_inversion, wk_TApp_inversion_v ].
  (* JFold *)
  solve [ eauto ].
  (* JNew *)
  eapply SeqAcceptableRed. solve [ eauto with red ].
  (* JRead *)
  progress_block.
  (* JWrite *)
  progress_block.
  (* JWriteTag *)
  progress_block.
  (* JSwitch *)
  classification.
  decomposition_TyCon.
  dca.
  jbs_inversion.
  progress_block_conclude.
  (* JGive *)
  classification.
  match goal with h1: jctop _ _ _ ?U, h2: adopts _ ?U |- _ =>
    unpack (decomposition_adoptee h1 h2); subst
  end.
  dca.
  progress_block_conclude.
  (* JTake *)
  (* We are looking at a term of the form [take y from x]. Because [y] has
     type [dynamic], it must be a valid memory location. *)
  classification.
  dca.
  (* This implies that there exists a memory block at address [y]. *)
  forwards: jh_heap_limit_maps_maps; eauto 2 with lt_lt SCAM_limit_split_heap. unpack.
  (* Now, either [y.adopter] is [x], or it isn't. In either case, one
     of the reduction rules associated with [take] applies. *)
  match goal with b: block |- seq_acceptable (Cf _ (TTake _ (TLoc ?x))) =>
    destruct b as [ ? adopter ? ];
    destruct (@abandon_succeeds_or_fails adopter x); [ subst | ]
  end;
  eapply SeqAcceptableRed; [ eapply RedTakeSuccess | eapply RedTakeFailure ];
  eauto using jh_heap_maps_right_to_left, agree_mo2dx, agree_Exclusive with SCAM_maps_maximal_heap.
  (* JFail *)
  eapply SeqAcceptableAnswer. eapply AnswerFail.
  (* JSpawn *)
  eapply SeqAcceptableSpawner.
  inversion hwk; subst.
  wk_inversion.
  match goal with
    | hw: wk zilch (TSpawn ?u1 ?u2) _,
      hu1: wk zilch ?u1 VALUE,
      hu2: wk zilch ?u2 VALUE |- _ =>
      forwards hsp: Spawner hu1 hu2
  end.
  eapply DeepEvalContextHole.
  simpl_subst_in hsp; auto.
  (* JNewLck *)
  eapply SeqAcceptableRed. solve [ eauto using RedNewLck ].
  (* JAcquire *)
  classification.
  dca.
  match goal with
    | h: jh _ _ |- _ => inversion h
  end.
  forwards hlck: (SAM_limit_maps concrete_lck_project).
    eapply eq_lt_lt_1.
      eauto using (SCAM_maps_limit lck_project).
      erewrite (@SCAM_agrees_limit _ _ concrete_lck_project _ _ _ _ _ lck_project lck_agree) by eauto;
      SCAM_limit_split lck_project_morphism.
  destruct hlck as (lck, hlck).
  destruct lck.
    eapply SeqAcceptableRed. eauto using RedAcquire.
    eapply SeqAcceptableWaits.
      inversion hwk; subst.
      wk_inversion.
      forwards hw: WaitAcquire; eauto using DeepEvalContextHole.
  (* JRelease *)
  classification.
  dca.
  match goal with
    | h: jh _ _ |- _ => inversion h
  end.
  forwards hlck: (SAM_limit_maps concrete_lck_project).
    eapply eq_lt_lt_1.
      eauto using (SCAM_maps_limit lck_project).
      erewrite (@SCAM_agrees_limit _ _ concrete_lck_project _ _ _ _ _ lck_project lck_agree) by eauto;
      SCAM_limit_split lck_project_morphism.
  destruct hlck as (lck, hlck).
  destruct lck.
    forwards: (@SCAM_agrees_maps_maps _ _ concrete_lck_project _ _ _ _ _ lck_project).
      eauto.
      eauto.
      SCAM_maps_maximal lck_project_morphism.
      lck_agree_inversion.
    eapply SeqAcceptableRed. eauto using RedRelease.
  (* JExistsElim *)
  eauto using SeqAcceptableRed, RedUnpackIn.
  (* JForallIntro *)
  jpi_inversion.
  (* The term [t] is closed, so [shift 0 t] is [t]. *)
  match goal with h: jt _ _ _ _ (shift 0 _) _ |- _ =>
    erewrite closed_lift_invariant in h by eauto using wk_closed, length_empty
  end.
  (* The result follows by the induction hypothesis. *)
  eauto.
Qed.

(* Weakening lemmas for [jcp]. *)

Lemma jcp_star_left:
  forall R Q t,
  jcp R t ->
  jcp (star R Q) t.
Proof.
  induction 1.
  (* Case Thread. *)
  eapply JcpThread.
  auto.
  apply JCut with Q R TyEmpty;
    [ eauto with j
    | eauto with jpi
    | eauto using @star_commutative].
  (* Case Par. *)
  subst.
  apply JcpPar with R1 (star R2 Q);
  solve [auto using @star_associative].
Qed.

Lemma jcp_star_right:
  forall R Q t,
  jcp Q t ->
  jcp (star R Q) t.
Proof.
  intros.
  rewrite star_commutative.
  auto using jcp_star_left.
Qed.

(* Every well-typed concurrent configuration is acceptable. *)

Lemma progress:
  forall c,
  jconf c ->
  acceptable c.
Proof.
  introv Hjconf.
  inversion Hjconf as [ h t R Ravl Rlck Hwkcp Hr Hjh Hlck Hjcp ]; subst.
  inversion Hwkcp as [ h' t' Hwkh Hwk ]; subst.
  dependent induction Hwk.
  (* Case [TVar x] where [x] denotes a term of king [CPROG]. *)
  (* This case is discharged since it assumes a mapping for [x] in an empty
     environment. *)
    false; eauto using lookup_empty_Some.
  (* Case [TThread t]. *)
  (* We prove that the sequential program [t] is acceptable by invoking the
     sequential progress lemma. *)
    clear IHHwk.
    inversion Hjcp; subst.
    forwards HSeqAcc: progress_preliminary.
      make_jt_hypothesis hjt. eauto.
      eauto.
      eapply JpiEmpty.
      eauto.
      eauto.
      eapply @core_right_unit.
    inversion HSeqAcc; subst.
      inversion H1; subst.
        econstructor. eapply AnswaitThread. econstructor. auto.
        econstructor. eapply AnswaitThread. eapply AnswaitFail.
      destruct c2. eapply AcceptableRed. eauto with red.
      inversion H1; subst.
        eapply AcceptableRed. eauto with red.
        eapply AcceptableAnswait. eapply AnswaitThread. eapply AnswaitWaits. auto.
  (* Case [TPar t1 t2]. *)
  (* This case is reduced to the conjunction of the induction hypotheses. *)
    inversion Hjcp; subst.
    forwards HAcc1: IHHwk1; eauto using JConf, jcp_star_left with wk.
    forwards HAcc2: IHHwk2; eauto using JConf, jcp_star_right with wk.
    inversion HAcc1; subst.
      inversion HAcc2; subst.
        eapply AcceptableAnswait. eapply AnswaitPar; eauto.
        destruct c2. eapply AcceptableRed. eauto with red.
      destruct c2. eapply AcceptableRed. eauto with red.
Qed.

(*

(* ---------------------------------------------------------------------------- *)

(* Progress. A clean re-statement. *)

Lemma progress:
  forall c,
  jconf c ->
  acceptable c.
Proof.
  inversion 1; subst.
  eapply progress_preliminary; eauto with j wk.
  rewrite @core_right_unit. reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

*)