Set Implicit Arguments.
Require Import Relation_Operators.
Require Import EquivDec.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Row.
Require Import WkInversion.
Require Import MSAAllocationMaps.
Require Import MSASAMOperations.

(* ---------------------------------------------------------------------------- *)

(* The small-step, call-by-value reduction semantics. *)

Inductive red : configuration -> configuration -> Prop :=

(* The usual [beta-v]. Reduction is permitted only if the argument is a
   value. The distinction between values and terms is performed by the
   well-kindedness judgement. *)

  | RedBeta:
      forall h t v t',
      wk zilch v VALUE ->
      subst v 0 t = t' ->
      red
        (Cf h (TApp (TAbs t) v))
        (Cf h t')

(* Allocation of a new memory block. *)

  | RedNew:
      forall h tag vs,
      red
        (Cf h (TNew tag vs))
        (Cf (concrete_heap_allocate h (Block tag Null vs)) (TLoc (concrete_heap_limit h)))

(* Reading a field in a memory block. *)

  | RedRead:
      forall h field l tag adopter v vs,
      row_context field vs ->
      concrete_heap_maps h l (Block tag adopter (fill_row vs v)) ->
      red
        (Cf h (TRead (TLoc l) field))
        (Cf h v)

(* Writing a field in a memory block. This operation produces a unit result. *)

  | RedWrite:
      forall h field l tag adopter v v' vs,
      row_context field vs ->
      concrete_heap_maps h l (Block tag adopter (fill_row vs v)) ->
      red
        (Cf h (TWrite (TLoc l) field v'))
        (Cf (concrete_heap_update h l (Block tag adopter (fill_row vs v'))) TUnit)

(* Writing a tag in a memory block. This operation produces a unit result. *)

  | RedWriteTag:
      forall h l tag tag' adopter vs,
      concrete_heap_maps h l (Block tag adopter vs) ->
      red
        (Cf h (TWriteTag (TLoc l) tag'))
        (Cf (concrete_heap_update h l (Block tag' adopter vs)) TUnit)

(* Dispatch over a tag. *)

  | RedSwitch:
      forall h l tag adopter vs t ts,
      row_context tag ts ->
      concrete_heap_maps h l (Block tag adopter vs) ->
      red
        (Cf h (TSwitch (TLoc l) (fill_row ts t)))
        (Cf h t)

(* Adoption. *)

  | RedGive:
      forall h lx ly tag vs,
      concrete_heap_maps h ly (Block tag Null vs) ->
      red
        (Cf h (TGive (TLoc ly) (TLoc lx)))
        (Cf (concrete_heap_update h ly (Block tag (NonNull lx) vs)) TUnit)

(* Abandon. *)

  | RedTakeSuccess:
      forall h lx ly tag vs,
      concrete_heap_maps h ly (Block tag (NonNull lx) vs) ->
      red
        (Cf h (TTake (TLoc ly) (TLoc lx)))
        (Cf (concrete_heap_update h ly (Block tag Null vs)) TUnit)

  | RedTakeFailure:
      forall h lx ly tag adopter vs,
      concrete_heap_maps h ly (Block tag adopter vs) ->
      adopter <> NonNull lx ->
      red
        (Cf h (TTake (TLoc ly) (TLoc lx)))
        (Cf h TFail)

(* Unpack. *)

  | RedUnpackIn:
      forall h t,
      red (Cf h (TUnpackIn t)) (Cf h t)

(* Reduction under a context. *)

  | RedContext:
      forall h h' E t t',
      eval_context E ->
      red (Cf h t) (Cf h' t') ->
      red (Cf h (fill E t)) (Cf h' (fill E t'))

  | RedContextFail:
      forall h E,
      eval_context E ->
      red
        (Cf h (fill E TFail))
        (Cf h TFail)

(* Reduction in a thread pool. *)

  | RedThread:
      forall h h' t t',
      red (Cf h t) (Cf h' t') ->
      red (Cf h (TThread t)) (Cf h' (TThread t'))

  | RedParLeft:
      forall h h' t1 t1' t2,
      red (Cf h t1) (Cf h' t1') ->
      red (Cf h (TPar t1 t2)) (Cf h' (TPar t1' t2))

  | RedParRight:
      forall h h' t1 t2 t2',
      red (Cf h t2) (Cf h' t2') ->
      red (Cf h (TPar t1 t2)) (Cf h' (TPar t1 t2'))

(* Spawning a new thread. *)

  | RedSpawn:
      forall h D v1 v2,
      wk zilch v1 VALUE ->
      wk zilch v2 VALUE ->
      deep_eval_context D ->
      red
        (Cf h (TThread (fill D (TSpawn v1 v2))))
        (Cf h (TPar (TThread (fill D TUnit)) (TThread (TApp v1 v2))))

(* Creating a new lock. *)

  | RedNewLck:
      forall h,
      red
        (Cf h TNewLck)
        (Cf (concrete_lck_allocate h FreeLock) (TLck (concrete_lck_limit h)))

(* Acquiring / Releasing a lock. *)

  | RedAcquire:
      forall h l,
      concrete_lck_maps h l FreeLock ->
      red
        (Cf h (TAcquire (TLck l)))
        (Cf (concrete_lck_update h l TakenLock) TUnit)

  | RedRelease:
      forall h l,
      concrete_lck_maps h l TakenLock ->
      red
        (Cf h (TRelease (TLck l)))
        (Cf (concrete_lck_update h l FreeLock) TUnit)

.

Hint Constructors red : red.

(* ---------------------------------------------------------------------------- *)

(* Basic properties of the reduction relation. *)

(* Reduction preserves well-kindedness. *)

(* For a long time, this property was not explicitly used; it was in part
   implicitly re-proved as part of the subject reduction statement. It is
   now explicitly exploited in the subject reduction proof, in the case of
   [JForallIntro]. Anyway, it is good to check up front that it holds. *)

Lemma red_wk:
  forall c1 c2,
  red c1 c2 ->
  wkc c1 ->
  wkc c2.
Proof.
  introv hred; dependent induction hred; subst;
  intro hwkc; dependent destruction hwkc;
  try open_eval_context;
  try solve [
    match goal with h: wk _ _ _ |- _ => revert h end;
    intro hwk; dependent induction hwk; wk_inversion;
    try solve [
          eauto using wk_substitution with wk insert_concat
        |
          forwards: IHhred; [ eauto with wk | wk_inversion; eauto with wk ]
        |
          match goal with
              |- context[SAM_update _ _ _ _ _] =>
                eauto 16 using SAM_everywhere_update, concrete_heap_sub_resource, concrete_lck_sub_resource
                         with wk heap_lck_lift unfold_concrete_sub_resource
          end
        |
          match goal with
              |- context[SAM_allocate _ _ _ _] =>
                eauto 13 using SAM_everywhere_allocate, concrete_heap_sub_resource, concrete_lck_sub_resource
                         with wk heap_lck_lift unfold_concrete_sub_resource
          end
        ]
      |
        match goal with h: wk _ _ _ |- _ => revert h end;
        intro hwk; dependent destruction hwk; wk_inversion;
        eauto 13 using SAM_everywhere_update, concrete_heap_sub_resource, wk_substitution
                 with wk heap_lck_lift unfold_concrete_sub_resource
      ].
Qed.

(* Reduction of concurrent programs preserve well-kindedness. *)

Lemma red_wkcp:
  forall c1 c2,
  red c1 c2 ->
  wkcp c1 ->
  wkcp c2.
Proof.
  introv hred; dependent induction hred; subst;
  intro hwkc; dependent destruction hwkc;
  wk_inversion.
  (* Two impossible cases. *)
  open_eval_context. wk_inversion.
  open_eval_context. wk_inversion.
  (* Reduction in a thread. *)
  forwards: red_wk; [ eauto | eauto with wk | wk_inversion; eauto with wk ].
  (* Reduction on the left or right of a [TPar]. *)
  forwards: IHhred; [ eauto with wk | wk_inversion; eauto with wk ].
  forwards: IHhred; [ eauto with wk | wk_inversion; eauto with wk ].
  (* Spawn. *)
  eauto 10 using wk_fill_deep_eval_context with wk.
Qed.

(* Values are irreducible. *)

Lemma red_value:
  forall K h1 t1 c2,
  red (Cf h1 t1) c2 ->
  wk K t1 VALUE ->
  False.
Proof.
  introv hred; dependent induction hred; intros; wk_inversion.
  (* Cases with an evaluation context *)
  eauto using eval_context_value.
  forwards: eval_context_value; eauto; wk_inversion.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The reflexive-transitive closure of the reduction relation. *)

Notation redstar :=
  (clos_refl_trans_1n _ red).

Hint Constructors clos_refl_trans_1n : redstar.

