Set Implicit Arguments.
Require Import EquivDec.
Require Import MyTactics.
Require Export MSA.
Require Export MSAProduct.
Require Export MSAMorphism.
Require Export MSAAllocationMaps.
Require Export MSASAMOperations.
(* The dep on [MSAAllocationMapOperations] is for the definition of [adopt].
   The definition of [adopt] is based on [MSASCAMOperations::heap_update],
   which for now depends on [MSAAllocationMapOperations]. *)
Require Export MSAAllocationMapOperations.
Require Export MSADXPermissions.
Require Export MSALPermissions.
Require Export MSAClosedAllocationMaps.
Require Export MSASCAMOperations.
Require Import TyCon.
Require Import Syntax.
Require Import MyList.

(* ---------------------------------------------------------------------------- *)

(* An orphan is a memory block whose [adopter] field is [null]. *)

(* We represent it as a pair of a tag and a list of fields; other representations
   would be possible. *)

Inductive orphan :=
| Orphan: nat -> term -> orphan.

(* An agreement relation between blocks and orphans. *)

Inductive block_orphan_agree : block -> orphan -> Prop :=
| BlockOrphanAgree:
    forall tag vs,
    block_orphan_agree (Block tag Null vs) (Orphan tag vs).

Ltac select_block_orphan_agree action :=
  match goal with h: block_orphan_agree _ _ |- _ => action h end.
Ltac block_orphan_agree_inversion :=
  gji select_block_orphan_agree.

(* ---------------------------------------------------------------------------- *)

(* Orphans have decidable equality. This property is implicitly used in order
   to check that permissions and resources (which are defined below) form
   monotonic separation algebras. *)

Lemma eq_orphan_dec:
  forall b1 b2 : orphan,
  { b1 = b2 } + { b1 <> b2 }.
Proof.
  repeat decide equality.
  eapply eq_tycon_dec.
Qed.

Obligation Tactic := idtac.

Program Instance equiv_dec_orphan: EqDec orphan eq.

Next Obligation.
  intros t1 t2. unfold equiv, complement.
  destruct (eq_orphan_dec t1 t2); eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Permissions. *)

(* A permission is the piece of information that we maintain for each address in
   the instrumented heap. It keeps track of an access right, one of [DX_NONE],
   [DX_EXCLUSIVE], and [DX_DUPLICABLE]. Depending upon this access right, it may
   also keep track of what is found at this address. More precisely, if the
   access right is [DX_NONE], then the permission does not contain any extra
   information. If it is [DX_EXCLUSIVE], then the permission keeps track of the
   contents of the memory block that is found at this address. If it is
   [DX_DUPLICABLE], then the permission also keeps track of the contents of the
   block, whose [adopter] field must be [null]. Thus, we encode the invariant
   that duplicable blocks do not have an adopter. *)

Notation heap_permission :=
  (dx_permission orphan block).

(* Permissions form a monotonic separation algebra. *)

Goal MSA heap_permission.
Proof.
  eauto with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A relation of agreement between blocks and permissions. *)

Notation heap_agree :=
  (dx_permission_agrees block_orphan_agree eq).

(* A permission determines a block. *)

Lemma heap_permission_determines_block:
  forall b1 b2 p,
  heap_agree b1 p ->
  heap_agree b2 p ->
  b1 = b2.
Proof.
  introv h1 h2.
  dependent destruction h1; dependent destruction h2;
  block_orphan_agree_inversion;
  congruence.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A block whose [adopter] field is [null] can be viewed either as duplicable
   or as exclusive. This leads us to defining the following function from a
   mode and an orphan to a permission. *)

Fixpoint mo2dx (m : mode) (o : orphan) : heap_permission :=
  match m, o with
  | Duplicable, o =>
      DX_DUPLICABLE o
  | Exclusive, Orphan tag vs =>
      DX_EXCLUSIVE (Block tag Null vs)
  end.

Notation mtagvs2dx m tag vs :=
  (mo2dx m (Orphan tag vs)).

Lemma mo2dx_injective:
  forall m1 m2 tag1 tag2 vs1 vs2,
  mo2dx m1 (Orphan tag1 vs1) = mo2dx m2 (Orphan tag2 vs2) ->
  m1 = m2 /\ tag1 = tag2 /\ vs1 = vs2.
Proof.
  destruct m1; destruct m2; simpl; intros; splits;
  try solve [
    (* Trying this first because [congruence] segfaults. Coq 8.3. *)
    injections; reflexivity
  | congruence
  ].
Qed.

Lemma consistent_mo2dx:
  forall m o,
  consistent (mo2dx m o).
Proof.
  destruct m; destruct o; simpl;
  eauto using @consistent_duplicable, @consistent_exclusive.
Qed.

Lemma maximal_mo2dx:
  forall m o,
  maximal (mo2dx m o).
Proof.
  destruct m; destruct o; simpl;
  eauto using @maximal_duplicable, @maximal_exclusive.
Qed.

Lemma stable_mo2dx:
  forall m o,
  stable (mo2dx m o).
Proof.
  destruct m; destruct o; simpl;
  eauto using @stable_duplicable, @stable_exclusive.
Qed.

Hint Resolve @maximal_duplicable @maximal_exclusive maximal_mo2dx : maps_maximal_hints.

Lemma agree_mo2dx:
  forall m tag vs,
  heap_agree (Block tag Null vs) (mtagvs2dx m tag vs).
Proof.
  destruct m; simpl; econstructor. econstructor. reflexivity.
Qed.

Lemma agree_mo2dx_implies_block_orphan_agree:
  forall b m o,
  heap_agree b (mo2dx m o) ->
  block_orphan_agree b o.
Proof.
  destruct m, o; inversion 1; subst; simpl; eauto. econstructor.
Qed.

Lemma agree_Exclusive:
  forall tag vs,
  heap_agree (Block tag Null vs) (DX_EXCLUSIVE (Block tag Null vs)).
Proof.
  eapply agree_mo2dx with (m := Exclusive).
Qed.

(* Note: the following property is false, because an exclusive block
   can have a non-null [adopter] pointer.

  forall b p,
  agree b p ->
  exists m o,
  p = mo2dx m o.

*)  

Lemma mo2dx_Exclusive:
  forall tag vs,
  DX_EXCLUSIVE (Block tag Null vs) = mtagvs2dx Exclusive tag vs.
Proof.
  reflexivity.
Qed.

Hint Resolve mo2dx_Exclusive : maps_maximal_reflexivity_hints.

(* ---------------------------------------------------------------------------- *)

(* Raw instrumented heap fragments, also known as raw resources. *)

(* A raw resource is a map of memory addresses to permissions. *)

Notation raw_heap_resource :=
  (aMap heap_permission).

(* Raw resources form a monotonic separation algebra. *)

Goal MSA raw_heap_resource.
Proof.
  eauto with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)

(* We now wish to restrict our attention to instrumented heap fragments that
   are closed under adopter edges. In other words, we do not allow splitting
   the heap in arbitrary ways; we only allow splitting it in ways that do not
   create dangling adopter pointers. *)

(* This guarantees that, whenever we own an adoptee, we also own its adopter.
   This implies, conversely, that whenever we own an adopter, we must own all
   of its adoptees. This fact is used when arguing that abandon is sound. *)

(* The predicate [permission_present] is used to instantiate the abstract
   predicate [present] that appears in the definition of closed allocation
   maps. It is true of exclusive permissions. This expresses the fact that,
   when an adopter edge points to some object [x], then we must own [x] --
   i.e. we must have an exclusive permission for [x]. This predicate also
   accepts [DX_ERROR], for a technical reason -- without this choice, the
   proof obligation [present_star_construction] would fail. This is not a
   problem because we are interested in consistent resources only. *)

(* The predicate [permission_present] is true also of [DX_DUPLICABLE]
   permissions. Indeed, it is safe for a duplicable object to have adoptees.
   These adoptees will never be abandoned, but it is safe nonetheless. This
   allows us to argue that an exclusive object can be frozen, even though it
   may have adoptees. *)
(* TEMPORARY clarify; used in [upd]; but we shouldn't need that? *)

Definition heap_permission_present (p : heap_permission) :=
  match p with
  | DX_ERROR
  | DX_EXCLUSIVE _
  | DX_DUPLICABLE _ =>
      True
  | DX_NONE =>
      False
  end.

(* The predicate [permission_present] is an instance of the type class
   [Present] used in the theory of closed allocation maps. *)

Program Instance present_heap_permission_present : Present heap_permission := {
  present := heap_permission_present
}.

Next Obligation. (* present_star_construction *)
  intros p1 p2; destruct p1; destruct p2; simpl; equiv_dec_by_cases.
Qed.

Next Obligation. (* present_rio *)
  intros p1 p2; destruct p1; destruct p2; simpl; solve [ tauto | congruence ].
Qed.

(* If [p] is present, then it is maximal. This means that when [x] is ``in
   the domain'' of [R1], it is also ``in the domain'' of a larger (consistent)
   resource [R1 * R2]. *)

Lemma present_maximal:
  forall p : heap_permission,
  present p ->
  maximal p.
Proof.
  destruct p; simpl; solve [
    tauto
  | eauto using @maximal_exclusive, @maximal_duplicable, @maximal_error
  ].
Qed.

Lemma present_mo2dx:
  forall m o,
  present (mo2dx m o).
Proof.
  destruct m, o; simpl; tauto.
Qed.

(* The predicate [permission_pointsto] is used to instantiate the abstract
   predicate [pointsto] that appears in the definition of closed allocation
   maps. It is true of [p] and [x] when the permission [p] describes a memory
   block whose adopter field points to the address [x]. *)

Definition heap_permission_pointsto (p : heap_permission) x :=
  match p with
  | DX_EXCLUSIVE (Block tag (NonNull adopter) vs) =>
      adopter = x
  | _ =>
      False
  end.

(* The predicate [permission_pointsto] is an instance of the type class
   [Pointsto] used in the theory of closed allocation maps. *)

Program Instance pointsto_heap_permission_pointsto : Pointsto heap_permission := {
  pointsto := heap_permission_pointsto
}.

Next Obligation. (* pointsto_star_inversion *)
  intros p1 p2; destruct p1; destruct p2; simpl; equiv_dec_by_cases.
Qed.

Next Obligation. (* pointsto_rio_reverse *)
  intros p1 p2; destruct p1; destruct p2; simpl; intros;
  try solve [ tauto | congruence ].
  (* One case remains. *)
  injections. tauto.
Qed.

Next Obligation. (* pointsto_is_exclusive *)
  intro p; destruct p; simpl; solve [ tauto | congruence ].
Qed.

(* ---------------------------------------------------------------------------- *)

(* Thanks to the above definitions, we now have a notion of an instrumented heap
   fragment that is closed under adopter edges. This is what we call a resource. *)

Notation heap_resource :=
  (ClosedAMap heap_permission).

(* Resources form a monotonic separation algebra. *)

Goal MSA heap_resource.
Proof.
  eauto with typeclass_instances.
Qed.

(* ----------------------------------------------------------------------------- *)

(* Locks. *)

(* A lck_perm is a term whose kind is supposed to be PERM. *)
(* Not clear yet whether the kinding check is to be made at this point. *)

Notation lck_perm := term.

(* Inductive lck_perm := *)
(* | LckPerm: term -> lck_perm. *)

Inductive free_lock_lck_perm_agree : lock -> lck_perm -> Prop :=
| FreeLockLckPermAgree:
    forall P,
    DeBruijn.closed 0 P ->
    free_lock_lck_perm_agree FreeLock P.

Inductive taken_lock_lck_perm_agree : lock -> lck_perm -> Prop :=
| TakenLockLckPermAgree:
    forall P,
    DeBruijn.closed 0 P ->
    taken_lock_lck_perm_agree TakenLock P.

Lemma eq_lck_perm_dec:
  forall p1 p2 : lck_perm,
  { p1 = p2 } + { p1 <> p2 }.
Proof.
  repeat decide equality.
  eapply eq_tycon_dec.
Qed.

Obligation Tactic := idtac.

Program Instance equiv_dec_lck_perm: EqDec lck_perm eq.

Next Obligation.
  intros t1 t2. unfold equiv, complement.
  destruct (eq_lck_perm_dec t1 t2); eauto.
Qed.

Notation lck_agree :=
  (l_permission_agrees free_lock_lck_perm_agree taken_lock_lck_perm_agree).

Notation lck_permission :=
  (l_permission lck_perm).

Goal MSA lck_permission.
Proof.
  eauto with typeclass_instances.
Qed.

Definition p2l (p : term) : lck_permission :=
  L_LOCK p.

Definition p2hl (p : term) : lck_permission :=
  L_HELD_LOCK p.

Definition is_locked_permission (lp: lck_permission) (P: term) : Prop :=
  lp = p2l P \/ lp = p2hl P.

Lemma consistent_lp_inversion:
  forall lp,
  consistent lp ->
  exists P,
  is_locked_permission lp P.
Proof.
  eauto using consistent_l_permission_inversion.
Qed.

Lemma agree_p2l:
  forall P,
  DeBruijn.closed 0 P ->
  lck_agree FreeLock (p2l P).
Proof.
  eauto using LPermissionAgreesLock, FreeLockLckPermAgree.
Qed.

Lemma agree_p2hl:
  forall P,
  DeBruijn.closed 0 P ->
  lck_agree TakenLock (p2hl P).
Proof. 
  eauto using LPermissionAgreesHeldLock, TakenLockLckPermAgree.
Qed.

Lemma consistent_p2l:
  forall p,
  consistent (p2l p).
Proof.
  simpl. tauto.
Qed.

Lemma consistent_p2hl:
  forall p,
  consistent (p2hl p).
Proof.
  simpl. tauto.
Qed.

Lemma maximal_p2hl:
  forall p,
  maximal (p2hl p).
Proof.
  eauto using maximal_held_lock.
Qed.

Lemma stable_p2l:
  forall p,
  stable (p2l p).
Proof.
  eauto using stable_lock.
Qed.

Lemma stable_p2hl:
  forall p,
  stable (p2hl p).
Proof.
  eauto using stable_held_lock.
Qed.

Lemma core_p2l:
  forall p,
  core (p2l p) = p2l p.
Proof.
  auto.
Qed.

Lemma core_p2hl_p2l:
  forall p,
  core (p2hl p) = p2l p.
Proof.
  auto.
Qed.

Lemma duplicable_p2l:
  forall p,
  duplicable (p2l p).
Proof.
  intro.
  rewrite <- core_p2l.
  eauto using @cores_are_duplicable.
Qed.

Lemma is_locked_permission_core:
  forall lp P,
  is_locked_permission lp P ->
  is_locked_permission (core lp) P.
Proof.
  intros lp P his; destruct lp; inversion his; inversion H; subst; simpl; auto. left. auto.
Qed.

Lemma is_locked_permission_core_reverse:
  forall lp P,
  is_locked_permission (core lp) P ->
  is_locked_permission lp P.
Proof.
  intros lp P his; destruct lp; inversion his; inversion H; subst; simpl in *; auto. right. auto.
Qed.

Lemma is_locked_permission_split_left:
  forall lp1 lp2 P,
  is_locked_permission (star lp1 lp2) P ->
  is_locked_permission lp1 P.
Proof.
  introv his; destruct lp1; destruct lp2; inversion his; inversion H; subst; simpl in *; equiv_dec_by_cases; try discriminate. inversion H0. econstructor. eauto.
Qed.

Lemma is_locked_permission_split_right:
  forall lp1 lp2 P,
  is_locked_permission (star lp1 lp2) P ->
  is_locked_permission lp2 P.
Proof.
  introv his.
  rewrite @star_commutative in his.
  eauto using is_locked_permission_split_left.
Qed.

Hint Resolve maximal_p2hl : maps_maximal_hints.

(* Raw lck resource is an allocation map. The complete resource will
   be a closed allocation map with trivial closedness properties, so that 
   we can use the SCAM library. *)

Notation raw_lck_resource :=
  (aMap lck_permission).

Goal MSA raw_lck_resource.
Proof.
  eauto with typeclass_instances.
Qed.

Definition lck_permission_present (l: lck_permission) := True.

Program Instance present_lck_permission_present: Present lck_permission := {
  present := lck_permission_present
}.

Next Obligation.
intros. unfold lck_permission_present. auto.
Qed.

Next Obligation.
intros. unfold lck_permission_present. auto.
Qed.

Definition lck_permission_pointsto (l: lck_permission) (x: nat) := False.

Program Instance pointsto_lck_permission_pointsto: Pointsto lck_permission := {
  pointsto := lck_permission_pointsto
}.

Next Obligation.
intros. tauto.
Qed.

Next Obligation.
intros. tauto.
Qed.

Next Obligation.
intros. tauto.
Qed.

Notation lck_resource :=
  (ClosedAMap lck_permission).

Goal MSA lck_resource.
Proof.
  eauto with typeclass_instances.
Qed.

Notation resource :=
  (heap_resource * lck_resource)%type.

Goal MSA resource.
Proof.
  eauto with typeclass_instances.
Qed.

Notation heap_project :=
  (@left_projection heap_resource lck_resource).

Notation heap_inject :=
  (@left_injection heap_resource lck_resource).

Notation lck_project :=
  (@right_projection heap_resource lck_resource).

Notation lck_inject :=
  (@right_injection heap_resource lck_resource).

Instance heap_project_morphism: MSAMorphism heap_project.
Proof.
  eapply left_projection_morphism.
Qed.

Instance heap_inject_bimorphism: MSABiMorphism heap_inject.
Proof.
  eapply left_injection_bimorphism.
Qed.

Instance lck_project_morphism: MSAMorphism lck_project.
Proof.
  eapply right_projection_morphism.
Qed.

Instance lck_inject_bimorphism: MSABiMorphism lck_inject.
Proof.
  eapply right_injection_bimorphism.
Qed.

Instance heap_sub_resource: SubStructure heap_project heap_inject.
Proof.
  split; intros; decompose_pairs; simpl in *; auto.
  match goal with
    | h: (?a, ?b) = (?c, ?d) |- _ => inversion h
  end.
  auto.
Qed.

Instance lck_sub_resource: SubStructure lck_project lck_inject.
Proof.
  split; intros; decompose_pairs; simpl in *; auto.
  match goal with
    | h: (?a, ?b) = (?c, ?d) |- _ => inversion h
  end.
  auto.
Qed.  

(* A few shortcuts with the right instances. *)

Notation heap_maps :=
  (SCAM_maps heap_project).

Notation heap_allocate :=
  (SCAM_allocate heap_project heap_inject).

Notation heap_update :=
  (SCAM_update heap_permission resource heap_project heap_inject).

Notation heap_limit :=
  (SCAM_limit heap_project).

Notation lck_maps :=
  (SCAM_maps lck_project).

Lemma lck_resource_maps_star:
  forall R R1 R2 l lp1 P,
  lck_maps R1 l lp1 ->
  is_locked_permission lp1 P ->
  star R1 R2 = R ->
  consistent R ->
  exists lp,
  lck_maps R l lp /\ is_locked_permission lp P.
Proof.
  introv hmaps hp hr hc. subst.
  forwards hl: (SCAM_maps_limit lck_project R1); [ eauto | ].
  erewrite <- (SCAM_limit_split_left lck_project_morphism R1) in hl by eauto.
  forwards hmaps': (SCAM_limit_maps lck_project_morphism); eauto.
  clear hl.
  destruct hmaps' as (lp, hmaps').
  exists lp; split; [ eauto | ].
  forwards: (SCAM_maps_consistent lck_project_morphism); eauto.
  clear hc.
  forwards: (SCAM_maps_star_inversion lck_project_morphism hmaps').
  unpack.
  clear hmaps'.
  forwards: (SCAM_maps_injective lck_project hmaps H0). subst.
  destruct x.
    inversion hp; subst; inversion H2; subst; destruct x0; equiv_dec_by_cases. right. auto.
    inversion hp; subst; inversion H2; subst; destruct x0; equiv_dec_by_cases.
    inversion H.
Qed.

Lemma lck_allocate_req:
  forall p: lck_permission,
  forall y: nat, ~ pointsto p y.
Proof.
  tauto.
Qed.

Implicit Arguments lck_allocate_req [ ].
  
Notation lck_allocate R p :=
  (SCAM_allocate lck_project lck_inject R p (lck_allocate_req p)).

Lemma lck_update_dreq:
  forall (R: raw_lck_resource) (s: nat) (p: lck_permission),
  present p \/ has_no_antecedents R s.
Proof.
  unfold has_no_antecedents.
  tauto.
Qed.

Implicit Arguments lck_update_dreq [ ].

Lemma lck_update_oreq:
  forall (R: raw_lck_resource) (s: nat) (p: lck_permission),
  forall y, pointsto p y -> domain (update R s p) y.
Proof.
  simpl.
  unfold lck_permission_pointsto.
  tauto.
Qed.  

Implicit Arguments lck_update_oreq [ ].

Notation lck_update R s p :=
  (SCAM_update lck_permission resource lck_project lck_inject R s p
               (lck_update_dreq (proj1_sig (lck_project R)) s p)
               (lck_update_oreq (proj1_sig (lck_project R)) s p)
  ).

Notation lck_limit :=
  (SCAM_limit lck_project).

Hint Extern 1
     (SCAM_maps _ _ _ _) => (SCAM_maps_maximal heap_project_morphism) : SCAM_maps_maximal_heap.

Hint Extern 1
     (SCAM_maps _ _ _ _) => (SCAM_maps_maximal lck_project_morphism) : SCAM_maps_maximal_lck.

Hint Extern 1 (_ = _) =>
  (instantiate; SCAM_limit_split heap_project_morphism) : SCAM_limit_split_heap.

Hint Extern 1 (_ = _) =>
  (instantiate; SCAM_limit_split lck_project_morphism) : SCAM_limit_split_lck.


(* ---------------------------------------------------------------------------- *)

(* The empty resource. *)

Notation void :=
  (CAM_empty heap_permission, CAM_empty lck_permission).

Lemma core_void:
  core void = void.
Proof.
  rewrite duplicable_implies_self_core
          by eauto using @duplicable_product, @duplicable_CAM_empty.
  reflexivity.
Qed.

Lemma duplicable_void:
  duplicable void.
Proof.
  rewrite <- core_void. eapply cores_are_duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Set up the [AAC] library and plugin to help reason about resources. *)

Require Import AAC.

Instance sc : Commutative (@eq resource) star.
Proof. unfold Commutative. eauto using @star_commutative. Qed.
Instance sa: Associative (@eq resource) star.
Proof. unfold Associative. eauto using @star_associative. Qed.

Instance shc : Commutative (@eq heap_resource) star.
Proof. unfold Commutative. eauto using @star_commutative. Qed.
Instance sha: Associative (@eq heap_resource) star.
Proof. unfold Associative. eauto using @star_associative. Qed.

Instance slc : Commutative (@eq lck_resource) star.
Proof. unfold Commutative. eauto using @star_commutative. Qed.
Instance sla: Associative (@eq lck_resource) star.
Proof. unfold Associative. eauto using @star_associative. Qed.

(* ---------------------------------------------------------------------------- *)

(* Set up a few hints and tactics that help reason modulo AC. *)

(* There is a lot of work to be done in this area. The following
   infrastructure is very naive and preliminary. *)

(* This hint causes [eauto] to invoke the [AAC] plugin when trying to
   prove equations between resources. Any meta-variables whose value
   has been determined by other branches of the search are instantiated
   first, because [aac_reflexivity] does not like meta-variables. *)

Hint Extern 1 (_ = _ :> resource) => (instantiate; aac_reflexivity) : aac.
Hint Extern 1 (_ = _ :> heap_resource) => (instantiate; aac_reflexivity) : aac.
Hint Extern 1 (_ = _ :> lck_resource) => (instantiate; aac_reflexivity) : aac.

Ltac consistent_aac_crush :=
  match goal with
    | hc: consistent ?R' |- consistent ?R =>
        let haac := fresh in
        aac_match (fun x => star x R) R' haac;
        rewrite <- haac in hc;
        eauto using @consistent_split_right
  end.

(* ---------------------------------------------------------------------------- *)

(* Miscellaneous consistency lemmas. *)

(* The following lemmas are proved once and for all, and used later in the
   soundness proof. There are two potential reasons for doing this: a) it may
   be more efficient; b) it may allow these facts to be used by [eauto] in a
   manner that would not otherwise be possible. *)

(* One could move these lemmas to [MSA], but I prefer
   to avoid creating a dependency between [MSA] and
   the AAC plugin. *)

Lemma consistent_12p3_13:
  forall R1 R2 R3 : resource,
  consistent (star (star R1 R2) R3) ->
  consistent (star R1 R3).
Proof.
  intros. consistent_aac_crush.
Qed.

Lemma consistent_12p3_23:
  forall R1 R2 R3 : resource,
  consistent (star (star R1 R2) R3) ->
  consistent (star R2 R3).
Proof.
  intros. consistent_aac_crush.
Qed.

Lemma consistent_1p23_12:
  forall R1 R2 R3 : resource,
  consistent (star R1 (star R2 R3)) ->
  consistent (star R1 R2).
Proof.
  intros. consistent_aac_crush.
Qed.

Lemma consistent_1p23_13:
  forall R1 R2 R3 : resource,
  consistent (star R1 (star R2 R3)) ->
  consistent (star R1 R3).
Proof.
  intros. consistent_aac_crush.
Qed.

Lemma consistent_drop_left:
  forall R1 R2 R3 R : resource,
  consistent (star R R3) ->
  star R1 R2 = R ->
  consistent (star R2 R3).
Proof.
  intros; subst. consistent_aac_crush.
Qed.

Lemma consistent_1234_24:
  forall R1 R2 R3 R4 : resource,
  consistent (star (star R1 R2) (star R3 R4)) ->
  consistent (star R2 R4).
Proof.
  intros. consistent_aac_crush.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Miscellaneous reconfiguration lemmas. *)

Lemma reconfigure00:
  forall R1 R2 R3 : resource,
    star (star R1 R2) R3 = star R2 (star R3 R1).
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure01:
  forall R0 R1 R2 R3 : resource,
  star (star R0 R1) (star R2 R3) = star (star R0 (star R1 R2)) R3.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure02:
  forall R0 R1 R2 R3 : resource,
  star (star R0 R1) (star R2 R3) = star (star R0 (star R2 R1)) R3.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure03:
  forall R0 R1 R2 R3 : resource,
  star (star R2 (star R1 R0)) R3 = star (star (star R1 R2) R0) R3.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure04:
  forall R1 R2 R3 : resource,
  star (star R1 R2) R3 = star (star R3 R2) R1.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure05:
  forall R0 R1 R2 : resource,
  star (star R0 R1) R2 = star (star R0 R2) R1.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure06:
  forall R1 R2 R3 : resource,
  star R1 (star R2 R3) = star R2 (star R1 R3).
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure07:
  forall R1 R2 R3 : resource,
    star (star R1 R2) R3 = star R2 (star R1 R3).
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure08:
  forall R0 R1 R2 R3 : resource,
  star R2 (star R0 (star R1 R3)) = star (star R0 (star R1 R2)) R3.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure09:
  forall R1 R2 R3 : resource,
  star (star R1 R2) R3 = star R1 (star R3 R2).
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure10:
  forall R0 R1 R2 R3 : resource,
  star R0 (star (star R1 R2) R3) = star (star R0 (star R2 R3)) R1.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure11:
  forall R0 R1 R2 R3: resource,
  star (star R0 R1) (star R2 R3) = star (star R0 R2) (star R1 R3).
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure12:
  forall R0 R1 R2 R3 : resource,
  star R0 (star R1 (star R2 R3)) = star (star R0 (star R1 R3)) R2.
Proof.
  intros; aac_reflexivity.
Qed.

Lemma reconfigure13:
  forall R0 R1 R2 R3 : resource,
  star R0 (star R1 (star R2 R3)) = star (star R0 (star R2 R3)) R1.
Proof.
  intros; aac_reflexivity.
Qed.

Lemma reconfigure14:
  forall R0 R1 R2 R3 : resource,
  star R0 (star R1 (star R2 R3)) = star (star R0 R2) (star R1 R3).
Proof.
  intros; aac_reflexivity.
Qed.

Lemma reconfigure15:
  forall R0 R1 R2 R3 : resource,
  star (star R1 R2) (star R3 R0) = star R0 (star R1 (star R2 R3)).
Proof.
  intros; aac_reflexivity.
Qed.

Lemma reconfigure16:
  forall R0 R1 R2 R3 R4 : resource,
  star R2 (star R4 (star R0 (star R1 R3))) = star (star R4 (star R0 (star R1 R2))) R3.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure17:
  forall R0 R1 R2 R3 R4 : resource,
  star (star (star R0 R1) (star R2 R3)) R4 = star (star R2 (star R1 R3)) (star R0 R4).
Proof.
  intros. aac_reflexivity.
Qed.

Lemma reconfigure18:
  forall R0 R1 R2 R3 R4 : resource,
  star (star (star R0 R1) (star R2 R3)) R4 = star (star R1 (star R0 (star R2 R3))) R4.
Proof.
  intros. aac_reflexivity.
Qed.

Lemma congruence00:
  forall R0 R1 R2 R : resource,
  star R2 (star R1 R0) = R ->
  star (star R1 R2) R0 = R.
Proof.
  intros. subst. aac_reflexivity.
Qed.

Lemma congruence01:
  forall R0 R1 R2 R : resource,
  R = star R2 (star R1 R0) ->
  R = star (star R1 R2) R0.
Proof.
  intros. subst. aac_reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Allocation in an instrumented heap. *)

(* The following short-hand is used when allocating new blocks. *)

Program Definition heap_alloc R (m: mode) (o: orphan): resource :=
  heap_allocate R (mo2dx m o) _.

Next Obligation.
  intros; destruct m; destruct o; simpl; tauto.
Qed.

(* From a passive thread's point of view, allocation is safe. *)

Lemma rio_allocate_core_mo2dx:
  forall R m o req,
  rio R (heap_allocate R (core (mo2dx m o)) req).
Proof.
  intros.
  eapply (SCAM_rio_allocate heap_inject_bimorphism heap_sub_resource).
    eauto using @consistent_core, consistent_mo2dx.
    eauto using @cores_are_duplicable.
Qed.

Lemma rio_allocate_core_p2l:
  forall R p req,
  rio R (SCAM_allocate lck_project lck_inject R (core (p2l p)) req).
Proof.
  intros.
  eapply (SCAM_rio_allocate lck_inject_bimorphism lck_sub_resource).
    eauto using @consistent_core, consistent_p2l.
    eauto using @cores_are_duplicable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Updating an instrumented heap. *)

(* The following short-hand is used when updating or freezing an exclusive
   block. *)

Program Definition heap_upd R l (m: mode) (o: orphan): resource :=
  heap_update R l (mo2dx m o) _ _.

Next Obligation.
  (* Exclusive or duplicable access is retained. *)
  left. destruct m, o; simpl; tauto.
Defined.

Next Obligation.
  (* The block's adopter field contains Null. *)
  destruct m, o; simpl; tauto.
Defined.

(* Reformulate some of the properties of [CAM_update] in terms of [upd],
   so as to avoid having to unfold [upd]. *)

Lemma heap_consistent_upd:
  forall R l m o,
  consistent R ->
  consistent (heap_upd R l m o).
Proof.
  unfold heap_upd. 
  eauto using
        (SCAM_consistent_update heap_project_morphism heap_inject_bimorphism),
        consistent_mo2dx.
Qed.

Lemma heap_maps_upd_there:
  forall R r s p m o,
  heap_maps R r p ->
  r = s ->
  heap_maps (heap_upd R s m o) r (mo2dx m o).
Proof.
  unfold heap_upd.
  eauto using (SCAM_maps_update_there heap_sub_resource).
Qed.

(* ------------------------------------------------------------------------- *)

(* A collection of lemmas that depend on our specific notion of resource. *)

(* If [R1] knows about [l] and [R2] knows about [l], then they know the
   same thing. (In fact, [m1] and [m2] must be [Duplicable] in this case;
   see the lemma [CAM_exclusive_exclusive_star] below.) *)

Lemma heap_maps_mtagvs2dx_injective:
  forall R1 R2 l m1 m2 tag1 tag2 vs1 vs2,
  heap_maps R1 l (mtagvs2dx m1 tag1 vs1) ->
  heap_maps R2 l (mtagvs2dx m2 tag2 vs2) ->
  consistent (star R1 R2) ->
  m1 = m2 /\ tag1 = tag2 /\ vs1 = vs2.
Proof.
  intros.
  apply mo2dx_injective.
  eapply (@SCAM_maps_injective _ _ _ _ _ heap_project (star R1 R2) l); SCAM_maps_maximal heap_project_morphism.
Qed.

(* If [R1] owns [l] and is compatible with [R2], then [R2] does not own [l]. *)

Lemma resource_total_projection:
  forall R,
  (heap_project R, lck_project R) = R.
Proof.
  intro R; destruct R; simpl; auto.
Qed.

Lemma maps_exclusive_star:
  forall R1 R2 l b,
  heap_maps R1 l (DX_EXCLUSIVE b) ->
  consistent (star R1 R2) ->
  heap_maps R2 l (@DX_NONE _ _).
Proof.
  introv hmaps hc.
  forwards h: (@SCAM_maps_star heap_permission);
    [ eauto with typeclass_instances | eauto | eauto | ].
  destruct h as (?, h). destruct h as (?, h).
  generalize (consistent_star_exclusive h); intro; subst.
  assumption.
Qed.

(* If [R1] owns [x1] and [R2] owns [x2], and the combination is consistent,
   then [x1] and [x2] must be distinct addresses. *)

Lemma exclusive_exclusive_star:
  forall R1 R2 x1 x2 (b1 b2 : block),
  heap_maps R1 x1 (DX_EXCLUSIVE b1) ->
  heap_maps R2 x2 (DX_EXCLUSIVE b2) ->
  consistent (star R1 R2) ->
  x1 <> x2.
Proof.
  introv h1 h2 hc ?. subst.
  generalize (SCAM_maps_star_construction heap_project_morphism _ _ _ _ _ h1 h2 hc); intro hm.
  generalize (SCAM_maps_consistent heap_project_morphism _ _ _ hm hc); intro he.
  simpl in he; tauto.
Qed.

(* The following tactic is used when we have an assumption [hm] of the form
   [CAM_maps (star R1 R2) x (DX_EXCLUSIVE _)]. Then, either [R1] or [R2] has
   this exclusive permission for [x], while the other has [DX_NONE]. *)

Ltac maps_split_exclusive hm :=
  (* If [R1 * R2] has an exclusive permission for [x], then each of
     [R1] and [R2] has part of this permission. *)
  unpack (@SCAM_maps_star_inversion _ _ _ _ _ _ heap_project _ _ _ _ _ hm);
  clear hm;
  (* But there are only two ways of splitting an exclusive permission. *)
  match goal with h: star _ _ = _ |- _ =>
    destruct (split_exclusive h) as [ [ ? ? ] | [ ? ? ] ]; clear h; subst
  end;
  (* If we end up with multiple facts about a single location, simplify them. *)
  try SCAM_maps_injective;
  (* Clean up. *)
  repeat match goal with
  | h: present (DX_EXCLUSIVE _) |- _ =>
      clear h
  | h: present DX_NONE |- _ =>
      tauto
  | h: ?p = ?p |- _ =>
      clear h
  end.

(* If [R1] evolves to [R2] along [rio] and [R2] has an exclusive
   permission for [x], then [R1] has the same permission for [x]. *)

Lemma maps_rio_exclusive_reverse:
  forall { R1 R2 x b },
  heap_maps R2 x (DX_EXCLUSIVE b) ->
  rio R1 R2 ->
  heap_maps R1 x (DX_EXCLUSIVE b).
Proof.
  introv hm hrio.
  (* If [R2] has an exclusive permission for [x], then either [x] existed in
     [R1] already, or [x] was newly allocated. *)
  destruct (@SCAM_maps_rio_reverse _ _ _ _ _ _ heap_project _ _ _ _ _ hm hrio); unpack.
  (* Sub-case: [x] existed in [R1] already. Then, because exclusive permissions
     don't appear out of nowhere, [R1] must have had the same permission for [x]. *)
  equates 1; [ eauto | symmetry ].
  eauto using @rio_exclusive_reverse.
  (* Sub-case: [x] was newly allocated. This contradicts the fact that
     locations that are newly allocated while a thread is asleep must
     receive duplicable permissions (from the sleeping thread's point
     of view). *)
  false. (* eauto using not_duplicable_exclusive *)
Qed.

(* If [R1] owns [l] and if [R2] and is compatible with [R1], then [R2]
   must map [l] to [DX_NONE], so it can evolve to [DX_DUPLICABLE a]
   along [rio]. *)

Lemma heap_rio_upd_duplicable:
  forall R1 R2 l tag1 tag2 vs,
  heap_maps R1 l (DX_EXCLUSIVE (Block tag1 Null vs)) ->
  consistent (star R1 R2) ->
  rio R2 (heap_upd R2 l Duplicable (Orphan tag2 vs)).
Proof.
   intros. eapply (SCAM_rio_update heap_inject_bimorphism heap_sub_resource).
   eauto using maps_exclusive_star.
   eapply rio_none_duplicable.
Qed.

Lemma heap_star_upd:
  forall R1 R2 s o,
  star (heap_upd R1 s Duplicable o) (heap_upd R2 s Duplicable o) =
  heap_upd (star R1 R2) s Duplicable o.
Proof.
  intros. eapply (SCAM_star_update heap_project_morphism heap_inject_bimorphism).
  eapply duplicable_duplicable.
Qed.

(* If [R1] owns [l] and is compatible with [R2], then [R1] can be updated
   at [l] without affecting [R2]. *)

Lemma star_update_exclusive_left:
  forall R1 R2 l b1 b2 dreq1 oreq1 dreq2 oreq2,
  heap_maps R1 l (DX_EXCLUSIVE b1) ->
  consistent (star R1 R2) ->
  star (heap_update R1 l (DX_EXCLUSIVE b2) dreq1 oreq1) R2 =
  heap_update (star R1 R2) l (DX_EXCLUSIVE b2) dreq2 oreq2.
Proof.
  intros.
  eapply (SCAM_star_update_core_left heap_project_morphism heap_inject_bimorphism heap_sub_resource).
  simpl.
  eapply maps_exclusive_star; eauto.
Qed.

Lemma star_update_exclusive_right:
  forall R1 R2 l b1 b2 dreq1 oreq1 dreq2 oreq2,
  heap_maps R2 l (DX_EXCLUSIVE b1) ->
  consistent (star R1 R2) ->
  star R1 (heap_update R2 l (DX_EXCLUSIVE b2) dreq1 oreq1) =
  heap_update (star R1 R2) l (DX_EXCLUSIVE b2) dreq2 oreq2.
Proof.
  intros.
  eapply (SCAM_star_update_core_right heap_project_morphism heap_inject_bimorphism heap_sub_resource).
  simpl.
  eapply maps_exclusive_star.
    eauto.
    rewrite star_commutative. eauto.
Qed.

Lemma star_upd_exclusive_left:
  forall R1 R2 l b1 o,
  heap_maps R1 l (DX_EXCLUSIVE b1) ->
  consistent (star R1 R2) ->
  star (heap_upd R1 l Exclusive o) R2 =
  heap_upd (star R1 R2) l Exclusive o.
Proof.
  intros. destruct o. eapply star_update_exclusive_left; eauto.
Qed.

Lemma star_upd_exclusive_right:
  forall R1 R2 l b1 o,
  heap_maps R2 l (DX_EXCLUSIVE b1) ->
  consistent (star R1 R2) ->
  star R1 (heap_upd R2 l Exclusive o) =
  heap_upd (star R1 R2) l Exclusive o.
Proof.
  intros. destruct o. eapply star_update_exclusive_right; eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* This auxiliary tactic assumes that the goal contains a subterm of the form
   [upd (star R1 R2) _ Exclusive _ _]. The goal may (and will actually, as we
   will see below) contain several occurrences of this subterm. The tactic
   determines which of [R1] and [R2] is concerned by the update, and uses a
   rewriting step to push [upd] into [R1] or [R2], as appropriate. *)

Ltac star_upd_exclusive_rec :=
  match goal with
  | h: heap_maps ?R _ (DX_EXCLUSIVE ?v)
    |- context[heap_upd (star ?R1 ?R2) _ Exclusive _] =>
      (* Accumulate consistency hypotheses as we go down, in order to avoid
         a quadratic overhead. *)
      let hc := fresh in
      assert (hc: consistent (star R1 R2)); [ solve [ eauto 3 with consistent_split ] | ];
      (* Determine whether we should push [upd] towards the left or the right. *)
      match R1 with
      | context[R] =>
          erewrite <- @star_upd_exclusive_left with (b1 := v); [ |
            SCAM_maps_maximal heap_project_morphism
          | eexact hc
          ];
          star_upd_exclusive_rec
      | _ =>
          match R2 with
          | context[R] =>
              erewrite <- @star_upd_exclusive_right with (b1 := v); [ |
                SCAM_maps_maximal heap_project_morphism
              | eexact hc
              ];
              star_upd_exclusive_rec
          | _ =>
              idtac "star_upd_exclusive: I don't know what to do"
          end
      end;
      clear hc
  | _ =>
      idtac (* we are presumably done *)
  end.

(* This tactic is a wrapper for the above tactic. Again, we assume that the
   goal contains a subterm of the form [update (star R1 R2) _ Exclusive _ _].
   We push [upd] down into the appropriate conjunct, and add a hypothesis
   stating that the transformed resource is consistent. *)

Ltac star_upd_exclusive :=
  match goal with
  | |- context[heap_upd ?R ?l Exclusive ?o] =>
      (* Add to the goal a hypothesis stating that [upd R l Exclusive tag vs]
         is consistent. This hypothesis will be rewritten as we go; thus, we
         keep track of the consistency of our resource as we transform it. *)
      cut (consistent (heap_upd R l Exclusive o)); [
        (* Transform the resource. *)
        star_upd_exclusive_rec;
        (* Move the consistency hypothesis from the goal into the hypotheses. *)
        intro
      | (* This should prove the initial consistency hypothesis. *)
        eapply (SCAM_consistent_update heap_project_morphism heap_inject_bimorphism);
          eauto using @consistent_mo2dx with consistent_split
      ]
  end.

(* ---------------------------------------------------------------------------- *)

(* This tactic inverts all hypotheses of the form [CAM_maps (upd _ _ _ _) _ _].
   Each such inversion gives to two sub-goals, which the tactic solves when the
   hypotheses are clearly infeasible. The property [CAM_maps_injective] is also
   used to simplify the sub-goals that remain. *)

Ltac heap_maps_update_inversion :=
  repeat match goal with
  h: heap_maps (heap_upd _ _ _ _) _ _ |- _ =>
    destruct (@SCAM_maps_update_inversion _ _ _ _ _ heap_project heap_inject _ _ _ _ _ _ _ _ h); clear h; unpack;
    subst;
    repeat SCAM_maps_injective;
    try solve [ simpl in *; congruence ]
  end.

(* ---------------------------------------------------------------------------- *)

(* Adoption in an instrumented heap. *)

(* The instrumented heap [adopt R y x tag vs] is the result of updating
   the memory location [y] with the block [Block tag (NonNull x) vs] in
   the instrumented heap [R]. There is a proof obligation: [x] and [y]
   must be distinct addresses, and [x] must be in the domain of [R]. *)

Notation owns R x :=
  (exists b, heap_maps R x (DX_EXCLUSIVE b)).

Program Definition adopt R y x tag vs
  (req : x <> y /\ owns R x)
  : resource :=
  heap_update R y (DX_EXCLUSIVE (Block tag (NonNull x) vs)) _ _.

Next Obligation.
  left. simpl. tauto.
Defined.

Next Obligation.
  (* intros. *)
  (* destruct R as (hR, lR). *)
  (* destruct hR. *)
  (* unfold domain. unfold SCAM_maps in *. simpl in * |-. subst. simpl. unpack. *)
  (* xplit. *)
  destruct R as (hR, lR); destruct hR; unfold domain, SCAM_maps. simpl. intros; subst. unpack.
  xplit.
  eapply maps_update_elsewhere. eauto. eauto.
  simpl. tauto.
Defined.

Implicit Arguments adopt [ ].

(* ---------------------------------------------------------------------------- *)

(* [adopt] preserves consistency. *)

Lemma consistent_adopt:
  forall R y x tag vs req,
  consistent R ->
  consistent (adopt R y x tag vs req).
Proof.
  unfold adopt.
  eauto using (SCAM_consistent_update heap_project_morphism heap_inject_bimorphism), @consistent_exclusive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The proof obligation associated with [adopt] is compatible with equality. *)

Lemma adopt_req_equality:
  forall R1 R2,
  R1 = R2 ->
  forall y x,
  (x <> y /\ owns R1 x) ->
  (x <> y /\ owns R2 x).
Proof.
  intros. subst. tauto.
Qed.

(* This allows us to rewrite under [adopt], as follows. We rewrite [R1] to
   [R2] and adapt the proof obligation. *)

Lemma adopt_equality:
  forall R1 R2 y x tag vs req1,
  forall eq : R1 = R2,
  adopt R1 y x tag vs req1 =
  adopt R2 y x tag vs (adopt_req_equality eq req1).
Proof.
  unfold adopt. 
  unfold SCAM_update.
  intros; subst.
  eapply injectA_equality.
  came.
  congruence.
Qed.

(* The following tactics show how the above lemma can be used. This seems
   crazy complicated, but I couldn't find a simpler way. *)

Ltac star_commutative_under_adopt :=
  match goal with |- context[adopt (star ?R1 ?R2) _ _ ?tag ?vs ?req] =>
    rewrite (
      adopt_equality tag vs req (@star_commutative _ _ R1 R2)
    )
  end.

Ltac star_associative_under_adopt :=
  match goal with |- context[adopt (star ?R1 (star ?R2 ?R3)) _ _ ?tag ?vs ?req] =>
    rewrite (
      adopt_equality tag vs req (@star_associative _ _ R1 R2 R3)
    )
  end.

Ltac reconfigure06_under_adopt :=
  match goal with |- context[adopt (star ?R1 (star ?R2 ?R3)) _ _ ?tag ?vs ?req] =>
    rewrite (
      adopt_equality tag vs req (@reconfigure06 R1 R2 R3)
    )
  end.

(* ---------------------------------------------------------------------------- *)

(* [adopt] can be pushed into [star]. If both [x] and [y] are owned
   by [R1], then [star (adopt R1 ...) R2] is [adopt (star R1 R2) ...]. *)

Lemma star_adopt_left:
  forall R1 R2,
  consistent (star R1 R2) ->
  forall y b,
  heap_maps R1 y (DX_EXCLUSIVE b) ->
  forall x tag vs req1 req2,
  star (adopt R1 y x tag vs req1) R2 =
  adopt (star R1 R2) y x tag vs req2.
Proof.
  intros.
  eapply (SCAM_star_update_core_left heap_project_morphism heap_inject_bimorphism heap_sub_resource). simpl.
  eapply maps_exclusive_star; eauto.
Qed.

(* The above lemma is agnostic as to how the proof obligations are constructed:
   both [req1] and [req2] are supplied by the client. We now define a directional
   version of this lemma, where [req1] is supplied by the client and [req2] is
   constructed using [req1]. *)

(* The proof obligation for [adopt] can be extended to a larger resource. *)

Lemma star_adopt_left_req_LR:
  forall R1 R2 y x,
  consistent (star R1 R2) ->
  (x <> y /\ owns R1 x) ->
  (x <> y /\ owns (star R1 R2) x).
Proof.
  intros. unpack. xplit. eauto.
  SCAM_maps_maximal heap_project_morphism.
Qed.

Implicit Arguments star_adopt_left_req_LR [ R1 R2 y x ].

(* [adopt] can be pushed into [star]. If both [x] and [y] are owned
   by [R1], then [star (adopt R1 ...) R2] is [adopt (star R1 R2) ...]. *)

Lemma star_adopt_left_LR:
  forall R1 R2 hc y x tag vs req p,
  heap_maps R1 y (DX_EXCLUSIVE p) ->
  star (adopt R1 y x tag vs req) R2 =
  adopt (star R1 R2) y x tag vs (star_adopt_left_req_LR hc req).
Proof.
  eauto using star_adopt_left.
Qed.

Implicit Arguments star_adopt_left_LR [ R1 R2 ].

(* ---------------------------------------------------------------------------- *)

(* Symmetric versions of the above lemmas. *)

Lemma star_adopt_right:
  forall R1 R2,
  consistent (star R1 R2) ->
  forall y p,
  heap_maps R2 y (DX_EXCLUSIVE p) ->
  forall x tag vs req1 req2,
  star R1 (adopt R2 y x tag vs req1) =
  adopt (star R1 R2) y x tag vs req2.
Proof.
  intros.
  eapply (SCAM_star_update_core_right heap_project_morphism heap_inject_bimorphism heap_sub_resource). simpl.
  eapply maps_exclusive_star; eauto using @consistent_star_commutative.
Qed.

Lemma star_adopt_right_req_LR:
  forall R1 R2 y x,
  consistent (star R1 R2) ->
  (x <> y /\ exists p, heap_maps R2 x (DX_EXCLUSIVE p)) ->
  (x <> y /\ exists p, heap_maps (star R1 R2) x (DX_EXCLUSIVE p)).
Proof.
  intros. unpack. xplit. eauto. SCAM_maps_maximal heap_project_morphism.
Qed.

Implicit Arguments star_adopt_right_req_LR [ R1 R2 y x ].

Lemma star_adopt_right_LR:
  forall R1 R2 hc y x tag vs req p,
  heap_maps R2 y (DX_EXCLUSIVE p) ->
  star R1 (adopt R2 y x tag vs req) =
  adopt (star R1 R2) y x tag vs (star_adopt_right_req_LR hc req).
Proof.
  eauto using star_adopt_right.
Qed.

Implicit Arguments star_adopt_right_LR [ R1 R2 ].

(* The following lemma cannot be true since L_LOCK is not maximal. *)

(* Lemma raw_lck_resource_maximal: *)
(*   forall (lR: raw_lck_resource), *)
(*   maximal lR. *)
(* Proof. *)
(*   introv hs hc. *)
(*   dam. *)
(*   induction_lists. *)
(*   auto. *)
(*   destruct a, l; try destruct (equiv_dec l0 l); tauto. *)
(*   eauto. *)
(* Qed. *)

(* Lemma lck_resource_maximal: *)
(*   forall (lR: lck_resource), *)
(*   maximal lR. *)
(* Proof. *)
(*   introv hs hc. *)
(*   came. *)
(*   eapply raw_lck_resource_maximal; eauto. *)
(* Qed. *)

(* ---------------------------------------------------------------------------- *)

(* The type soundness proof should not depend on the actual definition of
   resources, only on their properties. *)

Global Opaque star core consistent rio.

