Set Implicit Arguments.
Require Import MyTactics.
Require Import MyList.
Require Import TyCon.
Require Import Syntax.

(* The type system and the proof of type soundness are parameterized with a
   global signature, which fixes a number of algebraic data type definitions. *)

(* ---------------------------------------------------------------------------- *)

(* With each algebraic data type constructor, we associate a list of branches,
   where a branch is just a row of closed types. *)

Parameter tc_branches : tycon -> list term.

Definition wkbs (Tss : list term) :=
  Forall (fun Ts => wk zilch Ts (ROW TYPE)) Tss.

Parameter tc_branches_wk:
  forall tc,
  wkbs (tc_branches tc).

Lemma wkbs_inversion_head:
  forall Ts Tss,
  wkbs (Ts :: Tss) ->
  wk zilch Ts (ROW TYPE).
Proof.
  inversion 1; subst. assumption.
Qed.

Lemma wkbs_inversion_tail:
  forall Ts Tss,
  wkbs (Ts :: Tss) ->
  wkbs Tss.
Proof.
  inversion 1; subst. assumption.
Qed.

Hint Resolve tc_branches_wk wkbs_inversion_head wkbs_inversion_tail :
j_weakening j_substitution.

(* ---------------------------------------------------------------------------- *)

(* The valid tags for an algebraic data type constructor [tc] are just the
   numbers of the branches of [tc]. *)

Definition tc_tag tc tag Ts :=
  is_nth tag (tc_branches tc) Ts.

(* A ``record type'' is just an algebraic data type with one tag. *)

Definition tc_is_record tc Ts :=
  tc_branches tc = Ts :: nil.

(* Within each branch, the valid fields are just the numbers of the elements
   in [Ts]. *)

(* Thus, in this somewhat low-level calculus, both data constructors and
   fields are just numbers. *)

(* ---------------------------------------------------------------------------- *)

(* With each algebraic data type constructor, we associate a mode: duplicable
   or exclusive. *)

Parameter tc_mode : tycon -> mode.

(* ---------------------------------------------------------------------------- *)

(* With each algebraic data type constructor, we associate an adopts clause. *)

Parameter tc_adopts : tycon -> term.

Parameter tc_adopts_wk:
  forall tc,
  wk zilch (tc_adopts tc) TYPE.

Hint Resolve tc_adopts_wk : wk j_weakening j_substitution.

(* The type [tc_adopts tc] is typically either an exclusive type or [TyBottom]. *)

(* ---------------------------------------------------------------------------- *)

(* A few consequences of the above definitions. *)

Lemma tc_tag_wk:
  forall tc tag Ts,
  tc_tag tc tag Ts ->
  wk zilch Ts (ROW TYPE).
Proof.
  intros.
  eapply is_nth_Forall with (P := fun Ts => wk zilch Ts (ROW TYPE)).
  eassumption.
  eapply tc_branches_wk.
Qed.

Lemma tc_tag_injective:
  forall tc tag Ts1 Ts2,
  tc_tag tc tag Ts1 ->
  tc_tag tc tag Ts2 ->
  Ts1 = Ts2.
Proof.
  eauto using is_nth_injective.
Qed.

Lemma tc_is_record_implies_tc_tag:
  forall tc Ts,
  tc_is_record tc Ts ->
  tc_tag tc 0 Ts.
Proof.
  introv h. unfold tc_tag. rewrite h. econstructor.
Qed.

Lemma tc_is_record_wk:
  forall tc Ts,
  tc_is_record tc Ts ->
  wk zilch Ts (ROW TYPE).
Proof.
  eauto using tc_tag_wk, tc_is_record_implies_tc_tag.
Qed.

Hint Resolve tc_tag_wk tc_is_record_wk : wk.

Lemma tc_is_record_and_tc_tag:
  forall tc tag Ts1 Ts2,
  tc_is_record tc Ts1 ->
  tc_tag tc tag Ts2 ->
  tag = 0 /\ Ts2 = Ts1.
Proof.
  introv h. unfold tc_tag. rewrite h. clear h.
  introv h. dependent destruction h. eauto.
  dependent destruction h.
Qed.

