Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Reduction.
Require Import Row.
Require Import Intro.
Require Import Signature.
Require Import Dup.
Require Import Exclusive.
Require Import Adopts.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JDrop.
Require Import JMonotonicity.
Require Import JDuplication.
Require Import JClassification.
Require Import JDecomposition.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import JHeap.
Require Import JbsInversion.
Require Import WkInversion.
Require Import Give.
Require Import Take.
Require Import DerivedTyping.
Require Import Harmless.
Require Import JtDecomposition.

(* ---------------------------------------------------------------------------- *)

(* If the resource [R] justifies the permission [vs[f:=v] @ Ts[f:=T]], and if
   the type [T] is duplicable, then [core R] justifies the permission [v @ T].
   This lemma is used when reading a field in a memory block. *)

Lemma row_context_read:
  forall f vs,
  row_context f vs ->
  forall R K v T Ts,
  row_context f Ts ->
  jpi R K (zip (fill_row vs v) (fill_row Ts T)) ->
  dup T ->
  consistent R ->
  jctop (core R) K v T.
Proof.
  induction 1; inversion 1; introv hjpi ? ?; subst; simpl_subst_all; simpl in hjpi;
  repeat rewrite subst_lift in hjpi;
  dependent destruction hjpi;
  subst.
  (* Zero *)
  (* The permission [v @ T] is duplicated, hence we must use the lemma
     [jctop_duplicable]. *)
  rewrite core_star_left by assumption.
  solve [ eauto using jctop_duplicable with consistent_split ].
  (* Succ *)
  rewrite core_star_right by assumption.
  solve [ eauto with consistent_split ].
Qed.

Ltac row_context_read h :=
  match goal with
  h3: jpi _ _ (zip (fill_row _ _) (fill_row _ _)) |- _ =>
    forwards h: (>> row_context_read h3); [
      eauto
    | eauto
    | eauto
    | eauto with consistent_split
    | 
    ]
  end.

(* ---------------------------------------------------------------------------- *)

(* If the resource [R1] justifies the permission [vs[f:=v] @ Ts[f:=T]], and if
   the resource [R2] justifies [u @ U], then the conjunction [R1 * R2]
   justifies [vs[f:=u] @ Ts[f:=U]]. This lemma is used when writing a field in
   a memory block. *)

Lemma row_context_write:
  forall R2 K u U,
  jctop R2 K u U ->
  forall f vs,
  row_context f vs ->
  forall R1 v T Ts,
  row_context f Ts ->
  jpi R1 K (zip (fill_row vs v) (fill_row Ts T)) ->
  consistent (star R1 R2) ->
  jpi (star R1 R2) K (zip (fill_row vs u) (fill_row Ts U)).
Proof.
  introv ?; induction 1; inversion 1; introv hjpi ?; subst; simpl_subst_all;
  simpl in *;
  repeat rewrite subst_lift in hjpi;
  repeat rewrite subst_lift;
  dependent destruction hjpi;
  subst.
  (* Zero *)
  (* The permission [v @ T] is lost, hence we must use [j_drop_right]. *)
  solve [ eauto using jpi_drop_right, reconfigure04 with j ].
  (* Succ *)
  eauto using consistent_12p3_23 with j aac.
Qed.

Ltac row_context_write :=
  match goal with
  h4: jpi _ _ (zip (fill_row _ _) (fill_row _ _)) |- _ =>
    forwards: (>> row_context_write h4); [
      eauto
    | eauto
    | eauto
    | eauto using consistent_12p3_13 with consistent_split
    | 
    ]
  end.

(* ---------------------------------------------------------------------------- *)

(* In preparation for the statement of subject reduction, we now define a
   predicate [sr_target R1f K c2 T]. *)

(* This predicate will appear as the conclusion of the preliminary subject
   reduction statement. *)

(* To a first approximation, this judgement means that the configuration [c2]
   has type [T]. More precisely, this configuration is well-typed under some
   resource [R2] that corresponds exactly to the heap [h2], but not all of
   [R2] is required to type-check the term [t2]. The resource [R2] can be
   split as [star R2t R2f]. The component [R2t] is sufficient to type-check
   the term [t2], while the component [R2f] is unused, but we guarantee that
   this component is an evolution of [R1f] along [rio]. Thus, in short, the
   judgement [sr_target R1f K h2 t2 T] means ``the configuration [Cf h2 t2] has
   type [T], and if you framed out something that depended on [R1f], then you
   will be able to frame it back in''. (Phew!) *)

(* The term [t2] is closed, but a non-empty environment [K] is needed because
   the type [T] is not necessarily closed. We do not have a value restriction,
   so reduction may take place under [JForallIntro]. *)

Inductive sr_target : resource -> env kind -> configuration -> term -> Prop :=
| SrTarget:
    forall R1f K h2 t2 T R2t R2tlck R2tavl R2f R2 n,
    jh h2 R2 ->
    star R2t R2f = R2 ->
    rio R1f R2f ->
    star R2tlck R2tavl = R2t ->
    jpi R2tlck K (guarded_permission h2 R2) ->
    jt n R2tavl K TyEmpty t2 T ->
    sr_target R1f K (Cf h2 t2) T.

(* A variant of the constructor [SrTarget], where the type derivation
   happens to be canonical. Defining this variant up front is useful
   because the lemma [jctop_jt] is not easy to apply on the fly,
   due to the existential quantifier in its conclusion. *)

Lemma SrTargetCanonical:
  forall R1f K h2 t2 T R2t R2tlck R2tavl R2f R2,
  jh h2 R2 ->
  star R2t R2f = R2 ->
  rio R1f R2f ->
  star R2tlck R2tavl = R2t ->
  jpi R2tlck K (guarded_permission h2 R2) ->
  jctop R2tavl K t2 T ->
  sr_target R1f K (Cf h2 t2) T.
Proof.
  intros.
  forwards: jctop_jt. eauto. unpack.
  econstructor; eauto.
Qed.

(* The following lemma explains why execution under [JSubRight] is sound. *)

Lemma SrTargetSubRight:
  forall R1f c n K T1 T2,
  sub n K T1 T2 ->
  sr_target R1f K c T1 ->
  sr_target R1f K c T2.
Proof.
  (* The proof is immediate: it suffices to apply the subtyping rule. *)
  inversion 2; intros; subst.
  econstructor; eauto.
  eapply JSubRight; eauto.
Qed.

(* The following lemma explains why execution under [JFold] is sound. *)

Lemma SrTargetFold:
  forall R1f K c m tag Ts U tc,
  sr_target R1f K c (TyBlock m tag Ts U) ->
  tc_mode tc = m ->
  tc_tag tc tag Ts ->
  tc_adopts tc = U ->
  sr_target R1f K c (TyCon tc).
Proof.
  (* The proof is immediate: it suffices to apply [JFold]. *)
  inversion 1; intros; subst.
  econstructor; eauto.
  eapply JFold; eauto.
Qed.

(* The following lemma explains why execution under the frame rule is
    sound. *)

Lemma SrTargetFrame:
  forall R1f K c2 T R Q,
  (* If [R] justifies [Q], *)
  jpi R K Q ->
  (* and if the framed-out resource includes [R], *)
  sr_target (star R R1f) K c2 T ->
  consistent (star R R1f) ->
  (* then we are able to frame [Q] back in. *)
  sr_target R1f K c2 (TyMixStar T Q).
Proof.
  inversion 2; intros; subst.
  (* We are imposing an evolution along [rio] on the frame, which is a
     composite resource, [star R R1f]. Thus, [rio_star] applies. *)
  rio_star.
  (* There remains to check that everything fits together. *)
  econstructor.
    eauto.
    symmetry. rewrite star_associative. reflexivity.
    eauto.
    symmetry. rewrite <- star_associative. reflexivity. auto.
    (* eauto using j_drop_left. *)
    eapply JtTopCut.
      eauto with j.
      (* The monotonicity property [j_rio] guarantees that [Q] is still
         satisfied after [R] has evolved along [rio]. *)
      eapply j_rio;  eauto with j consistent_split.
      aac_reflexivity.
Qed.

(* The following lemma explains why reduction under the evaluation context
   [TApp t _] is sound. *)

(* This lemma is extremely close to the previous one, in its statement and
   in its proof. If the number of evaluation contexts grows, we will be
   tempted to share the common structure in these lemmas. For the moment,
   the duplication is tolerable (and simpler). *)

Lemma SrTargetApp:
  forall Rt R1f K h2 t u2 T U,
  (* If the framed-out resource includes [R], *)
  sr_target (star Rt R1f) K (Cf h2 u2) U ->
  consistent (star Rt R1f) ->
  (* and if [R] justifies that [t] is a function of suitable type, *)
  jctop Rt K t (TyArrow U T) ->
  (* then we are able to frame [t] back in and construct an application. *)
  sr_target R1f K (Cf h2 (TApp t u2)) T.
Proof.
  inversion 1; introv ? hjt; subst.
  forwards: jctop_value. eauto.
  dependent destruction hjt. subst.
  rio_star.
  econstructor.
    eauto.
    symmetry. rewrite star_associative. reflexivity.
    eauto.
    symmetry. rewrite <- star_associative. reflexivity. auto.
    eapply JtTopCut.
      eauto 2 with j.
      eapply j_rio; eauto with j consistent_split.
      aac_reflexivity.
Qed.

(* Reduction under [JForallIntro] is sound. The environment [K] is
   non-empty, but the term [t] must be closed. *)

Lemma SrTargetForallIntro:
  forall R1f kappa K h t T,
  sr_target R1f (insert 0 kappa K) (Cf h t) T ->
  wk zilch t PROG ->
  harmless t ->
  substitutable kappa ->
  sr_target R1f K (Cf h t) (TyForall kappa T).
Proof.
  inversion 1; intros; subst.
  econstructor; eauto.
  eapply jpi_closed; eauto using jh_guarded_permission_closed.
  eapply JForallIntroVariant; eauto using wk_empty.
  simpl_lift_goal.
  j_weakening_closed_tac.
Qed.

(*

(* If a configuration is a well-typed reduction target, it is a fortiori
   a well-typed configuration. *)

Lemma SrTargetConf:
  forall Rf c T,
  sr_target Rf zilch c T ->
  wkc c ->
  jconf c.
Proof.
  inversion 1; subst.
  forwards hc: jh_consistent; [ eauto | ].
    (* rewrite star_commutative in hc; *) rewrite <- star_associative in hc.
  econstructor.
    auto.
    eauto.
    rewrite star_commutative. rewrite star_associative. eauto.
    erewrite <- guarded_permission_drop_right. eauto. eauto using jh_consistent.
    rewrite star_commutative. rewrite star_associative. eauto.
    eapply j_drop_right. eauto. eauto. eauto with consistent_split.
Qed.
  (* (* Drop [Rf], and we are done. *) *)
  (* inversion 1; subst. econstructor; eauto using j_drop_right, jh_consistent. *)

*)

(* ---------------------------------------------------------------------------- *)

(* A group of lemmas that are useful for proving consistency goals. *)

Hint Resolve @SCAM_consistent_allocate @heap_consistent_upd
heap_project_morphism heap_inject_bimorphism
lck_project_morphism lck_inject_bimorphism
@consistent_adopt consistent_mo2dx @consistent_core consistent_p2l
@consistent_split_left @consistent_split_right : consistent.

(* ---------------------------------------------------------------------------- *)

(* Reduction preserves well-typedness. *)

Lemma subject_reduction_preliminary:
  (* The measure. *)
  forall n,
  (* The reduction hypothesis. *)
  forall h1 t1 c2,
  red (Cf h1 t1) c2 ->
  (* The typing hypothesis. The following three hypotheses are equivalent
     to [jt _ R1t K TyEmpty t1 T]. *)
  forall R1t1 K P T,
  jt n R1t1 K P t1 T ->
  forall R1t2 R1tavl,
  jpi R1t2 K P ->
  star R1t1 R1t2 = R1tavl ->
  forall R1tlck R1t,
  jpi R1tlck zilch (guarded_permission h1 R1t) ->
  star R1tlck R1tavl = R1t ->
  (* Even though [K] is non-empty, the program [t1] must be closed. (The
     environment [K] introduces only ghost variables, so to speak.) We
     require the configuration [Cf h1 t1] to be well-sorted: this property
     implies that [t1] is closed, and is preserved by reduction (see the
     lemma [red_wk]). An alternative approach would be to require just
     [wk zilch t1 PROG] and to re-prove on the fly that this property is
     preserved by reduction. *)
  wkc (Cf h1 t1) ->
  (* We have the entire heap [h1] at hand, of which the resource [R1t]
     describes only a part. The rest, which has been framed out already,
     is [R1f]. *)
  forall R1f R1,
  jh h1 R1 ->
  star R1t R1f = R1 ->
  (* If all this holds, then the configuration obtained after this
     reduction step is well-typed. *)
  sr_target R1f K c2 T.
Proof.
  (* Well-founded induction over the height of the type derivation. *)
  intro n. induction_wf ih: lt_wf n.
  (* Introduce the remaining hypotheses. *)
  introv hred hjt ? ? ? ? hwk. intros; subst.
  (* Reason by cases on the last typing rule that was used. *)
  dependent destruction hjt; subst;
  (* If this typing rule implies that [t1] must be a value, then
     we obtain a contradiction, because values are irreducible. *)
  try solve [ false; eauto using (@red_value zilch), red_value with wk ];
  (* Introduce the fact that [R1] is consistent. *)
  (forwards: jh_consistent; [ eassumption | ]).

  (* Case: JFrame *)
  decomposition.
  (* A permission [Q] is framed out. Apply the induction hypothesis, which
     allows us to take a type-preserving step in the absence of [Q], then
     apply the lemma [SrTargetFrame] to recover [Q]. *)
  eapply SrTargetFrame; eauto using reconfigure01, consistent_drop_left, @star_associative.
  eapply ih. eauto. eauto. eauto. eauto. eauto.
    rewrite star_associative in H1. rewrite star_associative in H1.
      erewrite guarded_permission_drop_right in H1.
      eauto. eapply consistent_split_left. eapply consistent_by_equality. eauto. 
      rewrite star_associative. rewrite star_associative. reflexivity.
    eauto. aac_reflexivity. auto. eauto. aac_reflexivity.
  (* solve [ eapply SrTargetFrame; eauto using reconfigure01, consistent_drop_left, @star_associative ]. *)

  (* Case: JCut *)
  (* Just apply the induction hypothesis. *)
  solve [ eauto with j aac ].

  (* Case: JSubLeft *)
  (* The lemma [jps_sound] allows us to transport the [jpi] hypothesis
     up to the level of the premise. There, we apply the induction
     hypothesis, and we are done. We do not even need to reconstruct
     an instance of [JSubLeft]. *)
  solve [ eauto 7 using jps_sound with consistent_split ].

  (* Case: JSubRight *)
  (* Just work under this instance of [JSubRight]. *)
  solve [ eapply SrTargetSubRight; eauto ].

  (* Case: JApplication *)
  (* Reason by cases on the last reduction rule that was used. In this
     rule, the redex must exhibit an application at the root. *)
  dependent destruction hred;
  decomposition.

    (* Sub-case: RedBeta *)
    clear ih.
    (* We have an application of the form [TApp (TAbs t) u]. Because of the form
       of the pre-condition in the conclusion of [JApplication], a permission of the form
       [(TAbs t @ U -> T) * Q] is known to hold. This implies that the term [t]
       has type [T] (under suitable assumptions) and that the permission [Q] holds. *)
    (* Because [t] is well-typed under [core R], it is also well-typed under [R].
       Here, the function [TAbs t] is applied just once, so there is no need to
       argue that it is well-typed under a duplicable resource. *)
    forwards: j_drop_right. eassumption. eapply core_left_unit. eauto with consistent_split.
    (* Because [u] is a value, and is well-typed according to [jt],
       it must be well-typed according to [jc] as well. *)
    forwards hjct: (>> canonicalization_preliminary hjt);
      eauto 4 using consistent_1p23_13, wk_empty with consistent_split.
    dependent destruction hjct.
    (* In this case, there is no side effect. Simplify the goal. *)
    erewrite <- guarded_permission_drop_right in H2 by eauto.
    eapply SrTarget; eauto using @rio_reflexive, jpi_zilch.
    (* We now need to prove that [subst u 0 t] is well-typed under a
       syntactic pre-condition of our choosing. *)
    eapply JtTopCut.
    (* By substituting [u] for the variable [0] throughout the type
       derivation for [t], we are able to justify that [subst u 0 t]
       is well-typed. *)
    solve [ eapply j_substitution; eauto with substitutable subst_lift ].
    (* There remains to compute which resources justify the pre-condition
       of this type derivation... *)
    simpl_subst_goal.
    subst_lift_hint.
    solve [ eauto with j ].
    (* ... and to check that these resources coincide with the resources
       that are at hand. *)
    aac_reflexivity.

    (* Sub-case: RedContext *)
    (* We are looking at the reduction of [ E[u1] ] to [ E[u2] ]. First decompose
       the context [E], whose only case is a reduction of [TApp t u1] to [TApp t u2].
       Then apply the induction hypothesis, which proves that the reduction of [u1]
       to [u2] is type-preserving. Finally, use [SrTargetApp] to reconstruct a type 
       derivation for [TApp t u2]. *)
    open_eval_context.
    injections.

    eapply SrTargetApp; [ | | eauto ].
      eapply ih. eauto. eauto. eauto. eauto. eauto.
      erewrite guarded_permission_drop_right in H2. eauto. eauto with consistent_split.
      symmetry. rewrite reconfigure12. eauto. auto.
      eauto using wkc_TApp_inversion. eauto.
    aac_reflexivity.
    consistent_aac_crush.

    (* Sub-case: RedContextAppRightFail *)
    (* This case is easy, since [TFail] is always well-typed. *)
    forwards: jctop_wk; [ eauto | wk_inversion ].
    eapply SrTarget; eauto using @rio_reflexive.
    erewrite guarded_permission_drop_right by eauto. auto using jpi_zilch.
    solve [ eapply (JFail 0); eauto ].

  (* Case: JFold *)
  solve [ eapply SrTargetFold; eauto ].

  (* Case: JNew *)
  clear ih.
  (* Only one reduction rule is applicable. *)
  (* Actually, three rules are now applicable, but a discrimination on
     the evaluation context removes the two new cases. *)
  dependent destruction hred; try open_eval_context_discriminate.
  (* Split the goal into a number of sub-goals. *)
  eapply SrTarget.
  (* Sub-goal 1. *)
  (* The new heap is well-typed. *)
  eapply jh_heap_allocate. eassumption.
  (* Sub-goal 2. *)
  (* Push [allocate] into [star]. *)
  eapply SCAM_star_allocate_core_right_unit; eauto with typeclass_instances.
  (* Sub-goal 3. *)
  (* The framed-out resource evolves along [rio]. It sees that a new
     memory location appears, but gets no permission for it, or more
     precisely, it receives a permission of the form [core _]. *)
  eapply rio_allocate_core_mo2dx.
  (* Additional subgoals. *)
  eapply SCAM_star_allocate_core_left_unit; eauto with typeclass_instances.
  rewrite guarded_permission_concrete_heap_allocate.
  rewrite guarded_permission_heap_alloc.
  erewrite guarded_permission_drop_right by eauto.
  eapply j_rio. eauto using jpi_zilch. eauto with consistent_split.
  eapply SCAM_rio_allocate.
    eapply heap_inject_bimorphism.
    eapply heap_sub_resource.
    eauto using @consistent_core, consistent_mo2dx.
    eauto using @cores_are_duplicable.
  (* Sub-goal 4. *)
  (* The new memory location is well-typed. *)
  (* The resource [R] is of no use in proving this fact, so we must
     first drop it. *)
  eapply j_drop_left; [ |
    eapply (SCAM_star_allocate_core_left_unit heap_project_morphism heap_inject_bimorphism)
  | eauto with consistent ].
  (* We can then apply the typing rule for memory locations. *)
  (* A resource of the form [allocate R (core p)] is used to check the values that
     are held in the fields, while a permission of the form [allocate (core R) p]
     is used to justify the ownership of the new memory location. The lemma
     [CAM_star_allocate_core] is used to indicate this. *)
  eapply (JBlock 0).
    eapply (SCAM_star_allocate_core heap_project_morphism heap_inject_bimorphism).
    solve [ eauto using wkc_TNew_inversion ].
    (* The values [vs] were well-typed under [Rvs], and must now be well-typed
       under [allocate Rvs (core p)]. The lemma [j_rio] justifies this. *)
    eapply j_rio; eauto using rio_allocate_core_mo2dx with consistent.
    (* A new object has no adoptees. *)
    eapply AbandonAllZero.
      eapply new_objects_have_no_adoptees.
        eauto with consistent.
        eapply jh_heap_limit; eauto.
        rewrite <- star_associative.
        rewrite (SCAM_limit_split_right heap_project_morphism); auto.
        eapply (SCAM_limit_limit_ad_hoc heap_project_morphism).
        consistent_aac_crush.
        rewrite star_associative. auto.
    eapply core_left_unit.
    eapply JpiEmpty.
    eauto with wk.
    (* Ownership of the new memory location. *)
    eapply SCAM_maps_allocate_there.
      eapply heap_project_morphism.
      eapply heap_sub_resource.
      eauto with consistent.
      eapply jh_heap_limit; eauto.
        rewrite <- star_associative.
        rewrite (SCAM_limit_split_right heap_project_morphism); auto.
        eapply (SCAM_limit_limit_ad_hoc heap_project_morphism).
        consistent_aac_crush.
        rewrite star_associative. auto.

  (* Case: JRead *)

  clear ih.
  (* Only one reduction rule is applicable. *)
  dependent destruction hred; try open_eval_context_discriminate.
  (* The memory location [l] is allocated in the heap, and is allocated in *)
  (*    the current resource. Furthermore, the heap and the current resource *)
  (*    agree. Draw the consequences. *)
  dca.
  jh_maps_maps.
  (* The value [v] found at address [l] is well-typed under some resource [Rv]. *)
  (*    Because this value admits a duplicable type, this derivation is in fact *)
  (*    valid under [core Rv]. *)
  row_context_read hcore.
  (* In this case, there is no side effect. Simplify the goal, and prepare to *)
  (*    construct a new type derivation, which happens to be canonical. *)
  eapply SrTargetCanonical; eauto using @rio_reflexive.
    erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
  (* Drop a resource that we do not need. *)
  eapply jpi_drop_right; [ | eapply star_commutative | eauto with consistent_split ].
  (* Apply the frame rule. *)
  eapply JcTopFrame.
    (* In the first premise, consume [core Rv]. *)
    eexact hcore.
    (* In the second premise, consume [Rv]. *)
    clear hcore. eapply JcTopLoc; eauto.
    (* Because [core Rv] and [Rv] combine to produce [Rv], this *)
  (*      application of the frame rule is valid. *)
    match goal with Rv: resource |- _ => aac_rewrite (core_left_unit Rv) end.
    reflexivity.

  (* Case: JWrite *)

  clear ih.
  (* Only one reduction rule is applicable. *)
  dependent destruction hred; try open_eval_context_discriminate.
  (* The memory location [l] is allocated in the heap, and is allocated in *)
  (*    the current resource. Furthermore, the heap and the current resource *)
  (*    agree. Draw the consequences. *)
  dca.
  jh_maps_maps.
  (* The value [v2] is well-typed. We do not care about the old value [v1], *)
  (*    since it is lost. Of course, it is well-typed too. *)
  (* Split the goal into a number of sub-goals. *)
  eapply SrTarget; try eapply rio_reflexive.
  (* Sub-goal 1. *)
  (* The heap remains well-typed. *)
  eapply jh_heap_update. eassumption.
  (* Sub-goal 2. *)
  (* Push [upd] into the appropriate side of [star]. The framed-out *)
  (*    resource does not evolve. *)
  eapply star_upd_exclusive_left; [ SCAM_maps_maximal heap_project_morphism | eauto ].
  (* Additional sub-goals. *)
  eapply star_upd_exclusive_right; [ SCAM_maps_maximal heap_project_morphism | eauto with consistent_split ].
  unfold heap_upd.
  unfold SCAM_update, SAM_update.
  rewrite <- guarded_permission_heap_inject_invariant.
  erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
  (* (* Sub-goal 3. *) *)
  (* (* Push [upd] down into the appropriate conjunct. *) *)
  (* star_upd_exclusive. *)
  (* We now prove that the reduct is well-typed. *)
  eapply JtTopUnitMixStar; [ | | eauto using heap_consistent_upd with consistent_split ].
    (* Sub-sub-goal 4a: the reference cell. *)
    row_context_write.
    wk_inversion.
    forwards: wk_TWrite_inversion. eauto.
    solve [ eauto 7 using JcTopLoc, wk_substitution,
                          abandon_all_update, @SCAM_maps_maximal_right, @maximal_exclusive,
                          heap_project_morphism,
                          star_upd_exclusive_right, heap_maps_upd_there ].
    (* Sub-sub-goal 4b: check our resources. *)
    star_upd_exclusive; [ f_equal; aac_reflexivity | eauto 6 with consistent_split ].

  (* Case: JWriteTag *)

  clear ih.
  (* Only one reduction rule is applicable. *)
  dependent destruction hred; try open_eval_context_discriminate.
  (* The memory location [l] is allocated in the heap, and is allocated in *)
  (*    the current resource. Furthermore, the heap and the current resource *)
  (*    agree. Draw the consequences. *)
  dca.
  jh_maps_maps.
  (* Treat separately the case where the new mode [m] is [Duplicable] -- *)
  (*    so the block becomes frozen -- and the case where [m] is [Exclusive] *)
  (*    -- so the block remains mutable. *)
  destruct m.

    (* Case: [m] is [Duplicable]. This is a ``freeze''. *)
    (* Split the goal into a number of sub-goals. *)
    eapply SrTarget.
    (* Sub-goal 1. *)
    (* The heap remains well-typed. *)
    eapply jh_heap_update; eauto.
    (* Sub-goal 2. *)
    eapply heap_star_upd.
    (* Sub-goal 3. *)
    (* We have exclusive permission over [l], so the framed-out resource *)
  (*      must have no permission over [l], that is, [DX_NONE]. This permission *)
  (*      evolves to [DX_DUPLICABLE _]. This update is permitted by [rio]. *)
    eapply heap_rio_upd_duplicable; [ eauto | consistent_aac_crush ].
    (* Additional goals. *)
    eapply heap_star_upd.
    unfold heap_upd at 2.
    unfold SCAM_update, SAM_update.
    rewrite <- guarded_permission_heap_inject_invariant.
    eapply j_rio.
      erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
      eauto with consistent_split.
      eapply heap_rio_upd_duplicable; [ eauto | consistent_aac_crush ].
    (* Sub-goal 4. *)
    (* Push [upd] down into [star], while keeping a consistency fact *)
  (*      along the way. *)
    match goal with |- jt _ ?R _ _ _ _ =>
      cut (consistent R); [ | eauto with consistent ]
    end.
    repeat rewrite <- heap_star_upd.
    intro.
    (* We now prove that the reduct is well-typed. *)
    eapply JtTopUnitMixStar; [ | eauto | eauto ].
    eapply JcTopLoc.
      eauto.
      eassumption.
      (* The fields. *)
      eapply j_rio; eauto 6 using heap_rio_upd_duplicable, @consistent_star_commutative with consistent.
      (* A duplicable object cannot have any adoptees, so we lose the *)
  (*        adoptees (if there were any). The typing rule [JBlock] is affine *)
  (*        by essence -- the resource that justifies the ownership of [l] *)
  (*        can cover more than just [l] -- so we do not even need to use *)
  (*        the lemma [j_drop] explicitly. *)
      eapply AbandonAllZero.
        eapply duplicable_objects_have_no_adoptees.
          eapply @SCAM_maps_update_there; eauto with typeclass_instances.
      eapply core_left_unit.
      eapply JpiEmpty.
      eauto with wk.
      (* The location [l]. *)
      eauto using heap_maps_upd_there.

    (* Case: [m] is [Exclusive]. This is a normal tag update. *)
    (* Split the goal into a number of sub-goals. *)
    eapply SrTarget; try eapply rio_reflexive.
    (* Sub-goal 1. *)
    (* The heap remains well-typed. *)
    eapply jh_heap_update. eassumption.
    (* Sub-goal 2. *)
    (* Push [upd] into the appropriate side of [star]. The framed-out *)
  (*      resource does not evolve. *)
    eapply star_upd_exclusive_left; [ SCAM_maps_maximal heap_project_morphism | eauto ].
  (* Additional sub-goals. *)
  eapply star_upd_exclusive_right; [ SCAM_maps_maximal heap_project_morphism | eauto with consistent_split ].
  unfold heap_upd.
  unfold SCAM_update, SAM_update.
  rewrite <- guarded_permission_heap_inject_invariant.
  erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
    (* Sub-goal 3. *)
    (* Push [update] down into the appropriate conjunct. *)
    star_upd_exclusive.
    (* We now prove that the reduct is well-typed. *)
    eapply JtTopUnitMixStar; [ | eauto | eauto ].
    solve [ eauto 6 using JcTopLoc,
                          abandon_all_update, @SCAM_maps_maximal_right, @maximal_exclusive,
                          heap_project_morphism,
                          star_upd_exclusive_right, heap_maps_upd_there ].

  (* Case: JSwitch *)

  clear ih.
  (* Only one reduction rule is applicable. *)
  dependent destruction hred; try open_eval_context_discriminate.
  (* The memory location [l] is allocated in the heap, and is allocated in *)
  (*    the current resource. Furthermore, the heap and the current resource *)
  (*    agree. Draw the consequences. *)
  decomposition.
  decomposition_TyCon.
  dca.
  jh_maps_maps.
  (* The branch that corresponds to the tag carried by this memory block *)
  (*    must be well-typed. *)
  jbs_inversion.
  (* In this case, there is no side effect. *)
  eapply SrTarget; eauto using @rio_reflexive.
  (* Additional sub-goal. *)
  erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
  (* Building a new derivation is easy. *)
  eapply JtTopCut; [ eassumption | | eapply star_commutative ].
    solve [ eauto using JpiAtomicEmpty, (JBlock 0) with jpi wk ].

  (* Case: JGive *)

  clear ih.
  (* Only one reduction rule is applicable. *)
  dependent destruction hred; try open_eval_context_discriminate.
  (* The memory locations [y] and [x] have types [U] and [T]. *)
  decomposition.
  (* The following consistency facts will be used later on (!). *)
  (* match goal with h: consistent (star (star ?Rtlck (star ?R (star ?R1 ?R2))) ?R1f) |- _ => *)
  (*   generalize h; clear h; intro hc1; *)
  (*   assert (hc2: consistent (star Rtlck (star R (star R1 R2)))); [ eauto with consistent_split | ]; *)
  (*   assert (hc3: consistent (star R1 R2)); [ eauto with consistent_split | ]; *)
  (*   assert (hc4: consistent (star R (star R1 R2))); [ eauto with consistent_split | ] *)
  (* end. *)
  (* [y] has an exclusive type, so it must in fact have a block type. *)
  match goal with hu: jctop _ _ _ ?U, ha: adopts ?T ?U |- _ =>
    unpack (decomposition_adoptee hu ha); subst
  end.
  (* The heap and the current resource agree. Draw the consequences. *)
  dca.
  jh_maps_maps.
  (* Split the goal. *)
  eapply SrTarget; try eapply rio_reflexive.
  (* Sub-goal 1. *)
  (* The modified heap is well-typed. *)
  eapply jh_adopt. eauto.
  (* Sub-goal 2. *)
  (* Push [adopt] into [star]. *)
  eapply star_adopt_left_LR. SCAM_maps_maximal heap_project_morphism.
  (* Additional sub-goals. *)
  eapply star_adopt_right_LR. SCAM_maps_maximal heap_project_morphism.
  unfold adopt.
  unfold SCAM_update, SAM_update.
  rewrite <- guarded_permission_heap_inject_invariant.
  erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
  (* Sub-goal 3. *)
  eapply JtTopUnitMixStar.
    (* The following incantation is quite crazy. *)
    (* [eapply give] does not work because it creates meta-variables *)
  (*      but does not stage them as sub-goals. In Coq 8.4, we could *)
  (*      grab them a posteriori, but I don't like this approach very much. *)
    match goal with ha: adopts ?T ?U, hx: jctop ?R2 _ _ ?T, hy: jctop ?R1 _ _ ?U |- _ =>
      eapply (give ha hx hy)
    end. SCAM_maps_maximal heap_project_morphism.
    eapply star_adopt_right_LR. SCAM_maps_maximal heap_project_morphism.
    eauto with consistent.

  (* Case: JTake *)

  clear ih.
  decomposition.
  (* There are two reduction rules to consider. *)
  dependent destruction hred; try open_eval_context_discriminate.
    (* Sub-case: RedTakeSuccess *)
    (* There is a block at address [y], whose [adopter] field contains *)
  (*      the address [x]. Because the heap and the instrumented heap agree, *)
  (*      the latter must map [y] to a permission [p] that is compatible with *)
  (*      this knowledge. *)
    forwards: (@SCAM_agrees_maps_left_to_right _ _ concrete_heap_project _ _ _ _ _ heap_project).
    match goal with
      | h: jh _ _ |- _ => inversion h; eauto
    end.
    eauto.
    unpack.
    (* What could the permission [p] be? It cannot be duplicable, because *)
  (*      a duplicable object never has a non-null [adopter] field. Hence, it *)
  (*      must be exclusive. *)
    match goal with h: heap_agree _ _ |- _ => revert h end; dependent_destruction;
    block_orphan_agree_inversion; subst.
    (* Now, because adoptees and adopter are never separated by a star, *)
  (*      the resource [R] that owns [x] also owns [y]. *)
    forwards: types_adopter_implies_owns_adoptee; eauto using reconfigure16 with exclusive.
    (* We are now in a position to apply the key abandon lemma. If [x] *)
  (*      abandons [y], then [x] and [y] are separately well-typed. *)
    forwards: take; [ eauto | eauto | eauto | eauto with consistent_split | unpack ].
    (* We can now build a new type derivation. *)
    eapply SrTarget; try eapply rio_reflexive.
    (* Sub-goal 1. *)
    (* The modified heap is well-typed. *)
    eapply jh_heap_update. eauto.
    (* Sub-goal 2. *)
    (* Push [upd] into [star]. *)
    eapply star_upd_exclusive_left; [ SCAM_maps_maximal heap_project_morphism | eauto ].
    (* Additional sub-goals. *)
    eapply star_upd_exclusive_right; [ SCAM_maps_maximal heap_project_morphism | eauto with consistent_split ].
    unfold heap_upd.
    unfold SCAM_update, SAM_update.
    rewrite <- guarded_permission_heap_inject_invariant.
    erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
    (* Sub-goal 3. *)
    eapply JtTopUnitMixStar.
      eauto with j.
      star_upd_exclusive. rewrite star_associative_symmetric. eauto with f_equal.
      eauto with consistent.

    (* Sub-case: RedTakeFailure *)
    (* This case is easy, since [TFail] is always well-typed. *)
    eapply SrTarget; eauto using @rio_reflexive.
    erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.    
    solve [ eapply (JFail 0); eauto 9 with wk ].

  (* Case: JFail *)
  (* This case is impossible, because [fail] does not reduce. However, the two
     cases with context inversion are still there. *)
  dependent destruction hred; open_eval_context_discriminate.

  (* Case: JSpawn *)
  dependent destruction hred; open_eval_context_discriminate.

  (* Case: JNewLck *)
  clear ih.
  dependent destruction hred; try open_eval_context_discriminate.
  eapply SrTarget.
    eapply jh_lck_allocate; eassumption.
    eapply SCAM_star_allocate_core_right_unit;
      eauto with typeclass_instances.
      eauto using rio_allocate_core_p2l.

    symmetry; rewrite <- reconfigure00; symmetry.
    eapply SCAM_star_allocate_core_right_unit;
      eauto with typeclass_instances.

    rewrite guarded_permission_new_lck by eauto.
    eapply j_rio.
    eapply JpiStar.
      erewrite guarded_permission_drop_right by eauto.
      eauto using jpi_zilch. eauto.
      eauto using @star_commutative.
      consistent_aac_crush.
      eapply SCAM_rio_allocate;
        eauto using consistent_p2l, duplicable_p2l with typeclass_instances.
      
      eapply (JLck 0).
        eauto.
        eapply SCAM_maps_allocate_there;
          eauto with typeclass_instances consistent_split.
          forwards hl: jh_lck_limit;
            [ eauto
            | eauto
            | rewrite hl; SCAM_limit_split lck_project_morphism ].
          rewrite core_p2l. left; auto.

(*
    erewrite <- @SCAM_star_allocate_core_right_unit by eauto with typeclass_instances.
    eapply JpiStar.
      erewrite guarded_permission_drop_right by eauto.
      eauto using jpi_zilch. eauto with consistent_split.
    Check @SCAM_star_allocate_core_right_unit.
      erewrite guarded_permission_drop_right by eauto.
      eapply JpiStar.
        eapply j_rio. eauto using jpi_zilch. eauto with consistent_split.
        eauto using rio_allocate_core_p2l.
        eauto.

    eapply SCAM_star_allocate_core_left_unit;
      eauto with typeclass_instances.

    rewrite guarded_permission_new_lck by eauto.
      erewrite guarded_permission_drop_right by eauto.
      eapply JpiStar.
        eapply j_rio. eauto using jpi_zilch. eauto with consistent_split.
        eauto using rio_allocate_core_p2l.
        eauto.

      (* eapply JpiStarEmpty. *)
      eapply j_rio. eauto using jpi_zilch. eauto with consistent_split.
        (* replace (core (p2hl P)) with (core (p2l P)); *)
          eauto using rio_allocate_core_p2l.
    eapply j_drop_left;
      [
      | eapply SCAM_star_allocate_core_left_unit; 
          eauto with typeclass_instances
      | eauto using consistent_p2l with consistent ].
      eapply (JLck 0).
        auto.
        eapply SCAM_maps_allocate_there;
          eauto with typeclass_instances consistent_split.
          forwards hl: jh_lck_limit;
            [ eauto
            | eauto
            | rewrite hl; SCAM_limit_split lck_project_morphism ].
        right; auto.
*)

  (* Case: JAcquire *)
  clear ih.
  dependent destruction hred; try open_eval_context_discriminate.
  decomposition.

  lck_agrees_maps_left_to_right.
  lck_agree_inversion.

  forwards gpa: guarded_permission_acquire.
    eauto. eauto. eassumption.
    rewrite star_associative. reflexivity.
    erewrite guarded_permission_drop_right by eauto. assumption.
  unpack. subst.

  lck_maps_lock_split.
  SCAM_maps_injective.
  is_locked_permission_inversion.

  eapply SrTarget;
    [ eapply jh_lck_update_Taken; eauto
    |
    | eapply @rio_reflexive
    | reflexivity
    | eauto using jpi_zilch
    | ].

  rewrite reconfigure18.
  rewrite <- (@SCAM_star_update_core_left _ _ _ _ _ _ _ _ lck_project_morphism lck_inject_bimorphism lck_sub_resource _ _ _ _ (lck_update_dreq _ _ _) (lck_update_oreq _ _ _)) by eauto.
  f_equal.
  eauto using (@SCAM_star_update_core_right _ _ _ _ _ _ _ _ lck_project_morphism lck_inject_bimorphism lck_sub_resource _ _ _ _ (lck_update_dreq _ _ _) (lck_update_oreq _ _ _)) with typeclass_instances.

  eapply JtTopUnitMixStar; [ | eapply @core_left_unit | ].
  rewrite <- (@SCAM_star_update_core_right _ _ _ _ _ _ _ _ lck_project_morphism lck_inject_bimorphism lck_sub_resource _ _ _ _ (lck_update_dreq _ _ _) (lck_update_oreq _ _ _)).
  econstructor; [ eauto using jpi_zilch | | eauto ].
    eapply JcTopLocked.
      instantiate (1 := P). eauto with wk.
      forwards: lck_resource_maps_star;
        [ eauto
        | econstructor; eauto
        | eauto
        | eauto with consistent_split
        | unpack; eapply (SCAM_maps_update_there lck_sub_resource); eauto ].
      eauto.
      eapply (SCAM_consistent_update lck_project_morphism lck_inject_bimorphism);
        [ consistent_aac_crush
        | eapply consistent_p2hl ].

  (* Case: JRelease *)
  clear ih.
  dependent destruction hred; try open_eval_context_discriminate.
  decomposition.

  lck_agrees_maps_left_to_right.
  lck_agree_inversion.
  lck_maps_held_lock_split.

  eapply SrTarget;
    [ eapply jh_lck_update_Free; eauto using SAM_maps_limit
    |
    | eapply @rio_reflexive
    | eauto
    | 
    | eapply JUnit ].

  rewrite <- (@SCAM_star_update_core_left _ _ _ _ _ _ _ _ lck_project_morphism lck_inject_bimorphism lck_sub_resource _ _ _ _ (lck_update_dreq _ _ _) (lck_update_oreq _ _ _)) by eassumption.
  repeat rewrite <- (@SCAM_star_update_core_right _ _ _ _ _ _ _ _ lck_project_morphism lck_inject_bimorphism lck_sub_resource _ _ _ _ (lck_update_dreq _ _ _) (lck_update_oreq _ _ _)) by eassumption.
  rewrite <- reconfigure14. eauto.

  forwards hgpr: guarded_permission_release_jps;
    [ eassumption
    | eassumption
    | SCAM_maps_maximal lck_project_morphism
    | named_unpack hgpr ].
  eapply jps_sound.
    eapply hgpr.
    econstructor.
      eauto.
      erewrite guarded_permission_drop_right by eauto; eauto using jpi_zilch.
      aac_reflexivity.
    consistent_aac_crush.
      
  (* Case: JExistsElim *)

  decomposition.
  (* We have a type derivation for the term [t] under the precondition [P],
     which refers to the name [x]. We know that [[v/x]P] holds. Hence, we
     substitute [v] for [x] throughout the type derivation. Note that this
     does not affect the term [t] or its type [T], because they do not refer
     to [x]. We are then able to apply the induction hypothesis. *)
  substitution.
  dependent destruction hred; try open_eval_context_discriminate.
  wk_inversion.
  eapply SrTarget;
    eauto using @rio_reflexive.
    erewrite guarded_permission_drop_right by eauto. eauto using jpi_zilch.
    eauto using @star_commutative with j.

  (* Case: JForallIntro *)

  (* This case is present because we do not have a value restriction. *)
  jpi_inversion.
  (* The term [t] is closed, so [shift 0 t] is [t]. *)
  match goal with h: jt _ _ _ _ (shift 0 _) _ |- _ =>
    erewrite closed_lift_invariant in h by eauto using wkc_wk, wk_closed, length_empty
  end.
  match goal with c: configuration |- _ => destruct c end.
  (* Apply the ad hoc lemma. *)
  eapply SrTargetForallIntro; eauto.
  (* We must now prove that the reduct is closed. This holds thanks
     to the fact that [wkc] is preserved by reduction. *)
  eauto using wkc_wk, red_wk.
  wk_inversion.
  eauto using harmless_red with wk.

Grab Existential Variables.
auto. auto. auto.
consistent_aac_crush.
consistent_aac_crush.
consistent_aac_crush.
auto.

Qed.

(*

(* ---------------------------------------------------------------------------- *)

(* Subject reduction. A clean re-statement. *)

Lemma subject_reduction:
  forall c1 c2,
  red c1 c2 ->
  jconf c1 ->
  jconf c2.
Proof.
  inversion 2; subst.
  eapply SrTargetConf; [ | eauto using red_wk with wk ].
  eapply subject_reduction_preliminary.
    eauto.
    eauto.
    eapply JpiEmpty.
    eauto.
    erewrite guarded_permission_drop_right.
      eauto.
      eauto using jh_consistent.
      eauto.
    rewrite @core_right_unit. aac_reflexivity.
    eauto.
    eauto.
    eauto using @core_right_unit.
Qed.
  (* inversion 2; subst. *)
  (* eauto using SrTargetConf, subject_reduction_preliminary, @core_right_unit, red_wk with j wk. *)
(* Qed. *)

*)

(* ---------------------------------------------------------------------------- *)

(* A concurrent version of [sr_target]. *)

Inductive csr_target : resource -> configuration -> Prop :=
| CsrTarget:
    forall R1f h2 t2 R2t R2tlck R2tavl R2f R2,
    jh h2 R2 ->
    star R2t R2f = R2 ->
    rio R1f R2f ->
    star R2tlck R2tavl = R2t ->
    jpi R2tlck zilch (guarded_permission h2 R2) ->
    jcp R2tavl t2 ->
    csr_target R1f (Cf h2 t2).

Lemma CsrTargetConf:
  forall Rf c,
  csr_target Rf c ->
  wkcp c ->
  jconf c.
Proof.
  (* Drop [Rf], and we are done. *)
  inversion 1; subst.
  forwards: jh_consistent; [ eauto | ].
  econstructor.
    auto.
    eauto.
    rewrite star_commutative. rewrite star_associative. eauto.
    erewrite <- guarded_permission_drop_right. eauto. eauto using jh_consistent.
    rewrite star_commutative. rewrite star_associative. eauto.
    eapply jcp_drop_right. auto. rewrite <- star_associative in H1. eauto with consistent_split.
Qed.

Lemma srtarget_csrtarget:
  forall R h t T,
  sr_target R zilch (Cf h t) T ->
  csr_target R (Cf h (TThread t)).
Proof.
  inversion 1; subst.
  eauto using CsrTarget, (get1 j_term) with jcp.
Qed.

(* A preliminary lemma for subject reduction of concurrent programs. *)
(* This lemma follows the same principle as [subject_reduction_preliminary]
   in [SubjectReduction.v]. *)

Lemma c_subject_reduction_preliminary:
  (* The reduction hypothesis. *)
  forall h1 t1 c2,
  red (Cf h1 t1) c2 ->
  (* The typing hypothesis. *)
  forall R1tavl R1tlck R1t,
  jcp R1tavl t1 ->
  jpi R1tlck zilch (guarded_permission h1 R1t) ->
  star R1tlck R1tavl = R1t ->  
  (* The program [t1] must be closed. We require the configuration [Cf h1 t1]
     to be well-sorted: this property implies that [t1] is closed, and is
     preserved by reduction (see the lemma [red_wkcp]). *)
  wkcp (Cf h1 t1) ->
  (* We have the entire heap [h1] at hand, of which the resource [R1t]
     describes only a part. The rest, which has been framed out already,
     is [R1f]. *)
  forall R1f R1,
  jh h1 R1 ->
  star R1t R1f = R1 ->
  (* If all this holds, then the configuration obtained after this
     reduction step is well-typed. *)
  csr_target R1f c2.
Proof.
  (* Induction on the reduction hypothesis. *)
  introv hred.
  dependent induction hred;
  (* Unfold the evaluation contexts and remove the cases where... *)
  introv hjcp hjpi hr1; intros; try open_eval_context;
  try solve [ discriminate (* ... this reveals a bad constructor *)
            | wk_inversion (* ... or well_kindedness hypotheses are absurd. *)
  ].

  (* Reduction in a thread. *)
  (* Retrieve well-kindedness information. *)
    wk_inversion.
  (* Destruct the concurrent typing hypothesis. *) 
    jcp_inversion.
    make_jt_hypothesis hjt.
  (* Use the preliminary subject reduction lemma for sequential terms. *)
    forwards: subject_reduction_preliminary;
      eauto using @core_right_unit with j wk.
    (* assert (hcore: R1t = star R1t (core R1t)); *)
    (*   [ eauto using @core_right_unit  *)
    (*   | rewrite hcore at 2; eauto ]. *)
  (* Conclude with the lemma connecting sequential and concurrent
     subject reduction. *)
    match goal with
      | h: sr_target _ _ _ _ |- _ =>
          inversion h; eauto using srtarget_csrtarget
    end.

  (* Reduction on the left. *)
    wk_inversion; jcp_inversion.
  (* Induction hypothesis. *)
    rewrite @star_associative in H0.
    rewrite @star_associative in hjpi.
    erewrite guarded_permission_drop_right in hjpi; eauto using jh_consistent with consistent_split.
    forwards: IHhred; eauto using @star_associative with wk; clear IHhred.
    match goal with
      | h: csr_target _ _ |- _ => inversion h; subst
    end.
    (* Proves consistency to allow the use of [rio_star]. *)
    assert (hcons: consistent (star R2 R1f)).
    match goal with
      | h: jh ?heap (star (star _ R2) R1f) |- _ =>
          rewrite <- star_associative in h;
          forwards: jh_consistent h;
          eauto using @consistent_split_right
    end.
    rio_star.
    (* Build the proof. *)
    eapply CsrTarget;
      [ eauto
      | symmetry; rewrite reconfigure01; eauto
      | eauto
      | eauto
      | eauto
      | ].
    eapply JcpPar;
    try (match goal with
      | h: jcp _ ?t |- wk zilch ?t CPROG => inversion h; eauto with wk
      end);
    eauto using jcp_rio, @consistent_split_left.

  (* Reduction on the right. *)
    wk_inversion; jcp_inversion.
  (* Induction hypothesis. *)
    assert (hr12: star R1 R2 = star R2 R1);
      [ eauto using @star_commutative
      | ].
    rewrite hr12 in hjpi. rewrite @star_associative in hjpi.
    rewrite hr12 in H0. rewrite @star_associative in H0.
    erewrite guarded_permission_drop_right in hjpi; eauto using jh_consistent with consistent_split.
    forwards: IHhred; try eauto with wk.
    rewrite @star_associative. eauto.
    match goal with
      | h: csr_target _ _ |- _ => inversion h; subst
    end.
    (* Proves consistency to allow the use of [rio_star]. *)
    assert (hcons: consistent (star R1 R1f)).
    match goal with
      | h: jh ?heap (star (star _ R1) R1f) |- _ =>
          rewrite <- star_associative in h;
          forwards: jh_consistent h;
          eauto using @consistent_split_right
    end.
    rio_star.
    (* Build the proof. *)
    eapply CsrTarget;
      [ eauto
      | symmetry; rewrite reconfigure01; eauto
      | eauto
      | eauto 
      | eauto
      | ].
    eapply JcpPar;
    try (match goal with
      | h: jcp _ ?t |- wk zilch ?t CPROG => inversion h; eauto with wk
      end);
    try rewrite @star_commutative;
    eauto using jcp_rio, @consistent_split_left.

  (* Spawn. *)
    subst.
    eapply CsrTarget;
      [ eauto
      | eauto
      | eapply rio_reflexive
      | eauto
      | erewrite guarded_permission_drop_right; eauto using jh_consistent with consistent_split
      | ].
    (* Retrieve well_kindedness and typing information. *)
    wk_inversion. jcp_inversion. make_jt_hypothesis hjt.
    (* Inverse the typing of spawn in a context. *)
    forwards hctx: spawn_in_context_decomposition.
    eauto.
    match goal with
      | h: wk _ v1 _ |- _ => eapply h
    end.
    match goal with
      | h: wk _ v2 _ |- _ => eapply h
    end.
    eauto.
    eauto with wk.
    eauto with j.
    eapply core_right_unit.
    eauto using jh_consistent with consistent_split.
    destruct hctx as (Rv, hctx).
    destruct hctx as (Tv1, hctx).
    destruct hctx as (Tv2, hctx).
    destruct hctx as (nd, hctx).
    destruct hctx as (Rd, hctx).
    destruct hctx as (Rpd, hctx).
    destruct hctx as (Pd, hctx).
    unpack.
    (* Build the proof. *)
    econstructor; eauto with wk.
      (* Typing judgement for the context thread. *)
      econstructor; [ eauto with wk | eauto with j aac ].
      (* Typing judgement for the spawned thread. *)
      jpi_inversion.
      make_jt_hypothesis hv1.
      make_jt_hypothesis hv2.
      econstructor; [ eauto with wk | ].
      eapply JCut.
        eapply JSubLeft.
        eapply JApplication; eauto using hv2.
        eauto with j.
      forwards: canonicalization_preliminary hv1;
        eauto 6 using jh_consistent with consistent_split j.
      auto with aac.

Grab Existential Variables.
auto. auto.

Qed.

(* Subject reduction for concurrent programs. *)

Lemma subject_reduction:
  forall c1 c2,
  red c1 c2 ->
  jconf c1 ->
  jconf c2.
Proof.
  inversion 2; subst.
  eapply CsrTargetConf; [ | eauto using red_wkcp with wk ].
  eapply c_subject_reduction_preliminary.
    eauto.
    eauto.
    erewrite guarded_permission_drop_right.
      eauto.
      eauto using jh_consistent.
      eauto.
    aac_reflexivity.
    eauto.
    eauto.
    eauto using @core_right_unit.
Qed.
