Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import MSAAllocationMaps.
Require Import MSASAMOperations.
Require Import MSAProduct.
Require Import MSAMorphism.
Require Import TyCon.

(* ---------------------------------------------------------------------------- *)

(* About term variables, type variables, and namespaces. *)

(* We wish to allow types to refer to terms: if [x] is a term variable, bound
   by an ordinary lambda-abstraction, then occurrences of [x] can appear both
   as part of a term and as part of a type. *)

(* This dictates that term variables and type variables belong to a single
   namespace. (Indeed, if we decided to maintain two distinct namespaces, then
   we would need to maintain a mapping of [x], viewed as a term variable, to
   [x], viewed as a type variable. I believe that this would be quite
   confusing.) *)

(* We wish to substitute terms for term variables, and types for type variables.
   However, as argued above, term variables and type variables belong to a single
   namespace. There follows that terms and types must belong to a single syntactic
   category. (I am not completely certain whether this is completely natural or a
   consequence of the design of my [DeBruijn] library. I am willing to believe it
   is natural.) *)

(* Thus, in the following, we define a single Coq inductive type which
   represents the syntax of terms and types. Later, we use kinds and a
   well-kindedness judgement to distinguish between terms and types. *)

(* Mixing terms and types might seem weird at first. I believe, in fact, that
   it is a good thing: it means that there is just one notion of lifting, one
   notion of substitution, and that fewer administrative lemmas (weakening,
   substitution, etc.) are needed. *)

(* ---------------------------------------------------------------------------- *)

(* Kinds. *)

(* Kinds play a double role in the type system and in its formalization.

   In the type system, kinds are used to distinguish between various ``kinds''
   of type variables. The following three kinds are used for this purpose:

     [TYPE] classifies ordinary types.
     [PERM] classifies permissions.
     [VALUE] classifies values (which exist at runtime); this kind is known
            as [term] in surface Mezzo

  In the formalization, kinds are also used to distinguish between terms and
  types (which are usually thought of as two disjoint syntactic categories)
  and to distinguish between terms and values (which are usually considered
  either two disjoint syntactic categories, or two syntactic categories in a
  subset relationship). The following kind is used for this purpose:

    [PROG] classifies terms (also known as programs)
    [CPROG] classifies concurrent programs made of one or more threads.

  We introduce the kind [ROW kappa] in order to represent lists of things
  of kind [kappa]. This is used by the type constructor [TyBlock], whose
  argument is a list of types, and by the term constructor [TNew], whose
  argument is a list of terms. This actually gives us more power than we
  need (and more power than is expressible in the surface syntax), but it
  is nevertheless a simple approach. *)

Inductive kind :=
  | VALUE:  kind
  | PROG : kind
  | CPROG: kind
  | PERM : kind
  | TYPE : kind
  | ROW  : kind -> kind
.

(* ---------------------------------------------------------------------------- *)

(* Some kinds classify "logic" things -- things that do not exist at runtime --
   and some kinds classify "physical" things. The following function performs
   a distinction between these kinds. *)

Fixpoint logic (kappa : kind) :=
  match kappa with
  | VALUE
  | PROG
  | CPROG =>
      False
  | PERM
  | TYPE =>
      True
  | ROW kappa =>
      logic kappa
  end.

(* ---------------------------------------------------------------------------- *)

(* A substitutable kind is one at which we may perform substitutions. This
   includes all logic kinds, as well as the kind [VALUE]. *)

Definition substitutable kappa :=
  logic kappa \/ kappa = VALUE.

Hint Unfold substitutable : substitutable.

(* ---------------------------------------------------------------------------- *)

(* Modes. *)

Inductive mode :=
  | Duplicable
  | Exclusive
.

(* ---------------------------------------------------------------------------- *)

(* The syntax of terms and types. *)

Inductive term :=

(* The following case is common to terms and types. *)

(* A variable has an arbitrary kind. Thus, it can be a term variable (one that
   exists at runtime and will be bound to a value) or a type variable (one
   that does not exist at runtime) of arbitrary kind. *)

  | TVar: nat -> term

(* The following cases are terms. *)

(* If [t] is a term, then [TAbs t] is a term as well. It represents a
   lambda-abstraction. *)

  | TAbs: term -> term

(* If [t1] is a value and [t2] is a term, then [TApp t1 t2] is a term. It
   represents a function application. *)

  | TApp: term -> term -> term

(* If [l] is a memory location, then [TLoc l] is a term. *)

  | TLoc: nat -> term

(* If [tag] is a natural integer and [t] is a row (that is, a list) of values,
   then [TNew tag t] is a term. It is the operation of allocating a fresh
   memory block whose tag is [tag] and whose fields are initialized with
   [t]. *)

  | TNew: nat -> term -> term

(* If [t] is a value and [f] is a natural integer, then [TRead t f] is a
   term. It is the operation of reading the field [f] in the memory block
   whose address is [t]. *)

  | TRead: term -> nat -> term

(* If [t] and [u] are values and [f] is a natural integer, then [TWrite t f u]
   is a term. It is the operation of writing the value [u] to the field [f] in
   the memory block whose address is [t]. *)

  | TWrite: term -> nat -> term -> term

(* If [t] is a value and [tag] is a natural integer, then [TWriteTag t tag]
   is a term. It is the operation of writing the tag [tag] to the memory
   block whose address is [t]. *)

  | TWriteTag: term -> nat -> term

(* If [t] is a value and [ts] are terms, then [TSwitch t ts] is a term.
   It is the operation of inspecting the tag of the memory block whose
   address is [t] and branching to one of the terms [ts]. This construct
   does not bind any variables. Within each branch, the field access
   operation [TRead] can be used if desired. *)

  | TSwitch: term -> term -> term

(* If [t] and [u] are values, then [TGive u t] is a term. It is the operation
   of giving [u] to [t], or in other words, having [t] adopt [u]. *)

  | TGive: term -> term -> term

(* If [t] and [u] are values, then [TTake u t] is a term. It is the operation
   of taking [u] from [t], or in other words, having [t] abandon [u]. *)

  | TTake: term -> term -> term

(* If [t] is a term, then [TUnpackIn t] is a term. *)

  | TUnpackIn: term -> term

(* [fail] is a term. It represents a runtime failure. It is well-typed in
   every context. *)

  | TFail: term

(* If [t] is a term, then [TThread t] is a thread and can be part of a 
   concurrent program. *)

  | TThread: term -> term

(* If [t] and [u] are two concurrent programs, then [TPar t u] is their
   parallel composition, which is still a concurrent program. *)

  | TPar: term -> term -> term

(* If [t] is a term, then [TSpawn t] is a term. *)

  | TSpawn: term -> term -> term

(* If [l] is a lock memory location, then [TLck l] is a term. *)

  | TLck: nat -> term

(* If [p] is a permission, then [TNewLock p] is a term. *)

  | TNewLck: term

(* If [t] is a value, then [TAcquire t] is a term. *)

  | TAcquire: term -> term

(* If [t] is a value, then [TRelease t] is a term. *)

  | TRelease: term -> term

(* The following cases are types. *)

(* Bottom is the empty type. *)

  | TyBottom: term

(* Universal types are used (quite obviously) to encode the universal
   types of the surface language. They are also (less obviously) used
   to encode the ``dependent types'' of the source language. *)

 | TyForall: kind -> term -> term

(* Existential permissions are used to express the focus rules. They are
   not explicitly manipulated by the end user. *)

  | TyExists: term -> term

(* If [t] has kind [VALUE], then [TySingleton t] is a type. The concrete
   syntax is [=t]. *)

  | TySingleton: term -> term

(* If [ty1] and [ty2] are types, then [TyArrow ty1 ty2] is a type. This
   is a standard function type. The concrete syntax is [t1 -> t2]. *)

  | TyArrow: term -> term -> term

(* If [ty] is a type and [P] is a permission, then [TyMixStar ty P] is a type.
   The concrete syntax is [ty | P]. This type is typically used to express
   function pre- and post-conditions. It also plays a role in the frame rule. *)

  | TyMixStar: term -> term -> term

(* If [m] is a mode, [tag] is a natural integer, [Ts] is a row of types, and
   [U] is a type, then [TyBlock m tag Ts U] is a type. It is the type of a
   memory location, and asserts that the memory block at this address carries
   the tag [tag] and contains the fields [Ts]. When [m] is [Duplicable], this
   type grants read-only access to the block. When [m] is [Exclusive], this
   type grants read-write access. Furthermore, the adoptees of this block have
   type [U]. *)

  | TyBlock: mode -> nat -> term -> term -> term 

(* If [tc] is the name of an algebraic data type constructor, then [TyCon tc]
   is a type. For the moment, algebraic data types do not have parameters. *)

  | TyCon: tycon -> term

(* [dynamic] is a type. It is the type of every memory location. *)

  | TyDynamic: term

(* If [p] is a permission, then [TyLck p] is a type. *)

  | TyLck: term -> term

(* [TyLocked] is an additional type for a lock that is being held. *)

  | TyLocked: term

(* If [t] has kind [VALUE] and [ty] is a type, then [TyAnchored t ty] is a
   permission, which we say is ``anchored'' at [t]. The concrete syntax is [t
   @ ty]. *)

  | TyAnchored: term -> term -> term

(* [TyEmpty] is the empty permission. It is a neutral element for [TyStar]. *)

  | TyEmpty: term

(* [TyStar] is the conjunction of permissions. *)

  | TyStar: term -> term -> term

(* [TyFalse] is an unsatisfiable permission. *)

  | TyFalse: term

(* [RowNil] is the empty row. *)

  | RowNil: term

(* [RowCons] builds a non-empty row. *)

  | RowCons: term -> term -> term

.

(* The unit term is an empty tuple. Note that this is a term, not a value.
   It reduces to a memory location. *)

Notation TUnit :=
  (TNew 0 RowNil).

(* The unit type is an empty duplicable tuple type. *)

Notation TyUnit :=
  (TyBlock Duplicable 0 RowNil TyBottom).

(* We don't need a sequencing construct, because we can use the following
   syntactic sugar instead. *)

Notation TLet u t :=
  (TApp (TAbs t) u).

(* If [v1] and [v2] have kind [VALUE], then [TyEquation v1 v2] is
   a permission. The concrete syntax is [v1 = v2]. It is a short-hand
   for the anchored permission [v1 @ =v2]. *)

Notation TyEquation v1 v2 :=
  (TyAnchored v1 (TySingleton v2)).

(* The empty environment. *)

Notation zilch :=
  (@empty kind).

(* A singleton environment. *)

Notation singleton a :=
  (insert 0 a zilch).


(* ---------------------------------------------------------------------------- *)

(* The binding structure of terms and types. *)

(* Because term variables and type variables form a single namespace, the
   definition of straightforward. Every binder counts. *)

Instance Var_term : Var term := {
  var := TVar (* avoid eta-expansion *)
}.

Fixpoint traverse_term (f : nat -> nat -> term) l t :=
  match t with
  | TVar x =>
      f l x
  | TAbs t =>
      TAbs (traverse_term f (1 + l) t)
  | TApp t1 t2 =>
      TApp (traverse_term f l t1) (traverse_term f l t2)
  | TLoc l =>
      TLoc l
  | TNew tag t =>
      TNew tag (traverse_term f l t)
  | TRead t field =>
      TRead (traverse_term f l t) field
  | TWrite t field u =>
      TWrite (traverse_term f l t) field (traverse_term f l u)
  | TWriteTag t tag =>
      TWriteTag (traverse_term f l t) tag
  | TSwitch t ts =>
      TSwitch (traverse_term f l t) (traverse_term f l ts)
  | TGive u t =>
      TGive (traverse_term f l u) (traverse_term f l t)
  | TTake u t =>
      TTake (traverse_term f l u) (traverse_term f l t)
  | TUnpackIn t =>
      TUnpackIn (traverse_term f l t)
  | TFail =>
      TFail
  | TThread t =>
      TThread (traverse_term f l t)
  | TPar t1 t2 =>
      TPar (traverse_term f l t1) (traverse_term f l t2)
  | TSpawn v1 v2 =>
      TSpawn (traverse_term f l v1) (traverse_term f l v2)
  | TLck l =>
      TLck l
  | TNewLck =>
      TNewLck
  | TAcquire t =>
      TAcquire (traverse_term f l t)
  | TRelease t =>
      TRelease (traverse_term f l t)
  (* | TStop => *)
  (*     TStop *)
  | TyBottom =>
      TyBottom
  | TyForall kappa T =>
      TyForall kappa (traverse_term f (1 + l) T)
  | TyExists T =>
      TyExists (traverse_term f (1 + l) T)
  | TySingleton T =>
      TySingleton (traverse_term f l T)
  | TyArrow T1 T2 =>
      TyArrow (traverse_term f l T1) (traverse_term f l T2)
  | TyMixStar T1 T2 =>
      TyMixStar (traverse_term f l T1) (traverse_term f l T2)
  | TyBlock m tag Ts U =>
      TyBlock m tag (traverse_term f l Ts) (traverse_term f l U)
  | TyCon tc =>
      TyCon tc
  | TyLck p =>
      TyLck (traverse_term f l p)
  | TyLocked =>
      TyLocked
  | TyDynamic =>
      TyDynamic
  | TyAnchored T1 T2 =>
      TyAnchored (traverse_term f l T1) (traverse_term f l T2)
  | TyEmpty =>
      TyEmpty
  | TyStar T1 T2 =>
      TyStar (traverse_term f l T1) (traverse_term f l T2)
  | TyFalse =>
      TyFalse
  | RowNil =>
      RowNil
  | RowCons T1 T2 =>
      RowCons (traverse_term f l T1) (traverse_term f l T2)
  end.

Instance Traverse_term : Traverse term term := {
  traverse := traverse_term
}.

Instance TraverseVarInjective_term : @TraverseVarInjective term _ term _.
Proof.
  constructor. prove_traverse_var_injective.
Qed.

Instance TraverseFunctorial_term : @TraverseFunctorial term _ term _.
Proof.
  constructor. prove_traverse_functorial.
Qed.

Instance TraverseRelative_term : @TraverseRelative term term _.
Proof.
  constructor. prove_traverse_relative.
Qed.

Instance TraverseIdentifiesVar_term : @TraverseIdentifiesVar term _ _.
Proof.
  constructor. prove_traverse_identifies_var.
Qed.

Instance TraverseVarIsIdentity_term : @TraverseVarIsIdentity term _ term _.
Proof.
  constructor. prove_traverse_var_is_identity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Well-kindedness. *)

(* In several places in the syntax of terms, we require a certain sub-term to
   be a value, as opposed to an arbitrary term. This is administrative normal
   form (ANF). This restriction can be worked around by adding extra [TLet]
   bindings.  It leads to a couple of simplifications: 1- the typing rules
   become simpler; 2- there are fewer evaluation contexts. *)

Inductive wk : env kind -> term -> kind -> Prop :=

(* A value is also a term. In other words, values are a subset of terms. This
   allows a value to appear directly within a term, without an explicit
   [return] construct. *)

  | WkValuesAreTerms:
      forall K v,
      wk K v VALUE ->
      wk K v PROG

(* The kind of a variable is dictated by the kind environment. *)

  | WkTVar:
      forall K x kappa,
      lookup x K = Some kappa ->
      wk K (TVar x) kappa

(* A lambda-abstraction is a value. The formal parameter [x] is considered a
   value, so it can appear within terms and within types. The body [t] must be
   a term. *)

  | WkTAbs:
      forall K t,
      wk (insert 0 VALUE K) t PROG ->
      wk K (TAbs t) VALUE

(* A function application is a term if its left-hand side is a value and its
   right-hand side is a term (ANF). *)

  | WkTApp:
      forall K v t,
      wk K v VALUE ->
      wk K t PROG ->
      wk K (TApp v t) PROG

(* A memory location is a value. *)

  | WkTLoc:
      forall K l,
      wk K (TLoc l) VALUE

(* The argument of a memory allocation operation is a row of values (ANF). *)

  | WkTNew:
      forall K tag vs,
      wk K vs (ROW VALUE) ->
      wk K (TNew tag vs) PROG

(* The argument of a read operation is a value (ANF). *)

  | WkTRead:
      forall K v field,
      wk K v VALUE ->
      wk K (TRead v field) PROG

(* The arguments of a write operation are values (ANF). *)

  | WkTWrite:
      forall K v1 field v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TWrite v1 field v2) PROG

(* The argument of a tag-write operation is value (ANF). *)

  | WkTWriteTag:
      forall K v tag,
      wk K v VALUE ->
      wk K (TWriteTag v tag) PROG

(* A [switch] construct is a term. *)

  | WkTSwitch:
      forall K v ts,
      wk K v VALUE ->
      wk K ts (ROW PROG) ->
      wk K (TSwitch v ts) PROG

(* A [give] construct is a term. *)

  | WkTGive:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TGive v1 v2) PROG

(* A [take] construct is a term. *)

  | WkTTake:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TTake v1 v2) PROG

(* A [unpack] construct is a term. *)

  | WkTUnpackIn:
      forall K t,
      wk K t PROG ->
      wk K (TUnpackIn t) PROG

(* [fail] is a term. *)

  | WkTFail:
      forall K,
      wk K TFail PROG

(* A [thread] construct on a term is a concurrent program. *)

  | WkTThread:
      forall K t,
      wk K t PROG ->
      wk K (TThread t) CPROG

(* A [par] construct on two concurrent programs is a concurrent program. *)

  | WkTPar:
      forall K t1 t2,
      wk K t1 CPROG ->
      wk K t2 CPROG ->
      wk K (TPar t1 t2) CPROG

(* The argument of a [spawn] operation is a value. *)

  | WkTSpawn:
      forall K v1 v2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      wk K (TSpawn v1 v2) PROG

(* A [lock] is a value. *)

  | WkTLck:
      forall K l,
      wk K (TLck l) VALUE

(* Creating a new [lock] requires a permission. *)

  | WkTNewLck:
      forall K,
      wk K TNewLck PROG

(* An [acquire] or a [release] construct is a term. *)

  | WkTAcquire:
      forall K v,
      wk K v VALUE ->
      wk K (TAcquire v) PROG

  | WkTRelease:
      forall K v,
      wk K v VALUE ->
      wk K (TRelease v) PROG

(* (* A the constant [stop] is a valid program. *) *)

(*   | WkTStop: *)
(*       forall K, wk K TStop PROG *)

(* Bottom is a type. *)

  | WkTyBottom:
      forall K,
      wk K TyBottom TYPE

(* The universal quantifier is used to bind a variable of substitutable kind within
   a type or a permission. *)

  | WkTyForallTYPE:
      forall K kappa1 T,
      substitutable kappa1 ->
      wk (insert 0 kappa1 K) T TYPE ->
      wk K (TyForall kappa1 T) TYPE

  | WkTyForallPERM:
      forall K kappa1 T,
      substitutable kappa1 ->
      wk (insert 0 kappa1 K) T PERM ->
      wk K (TyForall kappa1 T) PERM

(* The existential quantifier is used only to bind a variable that stands for a
   value within a permission. *)

  | WkTyExists:
      forall K P,
      wk (insert 0 VALUE K) P PERM ->
      wk K (TyExists P) PERM

(* The singleton type constructor transforms a logical value into an ordinary
   type. *)

  | WkTySingleton:
      forall K v,
      wk K v VALUE ->
      wk K (TySingleton v) TYPE

(* A function type is an ordinary type if its domain and codomain are ordinary
   types. *)

  | WkTyArrow:
      forall K T U,
      wk K T TYPE ->
      wk K U TYPE ->
      wk K (TyArrow T U) TYPE

(* [TyMixStar] builds a package of an ordinary type and a permission. The
   result is an ordinary type. *)

  | WkTyMixStar:
      forall K T P,
      wk K T TYPE ->
      wk K P PERM ->
      wk K (TyMixStar T P) TYPE

(* A memory block type [TyBlock _ _ Ts U] is an ordinary type if its content
   [Ts] is a row of ordinary types and if its adoptee type [U] is a type. *)

  | WkTyBlock:
      forall K m tag Ts U,
      wk K Ts (ROW TYPE) ->
      wk K U TYPE ->
      wk K (TyBlock m tag Ts U) TYPE

(* An algebraic data type is an ordinary type. *)

  | WkTyCon:
      forall K tc,
      wk K (TyCon tc) TYPE

(* [dynamic] is a type. *)

  | WkTyDynamic:
      forall K,
      wk K TyDynamic TYPE

(* [TyLck] builds a type of a permission. *)

  | WkTyLck:
      forall K p,
      wk K p PERM ->
      wk K (TyLck p) TYPE

(* [TyLocked] is a type. *)

  | WkTyLocked:
      forall K,
      wk K TyLocked TYPE

(* An anchored permission [t @ ty] is a permission if [t] is a logical value
   and [ty] is an ordinary type. *)

  | WkTyAnchored:
      forall K v T,
      wk K v VALUE ->
      wk K T TYPE ->
      wk K (TyAnchored v T) PERM

(* The empty permission is a permission. *)

  | WkTyEmpty:
      forall K,
      wk K TyEmpty PERM

(* The conjunction of two permissions is a permission. *)

  | WkTyStar:
      forall K P Q,
      wk K P PERM ->
      wk K Q PERM ->
      wk K (TyStar P Q) PERM

(* The unsatisfiable permission is a permission. *)

  | WkTyFalse:
      forall K,
      wk K TyFalse PERM

(* The empty row is a valid row. *)

  | WkRowNil:
      forall kappa K,
      wk K RowNil (ROW kappa)

(* A nonempty row is a valid row if its head is a valid element
   and its tail is a valid row. *)

  | WkRowCons:
      forall K kappa t ts,
      wk K t kappa ->
      wk K ts (ROW kappa) ->
      wk K (RowCons t ts) (ROW kappa)

.

Hint Constructors wk : wk.

(* ---------------------------------------------------------------------------- *)

(* Basic properties of the well-kindedness judgement. *)

(* If the environment [K2] extends the environment [K1], then a term
   that is well-kinded under [K1] is also well-kinded under [K2]. *)

Lemma wk_agree:
  forall K1 t kappa,
  wk K1 t kappa ->
  forall K2 k,
  length K1 k ->
  agree K1 K2 k ->
  wk K2 t kappa.
Proof.
  induction 1; intros; eauto 6 with wk length agree omega.
Qed.

(* A corollary of the previous lemma: a term that is well-kinded
   under the empty environment is well-kinded under an arbitrary
   environment. *)

Lemma wk_empty:
  forall K t kappa,
  wk (@empty _) t kappa ->
  wk K t kappa.
Proof.
  eauto using wk_agree with length agree.
Qed.

(* If [t] has kind [kappa2] under [K2] and [t] is also well-kinded
   under a shorter environment [K1], then [t] has kind [kappa2]
   under [K1]. *)

Lemma wk_agree_reverse:
  forall K2 t kappa2,
  wk K2 t kappa2 ->
  forall K1 k,
  length K1 k ->
  agree K1 K2 k ->
  forall kappa1,
  wk K1 t kappa1 ->
  wk K1 t kappa2.
Proof.
  induction 1; introv hl ha h; dependent induction h;
  eauto using (WkRowNil VALUE) with wk length agree omega.
  (* Case TVar/TVar. *)
  econstructor.
  forwards: agree_below; eauto.
  congruence.
Qed.

(* A corollary of the previous lemma. *)

Lemma wk_empty_reverse:
  forall K t kappa,
  wk K t kappa ->
  (exists kappa, wk zilch t kappa) ->
  wk zilch t kappa.
Proof.
  intros. unpack. eauto using wk_agree_reverse with length agree.
Qed.

(* A term that is well-kinded under the empty environment is closed. *)

Lemma wk_closed:
  forall K t kappa,
  wk K t kappa ->
  forall k,
  length K k ->
  closed k t.
Proof.
  induction 1; intros; construction_closed.
Qed.

(* Well-kindedness is preserved by the insertion of a new variable. *)

Lemma wk_weakening:
  forall K T2 kappa2,
  wk K T2 kappa2 ->
  forall x1 okappa1 K',
  raw_insert x1 okappa1 K = K' ->
  wk K' (shift x1 T2) kappa2.
Proof.
  induction 1; intros; subst; simpl_lift_goal;
  try econstructor (solve [ eauto with lookup_insert insert_insert insert_concat ]).
Qed.

Hint Resolve wk_weakening : j_weakening j_substitution.

(* Well-kindedness is preserved by the removal of an unused variable. *)

Lemma wk_weakening_inverse:
  forall T2 K' x1 kappa2,
  wk K' (shift x1 T2) kappa2 ->
  forall kappa1 K,
  insert x1 kappa1 K = K' ->
  wk K T2 kappa2.
Proof.
  intro T2; introv h; dependent induction h; intros; subst;
  (* This deals with the non-syntax-directed rules -- that is,
     the sub-kinding rules. *)
  try solve [ eauto with wk ];
  (* Every other case. *)
  (* Perform case analysis on [T2]; only one case is possible, because
     the well-kindedness rule exhibits an explicit head constructor;
     exploit this to simplify [shift x1 T2]. *)
  destruct T2; try discriminate; simpl_lift_all;
  (* Decompose the equation (if possible). *)
  injections;
  (* Conclude. *)
  econstructor (solve [ eauto with lookup_insert insert_insert insert_concat ]).
  (* Because removing an unused variable amounts to replacing it with an
     arbitrary type of the appropriate kind, this lemma would be a consequence
     of [wk_substitution] below, if we were willing to exploit the fact that
     every kind is inhabited. *)
Qed.

(* Well-kindedness is preserved by substitution. *)

Lemma wk_substitution:
  forall K' T2 kappa2,
  wk K' T2 kappa2 ->
  forall x kappa1 K,
  insert x kappa1 K = K' ->
  forall T1,
  wk K T1 kappa1 ->
  wk K (subst T1 x T2) kappa2.
Proof.
  induction 1; intros; subst; simpl_subst_goal;
  (* General rule. *)
  try econstructor (solve [
    eauto using wk_weakening with insert_insert insert_concat
  ]).
  (* Case TVar. *)
  unfold subst_idx; dblib_by_cases; lookup_insert_all; eauto with wk.
Qed.

Hint Resolve wk_substitution : j_substitution.

(* Substituting an object [t1] of logical kind into an object [t2] of
   physical kind has no effect -- the result is [t2], up to a shift. *)

Lemma wk_substitution_logic:
  forall K' t2 kappa2,
  wk K' t2 kappa2 ->
  ~ logic kappa2 ->
  forall t1 x K kappa1,
  insert x kappa1 K = K' ->
  wk K t1 kappa1 ->
  logic kappa1 ->
  shift x (subst t1 x t2) = t2.
Proof.
  induction 1; intros; subst; simpl_subst_goal; simpl_lift_goal;
  eauto using wk_weakening with f_equal insert_insert.
  (* TVar *)
  unfold subst_idx. dblib_by_cases; try solve [
    do 2 simpl_lift_goal; eauto with f_equal omega
  ].
  subst. lookup_insert_all. tauto. (* Contradiction. *)
Qed.

(* ---------------------------------------------------------------------------- *)

(* Evaluation contexts. *)

(* Evaluation contexts are defined as special terms, in which the hole is
   represented by a variable of index 0. Evaluation is on the right of a
   value. *)

Inductive eval_context : term -> Prop :=

  | EvalContextAppRight:
      forall v,
      wk zilch v VALUE ->
      eval_context (TApp v (TVar 0))

.

(* Fill the hole of a context. *)

Notation fill E t :=
  (subst t 0 E).

(* Deep evaluation contexts are defined as an iteration of evaluation
   contexts. *)

Inductive deep_eval_context : term -> Prop :=

  | DeepEvalContextHole:
      deep_eval_context (TVar 0)

  | DeepEvalContextInd:
      forall E D,
      eval_context E ->
      deep_eval_context D ->
      deep_eval_context (fill E D)

.

(* Substitution in a term that is well-kinded under [zilch] environment. *)

Lemma fill_zilch:
  forall v k t,
  wk zilch v k ->
  fill v t = v.
Proof.
  eauto using closed_subst_invariant, wk_closed, length_empty.
Qed.

Ltac rewrite_fill_zilch :=
  repeat match goal with
  | hzilch: wk zilch ?v _ |- context[fill ?v ?t] =>
      rewrite (fill_zilch t hzilch)
end.

Ltac rewrite_fill_zilch_all :=
  rewrite_fill_zilch;
  repeat match goal with
  | hzilch: wk zilch ?v _, hf: context[fill ?v ?t] |- _ =>
      rewrite (fill_zilch t hzilch) in hf
end.

(* Tactics to open and substitute in an evaluation context. *)

Ltac open_eval_context :=
  match goal with
    | h: eval_context _ |- _ =>
        destruct h; simpl_subst_all; rewrite_fill_zilch_all
  end.

Ltac open_eval_context_discriminate :=
  try open_eval_context; discriminate.

(* An evaluation context is a deep evaluation context. *)

Lemma shallow_is_deep:
  forall E,
  eval_context E ->
  deep_eval_context E.
Proof.
  inversion 1.
  forwards: DeepEvalContextInd; eauto using DeepEvalContextHole.
  subst; open_eval_context; auto.
Qed.

(* TLet defines an evaluation context. *)

Lemma tlet_eval_context:
  forall t,
  wk (singleton VALUE) t PROG ->
  eval_context (TLet (TVar 0) t).
Proof.
  intros; eapply EvalContextAppRight; eauto with wk.
Qed.

(* Evaluation and deep evaluation contexts are well-kinded programs. *)

Lemma wk_eval_context:
  forall E K,
  eval_context E ->
  wk (insert 0 PROG K) E PROG.
Proof.
  intros.
  open_eval_context.
  eauto using wk_empty with wk.
Qed.

Lemma wk_deep_eval_context:
  forall D K,
  deep_eval_context D ->
  wk (insert 0 PROG K) D PROG.
Proof.
  induction 1.
  eauto with wk.
  open_eval_context; eauto using wk_empty with wk.
Qed.  

(* A context cannot be a value. *)

Lemma eval_context_value:
  forall E K t,
  eval_context E ->
  wk K (fill E t) VALUE ->
  wk K t VALUE.
Proof.
  introv HE HEt. 
  open_eval_context.
  inversion HEt.
Qed.

(* Filling an evaluation context changes only its hole. *)

Lemma fill_eval_context:
  forall v u t,
  eval_context (TApp v u) ->
  fill (TApp v u) t = TApp v t.
Proof.
  introv Hev; dependent induction Hev; simpl_subst_goal.
  rewrite_fill_zilch; auto.
Qed.

Lemma fill_nested_context:
  forall D1 D2 t,
  deep_eval_context D1 ->
  deep_eval_context D2 ->
  fill (fill D1 D2) t = fill D1 (fill D2 t).
Proof.
  induction 1; intros.
  auto with simpl_subst_goal.
  open_eval_context.
  rewrite IHdeep_eval_context; eauto.
Qed.

(* Filling a context preserves well-kindedness. *)

Lemma wk_fill_eval_context:
  forall E K t,
  eval_context E ->
  wk K t PROG ->
  wk K (fill E t) PROG.
Proof.
  eauto using wk_substitution, wk_eval_context.
Qed.

Lemma wk_fill_deep_eval_context:
  forall D K t,
  deep_eval_context D ->
  wk K t PROG ->
  wk K (fill D t) PROG.
Proof.
  eauto using wk_substitution, wk_deep_eval_context.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Heaps and configurations. *)

(* A pointer is either null or a memory location. *)

Inductive pointer :=
|    Null:        pointer
| NonNull: nat -> pointer.

(* A memory block consists of:
   - a tag (a natural integer);
   - an adopter field (a pointer);
   - ordinary fields (a row of closed values). *)

Inductive block :=
| Block: nat -> pointer -> term -> block.

(* A heap is an allocation map: it keeps track of which memory locations are
   currently allocated, and maps every allocated memory location to a memory
   block. *)

Definition concrete_heap := aMap block.

(* The empty heap. *)

Definition concrete_empty_heap : concrete_heap :=
  (empty_map _).

(* Locks are defined in a special part of the heap. *)

(* Inductive lock := *)
(* | Lock: bool -> lock. *)

(* Notation taken_lck := false. *)
(* Notation free_lck := true. *)

Inductive lock :=
| FreeLock
| TakenLock.


Definition concrete_lck := aMap lock.

Definition concrete_empty_lck : concrete_lck :=
  (empty_map _).

Notation concrete_resource :=
  (concrete_heap * concrete_lck)%type.

Definition concrete_empty : concrete_resource :=
  (concrete_empty_heap, concrete_empty_lck).

Notation concrete_heap_project :=
  (@left_projection concrete_heap concrete_lck).

Notation concrete_heap_inject :=
  (@left_injection concrete_heap concrete_lck).

Notation concrete_lck_project :=
  (@right_projection concrete_heap concrete_lck).

Notation concrete_lck_inject :=
  (@right_injection concrete_heap concrete_lck).

Instance concrete_heap_sub_resource:
  SubStructure concrete_heap_project concrete_heap_inject.
Proof.
  split; intros; decompose_pairs; simpl in *; auto.
  match goal with
    | h: (?a, ?b) = (?c, ?d) |- _ => inversion h
  end.
  auto.
Qed.

Instance concrete_lck_sub_resource:
  SubStructure concrete_lck_project concrete_lck_inject.
Proof.
  split; intros; decompose_pairs; simpl in *; auto.
  match goal with
    | h: (?a, ?b) = (?c, ?d) |- _ => inversion h
  end.
  auto.
Qed.

Notation concrete_heap_maps :=
  (SAM_maps concrete_heap_project).

Notation concrete_heap_allocate :=
  (SAM_allocate concrete_heap_project concrete_heap_inject).

Notation concrete_heap_update :=
  (SAM_update concrete_heap_project concrete_heap_inject).

Notation concrete_heap_limit :=
  (SAM_limit concrete_heap_project).

Notation concrete_lck_maps :=
  (SAM_maps concrete_lck_project).

Notation concrete_lck_allocate :=
  (SAM_allocate concrete_lck_project concrete_lck_inject).

Notation concrete_lck_update :=
  (SAM_update concrete_lck_project concrete_lck_inject).

Notation concrete_lck_limit :=
  (SAM_limit concrete_lck_project).

Lemma independent_heap_lck:
  forall h lh,
  concrete_heap_project (concrete_lck_inject lh h) = concrete_heap_project h.
Proof.
  decompose_pairs. auto.
Qed.

Lemma independent_lck_heap:
  forall h hh,
  concrete_lck_project (concrete_heap_inject hh h) = concrete_lck_project h.
Proof.
  decompose_pairs. auto.
Qed.

Create HintDb unfold_concrete_sub_resource.

Hint Extern 4 =>
  match goal with
    | h: context[SAM_everywhere _ _ _] |- _ =>
        unfold SAM_everywhere in h
    | |- context[SAM_everywhere _ _ _] =>
        unfold SAM_everywhere
    | h: context[SAM_maps _ _ _ _] |- _ =>
        unfold SAM_maps in h
    | |- context[SAM_maps _ _ _ _] =>
        unfold SAM_maps
    | h: context[SAM_update _ _ _ _ _] |- _ =>
        unfold SAM_update in h
    | |- context[SAM_update _ _ _ _ _] =>
        unfold SAM_update
    | h: context[SAM_allocate _ _ _ _] |- _ =>
        unfold SAM_allocate in h
    | |- context[SAM_allocate _ _ _ _] =>
        unfold SAM_allocate
  end:
unfold_concrete_sub_resource.

Create HintDb heap_lck_lift.

Hint Extern 4 =>
  match goal with
    | h: context[concrete_heap_project (concrete_lck_inject _ _)] |- _ =>
        rewrite independent_heap_lck in h
    | |- context[concrete_heap_project (concrete_lck_inject _ _)] =>
        rewrite independent_heap_lck
    | h: context[concrete_lck_project (concrete_heap_inject _ _)] |- _ =>
        rewrite independent_lck_heap in h
    | |- context[concrete_lck_project (concrete_heap_inject _ _)] =>
        rewrite independent_lck_heap
    | h: context[concrete_lck_project (concrete_lck_inject _ _)] |- _ =>
        rewrite (pi_inverse concrete_lck_sub_resource) in h
    | |- context[concrete_lck_project (concrete_lck_inject _ _)] =>
        rewrite (pi_inverse concrete_lck_sub_resource)
    | h: context[concrete_lck_inject (concrete_lck_project _) _] |- _ =>
        rewrite (ip_inverse concrete_lck_sub_resource) in h
    | |- context[concrete_lck_inject (concrete_lck_project _) _] =>
        rewrite (ip_inverse concrete_lck_sub_resource)
  end:
heap_lck_lift.


(* A configuration is a pair of a heap and a term. *)

Inductive configuration :=
| Cf: concrete_resource -> term -> configuration.

(* ---------------------------------------------------------------------------- *)

(* Well-kindedness for blocks, heaps, and configurations. *)

Inductive wkb : block -> Prop :=
| Wkb:
    forall tag adopter vs,
    wk zilch vs (ROW VALUE) ->
    wkb (Block tag adopter vs).

Inductive wkl : lock -> Prop :=
| WkTakenLock: wkl TakenLock
| WkFreeLock: wkl FreeLock.

Inductive wkh : concrete_resource -> Prop :=
| Wkh:
    forall h,
    SAM_everywhere concrete_heap_project wkb h ->
    SAM_everywhere concrete_lck_project wkl h ->
    wkh h.

Inductive wkc : configuration -> Prop :=
| Wkc:
    forall h t,
    wkh h ->
    wk zilch t PROG ->
    wkc (Cf h t).

Inductive wkcp : configuration -> Prop :=
| Wkcp:
    forall h t,
    wkh h ->
    wk zilch t CPROG ->
    wkcp (Cf h t).

Hint Constructors wkb wkl wkh wkc wkcp : wk.

(* ---------------------------------------------------------------------------- *)

(* A number of little inversion lemmas. *)

Lemma wkc_wk:
  forall h t,
  wkc (Cf h t) ->
  wk zilch t PROG.
Proof.
  inversion 1; eauto.
Qed.

Lemma wkcp_wk:
  forall h t,
  wkcp (Cf h t) ->
  wk zilch t CPROG.
Proof.
  inversion 1; eauto.
Qed.

Lemma wk_TAbs_inversion:
  forall K t,
  wk K (TAbs t) VALUE ->
  wk (insert 0 VALUE K) t PROG.
Proof.
  introv Hwk; dependent induction Hwk; eauto.
Qed.

Lemma wk_TApp_inversion:
  forall K v t,
  wk K (TApp v t) PROG ->
  wk K t PROG.
Proof.
  introv h. dependent destruction h; [ inversion h | assumption ].
Qed.

Lemma wk_TApp_inversion_v:
  forall K v t,
  wk K (TApp v t) PROG ->
  wk K v VALUE.
Proof.
  introv h. dependent destruction h; [ inversion h | assumption ].
Qed.

Lemma wkc_TApp_inversion:
  forall h v t,
  wkc (Cf h (TApp v t)) ->
  wkc (Cf h t).
Proof.
  inversion 1; subst. econstructor; eauto using wk_TApp_inversion.
Qed.

Lemma wk_eval_context_inversion:
  forall E K t,
  eval_context E ->
  wk K (fill E t) PROG ->
  wk K t PROG.
Proof.
  intros.
  open_eval_context.
  eauto using wk_TApp_inversion.
Qed.

Lemma wk_deep_eval_context_inversion:
  forall D K t,
  deep_eval_context D ->
  wk K (fill D t) PROG ->
  wk K t PROG.
Proof.
  induction 1; intros.
  simpl_subst_all; auto.
  rewrite fill_nested_context in * by eauto using shallow_is_deep.
  eauto using wk_eval_context_inversion.
Qed.

Lemma wkc_TNew_inversion:
  forall h tag vs,
  wkc (Cf h (TNew tag vs)) ->
  wk zilch vs (ROW VALUE).
Proof.
  dependent_destruction.
  match goal with h: wk _ _ _ |- _ => inversion h; subst end.
  match goal with h: wk _ _ _ |- _ => inversion h; subst end.
  assumption.
Qed.

Lemma wk_TWrite_inversion:
  forall v1 field v2,
  wk zilch (TWrite v1 field v2) PROG ->
  wk zilch v2 VALUE.
Proof.
  inversion 1; subst.
  match goal with h: wk _ _ _ |- _ => inversion h; subst end.
  assumption.
Qed.

Lemma wk_TySingleton_inversion:
  forall K v,
  wk K (TySingleton v) TYPE ->
  wk K v VALUE.
Proof.
  inversion 1. auto.
Qed.

Lemma wk_TyBlock_inversion:
  forall K m tag Ts U,
  wk K (TyBlock m tag Ts U) TYPE ->
  wk K Ts (ROW TYPE).
Proof.
  inversion 1. auto.
Qed.

Hint Resolve wk_TySingleton_inversion wk_TyBlock_inversion : wk_inversion.

