Set Implicit Arguments.
Require Import List.
Require Import Permutation.
Require Import MyTactics.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Exclusive.
Require Import Adopts.
Require Import Row.
Require Import Typing.
Require Import Top.
Require Import JBasic.
Require Import JDecomposition.
Require Import JDrop.
Require Import Give.

(* This file contains the key subject reduction lemma for the [take] operation.
   This lemma relies heavily on the theory of [abandon_all], which is developed
   in the file [Abandon.v]. *)

(* ---------------------------------------------------------------------------- *)

(* This is a version of the lemma [owns_adopter_implies_owns_adoptee] where the
   hypothesis about [x] takes the form of a typing hypothesis. *)

Lemma types_adopter_implies_owns_adoptee:
  forall R1 R2 R K x y tagy vsy T,
  jctop R1 K (TLoc x) T ->
  exclusive T ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  consistent R ->
  star R1 R2 = R ->
  heap_maps R1 y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)).
Proof.
  intros. subst.
  forwards: exclusive_type_implies_maps_exclusive; eauto with consistent_split. unpack.
  eauto using owns_adopter_implies_owns_adoptee.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The typing rule [JBlock] contains a premise of the form [jpi R zilch (zap ys U)],
   which means that all of the memory locations in the list [ys] have type [U]. *)

(* This property is stable under a permutation of the list. *)

Lemma jpi_zap_permutation:
  forall ys zs,
  Permutation ys zs ->
  forall R K U,
  jpi R K (zap ys U) ->
  jpi R K (zap zs U).
Proof.
  induction 1; simpl; intros; jpi_inversion; eauto with j aac.
Qed.

(* Provided the type [U] is exclusive, this property implies that [R] owns
   every element of the list [ys]. This fact is used when applying the
   lemma [star_abandon_all_reverse]. *)

Lemma jpi_zap_exclusive:
  forall ys R K U,
  jpi R K (zap ys U) ->
  consistent R ->
  exclusive U ->
  Forall (fun y =>
    exists tagy vsy,
    heap_maps R y (DX_EXCLUSIVE (Block tagy Null vsy))
  ) ys.
Proof.
  induction ys; simpl; dependent_destruction; intros; subst; econstructor.
  eapply exclusive_type_implies_maps_exclusive; eauto using jpi_drop_right.
  eauto using jpi_drop_right, @star_commutative.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The following is the key lemma, stated in the special case where [x] has a
   [TyBlock] type. *)

(* If [x] has an exclusive block type, which claims that [x] adopts objects of
   some exclusive type [U], and if [y.adopter] is [x], then, after [x] abandons
   [y], separately [y] has type [U] and [x] retains its original type. *)

Lemma take_TyBlock:
  forall R K x y tagx tagy vsy Tsx U,
  jctop R K (TLoc x) (TyBlock Exclusive tagx Tsx U) ->
  exclusive U ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  consistent R ->
  exists R1 R2,
  jctop R1 K (TLoc y) U /\
  jctop R2 K (TLoc x) (TyBlock Exclusive tagx Tsx U) /\
  heap_upd R y Exclusive (Orphan tagy vsy) = star R1 R2.
Proof.
  (* Decompose the hypothesis about [x]. *)
  intros. dca.
  (* The hypothesis about [y] can then be made more precise. *)
  forwards: owns_adopter_implies_owns_adoptee; eauto using @star_commutative.
  match goal with h1: heap_maps _ ?y _, h2: heap_maps _ ?y _ |- _ => clear h1 end.
  (* [y] must appear among the adoptees of [x]. *)
  forwards: no_lost_sheep; eauto.
  (* Bring it to the front of the list. *)
  forwards: the_last_will_be_the_first; eauto. unpack.
  (* Add some administrative consistency facts. *)
  forwards: consistent_abandon; [ eauto | eauto with consistent_split | ].
  (* The fact that all adoptees have type [U] remains true after [y] is
     brought to the front, and implies that separately [y] has type [U]
     and all of the remaining adoptees have type [U]. *)
  forwards hzap: jpi_zap_permutation; eauto. dependent destruction hzap. subst.
  (* Now, the resource [R1] which guarantees that [y] has type [U] must be
     hoisted out of [abandon_all]. *)
  forwards: star_abandon_all_reverse.
    eauto. 
    eapply star_associative_symmetric.
    eauto.
    eauto using SCAM_maps_maximal_right, heap_project_morphism, @maximal_exclusive, consistent_12p3_23.
    eauto using jpi_zap_exclusive, jpi_drop_right, consistent_12p3_23.
  unpack.
  (* Do not [subst] here; this changes the order of the hypotheses, which matters. *)
  (* Conclude. *)
  xplit.
  (* [y] is well-typed. *)
  eauto.
  (* [x] is well-typed. *)
  eapply JcTopLoc; eauto.
  (* And the resources match. *)
  subst.
  star_upd_exclusive.
  erewrite <- abandon_as_update by eauto. (* ha! *)
  aac_reflexivity.
Qed. (* This takes a very long time... about 20 seconds. *)
     (* Update: checking this file now seems to require about 5 minutes on my laptop! *)

(* ---------------------------------------------------------------------------- *)

(* There remains to generalize the above lemma to the case where [x] has an
   arbitrary exclusive type, as opposed to an exclusive [TyBlock] type. *)

Lemma take:
  forall R K x y tagy vsy T U,
  jctop R K (TLoc x) T ->
  adopts T U ->
  heap_maps R y (DX_EXCLUSIVE (Block tagy (NonNull x) vsy)) ->
  consistent R ->
  exists R1 R2,
  jctop R1 K (TLoc y) U /\
  jctop R2 K (TLoc x) T /\
  heap_upd R y Exclusive (Orphan tagy vsy) = star R1 R2.
Proof.
  introv hx ha hy hc. dependent destruction ha; subst.
  (* Case: [T] is [TyCon tc]. *)
  decomposition_TyCon.
  match goal with h: _ = Exclusive |- _ => rewrite h in * end.
  forwards: take_TyBlock; eauto. unpack.
  solve [ eauto 6 using JcTopFold ].
  (* Case: [T] is [TyBlock _ _ _ _]. *)
  solve [ eauto using take_TyBlock ].
Qed.

