Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Signature.
Require Import Dup.
Require Import Typing.
Require Import JBasic.
Require Import JDrop.
Require Import Harmless.

(* Typing judgements for closed values and closed terms. *)

(* ---------------------------------------------------------------------------- *)

(* The judgement [jctop R K v T] states that the closed value [v] has type [T]
   under resource [R] and under the environment [K]. A non-empty environment
   is required in order to deal with polymorphic values, i.e., values that are
   well-typed in the presence of certain type variables.*)

(* We don't need to invent new rules for this judgement: we posit that, by
   definition, this judgement holds if and only if the resource [R] justifies
   the permission [v @ T]. *)

Notation jctop R K v T :=
  (jpi R K (TyAnchored v T)).

(* ---------------------------------------------------------------------------- *)

(* In a judgement of the form [jctop R K v T], [v] must be a closed value,
   and [T] must be a closed type. *)

Lemma jctop_value:
  forall R K v T,
  jctop R K v T ->
  wk K v VALUE.
Proof.
  intros. jpi_inversion. assumption.
Qed.

Lemma jctop_wk:
  forall R K v T,
  jctop R K v T ->
  wk K T TYPE.
Proof.
  intros. jpi_inversion. eauto with wk.
Qed.

Hint Resolve jctop_value jctop_wk : wk.

(* ---------------------------------------------------------------------------- *)

(* The definition of [jctop] in terms of [jpi] makes it rather inconvenient
   to build [jctop] judgements. In order to facilitate this task, we prove
   a number of derived rules. *)

(* Singleton. *)

Lemma JcTopSingleton:
  forall R K v,
  wk K v VALUE ->
  jctop R K v (TySingleton v).
Proof.
  eauto using JpiAtomicEmpty, (JSingleton 0).
Qed.

Lemma JcTopSingletonEq:
  forall R K v1 v2,
  v1 = v2 ->
  wk K v1 VALUE ->
  jctop R K v1 (TySingleton v2).
Proof.
  intros; subst. eauto using JcTopSingleton.
Qed.

(* Frame. *)

(* R1 |- v : T
   R2 |- Q
   --------------------
   R1 * R2 |- v : T | Q *)

Lemma JcTopFrame:
  forall R1 R2 K R Q v T,
  jctop R1 K v T ->
  jpi R2 K Q ->
  star R1 R2 = R ->
  jctop R K v (TyMixStar T Q).
Proof.
  intros. jpi_inversion. econstructor; eauto with j aac.
Qed.

(* Memory locations. *)

Lemma JcTopLoc:
  forall R R1 R2 R2' R2'a R2'b K l m tag vs ys Ts U,
  star R1 R2 = R ->
  wk zilch vs (ROW VALUE) ->
  jpi R1 K (zip vs Ts) ->
  abandon_all R2 l ys R2' ->
  star R2'a R2'b = R2' ->
  jpi R2'a K (zap ys U) ->
  wk K U TYPE ->
  heap_maps R2'b l (mtagvs2dx m tag vs) ->
  jctop R K (TLoc l) (TyBlock m tag Ts U).
Proof.
  eauto using JpiAtomicEmpty, (JBlock 0) with wk.
Qed.

Lemma JcTopFold:
  forall R K v m tag Ts tc U,
  jctop R K v (TyBlock m tag Ts U) ->
  tc_mode tc = m ->
  tc_tag tc tag Ts ->
  tc_adopts tc = U ->
  jctop R K v (TyCon tc).
Proof.
  intros. jpi_inversion. econstructor; eauto with j.
Qed.

Lemma JcTopDynamic:
  forall R K l,
  l < heap_limit R ->
  jctop R K (TLoc l) TyDynamic.
Proof.
  eauto using JpiAtomicEmpty, (JDynamic 0) with wk.
Qed.

Lemma JcTopLck:
  forall R K l lp P,
  wk zilch P PERM ->
  lck_maps R l lp ->
  is_locked_permission lp P ->
  jctop R K (TLck l) (TyLck P).
Proof.
  eauto using JpiAtomicEmpty, (JLck 0) with wk.
Qed.

Lemma JcTopLocked:
  forall R K l P,
  wk zilch P PERM ->
  lck_maps R l (p2hl P) ->
  jctop R K (TLck l) TyLocked.
Proof.
  eauto using JpiAtomicEmpty, (JLocked 0) with wk.
Qed.  

Lemma JcTopForallIntro:
  forall R K v kappa T,
  wk K v VALUE ->
  jctop R (insert 0 kappa K) (shift 0 v) T ->
  substitutable kappa ->
  jctop R K v (TyForall kappa T).
Proof.
  intros. jpi_inversion.
  (* There is a subtle technical point here. The permission [P]
     which justifies that [shift 0 v] has type [T] may mention
     the variable [0]. Yet, in order to obtain the desired
     conclusion, we need a permission that does not mention this
     variable. The trick is to universally quantify [P]. Thus, we
     seem to need universal permissions -- and their interpretation
     rule, [JpiForall] -- in order to justify universal types. *)
  econstructor.
    eauto with wk.
    eapply JForallIntro; eauto with wk.
    eapply harmless_value; eauto.
    eapply JpiForall; eauto.
    eauto.
Qed.

(* A lemma that helps adjust the resource that appears in a typing judgement. *)

Lemma JcTopSameResource:
  forall R1 R2 K t T,
  jctop R1 K t T ->
  R2 = R1 ->
  jctop R2 K t T.
Proof.
  intros. subst. eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A derivation for a judgement of the form [jt _ _ _ TyEmpty t T] often begins
   with an application of [JCut]. In order to facilitate its use, we prove a
   derived rule. *)

Lemma JtTopCut:
  forall n R1 R2 R K P t T,
  jt n R2 K P t T ->
  jpi R1 K P ->
  star R1 R2 = R ->
  jt (S (S n)) R K TyEmpty t T.
Proof.
  eauto using JCut, (@JSubLeft 0), JpsDisappearsRight.
Qed.

(* A derived typing rule for the unit term. *)

Lemma JtTopUnitMixStar:
  forall R1 R2 R K P,
  jpi R2 K P ->
  star R1 R2 = R ->
  consistent R ->
  jt 3 R K TyEmpty TUnit (TyMixStar TyUnit P).
Proof.
  intros; subst.
  eapply j_drop_right.
    eapply JtTopCut.
      eapply JFrame. eapply JNew; eauto with wk row_length.
      eauto with j.
      eapply core_right_unit.
   eapply star_commutative.
   eauto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* A well-typed closed value is a well-typed closed term. *)

Lemma jctop_jt:
  forall R K v T,
  jctop R K v T ->
  exists n,
  jt n R K TyEmpty v T.
Proof.
  dependent_destruction; intros. subst.
  eexists. eapply JtTopCut; eauto using (get1 jc_jt) with aac.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Another corollary of [j_wk]. *)

Lemma jctop_substitutable:
  forall R K v kappa T,
  jctop R K v (TyForall kappa T) ->
  substitutable kappa.
Proof.
  dependent_destruction; intros; subst.
  forwards: (get4 j_wk); [ eauto | eauto with wk | ]. wk_inversion.
  assumption.
Qed.

Hint Resolve jctop_substitutable : substitutable.

