Set Implicit Arguments.
Require Import Relation_Operators.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Typing.
Require Import Top.
Require Import JHeap.
Require Import Reduction.
Require Import SubjectReduction.
Require Import Progress.
Require Import DataRaceFreedom.

(* Final corollaries of the subject reduction and progress theorems. *)

(* ---------------------------------------------------------------------------- *)

(* The initial configuration is well-typed. *)

Lemma initial_configuration:
  forall t,
  jcp void t ->
  jconf (Cf concrete_empty t).
Proof.
  introv hcjp.
  econstructor; eauto using @core_right_unit, jh_empty.
  (* There remains to show that the initial configuration is well-kinded. *)
  econstructor.
    econstructor.
      replace concrete_empty with (SAM_empty concrete_heap_inject concrete_empty).
      eapply SAM_everywhere_empty. eauto with typeclass_instances.
      auto.
      replace concrete_empty with (SAM_empty concrete_lck_inject concrete_empty).
      eapply SAM_everywhere_empty. eauto with typeclass_instances.
      auto.
    inversion hcjp; eauto with wk.
    unfold guarded_permission. simpl. eauto with j.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Type soundness in small steps. *)

(* Well-typedness is preserved throughout reduction. *)

Lemma jconf_preserved:
  forall c1 c2,
  redstar c1 c2 ->
  jconf c1 ->
  jconf c2.
Proof.
  induction 1; eauto using subject_reduction.
Qed.

(* If a configuration [c] is well-typed, then executing it for an arbitrary
   number of steps can only lead to acceptable configurations. *)

Lemma type_soundness_small_step:
  forall c1,
  jconf c1 ->
  forall c2,
  redstar c1 c2 ->
  acceptable c2.
Proof.
  eauto using progress, jconf_preserved.
Qed.

Lemma data_race_freedom_small_step:
  forall c1,
  jconf c1 ->
  forall h t,
  redstar c1 (Cf h t) ->
  data_race t ->
  False.
Proof.  
  eauto using data_race_freedom, jconf_preserved.
Qed.

(*

(* ---------------------------------------------------------------------------- *)

(* A definition of reducing to an answer. *)

Notation reduces_to_an_answer c1 :=
  (exists h2 t2, redstar c1 (Cf h2 t2) /\ answer t2).

(* A definition of divergence. *)

CoInductive diverges : configuration -> Prop :=
| DivergesCons:
    forall c1 c2,
    red c1 c2 ->
    diverges c2 ->
    diverges c1.

(* A definition of reducing to an answer or a spawner. *)

Notation reduces_to_a_spawswer c1 :=
  ((exists h2 t2, redstar c1 (Cf h2 t2) /\ answer t2) \/
  (exists h2 t2, redstar c1 (Cf h2 t2) /\ spawner t2)).

(* A definition of reducing to an answer, a spawner, or a waiting configuration. *)

Notation reduces_to_a_waispawswer c1 :=
  ((exists h2 t2, redstar c1 (Cf h2 t2) /\ answer t2)    \/
   (exists h2 t2, redstar c1 (Cf h2 t2) /\ spawner t2)   \/
   (exists h2 t2, redstar c1 (Cf h2 t2) /\ waits t2 h2)).

(* If a well-typed configuration is unable to reach an answer, then it
   must diverge. This formulation is amenable to an intuitionistic
   co-inductive proof. *)

Lemma type_soundness_large_step_preliminary:
  forall c1,
  jconf c1 ->
  (forall h2 t2, ~ (redstar c1 (Cf h2 t2) /\ answer t2)) ->
  (forall h2 t2, ~ (redstar c1 (Cf h2 t2) /\ spawner t2)) ->
  (forall h2 t2, ~ (redstar c1 (Cf h2 t2) /\ waits t2 h2)) ->
  diverges c1.
Proof.
  cofix. introv wt noanswer nospawner nowaits. destruct c1 as [ h1 t1 ].
  (* [c1] is well-typed, so either it is an answer, or it steps. *)
  forwards hp: progress; try eassumption; simpl in *.
  dependent destruction hp.
  (* In the former case, we have a contradiction, since we have assumed
     that [c1] is unable to reach an answer. *)
  false. eapply noanswer. eauto with redstar.
  (* In the latter case, we have produced one step, and the co-inductive
     hypothesis provides the rest. *)
  econstructor; [ eassumption | ].
  eapply type_soundness_large_step_preliminary;
    intuition eauto using subject_reduction with redstar.
  (* In the last case we have a spawner and a contradiction. *)
  false. eapply nospawner. eauto with redstar.
  (* Wait, there is still one case. *)
  false. eapply nowaits. eauto with redstar.
Qed.

(* A reformulation in classical logic: a well-typed configuration must either
   yield a value or diverge. *)

Require Import Classical.

Lemma type_soundness_large_step:
  forall c1,
  jconf c1 ->
  reduces_to_a_waispawswer c1 \/ diverges c1.
Proof.
  intros.
  (* Assume the negation of the left-hand disjunct, and prove the right-hand disjunct. *)
  match goal with |- ?A \/ ?B => destruct (classic A) as [ Hred | Hnred ] end. tauto. right.
  (* Apply the previous result. *)
  eapply type_soundness_large_step_preliminary;
  [ auto
  | intros; intro; apply Hnred; left; eauto
  | intros; intro; apply Hnred; right; left; eauto
  | intros; intro; apply Hnred; right; auto; eauto
  ].
Qed.

(* ---------------------------------------------------------------------------- *)

(* We instantiate the above result with the initial configuration. *)

(* If a closed term [t] is well-typed, then placing it in an empty heap and
   executing it for an arbitrary number of steps either eventually leads to
   either an answer or divergence. *)

Lemma type_soundness_large_step_initial:
  forall n t T,
  jt n void zilch TyEmpty t T ->
  let c := Cf concrete_empty t in
  reduces_to_a_waispawswer c \/ diverges c.
Proof.
  intros. subst c. eauto using type_soundness_large_step, initial_configuration.
Qed.

*)