Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Signature.
Require Import Dup.
Require Import Exclusive.
Require Import Adopts.
Require Import Harmless.

(* ---------------------------------------------------------------------------- *)

(* In the following, there are two variants of the main typing judgement, [j].
   In the [Canonical] variant, the use of the permission subsumption rule
   [JSubLeft] and of the subtyping rule [JSubRight] is forbidden. In the
   [Unrestricted] variant, it is allowed. The canonical variant is used to
   define the semantic interpretation of permissions, which gives rise to a
   semantic notion of permission subsumption. The syntactic notion of
   permission subsumption, represented by the judgements [jps] and [sub], is
   then proved sound with respect to this semantic notion. The use of [jps]
   and [sub] is allowed, via [JSubLeft] and [JSubRight], in [Unrestricted]
   type derivations. *)

Inductive flag :=
  | Canonical (* subsumption disabled *)
  | Unrestricted (* subsumption enabled *)
.

(* ---------------------------------------------------------------------------- *)

(* Permission subsumption. *)

(* The subsumption judgements [jps], [sub], and [row_sub] are mutually
   recursive. The typing judgements [j], [jbs], and [jpi], are mutually
   recursive (and depend on the subsumption judgements). Finally, comes the
   typing judgement for soups, [jcp]. We make all of these judgements mutually
   inductive, because this allows us to define a combined induction scheme,
   which is convenient. *)

(* The judgements [jps] and [sub] carry an integer index, which measures the
   height of the subtyping derivation. The proof of soundness of subsumption
   involves a well-founded induction over this index. As far as the end user
   is concerned, this index has no meaning. *)

(* The judgement [jps] takes the form [K |- P <= Q].

   This judgement means that the permissions [P] can be re-arranged to yield
   the permissions [Q].

   The set of axioms and rules that defines [jps] is not minimal: some rules
   can be derived from others. This is not a problem: it is easier to prove
   these rules sound directly than to prove that they can be derived from a
   number of other rules. Indeed, we will prove that [jps] is sound with
   respect to a semantic notion of permission subsumption. *)

Inductive jps : nat -> env kind -> term -> term -> Prop :=

(* Permission subsumption is reflexive and transitive. *)

| JpsReflexive:
    forall n K P,
    jps n K P P

| JpsTransitive:
    forall n K P1 P2 P3,
    jps n K P1 P2 ->
    jps n K P2 P3 ->
    jps (S n) K P1 P3

(* Every permission is affine. This is indicated by the following
   rule, which states that every permission [P] can be replaced
   with [empty]. In other words, [empty] is the top element in
   the ordering of permissions. *)

| JpsEmptyTop:
    forall n K P,
    jps n K
      P
      TyEmpty

(* [TyEmpty] is a left and right unit for [TyStar]. This is indicated
   by the following four permission subsumption rules, which state
   that [P * empty] and [empty * P] are interconvertible with [P]. *)

| JpsEmptyAppearsLeft:
    forall n K P,
    jps n K
      P
      (TyStar TyEmpty P)

| JpsEmptyAppearsRight:
    forall n K P,
    jps n K
      P
      (TyStar P TyEmpty)

| JpsDisappearsLeft:
    forall n K P Q,
    jps n K
      (TyStar P Q)
      Q

| JpsDisappearsRight:
    forall n K P Q,
    jps n K
      (TyStar P Q)
      P

(* Conjunction is commutative and associative. *)

| JpsStarCommutative:
    forall n K P1 P2,
    jps n K
      (TyStar P1 P2)
      (TyStar P2 P1)

| JpsStarAssociative:
    forall n K P1 P2 P3,
    jps n K
      (TyStar P1 (TyStar P2 P3))
      (TyStar (TyStar P1 P2) P3)

(* Equality is reflexive. That is, if [v] is a value, a permission of the form
   [v = v] can spontaneously appear during a permission subsumption step. I
   use [TyEmpty] on the left-hand side, even though an arbitrary permission [P]
   could be used here, because this makes it clearer that the equation appears
   out of nowhere. *)

| JpsEqualityReflexive:
    forall n K v,
    wk K v VALUE ->
    jps n K TyEmpty (TyEquation v v)

(* Equality is symmetric and transitive. *)

(* I believe that these rules could be derived out of [JpsReflexive],
   [JpsEqualsForEquals], and the rules that govern [TyEmpty] and [TyStar].
   However, as explained above, it is less costly for us to give these rules
   than to derive them. *)

| JpsEqualitySymmetric:
    forall n K v1 v2,
    jps n K
      (TyEquation v1 v2)
      (TyEquation v2 v1)

| JpsEqualityTransitive:
    forall n K v1 v2 v3,
    jps n K
      (TyStar (TyEquation v1 v2) (TyEquation v2 v3))
      (TyEquation v1 v3)

(* Equals can be substituted for equals. *)

(* This is stated as follows: in the presence of the equation [v1 = v2],
   the permissions [P(v1)] and [P(v2)] are interconvertible. Here, [v1]
   and [v2] are logical values, and [P] is a permission context with a
   hole of kind [VALUE]. *)

(* For convenience, we give multiple variants of this rule, including variants
   where we go from [v2] to [v1] instead of from [v1] to [v2] and variants
   where the equation is dropped. *)

| JpsEqualsForEquals:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v1 v2) P1)
      (TyStar (TyEquation v1 v2) P2)

| JpsEqualsForEqualsReverse:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v2 v1) P1)
      (TyStar (TyEquation v2 v1) P2)

| JpsEqualsForEqualsDrop:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v1 v2) P1)
      P2

| JpsEqualsForEqualsReverseDrop:
    forall n K v1 v2 P P1 P2,
    P1 = subst v1 0 P ->
    P2 = subst v2 0 P ->
    wk (insert 0 VALUE K) P PERM ->
    jps n K
      (TyStar (TyEquation v2 v1) P1)
      P2

(* A duplicable permission can be duplicated. *)

| JpsDuplicate:
    forall n K P,
    dup P ->
    jps n K
      P
      (TyStar P P)

(* A function that requires a duplicable permission [P] can be turned into a
   function that does NOT require this permission, provided [P] is available
   now. *)

| JpsHideDuplicablePrecondition:
    forall n K v T U P,
    dup P ->
    jps n K
      (TyStar (TyAnchored v (TyArrow (TyMixStar T P) U)) P)
      (TyAnchored v (TyArrow T U))

(* A package of a type and a permission can be implicitly introduced
   and eliminated. That is, the permissions [x @ (T | P)] and
   [(x @ T) * P] are inter-convertible. *)

| JpsMixStarIntro:
    forall n K v T P,
    jps n K
      (TyStar (TyAnchored v T) P)
      (TyAnchored v (TyMixStar T P))

| JpsMixStarElim:
    forall n K v T P,
    jps n K
      (TyAnchored v (TyMixStar T P))
      (TyStar (TyAnchored v T) P)

(* An existential permission can be introduced via the standard rule. *)

| JpsExistsIntro:
    forall n K v P Q,
    subst v 0 Q = P ->
    wk K v VALUE ->
    wk (insert 0 VALUE K) Q PERM ->
    jps n K
      P
      (TyExists Q)

(* An existential permission can be hoisted out of a conjunction. *)

| JpsExistsHoistLeft:
    forall n K P Q Q',
    shift 0 Q = Q' ->
    jps n K
      (TyStar (TyExists P) Q)
      (TyExists (TyStar P Q'))

(* A permission at a universal type is equivalent to a universal permission. *)

| JpsForallType2Perm:
    forall n K v v' kappa T,
    shift 0 v = v' ->
    jps n K
      (TyAnchored v (TyForall kappa T))
      (TyForall kappa (TyAnchored v' T))

| JpsForallPerm2Type:
    forall n K v v' kappa T,
    shift 0 v = v' ->
    jps n K
      (TyForall kappa (TyAnchored v' T))
      (TyAnchored v (TyForall kappa T))

(* A universal type or permission can be eliminated. *)

| JpsForallElimType:
    forall n K v kappa T U T',
    subst U 0 T = T' ->
    wk K U kappa ->
    jps n K
      (TyAnchored v (TyForall kappa T))
      (TyAnchored v T')

| JpsForallElimPerm:
    forall n K kappa P U P',
    subst U 0 P = P' ->
    wk K U kappa ->
    jps n K
      (TyForall kappa P)
      P'

(* A vacuous universal quantifier can be introduced. *)

| JpsVacuousForallIntro:
    forall n K kappa P P',
    substitutable kappa ->
    shift 0 P = P' ->
    jps n K
      P
      (TyForall kappa P')

(* Structural permissions can be decomposed and recomposed. *)

(* This is expressed informally as follows, where [Ts] is a row
   context with one hole: *)

(* v @ block m tag Ts[T]
   <=
   exists x, v @ block m tag Ts[x] * x @ T *)

| JpsDecomposeBlock:
    forall field n K v v' m tag T Ts T' Ts' Us Us' U U',
    shift 0 v = v' ->
    shift 0 T = T' ->
    shift 0 U = U' ->
    shift 1 Ts = Ts' ->
    fill_row Ts T = Us ->
    fill_row Ts' (TySingleton (TVar 0)) = Us' ->
    row_context field Ts ->
    wk K T TYPE ->
    wk (insert 0 TYPE K) Ts (ROW TYPE) ->
    jps n K
      (TyAnchored v (TyBlock m tag Us U))
      (TyExists
        (TyStar
          (TyAnchored v' (TyBlock m tag Us' U'))
          (TyAnchored (TVar 0) T')
        )
      )

(* v1 @ block m tag Ts[v2] * v2 @ T
   <=
   v1 @ block m tag Ts[T] *)

| JpsRecomposeBlock:
    forall field n K v1 v2 m tag T Ts TsT Tsv2 U,
    fill_row Ts T = TsT ->
    fill_row Ts (TySingleton v2) = Tsv2 ->
    row_context field Ts ->
    wk (insert 0 TYPE K) Ts (ROW TYPE) ->
    jps n K
      (TyStar
        (TyAnchored v1 (TyBlock m tag Tsv2 U))
        (TyAnchored v2 T)
      )
      (TyAnchored v1 (TyBlock m tag TsT U))

(* v @ block m1 tag1 Ts[f = v1] *
   v @ block m2 tag2 Ts[f = v2]
   imply
   v1 = v2 *)

(* If we have a conjunction of two block permissions about [v], as above, then
   necessarily [m1] and [m2] must be [Duplicable], and [tag1] and [tag2] must
   be equal, otherwise this conjunction implies [TyFalse]. We could add a large
   number of subsumption rules whose right-hand side is [TyFalse]. We do not
   bother. *)

| JpsUnifyBlock:
    forall field n K v v1 v2 m1 tag1 m2 tag2 Ts1 Ts2 Tsv1 Tsv2 U1 U2,
    fill_row Ts1 (TySingleton v1) = Tsv1 ->
    row_context field Ts1 ->
    wk (insert 0 TYPE K) Ts1 (ROW TYPE) ->    
    fill_row Ts2 (TySingleton v2) = Tsv2 ->
    row_context field Ts2 ->
    wk (insert 0 TYPE K) Ts2 (ROW TYPE) ->    
    jps n K
      (TyStar
        (TyAnchored v (TyBlock m1 tag1 Tsv1 U1))
        (TyAnchored v (TyBlock m2 tag2 Tsv2 U2))
      )
      (TyStar (TyStar
        (TyAnchored v (TyBlock m1 tag1 Tsv1 U1))
        (TyAnchored v (TyBlock m2 tag2 Tsv2 U2)))
        (TyEquation v1 v2)
      )

(* A structural permission can be turned into a nominal permission,
   provided the former is the unfolding of the latter. *)

| JpsFold:
    forall n K v m tag Ts U tc,
    tc_mode tc = m ->
    tc_tag tc tag Ts ->
    tc_adopts tc = U ->
    jps n K
      (TyAnchored v (TyBlock m tag Ts U))
      (TyAnchored v (TyCon tc))

(* The converse transformation is also possible, provided this algebraic data
   type has exactly one branch. If it has (zero or) several branches, then a
   [TSwitch] construct must usually be used instead. The subsumption rule
   [JpsDirectedUnfold] can also be used when there are several tags and the
   correct tag happens to be known. *)

| JpsUnfold:
    forall n K v m Ts U tc,
    tc_mode tc = m ->
    tc_is_record tc Ts ->
    tc_adopts tc = U ->
    jps n K
      (TyAnchored v (TyCon tc))
      (TyAnchored v (TyBlock m 0 Ts U))

(* [JpsDirectedUnfold] allows unfolding a nominal permission, even in the case
   where there are multiple branches, provided the correct tag is known.  This
   can be useful in practice. A typical scenario involves passing an immutable
   [Cons] cell to the list length function; even though the function only
   returns a [list] permission, since we have retained the information that
   the cell is tagged [Cons], we are able to unfold the [list] permission
   produced by the call. *)

| JpsDirectedUnfold:
    forall n K v m1 m tag Ts1 Ts U1 U tc,
    tc_mode tc = m ->
    tc_tag tc tag Ts ->
    tc_adopts tc = U ->
    jps n K
      (TyStar
        (TyAnchored v (TyBlock m1 tag Ts1 U1))
        (TyAnchored v (TyCon tc))
      )
      (TyStar
        (TyAnchored v (TyBlock m1 tag Ts1 U1))
        (TyAnchored v (TyBlock m tag Ts U))
      )

(* Because every memory location has type [dynamic], every type whose
   inhabitants are memory locations is a subtype of [dynamic]. This includes
   [TyBlock ...] and [TyCon ...]. Instead of explicitly listing these types,
   we write that any exclusive type is a subtype of [TyDynamic]. This should
   be a more forward-compatible definition. *)

| JpsDynamicAppears:
    forall n K v T,
    exclusive T ->
    jps n K
      (TyAnchored v T)
      (TyStar (TyAnchored v T) (TyAnchored v TyDynamic))

(* Bottom is the least type. This rule is useful, in particular, because
   JNew creates a memory block with a clause ``adopts bottom'', so [JpsCoBlock]
   and [JpsBottom] are needed to change this clause to a more useful type, if
   desired. *)

| JpsBottom:
    forall n K v Q,
    wk K Q PERM ->
    jps n K (TyAnchored v TyBottom) Q

(* Functions are contravariant in their domain and covariant in their
   codomain. This rule (and others like it) make the judgements [jps]
   and [sub] mutually recursive. *)

(* The rule [JpsCoArrow] is a little bit special in that it is
   considered to have zero height, regardless of the height of
   its premises. This is not important -- we could use [S m]
   in the conclusion instead of [n] -- but it emphasizes the
   fact that the proof of the lemma [jps_sub_sound] does not
   exploit the induction hypothesis in the case of [JpsCoArrow]. *)

| JpsCoArrow:
    forall m n K v T1 T2 U1 U2,
    sub m K U1 T1 ->
    sub m K T2 U2 ->
    wk K U1 TYPE ->
    jps n K
      (TyAnchored v (TyArrow T1 T2))
      (TyAnchored v (TyArrow U1 U2))

(* Memory blocks are covariant, regardless of whether they are mutable or
   immutable. The [adopts] clause is also covariant. *)

| JpsCoBlock:
    forall n K v m tag Ts Us T U,
    row_sub n K Ts Us ->
    sub n K T U ->
    jps (S n) K
      (TyAnchored v (TyBlock m tag Ts T))
      (TyAnchored v (TyBlock m tag Us U))

(* The type constructor [TyAnchored] is covariant. This is in fact stated
   by the rule [Sub] below, which defines [sub] in terms of [jps]. See
   the lemma [JpsCoAnchored] in the file [DerivedSubtyping]. *)

(* Permission conjunction is covariant. *)

| JpsCoStar:
    forall n K P1 P2 Q1 Q2,
    jps n K P1 P2 ->
    jps n K Q1 Q2 ->
    jps (S n) K (TyStar P1 Q1) (TyStar P2 Q2)

(* Universal quantification is covariant. *)

| JpsCoForall:
    forall n K kappa P Q,
    jps n (insert 0 kappa K) P Q ->
    jps (S n) K (TyForall kappa P) (TyForall kappa Q)

(* ---------------------------------------------------------------------------- *)

(* Subtyping is defined in terms of permission subsumption. *)

(* If, for a fresh variable [x], the permission [x @ T1] can be transformed
   to [x @ T2], then the type [T1] is considered a subtype of [T2]. *)

(* In other words, this rule temporarily invents a fresh name [x] for the
   root value, and this allows us to think in terms of permissions instead
   of in terms of types. This is handy because we already have many rules
   for reasoning about permission subsumption. *)

with sub : nat -> env kind -> term -> term -> Prop :=

(* K; x : VALUE |- (x @ T) <= (x @ U)
   ----------------------------------------
   K |- T <= U *)

| Sub:
    forall n K T U,
    jps n
      (insert 0 VALUE K)
      (TyAnchored (TVar 0) (shift 0 T))
      (TyAnchored (TVar 0) (shift 0 U)) ->
    sub (S n) K T U

(* ---------------------------------------------------------------------------- *)

(* Subtyping of rows is defined pointwise. *)

with row_sub : nat -> env kind -> term -> term -> Prop :=

| RowSubNil:
    forall n K,
    row_sub n K RowNil RowNil

| RowSubCons:
    forall n K T U Ts Us,
    sub n K T U ->
    row_sub n K Ts Us ->
    row_sub (S n) K (RowCons T Ts) (RowCons U Us)

(* ---------------------------------------------------------------------------- *)

(* The typing judgements. *)

(* There are two judgements, only one of which is needed when type-checking
   source programs.

     [j] concerns values and terms.
       We write [jc] for the canonical variant, which is used in the
       interpretation of permissions, and [jt] for the unrestricted
       variant, which is used to type-check source values and terms.
     [jpi] defines the semantic interpretation of permissions.

   [j] depends on [jpi] because a semantic pre-condition (a resource) can be
   turned into a syntactic pre-condition (a permission): [JCut].

   [jpi] depends on [jc] because the permission [v @ T] is considered true
   when the value [v] has type [T]: [JpiAtomic]. *)

(* The judgement [j] takes the form [R, K |- { P } t : T], where [R] is a
   resource, [K] is a kind environment (which maps variables to kinds), [t] is
   a term, [P] is a permission, and [T] is a type.

   The components [R] and [P] play the role of pre-conditions. [R] can be viewed
   as a semantic pre-condition, a resource that must be held for the term to be
   well-typed; and [P] can be viewed as a syntactic pre-condition, a permission
   that must be held for the term to be well-typed. Permissions are ultimately
   interpreted in terms of resources, so [R] and [P] serve analogous purposes,
   and indeed we provide a rule, [JCut], which allows turning a syntactic
   pre-condition into a semantic one (or vice-versa, depending on whether the rule
   is read bottom up or top down).

   The type [T] is a post-condition: it describes the value produced by
   executing [t] as well as the permissions that exist at the end of the
   execution.

   Instead of a type [T] on the right-hand side, one could prefer to use a
   permission [Q], which would serve as a post-condition, and where one more
   variable would be bound, allowing the post-condition to refer to the result
   of evaluating [t]. The two options should ultimately be equivalent, and
   this one seems simpler. *)

(* ---------------------------------------------------------------------------- *)

(* Typing rules for values and terms. *)

(* The judgement [j] carries an integer index [n], which measures the height
   of the typing derivation. The proof of canonicalization involves a
   well-founded induction over this index. As far as the end user is
   concerned, this index has no meaning. *)

with j : flag -> nat -> resource -> env kind -> term -> term -> term -> Prop :=

(* If [v] is a value, then [v] admits the singleton type [=v]. *)

  | JSingleton:
      forall n f R K P v1 v2,
      v1 = v2 ->
      wk K v2 VALUE ->
      j f n R K P v1 (TySingleton v2)

(* An introduction rule for [TyMixStar]. *)

(* [TyMixStar] can be introduced and eliminated via permission subsumption:
   see [JpsMixStarIntro] and [JpsMixStarElim]. Nevertheless, we do need an
   introduction rule for [TyMixStar] as part of the judgement [j]. Indeed, we
   must prove that permission subsumption is sound with respect to [jpi] (see
   the lemma [jps_sound]), and because the definition of [jpi] relies on [j],
   we need [j] to justify that a value [v] can have a type of the form
   [TyMixStar T P]. *)

(* Furthermore, this rule is useful when type-checking source programs: it is
   the frame rule. *)

  | JFrame:
      forall n f R K P Q t T,
      j f n R K P t T ->
      j f (S n) R K (TyStar P Q) t (TyMixStar T Q)

(* The judgement [j] is parameterized with a resource [R] and a permission
   [P], which both serve as pre-conditions. There is a communication between
   the two: (part of) the semantic pre-condition [R] can be used to discharge
   (part of) the syntactic pre-condition [P]. *)

(* This rule is required in order to type-check source programs. It is in fact
   the only branching rule, that is, the only rule that has two typing
   judgements among its premises. *)

  | JCut:
      forall n f R1 R2 R K P1 P2 t T,
      j f n R2 K (TyStar P1 P2) t T ->
      jpi R1 K P1 ->
      star R1 R2 = R ->
      j f (S n) R K P2 t T

(* A lambda-bound variable has kind [VALUE]. This indicates that it is
   a value and that it exists at runtime. *)

(* A permission [x @ T1] for the formal parameter [x] is available in the
   function body [t]. If a duplicable permission [P] is available outside the
   function body, it is also available inside. The resource [R] that is
   available outside is replaced with [core R] inside. A closure can capture
   only duplicable resources and permissions. *)

(* Inside the function body, the use of permission subsumption is allowed: the
   inside flag is [Unrestricted], regardless of the outside flag [f]. *)

  | JFunction:
      forall n f R K P T t U,
      wk K T TYPE ->
      dup P -> (* tex: \\ *)
      j
        Unrestricted
        n
        (core R)
        (insert 0 VALUE K)
        (intro P T)
        t
        (shift 0 U)
      ->
      j f (S n) R K P (TAbs t) (TyArrow T U)

(* At any point in a program, the currently available permissions
   can be re-organized, as per the permission subsumption rules. *)

(* This rule is enabled only when the flag is [Unrestricted]. *)

  | JSubLeft:
      forall m n R K P1 P2 t T,
      j Unrestricted n R K P2 t T ->
      jps m K P1 P2 ->
      j Unrestricted (S n) R K P1 t T

(* Subtyping on the right-hand side. *)

(* This rule is enabled only when the flag is [Unrestricted]. *)

(* If one expands the definition of [sub] in terms of [jps], one finds that
   this rule allows inventing a temporary name [x] for the return value,
   re-arranging the return permissions using [jps], and finally forgetting the
   name [x] again. *)

  | JSubRight:
      forall m n R K P t T1 T2,
      j Unrestricted n R K P t T1 ->
      sub m K T1 T2 ->
      j Unrestricted (S n) R K P t T2

(* The function application rule requires the left-hand term, [v], to be a
   value.

   There is no loss of generality in doing so; we can convert terms to
   administrative normal form before type-checking them.

   Because [v] is required to be a value, we are allowed to mention [v] within
   a type. More precisely, [v @ T -> U] is a well-formed permission, which
   states that [v] is a function. Thus, instead of using a premise that
   explicitly requires [v] to have type [T -> U], we simply make the
   permission [v @ T -> U] part of the pre-condition that appears in the
   conclusion.

   Why adopt this formulation? By placing the requirement [v @ T -> U] in the
   (syntactic) pre-condition, we avoid the need to split up the resource [R]
   between [v] and [t]. All of [R] goes to [t]. Thus, the only typing rule
   where [R] is split is [JCut]. *)

  | JApplication:
      forall n R K Q v t T U,
      wk K v VALUE ->
      j Unrestricted n R K Q t T ->
      j Unrestricted (S n) R K (TyStar (TyAnchored v (TyArrow T U)) Q) (TApp v t) U

(* The typing rule [JBlock] assigns a structural type to a memory location
   and at the same time checks that its adoptees are well-typed. *)

  | JBlock:
      forall n f R R1 R2 R'2 R'21 R'22 P K l m vs lys tag Ts U,
      (* We check that [R] can be split in two parts. *)
      star R1 R2 = R -> (* tex: \\ *)
      (* The values [vs] must be closed, but their types [Ts] need not be closed. *)
      wk zilch vs (ROW VALUE) ->
      (* The first part, [R1], justifies that the fields [vs] have types [Ts]. *)
      jpi R1 K (zip vs Ts) ->
      (* The second part, [R2], covers the memory block at address [l] and
         its adoptees. *)
      (* The adoptees of the object [l] form a list [lys] of memory locations.
         (If [m] is [Duplicable], then this list must be empty.)
         In [R2], every member [y] of this list has [y.adopter == l]. Now, by
         a thought experiment, we abandon all of them. This yields a resource
         [R'2] where every [y.adopter] is [null]. *)
      abandon_all R2 l lys R'2 -> (* tex: \\ *)
      (* Now, we check that [R'2] can be split in two parts, where: *)
      star R'21 R'22 = R'2 -> (* tex: \\ *)
      (* [R'21] justifies that every adoptee has type [U]. *)
      jpi R'21 K (zap lys U) ->
      wk K U TYPE ->
      (* [R'22] justifies the ownership of the memory block at address [l].
         We require this block to have no adopter. *)
      heap_maps R'22 l (mtagvs2dx m tag vs) ->
      (* The conclusion! *)
      j f n R K P (TLoc l) (TyBlock m tag Ts U)

(* A value that admits a structural type possibly also admits a nominal type,
   provided the former is the unfolding of the latter. *)

  | JFold:
      forall n f R K P v m tag Ts U tc,
      j f n R K P v (TyBlock m tag Ts U) -> (* tex: \\ *)
      tc_mode tc = m ->
      tc_tag tc tag Ts ->
      tc_adopts tc = U ->
      j f (S n) R K P v (TyCon tc)

(* A memory location [l] has type [dynamic] if and only if the block at address
   [l] has an [adopter] field. At present, every memory block has an [adopter]
   field, so every allocated memory location has type [dynamic]. This makes
   things particularly simple. *)

  | JDynamic:
      forall n f R K P l,
      l < heap_limit R ->
      j f n R K P (TLoc l) TyDynamic

(* Applying [new] to a tag [tag] and to a list of values of types [Ts] creates
   a memory block of type [TyBlock m tag Ts TyBottom]. The permissions for the
   values [vs] are consumed. The mode [m], one of [Duplicable] or [Exclusive],
   can be arbitrarily chosen. *)

  | JNew:
      forall n R K P m tag k vs Ts,
      wk K vs (ROW VALUE) ->
      wk K Ts (ROW TYPE) ->
      zip vs Ts = P ->
      row_length k vs ->
      row_length k Ts ->
      j Unrestricted n R K P (TNew tag vs) (TyBlock m tag Ts TyBottom)

(* Reading the field [field] in a memory block of type [TyBlock m tag Ts U]
   produces a value of type [T] if [T] is the [field]-th component of the row
   [Ts]. The permission for the memory block is not consumed. The type [T]
   must be duplicable. *)

  | JRead:
      forall n R K P m tag v field T Ts Ts' U,
      wk K v VALUE ->
      P = TyAnchored v (TyBlock m tag Ts' U) ->
      fill_row Ts T = Ts' ->
      row_context field Ts ->
      dup T ->
      j Unrestricted n R K P (TRead v field) (TyMixStar T P)

(* Writing a value of type [U] to the field [field] of a memory block of type
   [TyBlock Exclusive tag Ts[field:=T] U] changes the type of the memory block to
   [TyBlock Exclusive tag Ts[field:=U] U] and produces a unit result. *)

  | JWrite:
      forall n R K tag v1 field v2 T1 T2 Ts TsT1 TsT2 U,
      fill_row Ts T1 = TsT1 ->
      fill_row Ts T2 = TsT2 ->
      row_context field Ts ->
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      j Unrestricted n R K
        (TyStar (TyAnchored v1 (TyBlock Exclusive tag TsT1 U)) (TyAnchored v2 T2))
        (TWrite v1 field v2)
        (TyMixStar TyUnit (TyAnchored v1 (TyBlock Exclusive tag TsT2 U)))

(* Updating the tag of a mutable memory block changes its type from
   [TyBlock Exclusive tag1 Ts U] to [TyBlock m tag2 Ts U] and
   produces a unit result. The new mode [m] can be [Exclusive]
   or [Duplicable] -- in the latter case, the block is frozen. *)

  | JWriteTag:
      forall n R K m tag1 tag2 v Ts U,
      wk K v VALUE ->
      j Unrestricted n R K
        (TyAnchored v (TyBlock Exclusive tag1 Ts U))
        (TWriteTag v tag2)
        (TyMixStar TyUnit (TyAnchored v (TyBlock m tag2 Ts U)))

(* The construct [TSwitch v ts] requires a permission for the value [v] at
   some algebraic data type [TyCon tc]. In each branch, this permission is
   refined, that is, replaced with a permission for [v] at a structural
   type [TyBlock m tag Ts U]. *)

(* The premise is unfortunately not extremely readable. It uses an auxiliary
   judgement [jbs], which checks the list of branches [ts] against the
   algebraic data type definition [tc_branches tc]. *)

  | JSwitch:
      forall n R K P v ts tc T,
      jbs Unrestricted n R K (fun tag Ts =>
        (* This is the refined pre-condition. *)
        TyStar P (TyAnchored v (TyBlock (tc_mode tc) tag Ts (tc_adopts tc)))
      ) 0 (tc_branches tc) ts T ->
      wk K v VALUE ->
      wk K T TYPE ->
      j Unrestricted (S n) R K
        (TyStar P (TyAnchored v (TyCon tc)))
        (TSwitch v ts)
        T

(* The construct [TGive u t] requires the permissions [t @ T] and [u @ U],
   where the types [T] and [U] are in the relation [adopts]. It consumes
   the latter permission. *)

  | JGive:
      forall n R K v1 v2 T1 T2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      adopts T2 T1 ->
      j Unrestricted n R K
        (TyStar (TyAnchored v1 T1) (TyAnchored v2 T2))
        (TGive v1 v2)
        (TyMixStar TyUnit (TyAnchored v2 T2))

(* The construct [TTake u t] requires the permissions [t @ T] and [u @ dynamic].
   It requires the type [T] to be in the relation [adopts] with some type [U].
   It produces the permission [u @ U]. *)

  | JTake:
      forall n R K v1 v2 T1 T2,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      adopts T2 T1 ->
      j Unrestricted n R K
        (TyStar (TyAnchored v1 TyDynamic) (TyAnchored v2 T2))
        (TTake v1 v2)
        (TyMixStar TyUnit (TyStar (TyAnchored v1 T1) (TyAnchored v2 T2)))

(* The [fail] construct is well-typed in every context. It represents a
   runtime failure that the type system cannot detect. The [take] operation
   reduces to [fail] when it fails. *)

  | JFail:
      forall n R K P T,
      wk K T TYPE ->
      j Unrestricted n R K P TFail T

(*
(* The construct [TSpawn v] requires the expression [v] to be a value,
   thus [v @ T -> U] is a well-formed permission. Here we also state that
   the type [T] has the form [TyUnit | P] for some permission [P]. *)
*)

  | JSpawn:
      forall n f R K v1 v2 T U,
      wk K v1 VALUE ->
      wk K v2 VALUE ->
      j f n R K (TyStar (TyAnchored v1 (TyArrow T U)) (TyAnchored v2 T))
                (TSpawn v1 v2)
                TyUnit

(* The construct [TLck l] is well-typed if the lock resource is defined on
   [l]. *)

  | JLck:
      forall n f R K Q l lp P,
      wk zilch P PERM ->
      lck_maps R l lp ->
      is_locked_permission lp P ->
      j f n R K Q (TLck l) (TyLck P)

  | JLocked:
      forall n f R K Q l P,
      wk zilch P PERM ->
      lck_maps R l (p2hl P) ->
      j f n R K Q (TLck l) TyLocked

(* The construct [TNewLck] is well-typed in every context. *)
(* In this version, the lock is created free, thus we have to provide
   and consume the permission P. *)

  | JNewLck:
      forall n R K P,
      wk zilch P PERM -> (* [zilch] should be [K] here! fixed in reconstruction. *)
      j Unrestricted n R K P TNewLck (TyLck P)

(*
We would like to have an existential type here.
This does not exist yet in the formalization.
  | JNewLck:
      forall n R K P Q,
      wk zilch P PERM ->
      j Unrestricted n R K Q TNewLck (TyExists
                                      (TyMixStar (TySingleton (TLck 0))
                                       (TyStar (TyAnchored (TLck 0) (TyLck (shift 0 P)))
                                               (TyAnchored (TLck 0) (TyLocked)))))
*)

(* Acquiring a lock grants a permission. *)
(* The argument [v] of acquire being required to be a value, the permission
   [v @ TyLck P] is well_formed. *)

  | JAcquire:
      forall n f R K P v,
      wk K v VALUE ->
      j f n R K (TyAnchored v (TyLck P))
                (TAcquire v)
                (TyMixStar TyUnit (TyStar P (TyAnchored v TyLocked)))

(* Releasing a lock consumes a permission. *)

  | JRelease:
      forall n f R K P v,
      wk K v VALUE ->
      j f n R K (TyStar P (TyStar (TyAnchored v (TyLck P)) (TyAnchored v TyLocked)))
                (TRelease v) 
                TyUnit

(* Eliminating an existential permission introduces a new name. For this
   reason, it cannot be considered a special case of [JSubLeft]. *)

  | JExistsElim:
      forall n R K P t T,
      j Unrestricted n R (insert 0 VALUE K) P (shift 0 t) (shift 0 T) ->
      j Unrestricted (S n) R K (TyExists P) (TUnpackIn t) T

(* The introduction of a universal quantifier is restricted to harmless
   terms. This is a class of expressions whose execution definitely
   cannot allocate a new lock. *)

  | JForallIntro:
      forall f n R K P t kappa T,
      wk K t PROG ->
      harmless t ->
      j f n R (insert 0 kappa K) P (shift 0 t) T ->
      j f (S n) R K (TyForall kappa P) t (TyForall kappa T)

(* ---------------------------------------------------------------------------- *)

(* The auxiliary judgement [jbs f n R K P tag Tss ts T] is used to check a row
   of terms [ts], which represent the branches of a [TSwitch] construct,
   against a list of branches [Tss], which represent the definition of an
   algebraic data type. The pre-condition [P] is a *function* of the tag [tag]
   and fields [Ts] in each branch. *)

(* One could give a more readable definition by defining a higher-order predicate
   [in_each_branch] independently of the judgement [j]. Unfortunately, that would
   make it difficult for us to construct a mutual induction principle. *)

with jbs : flag -> nat -> resource -> env kind -> (* f, n, R, K *)
           (nat -> term -> term) ->               (* the parameterized pre-condition *)
           nat ->                                 (* the tag, which counts up *)
           list term ->                           (* the branches [Tss] *)
           term ->                                (* the branches [ts] *)
           term ->                                (* the return type [T] *)
           Prop :=

  | JbsZero:
      forall f n R K P tag T,
      jbs f n R K P tag nil RowNil T

  | JbsSucc:
      forall f n R K P tag Ts Tss t ts T,
      jbs f n R K P          (S tag) Tss         ts             T ->
      j   f n R K (P tag Ts)                     t              T ->
      jbs f n R K P             tag  (Ts :: Tss) (RowCons t ts) T

(* ---------------------------------------------------------------------------- *)

(* A specific judgement [jcp R t] states that a parallel program [t] is
   well-typed, given the resources [R]. *)

with jcp : resource -> term -> Prop :=

  | JcpThread:
      forall n f R t T,
      wk zilch t PROG ->
      j f n R zilch TyEmpty t T ->
      jcp R (TThread t)

(* The construct [TPar t u] requires the programs [t] and [u] to be well-typed
   in independent contexts. *)

  | JcpPar:
      forall R R1 R2 t1 t2,
      wk zilch t1 CPROG ->
      wk zilch t2 CPROG ->
      star R1 R2 = R ->
      jcp R1 t1 ->
      jcp R2 t2 ->
      jcp R (TPar t1 t2)

(* ---------------------------------------------------------------------------- *)

(* A semantic interpretation of permissions. *)

(* A permission is analogous to a separation logic formula: it is an assertion
   about the variables and about the heap, or more generally, about a resource. *)

(* One might think that we need only give meaning to closed permissions, and
   the meaning of open permissions will follow. In fact, in an early version
   of this proof, the judgement [jpi] was only about closed permissions.
   However, now that the judgement [j] depends on [jpi], we need to interpret
   open permissions as well. This seems to be necessary in order to allow a
   smooth transition (by substitution) between type derivations for source
   programs and type derivations for programs under execution.

   Because we give meaning to open permissions, the judgement [jpi] is
   parameterized with an environment [K]. This environment is effectively used
   only by [JpiAtomic], the rule that states that the permission [v @ T] if
   [v] has type [T]. There, [v] can have free variables. *)

(* A permission is an assertion about the current resource. A permission [P]
   is either true or false of a resource [R]. This is the interpretation of
   permissions. It is defined by the judgement [jpi R K P]. *)

(* This predicate is not used to type-check source programs. It plays a role
   in the type soundness proof only. *)

(* Technically, we could either build the consistency of [R] into the
   definition of [jpi], so that [jpi R K P] implies [consistent R], or not
   build it in, in which case we have to explicitly require both [jpi R K P]
   and [consistent R] in later lemmas and definitions, such as [jps_sound] and
   [jconf]. It seems that the proof of [jps_sound] is easier if we do not
   build consistency in -- even though this lemma needs a consistency
   hypothesis. *)

with jpi : resource -> env kind -> term -> Prop :=

(* The empty permission is true of an arbitrary resource. *)

| JpiEmpty:
    forall R K,
    jpi R K TyEmpty

(* A conjunction of two permissions is true of a resource [R] if and
   only [R] can be split as [star R1 R2] and each of the conjuncts is
   true of one fragment. *)

| JpiStar:
    forall R1 R2 R K P1 P2,
    jpi R1 K P1 ->
    jpi R2 K P2 ->
    star R1 R2 = R ->
    jpi R K (TyStar P1 P2)

(* Universal permissions are interpreted in the obvious manner. *)

| JpiForall:
    forall R kappa K P,
    substitutable kappa ->
    jpi R (insert 0 kappa K) P ->
    jpi R K (TyForall kappa P)

(* Existential permissions are interpreted in the obvious manner. *)

| JpiExists:
    forall R K v P,
    jpi R K (subst v 0 P) ->
    wk K v VALUE ->
    wk (insert 0 VALUE K) P PERM ->
    jpi R K (TyExists P)

(* A permission of the form [v @ T], where [v] is a value, is true if
   [v] has type [T] under a permission [P] which itself is true. *)

(* We cannot require [P] to be [TyEmpty]. This would be too demanding. The
   derivation of the premise [jc _ _ P v T] must be canonical, so it cannot
   exploit [JSubLeft], and as a result, we cannot control the syntactic
   shape of [P]. *)

| JpiAtomic:
    forall n R1 R2 R K P v T,
    wk K v VALUE ->
    j Canonical n R1 K P v T ->
    jpi R2 K P ->
    star R1 R2 = R ->
    jpi R K (TyAnchored v T)
.

Hint Constructors jps sub row_sub j jbs jpi : j.

Hint Constructors jpi : jpi.

Hint Constructors jcp : jcp.

Notation jt := (j Unrestricted).

Notation jc := (j Canonical).

