Set Implicit Arguments.
Generalizable All Variables.
Require Import Relation_Definitions.
Require Import EquivDec.
Require Import MyTactics.
Require Import MSA.

Section Definitions.

Context `{EqDec A eq, B}.

(* ------------------------------------------------------------------------- *)

(* Duplicable-or-exclusive (DX) permissions. *)

(* A duplicable-or-exclusive permission on an object is one of: [DX_NONE],
   which means we have no permission on the object, and know nothing about its
   current state; [DX_DUPLICABLE a], which means we know that the object is in
   state [a], others may know it as well, and this state is fixed forever;
   [DX_EXCLUSIVE b], which means we have an exclusive permission over the
   object, and the object is currently in state [b]; and [DX_ERROR], which
   represents an inconsistent permission. *)

(* The descriptions associated with duplicable objects have type [A], while
   the descriptions associated with exclusive objects have type [B]. There
   is no requirement that the types [A] and [B] coincide. *)

(* Note: I initially allowed [DX_NONE] and [DX_DUPLICABLE] to co-exist. But
   this breaks [compatible_implies_common_core_left], which states that two
   compatible resources have the same core. Thus, I do not allow them to
   co-exist. If an active thread changes its permission from [DX_EXCLUSIVE]
   to [DX_DUPLICABLE], then every inactive thread will have to adapt by
   changing their permission from [DX_NONE] to [DX_DUPLICABLE]. *)

Inductive dx_permission :=
| DX_NONE: dx_permission
| DX_DUPLICABLE: A -> dx_permission
| DX_EXCLUSIVE: B -> dx_permission
| DX_ERROR: dx_permission.

(* DX permissions form a monotonic separation algebra. *)

(* Equality over [A] must be decidable. This is exploited in the definition of
   the function [star], where one needs to determine whether the two sides
   agree on a single value [a]. *)

Obligation Tactic := idtac.

Global Program Instance msa_dx_permission : MSA dx_permission := {

  star p1 p2 :=
    match p1, p2 return _ with
    | DX_ERROR, _
    | _, DX_ERROR
    | DX_EXCLUSIVE _, DX_EXCLUSIVE _
    | DX_EXCLUSIVE _, DX_DUPLICABLE _
    | DX_DUPLICABLE _, DX_EXCLUSIVE _
    | DX_DUPLICABLE _, DX_NONE
    | DX_NONE, DX_DUPLICABLE _ =>
        DX_ERROR
    | DX_DUPLICABLE a1, DX_DUPLICABLE a2 =>
        (* [DX_DUPLICABLE a] is compatible with itself, and is idempotent:
           it is duplicable. *)
        if equiv_dec a1 a2 then DX_DUPLICABLE a1 else DX_ERROR
    | DX_NONE, DX_NONE =>
        (* [DX_NONE] is also idempotent. *)
        DX_NONE
    | DX_NONE, DX_EXCLUSIVE b
    | DX_EXCLUSIVE b, DX_NONE =>
        (* [DX_NONE] is compatible with, and a neutral element for,
           [DX_EXCLUSIVE b]. *)
        DX_EXCLUSIVE b
    end;

  consistent p :=
    match p return _ with
    | DX_ERROR =>
        False
    | _ =>
        True
    end;

  core p :=
    match p return _ with
    | DX_ERROR =>
        DX_ERROR
    | DX_DUPLICABLE a =>
        (* [DX_DUPLICABLE a] is preserved by [core], because it represents
           duplicable information. *)
        DX_DUPLICABLE a
    | DX_NONE
    | DX_EXCLUSIVE _ =>
        DX_NONE
    end;

  (* To a first approximation, the relation [rio] over permissions is
     equality: when a thread is inactive, the permissions that it holds do not
     change. There is one exception, though: [DX_NONE] can change to
     [DX_DUPLICABLE], which means that the active thread has changed its
     permission from exclusive to duplicable, and the passive thread must
     adapt. *)

  rio p1 p2 :=
    match p1, p2 return _ with
    | DX_NONE, DX_DUPLICABLE _ =>
        True
    | _, _ =>
        p1 = p2
    end

}.

Ltac dx :=
  intros;
  repeat match goal with R: dx_permission |- _ => destruct R end;
  simpl in *;
  equiv_dec_by_cases.

Next Obligation. (* star_commutative *)
  dx.
Qed.

Next Obligation. (* star_associative *)
  dx.
Qed.

Next Obligation. (* consistent_split_left *)
  dx.
Qed.

Next Obligation. (* core_right_unit *)
  dx.
Qed.

Next Obligation. (* compatible_implies_common_core_left *)
  dx; congruence.
Qed.

Next Obligation. (* duplicable_implies_self_core *)
  dx; congruence.
Qed.

Next Obligation. (* cores_are_duplicable *)
  dx.
Qed.

Next Obligation. (* duplicable_split_left *)
  dx; congruence.
Qed.

Next Obligation. (* rio_reflexive *)
  dx.
Qed.

Next Obligation. (* rio_consistent *)
  dx; congruence.
Qed.

Next Obligation. (* rio_star *)
  intros.
  case_eq R1a; case_eq R1b; case_eq R2; intros;
  try solve [
    (* In most cases, [rio] implies that [R1] and [R2] are equal,
       and the result is immediate: just take [R2a] and [R2b] to
       be [R1a] and [R1b]. This solves 63 out of 64 cases. *)
    exists R1a R1b; subst; equiv_dec_by_cases; eauto
  ].
  (* In the remaining case, [R1] is [DX_NONE] and [R2] is [DX_DUPLICABLE _]. *)
  exists R2 R2. subst. equiv_dec_by_cases.
Qed.

Next Obligation. (* rio_core *)
  dx; congruence.
Qed.

(* ------------------------------------------------------------------------- *)

(* The permission [DX_DUPLICABLE _] is duplicable. *)

Lemma duplicable_duplicable:
  forall a,
  duplicable (DX_DUPLICABLE a).
Proof.
  dx.
Qed.

(* The permission [DX_EXCLUSIVE _] is not duplicable. *)

Lemma not_duplicable_exclusive:
  forall b,
  ~ duplicable (DX_EXCLUSIVE b).
Proof.
  dx. congruence.
Qed.

(* An evolution from [DX_NONE] to [DX_DUPLICABLE] is permitted by [rio]. *)

Lemma rio_none_duplicable:
  forall a,
  rio DX_NONE (DX_DUPLICABLE a).
Proof.
  dx.
Qed.

(* If [p] evolves to [DX_EXCLUSIVE b] along [rio], then [p] must be equal
   to [DX_EXCLUSIVE b]. *)

Lemma rio_exclusive_reverse:
  forall { p b },
  rio p (DX_EXCLUSIVE b) ->
  p = DX_EXCLUSIVE b.
Proof.
  dx.
Qed.

(* ------------------------------------------------------------------------- *)

(* In the algebra of duplicable-or-exclusive permissions, [DX_DUPLICABLE a]
   and [DX_EXCLUSIVE a] are consistent, maximal, and stable elements. *)

Lemma consistent_duplicable:
  forall a,
  consistent (DX_DUPLICABLE a).
Proof.
  dx.
Qed.

Lemma consistent_exclusive:
  forall b,
  consistent (DX_EXCLUSIVE b).
Proof.
  dx.
Qed.

Lemma maximal_duplicable:
  forall a,
  maximal (DX_DUPLICABLE a).
Proof.
  unfold maximal. intros. dx; congruence.
Qed.

Lemma maximal_exclusive:
  forall b,
  maximal (DX_EXCLUSIVE b).
Proof.
  unfold maximal. intros. dx; congruence.
Qed.

Lemma stable_duplicable:
  forall a,
  stable (DX_DUPLICABLE a).
Proof.
  unfold stable. intros. dx.
Qed.

Lemma stable_exclusive:
  forall b,
  stable (DX_EXCLUSIVE b).
Proof.
  unfold stable. intros. dx.
Qed.

Lemma maximal_error:
  maximal DX_ERROR.
Proof.
  unfold maximal. dx.
Qed.

(* ------------------------------------------------------------------------- *)

(* The only way of splitting a duplicable permission is into two copies of
   itself. *)

Lemma split_duplicable:
  forall { p1 p2 a },
  star p1 p2 = DX_DUPLICABLE a ->
  p1 = DX_DUPLICABLE a /\ p2 = DX_DUPLICABLE a.
Proof.
  dx; congruence.
Qed.

(* The only way of splitting an exclusive permission is into itself and
   [DX_NONE]. *)

Lemma split_exclusive:
  forall { p1 p2 b },
  star p1 p2 = DX_EXCLUSIVE b ->
  p1 = DX_EXCLUSIVE b /\ p2 = DX_NONE \/
  p1 = DX_NONE /\ p2 = DX_EXCLUSIVE b.
Proof.
  dx; congruence.
Qed.

(* ------------------------------------------------------------------------- *)

(* If a permission [p] is consistent with [DX_EXCLUSIVE b], then [p] must be
   [DX_NONE]. *)

Lemma consistent_star_exclusive:
  forall { b p },
  consistent (star (DX_EXCLUSIVE b) p) ->
  p = DX_NONE.
Proof.
  dx.
Qed.

(* ------------------------------------------------------------------------- *)

End Definitions.

Implicit Arguments dx_permission [ ].

Section Agreement.

Context `{A : Type, B : Type, C : Type}.
Context `{d_agrees : C -> A -> Prop, x_agrees : C -> B -> Prop}.
Context `{H : EqDec A eq}.

(* ------------------------------------------------------------------------- *)

(* The relation [dx_permission_agrees c p] indicates that the permission [p]
   is compatible with the underlying value [c]. *)

Inductive dx_permission_agrees : C -> dx_permission A B -> Prop :=
| DxPermissionAgreesDuplicable:
    forall c a,
    d_agrees c a ->
    dx_permission_agrees c (DX_DUPLICABLE a)
| DxPermissionAgreesExclusive:
    forall c b,
    x_agrees c b ->
    dx_permission_agrees c (DX_EXCLUSIVE b).

Lemma dx_permission_agrees_consistent:
  forall c p,
  dx_permission_agrees c p ->
  consistent p.
Proof.
  inversion 1; subst; simpl; tauto.
Qed.

(* ------------------------------------------------------------------------- *)

End Agreement.

Implicit Arguments dx_permission_agrees [ A B C ].

