Set Printing All.
Set Implicit Arguments.
Generalizable All Variables.
Require Import MyProofIrrelevance.
Require Import MyTactics.
Require Import MyList.
Require Import MSA.
Require Import MSAAllocationMaps.
Require Import MSASAMOperations.
Require Import MSAAllocationMapOperations.
Require Import MSAClosedAllocationMaps.
Require Import MSAMorphism.

(* Let us write CAM for a closed allocation map. *)

(* ---------------------------------------------------------------------------- *)

Ltac came :=
  intros;
  repeat match goal with R : ClosedAMap _ |- _ => destruct R end;
  simpl in *;
  exist_exist_hypotheses;
  exist_exist_goal;
  simpl.

Obligation Tactic := idtac.

(* ---------------------------------------------------------------------------- *)

(* We need to transport to the level of CAMs each of the operations that existed
   at the level of allocation maps. *)

Section Operations.

Variable A : Type.

Context `{MA: MSA A, @Present A MA, @Pointsto A MA}.

Variable B : Type.

Context `{MB: MSA B}.

Context (project: B -> ClosedAMap A).
Context (inject: ClosedAMap A -> B -> B).
Context (MSAP: MSAMorphism project).
Context (MSAI: MSABiMorphism inject).
Context (sub_struct: SubStructure project inject).

(* Instance UAM: MSA (aMap A). *)
(* Proof. *)
(* eauto with typeclass_instances. *)
(* Defined. *)

(* Goal MSA (aMap A). *)
(* eauto with typeclass_instances. *)
(* Qed. *)

(* Instance UCAM: MSA (ClosedAMap A). *)
(* Proof. *)
(* eauto with typeclass_instances. *)
(* Defined. *)

(* (* Goal MSA (ClosedAMap A). *) *)
(* (* eauto with typeclass_instances. *) *)
(* (* Qed. *) *)

(* ---------------------------------------------------------------------------- *)

(* Empty. *)

Program Definition CAM_empty : ClosedAMap A :=
  empty_map A.

Next Obligation. (* the empty heap fragment is closed *)
  introv h. false. eauto using maps_empty.
Qed.

(* Notation SCAM_empty R := *)
(*   (inject CAM_empty R). *)

Program Definition SCAM_empty (R: B): B :=
  inject CAM_empty R.

(* Next Obligation. *)
(*   intros. *)
(*   introv h. false. eauto using maps_empty. *)
(* Qed. *)

Lemma consistent_SCAM_empty:
  forall B,
  consistent B ->
  consistent (SCAM_empty B).
Proof.
  intros.
  eapply bimorph_consistent; [ econstructor | eauto ].
Qed.

Lemma duplicable_CAM_empty:
  duplicable CAM_empty.
Proof.
  came. eapply duplicable_empty_map.
Qed.

Lemma duplicable_SCAM_empty:
  forall B,
  duplicable B ->
  duplicable (SCAM_empty B).
Proof.
  introv hdB.
  unfold SCAM_empty.
  rewrite <- bimorph_star.
  rewrite duplicable_CAM_empty.
  rewrite hdB.
  simpl in *.
  auto.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Definition of the operations. *)

Definition CAM_maps (R : ClosedAMap A) (r : nat) (a : A) :=
  let (R, _) := R in
  maps R r a.

Definition SCAM_maps R r a :=
  (CAM_maps (project R) r a).


Program Definition CAM_allocate
  (R : ClosedAMap A)
  (a : A)
  (req: forall y, ~ pointsto a y)
: ClosedAMap A :=

    allocate R a.

Next Obligation. (* allocation preserves closedness *)
  intros [ R hR ]. simpl.
  (* Consider an edge from [x] to [y] in [allocate R a]. *)
  repeat intro.
  (* Either [x] is an old address, or it is the newly allocated address. *)
  forwards [ | ]: maps_allocate_inversion; [ eauto | | ]; intros.
  (* Sub-case: [x] was present in [R]. *)
  (* We exploit the fact that [R] is closed, and learn that [y] was present
     in [R]. *)
  forwards: hR; [ eauto | eauto | unfold domain in *; unpack ].
  (* Thus, [y] is still present in [allocate R a]. The result follows. *)
  eauto using maps_allocate.
  (* Sub-case: [x] is new. *)
  (* There cannot be an edge from [x] to [y], because we have required the
     permission [a], which describes [x], to not carry an edge. Thus, this
     case is impossible. *)
  subst. false. unfold not in *. eauto.
Qed.

Definition SCAM_allocate (R: B) (a: A) (req: forall y, ~ pointsto a y): B :=
  inject (CAM_allocate (project R) a req) R.

Program Fixpoint CAM_limit (R : ClosedAMap A) :=
  @limit A R.

Definition SCAM_limit (R : B) :=
  CAM_limit (project R).

(* The predicate [has_no_antecedents R y] means that there no edges whose
   origin is in the domain of [R] and whose target is [y]. *)

Definition has_no_antecedents R y :=
  forall x a,
  maps R x a ->
  pointsto a y ->
  False.

(* For update to preserve closedness, we must impose two requirements. First,
   in order to deal with the pre-existing edges whose destination is [s], we
   require that either [s] remains present after the update, or there were no
   such edges in the first place. Second, in order to deal with the newly
   created edges whose origin is [s], we require that the destination of these
   edges is in the domain of the heap fragment after the update (this allows
   the creation of self-loops, should they be useful). *)

Unset Implicit Arguments.


(* Definition SCAM_has_no_antecedents (R: B) y := *)
(*   forall x a, *)
(*   SCAM_maps R x a -> *)
(*   pointsto a y -> *)
(*   False. *)

(* Lemma dReq_project: *)
(*   forall (R: B) (s: nat) (a: A), *)
(*   present a \/ SCAM_has_no_antecedents R s -> *)
(*   present a \/ has_no_antecedents (proj1_sig (project R)) s. *)
(* Proof. *)
(*   introv hdisj. *)
(*   destruct hdisj as [ hp | hhna ]. *)
(*     left. auto. *)
(*     right. intros ? ? ? ?. eapply hhna. *)
(*       unfold SCAM_maps; destruct (project R); came; eauto. *)
(*       eauto. *)
(* Qed. *)

(* Definition SCAM_domain (R: B) y := *)
(*   exists b, *)
(*   SCAM_maps R y b /\ present b. *)

(* Lemma oReq_project: *)
(*   forall (R: B) (s: nat) (a: A), *)
(*   (forall y, pointsto a y -> SCAM_domain ( *)


Program Definition CAM_update
  (R : ClosedAMap A) (s : nat) (a : A)
  (destinationReq: present a \/ has_no_antecedents R s)
  (originReq: forall y, pointsto a y -> domain (update R s a) y)
: ClosedAMap A :=
  update R s a.

Next Obligation. (* update preserves closedness *)
  intros [ R hR ]. simpl.
  (* Consider an edge from [x] to [y] in [update R s a]. *)
  repeat intro.
  (* Is [x] the address that was updated? *)
  forwards [ | ]: maps_update_inversion; [ eauto | | ]; intros; unpack.
  (* Case: [x] and [s] are distinct. *)
  (* Because [R] is closed, [y] is in the domain of [R]. *)
  forwards: hR; [ eauto | eauto | unpack ].
  (* Is [y] the address that was updated? *)
  case (eq_nat_dec y s); intros; [ subst y | ].
    (* Sub-case: yes. *)
    (* We are potentially in trouble. There was an edge from [x] to [s],
       and we have just modified the permission for [s]. We must prove
       that the new permission for [s] is still ``present'', OR argue
       that this case is impossible, i.e. [s] did not have any antecedents. *)
    destruct destinationReq.
      (* Sub-sub-case: [s] is still present. All is well. *)
      unfold domain in *. unpack.
      solve [ eauto using maps_update_there ].
      (* Sub-sub-case: [s] has no adoptees. *)
      false. solve [ eauto ].
    (* Sub-case: no. *)
    unfold domain in *. unpack.
    solve [ eauto using maps_update_elsewhere ].
  (* Case: [x] and [s] coincide. *)
  subst x. subst.
  (* The edge from [s] to [y] that we are considering may be a newly
     created edge. We exploit the hypothesis that any newly created
     edges point into the heap fragment. *)
  eapply originReq. eassumption.
Defined.

(* Ici ça plante si on demande le type [ClosedAMap] pour [R1] et [R2].
   Ça paraît logique puisque [has_no_antecedents] demande un [aMap].
   Mais pourquoi est-ce que celà passe au-dessus ? *)

Lemma CAM_update_dreq_equality:
  forall (R1 R2: aMap A),
  R1 = R2 ->
  forall (s: nat) (a: A),
  (present a \/ has_no_antecedents R1 s) ->
  (present a \/ has_no_antecedents R2 s).
Proof.
  intros. subst. tauto.
Qed.
  
Lemma CAM_update_oreq_equality:
  forall (R1 R2: aMap A),
  R1 = R2 ->
  forall (s: nat) (a: A),
  (forall y, pointsto a y -> domain (update R1 s a) y) ->
  (forall y, pointsto a y -> domain (update R2 s a) y).
Proof.
  introv ? ?. subst. tauto.
Qed.

Lemma proj1_sig_equality:
  forall (R1 R2: ClosedAMap A),
  R1 = R2 ->
  proj1_sig R1 = proj1_sig R2.
Proof.
  intros; subst; tauto.
Qed.

Lemma CAM_update_equality:
  forall R1 R2 s a dreq1 oreq1,
  forall eq : R1 = R2,
  CAM_update R1 s a dreq1 oreq1 =
  CAM_update R2 s a
             (CAM_update_dreq_equality (proj1_sig R1)
                                       (proj1_sig R2)
                                       (proj1_sig_equality R1 R2 eq)
                                       s a dreq1)
             (CAM_update_oreq_equality (proj1_sig R1)
                                       (proj1_sig R2)
                                       (proj1_sig_equality R1 R2 eq)
                                       s a oreq1).
Proof.
  unfold CAM_update. came. congruence.
Qed.

Definition SCAM_update (R: B) (s: nat) (a: A) dreq oreq: B :=
  inject (CAM_update (project R) s a dreq oreq) R.

Lemma project_equality:
  forall (R1 R2: B),
  R1 = R2 ->
  project R1 = project R2.
Proof.
  intros; subst; tauto.
Qed.

Lemma project_sig_equality:
  forall (R1 R2: B),
  R1 = R2 ->
  proj1_sig (project R1) = proj1_sig (project R2).
Proof.
  intros; subst; tauto.
Qed.

Lemma SCAM_update_equality:
  forall R1 R2 s a dreq1 oreq1,
  forall eq : R1 = R2,
  SCAM_update R1 s a dreq1 oreq1 =
  SCAM_update R2 s a
             (CAM_update_dreq_equality (proj1_sig (project R1))
                                       (proj1_sig (project R2))
                                       (project_sig_equality R1 R2 eq)
                                       s a dreq1)
             (CAM_update_oreq_equality (proj1_sig (project R1))
                                       (proj1_sig (project R2))
                                       (project_sig_equality R1 R2 eq)
                                       s a oreq1).
Proof.
  unfold SCAM_update, CAM_update; intros; subst.
  eapply injectA_equality.
  came.
  auto.
Qed.

Set Implicit Arguments.

Ltac scame :=
  unfold SCAM_empty, SCAM_maps, SCAM_allocate, SCAM_limit, SCAM_update;
  intros;
  repeat match goal with
        | hpi: context[project (inject _ _)] |- _ =>
            rewrite pi_inverse in hpi
        | |- context[project (inject _ _)] =>
            rewrite pi_inverse
      end;
  try match goal with
           | hc: consistent _ |- _ =>
               forwards: (@morph_consistent B _ _ _ project _) hc
         end;
  try match goal with
           | hr: rio _ _ |- _ =>
               forwards: morph_rio hr
         end;
  repeat match goal with
        | hc: context[project (core _)] |- _ =>
            rewrite (@morph_core _ _ _ _ project _) in hc
        | |- context[project (core _)] =>
            rewrite (@morph_core _ _ _ _ project _)
      end;
  repeat match goal with
        | hs: context[project (star _ _)] |- _ =>
            rewrite (@morph_star _ _ _ _ project _) in hs
        | |- context[project (star _ _)] =>
            rewrite (@morph_star _ _ _ _ project _)
      end;
  repeat match goal with
           | h: context[project ?r] |- _ =>
               destruct (project r)
           | |- context[project ?r] =>
               destruct (project r)
         end;
  try match goal with
        | |- consistent (project _) => eapply (@morph_consistent _ _ _ _ project _); eauto
        | |- consistent (inject _ _) => eapply (@bimorph_consistent _ _ _ _ inject _); eauto
      end;
  came.


Ltac destruct_maps :=
  match goal with
    | h: maps _ _ _ |- _ => dependent destruction h
  end.
(* This tactics correctly matches the goal, and then fails... *)

(* Access. *)

(* A consistent resource maps every address to a consistent value. *)

Lemma SCAM_maps_consistent:
  forall R r a,
  SCAM_maps R r a ->
  consistent R ->
  consistent a.
Proof.
  scame.
  inversion H1; dam. eapply is_nth_Forall; eauto.
Qed.

Lemma SCAM_maps_star_construction:
  forall R1 R2 l a1 a2,
  SCAM_maps R1 l a1 ->
  SCAM_maps R2 l a2 ->
  consistent (star R1 R2) ->
  SCAM_maps (star R1 R2) l (star a1 a2).
Proof.
  scame.
  dependent destruction H1; dependent destruction H2; dam.
  econstructor. eauto using is_nth_zip.
Qed.

Lemma SCAM_maps_star:
  forall { R1 R2 l a1 },
  SCAM_maps R1 l a1 ->
  consistent (star R1 R2) ->
  exists a2,
  SCAM_maps R2 l a2 /\ consistent (star a1 a2).
Proof.
  scame.
  dependent destruction H1.
  dam.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eexists. split.
    econstructor; eauto.
    eapply is_nth_Forall; eauto using is_nth_zip.
Qed.

Lemma SCAM_maps_star_inversion:
  forall { R1 R2 l a },
  SCAM_maps (star R1 R2) l a ->
  exists a1 a2,
  SCAM_maps R1 l a1 /\ SCAM_maps R2 l a2 /\ star a1 a2 = a.
Proof.
  scame.
  dependent destruction H1. dam.
  forwards: length_is_nth.
    erewrite <- zip_length_right by eauto. eauto using is_nth_length.
  forwards: length_is_nth.
    erewrite <- zip_length_left by eauto. eauto using is_nth_length.
  unpack.
  repeat econstructor; eauto using is_nth_injective, is_nth_zip. (* cool *)
Qed.

Lemma SCAM_maps_maximal_left:
  forall { R1 R2 l a },
  SCAM_maps R1 l a ->
  consistent (star R1 R2) ->
  maximal a ->
  SCAM_maps (star R1 R2) l a.
Proof.
  scame.
  dependent destruction H1; dam. econstructor.
  forwards: is_nth_common_length; [ eauto | eauto | unpack ].
  eapply is_nth_zip_eq; eauto.
  match goal with
    | h: maximal _ |- _ =>
        unfold maximal in h;
        eapply h;
        [ eauto | ]
  end.
  eapply is_nth_Forall; eauto using is_nth_zip.
Qed.

Lemma SCAM_maps_maximal_right:
  forall { R1 R2 l a },
  SCAM_maps R2 l a ->
  consistent (star R1 R2) ->
  maximal a ->
  SCAM_maps (star R1 R2) l a.
Proof.
  intros.
  rewrite star_commutative.
  eapply SCAM_maps_maximal_left; eauto.
  rewrite star_commutative.
  eauto.
Qed.

Lemma SCAM_maps_stable:
  forall R1 R2 l a,
  SCAM_maps R1 l a ->
  rio R1 R2 ->
  stable a ->
  SCAM_maps R2 l a.
Proof.
  scame.
  dependent destruction H1; dam. econstructor.
  forwards: growth_is_nth; [ eauto | eauto | unpack ].
  match goal with
    | h: stable _ |- _ =>
        unfold stable in h; erewrite h; eauto
  end.
Qed.

Lemma SCAM_maps_core:
  forall R r a b,
  SCAM_maps R r a ->
  core a = b ->
  SCAM_maps (core R) r b.
Proof.
  scame.
  inversion H1; dam.
  econstructor; eauto using is_nth_map.
Qed.

Lemma SCAM_maps_injective:
  forall { R r a1 a2 },
  SCAM_maps R r a1 ->
  SCAM_maps R r a2 ->
  a1 = a2.
Proof.
  scame.
  dependent destruction H1; inversion H2.
  eauto using is_nth_injective.
Qed.

Lemma SCAM_maps_rio:
  forall { R1 R2 r a1 },
  SCAM_maps R1 r a1 ->
  rio R1 R2 ->
  exists a2,
  SCAM_maps R2 r a2 /\ rio a1 a2.
Proof.
  scame.
  dependent destruction H1; dam.
  forwards: growth_is_nth. eauto. eauto. unpack.
  eauto using Maps.
Qed.

Lemma SCAM_maps_rio_reverse:
  forall { R1 R2 r a2 },
  SCAM_maps R2 r a2 ->
  rio R1 R2 ->
  (exists a1, SCAM_maps R1 r a1 /\ rio a1 a2) \/
  (consistent a2 /\ duplicable a2).
Proof.
  scame.
  dependent destruction H1; dam.
  forwards [ ? | ? ]: growth_is_nth_reverse; [ eauto | eauto | unpack | unpack ];
  eauto using Maps.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Allocation. *)

(* For allocation to preserve closedness, we must require the newly allocated
   address to carry a permission [x] that does not have an outgoing edge. *)

Lemma SCAM_consistent_allocate:
  forall R a req,
  consistent R ->
  consistent a ->
  consistent (SCAM_allocate R a req).
Proof.
  scame.
  dam. eauto using Forall_app.
Qed.

Lemma SCAM_star_allocate:
  forall R1 R2 a1 a2 req1 req2 req,
  star (SCAM_allocate R1 a1 req1) (SCAM_allocate R2 a2 req2) =
  SCAM_allocate (star R1 R2) (star a1 a2) req.
Proof.
  scame.
  rewrite <- (@bimorph_star _ _ _ _ inject _).
  eapply injectA_equality.
  came.
  dam;  try solve [ do 2 rewrite app_length in *; simpl in *; false; omega ].
  eapply zip_app; eauto.
Qed.

(* A particular case, which shows that [allocate R a] can be decomposed as a
   [star]-combination of a suitably-extended [R] and a suitably-extended [a]. *)

(* Using the proof term [not_pointsto_core a], instead of an arbitrary proof
   term, is less general, but this is actually a good thing in some situations
   where we have a meta-variable and wish to instantiate it. *)

Lemma SCAM_star_allocate_core:
  forall R a req,
  star
    (SCAM_allocate R (core a) (not_pointsto_core _ a))
    (SCAM_allocate (core R) a req) =
  SCAM_allocate R a req.
Proof.
  scame.
  rewrite <- (core_right_unit R) at 3.
  rewrite <- (@bimorph_star _ _ _ _ inject _).
  eapply injectA_equality.
  came.
  rewrite <- (core_right_unit x) at 3.
  rewrite <- (core_left_unit a) at 3.
  dam;  try solve [ do 2 rewrite app_length in *; simpl in *; false; omega ].
  eapply zip_app; eauto.
Qed.

(* A particular case, where the resource [a] is split into [a] and [core a].
   Roughly speaking, all of [a] is sent into one side. *)

Lemma SCAM_star_allocate_core_left_unit:
  forall R1 R2 a req,
  star
    (SCAM_allocate R1 (core a) (not_pointsto_core _ a))
    (SCAM_allocate R2 a req) =
  SCAM_allocate (star R1 R2) a req.
Proof.
  scame.
  rewrite <- (@bimorph_star _ _ _ _ inject _).
  eapply injectA_equality.
  came.
  rewrite <- (core_left_unit a) at 3.
  dam;  try solve [ do 2 rewrite app_length in *; simpl in *; false; omega ].
  eapply zip_app; eauto.
Qed.

Lemma SCAM_star_allocate_core_right_unit:
  forall R1 R2 a req,
  star
    (SCAM_allocate R1 a req)
    (SCAM_allocate R2 (core a) (not_pointsto_core _ a)) =
  SCAM_allocate (star R1 R2) a req.
Proof.
  scame.
  rewrite <- (@bimorph_star _ _ _ _ inject _).
  eapply injectA_equality.
  came.
  rewrite <- (core_right_unit a) at 3.
  dam;  try solve [ do 2 rewrite app_length in *; simpl in *; false; omega ].
  eapply zip_app; eauto.
Qed.

Lemma SCAM_rio_allocate:
  forall R a req,
  consistent a ->
  duplicable a ->
  rio R (SCAM_allocate R a req).
Proof.
  introv hc hd.
  unfold SCAM_allocate.
  rewrite <- (ip_inverse R) at 1.
  eapply (@bimorph_rio _ _ _ _ inject _); [ | eauto using rio_reflexive ].
  destruct (project R).
  came. 
  dam.
  eauto using growth_app, @rio_reflexive.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Limit. *)

Lemma SCAM_limit_core:
  forall R,
  SCAM_limit (core R) = SCAM_limit R.
Proof.
  scame. dam. eapply map_length.
Qed.

Lemma SCAM_maps_allocate_there:
  forall R r a req,
  consistent R ->
  r = SCAM_limit R ->
  SCAM_maps (SCAM_allocate R a req) r a.
Proof.
  scame.
  dam.
  eauto using is_nth_app_far, Maps with is_nth.
Qed.

Lemma SCAM_maps_allocate_elsewhere:
  forall R l b,
  SCAM_maps R l b ->
  forall a req,
  SCAM_maps (SCAM_allocate R a req) l b.
Proof.
  scame.
  dam.
  dependent destruction H1.
  eauto using is_nth_app_near, Maps.
Qed.

Lemma SCAM_maps_allocate_inversion:
  forall { R r a b req },
  SCAM_maps (SCAM_allocate R a req) r b ->
  SCAM_maps R r b \/ a = b.
Proof.
  scame. dam.
  dependent destruction H1.
  forwards [ | ]: is_nth_inversion; eauto using Maps.
Qed.

Lemma SCAM_limit_split_left:
  forall R1 R2,
  consistent (star R1 R2) ->
  SCAM_limit (star R1 R2) = SCAM_limit R1.
Proof.
  scame.
  dam.
  rewrite zip_length_left; eauto.
Qed.

Lemma SCAM_limit_split_right:
  forall R1 R2,
  consistent (star R1 R2) ->
  SCAM_limit (star R1 R2) = SCAM_limit R2.
Proof.
  scame.
  dam.
  rewrite zip_length_right; eauto.
Qed.

(* An ad hoc lemma. *)

Lemma SCAM_limit_limit_ad_hoc:
  forall R1 R2 R3,
  consistent (star (star R1 R2) R3) ->
  SCAM_limit (star (star R1 R2) R3) = SCAM_limit (core R2).
Proof.
  scame.
  dam.
  rewrite zip_length_left; eauto.
  rewrite zip_length_right; eauto.
  rewrite map_length.
  reflexivity.
Qed.

Lemma SCAM_limit_rio:
  forall R1 R2,
  rio R1 R2 ->
  SCAM_limit R1 <= SCAM_limit R2.
Proof.
  scame.
  dam; eauto using growth_length.
Qed.

Lemma SCAM_limit_maps:
  forall R l,
  l < SCAM_limit R ->
  consistent R ->
  exists a,
  SCAM_maps R l a.
Proof.
  scame.
  dam.
  forwards: length_is_nth. eauto. unpack. eauto using Maps.
Qed.

Lemma SCAM_maps_limit:
  forall R l a,
  SCAM_maps R l a ->
  l < SCAM_limit R.
Proof.
  scame.
  dependent destruction H1.
  dam.
  eauto using is_nth_length.
Qed.

(* ---------------------------------------------------------------------------- *)

(* Update. *)

Lemma SCAM_consistent_update:
  forall R l a dreq oreq,
  consistent R ->
  consistent a ->
  consistent (SCAM_update R l a dreq oreq).
Proof.
  scame.
  dam.
  eauto using Forall_set_nth.
Qed.

Lemma SCAM_maps_update_there:
  forall { R r s a1 a2 dreq oreq },
  SCAM_maps R r a1 ->
  r = s ->
  SCAM_maps (SCAM_update R s a2 dreq oreq) r a2.
Proof.
  scame.
  dependent destruction H1.
  econstructor; subst.
  eauto using is_nth_set_nth_there.
Qed.

Lemma SCAM_maps_update_elsewhere:
  forall R r s a1 a2 dreq oreq,
  SCAM_maps R r a1 ->
  r <> s ->
  SCAM_maps (SCAM_update R s a2 dreq oreq) r a1.
Proof.
  scame.
  dependent destruction H1.
  econstructor.
  eauto using is_nth_set_nth_elsewhere.
Qed.

Lemma SCAM_maps_update_inversion:
  forall { R r s a b dreq oreq },
  SCAM_maps (SCAM_update R r a dreq oreq) s b ->
  (r <> s /\ SCAM_maps R s b) \/ (r = s /\ a = b).
Proof.
  scame.
  dependent destruction H1.
  dam.
  forwards [ | ]: is_nth_set_nth_inversion; [ eauto | left | right ]; unpack;
  eauto using Maps.
Qed.

Lemma SCAM_rio_update:
  forall R s a1 a2 dreq oreq,
  SCAM_maps R s a1 ->
  rio a1 a2 ->
  rio R (SCAM_update R s a2 dreq oreq).
Proof.
  unfold SCAM_maps, SCAM_update.
  intros.
  rewrite <- (ip_inverse R) at 1.
  eapply (@bimorph_rio _ _ _ _ inject _); [ | eauto using rio_reflexive ].
  scame.
  dependent destruction H1.
  dam.
  eapply growth_set_nth; eauto.
    eapply rio_reflexive.
Qed.

Lemma SCAM_rio_update_update:
  forall R1 R2 s a a1 a2 dreq1 oreq1 dreq2 oreq2,
  SCAM_maps R1 s a ->
  rio R1 R2 ->
  rio a1 a2 ->
  rio (SCAM_update R1 s a1 dreq1 oreq1) (SCAM_update R2 s a2 dreq2 oreq2).
Proof.
  scame.
  eapply (@bimorph_rio _ _ _ _ inject _); [ | eauto using rio_reflexive ].
  dependent destruction H1.
  dam.
  eauto using growth_set_nth_set_nth.
Qed.

Lemma SCAM_star_update:
  forall R1 R2 s a1 a2 a dreq1 oreq1 dreq2 oreq2 dreq oreq,
  star a1 a2 = a ->
  star (SCAM_update R1 s a1 dreq1 oreq1) (SCAM_update R2 s a2 dreq2 oreq2) =
  SCAM_update (star R1 R2) s a dreq oreq.
Proof.
  unfold SCAM_update; intros.
  rewrite <- (@bimorph_star _ _ _ _ inject _).
  eapply injectA_equality.
  came.
  scame.
  dam; try solve [ do 2 rewrite length_set_nth in *; simpl in *; false; omega ].
  eapply zip_set_nth.
Qed.

Lemma SCAM_update_identity:
  forall R s a dreq oreq,
  SCAM_maps R s a ->
  SCAM_update R s a dreq oreq = R.
Proof.
  unfold SCAM_maps, SCAM_update.
  intros.
  rewrite <- (ip_inverse R) at 3.
  eapply injectA_equality.
  destruct (project R).
  came.
  dependent destruction H1.
  dam.
  eauto using set_nth_identity.
Qed.

(*
Lemma SCAM_update_overwrite:
  forall R r b1 b2 dreq oreq dreq1 oreq1 dreq2 oreq2,
  SCAM_update (SCAM_update R r b1 dreq1 oreq1) r b2 dreq2 oreq2
  = SCAM_update R r b2 dreq oreq.
Proof.
  unfold SCAM_update.
  intros.

  rewrite inject_erase.
  eapply injectA_equality.

  erewrite (CAM_update_equality
             (project (inject (CAM_update (project R) r b1 dreq1 oreq1) R))
             (CAM_update (project R) r b1 dreq1 oreq1)
             _ _ _ _ 
             (@pi_inverse _ _ _ _ _ _ _)).

  unfold CAM_update.
  came.

  destruct (project R).
  came.
  dam.

  eauto using set_nth_overwrite.

Qed.

Lemma SCAM_update_overwrite_ex:
  forall R r b1 b2 dreq1 oreq1 dreq2 oreq2,
  exists dreq oreq,
  SCAM_update (SCAM_update R r b1 dreq1 oreq1) r b2 dreq2 oreq2
  = SCAM_update R r b2 dreq oreq.
Proof.
  intros. xplit.
  erewrite SCAM_update_overwrite.
  eauto.

Grab Existential Variables.

intros.

skip.
skip.

Qed.
*)
(* Lemma SCAM_update_overwrite_ex: *)
(*   forall R r b1 b2 dreq1 oreq1 dreq2 oreq2, *)
(*   exists dreq oreq, *)
(*   SCAM_update (SCAM_update R r b1 dreq1 oreq1) r b2 dreq2 oreq2 *)
(*   = SCAM_update R r b2 dreq oreq. *)
(* Proof. *)
(*   unfold SCAM_update. *)
(*   intros. *)

(*   rewrite inject_erase. *)
(*   eexists. eexists. *)
(*   eapply injectA_equality. *)

(*   Check CAM_update_equality. *)
(*   Check @pi_inverse. *)

(*   erewrite (CAM_update_equality *)
(*              (project (inject (CAM_update (project R) r b1 dreq1 oreq1) R)) *)
(*              (CAM_update (project R) r b1 dreq1 oreq1) *)
(*              _ _ _ _  *)
(*              (@pi_inverse _ _ _ _ _ _ _ _ _ _ _)). *)

(*   unfold CAM_update. *)
(*   came. *)

(*   destruct (project R). *)
(*   came. *)
(*   dam. *)

(*   eauto using set_nth_overwrite. *)

(* Grab Existential Variables. *)

(* rewrite (@pi_inverse _ _ _ _ project inject _ _ _) in oreq2. *)
(* came. *)
(* rewrite update_overwrite in oreq2. eauto. *)

(* rewrite (@pi_inverse _ _ _ _ project inject _ _ _) in oreq2. *)
(* rewrite (@pi_inverse _ _ _ _ project inject _ _ _) in dreq2. *)
(* came. *)
(* destruct (project R). *)
(* came. *)

(* clear oreq1 oreq2. *)
(* destruct dreq1; destruct dreq2; eauto. *)
(* right. *)
(* introv ? ?. unfold has_no_antecedents in H4. *)
(* eapply H4; eauto. *)
(* eapply maps_update_elsewhere; eauto. *)
(* intro; subst. *)



(* Qed. *)


(* Lemma SCAM_star_update_core_left: *)
(*   forall R1 R2 s a dreq1 oreq1 dreq oreq, *)
(*   SCAM_maps R2 s (core a) -> *)
(*   star (SCAM_update R1 s a dreq1 oreq1) R2 = SCAM_update (star R1 R2) s a dreq oreq. *)
(* Proof. *)
(*   unfold SCAM_maps, SCAM_update; intros. *)
(*   rewrite <- (ip_inverse R2) at 1. *)
(*   rewrite <- (@bimorph_star _ _ _ _ inject _). *)
(*   eapply injectA_equality. *)

Require Import Classical.

Lemma SCAM_star_update_core_left:
  forall R1 R2 s a dreq1 oreq1 dreq oreq,
  SCAM_maps R2 s (core a) ->
  star (SCAM_update R1 s a dreq1 oreq1) R2 = SCAM_update (star R1 R2) s a dreq oreq.
Proof.
  introv hmaps.
  rewrite <- (SCAM_update_identity R2) at 1.
    eapply SCAM_star_update; eapply core_right_unit.
    eauto.

Grab Existential Variables.

intros. false. eauto using not_pointsto_core.

destruct (classic (present (core a))) as [ hpr | hnpr ].
  tauto.
  right.
    introv hm hpts.
    eapply hnpr.
    unfold SCAM_maps in hmaps.
    destruct (project R2).
    came.
    dam.
    inversion hmaps.
    forwards hdm: (c x a0) hm hpts.
    destruct hdm as (b, hdm).
    destruct hdm as (hdmm, hdmp).
    forwards: maps_injective hmaps hdmm.
    subst.
    auto.
Qed.

Lemma SCAM_star_update_core_right:
  forall R1 R2 s a dreq2 oreq2 dreq oreq,
  SCAM_maps R1 s (core a) ->
  star R1 (SCAM_update R2 s a dreq2 oreq2) = SCAM_update (star R1 R2) s a dreq oreq.
Proof.
  introv hmaps.
  rewrite <- (SCAM_update_identity R1) at 1.
    eapply SCAM_star_update; eapply core_left_unit.
    eauto.

Grab Existential Variables.

intros. false. eauto using not_pointsto_core.

destruct (classic (present (core a))) as [ hpr | hnpr ].
  tauto.
  right.
    introv hm hpts.
    eapply hnpr.
    unfold SCAM_maps in hmaps.
    destruct (project R1).
    came.
    dam.
    inversion hmaps.
    forwards hdm: (c x a0) hm hpts.
    destruct hdm as (b, hdm).
    destruct hdm as (hdmm, hdmp).
    forwards: maps_injective hmaps hdmm.
    subst.
    auto.
Qed.

Lemma SCAM_update_reverse:
  forall R R' s a b dreq oreq dreq' oreq',
  SCAM_maps R s a ->
  SCAM_update R s b dreq oreq = R' ->
  R = SCAM_update R' s a dreq' oreq'.
Proof.

intros; subst.
symmetry.
unfold SCAM_maps, SCAM_update in *.
rewrite inject_erase.
rewrite <- (ip_inverse R) at 4.
eapply injectA_equality.

destruct (project R).
came.
erewrite (@pi_inverse _ _ _ _ sub_struct).
erewrite (@pi_inverse _ _ _ _ sub_struct) in dreq'.
erewrite (@pi_inverse _ _ _ _ sub_struct) in oreq'.
came.

erewrite update_overwrite.
eauto using update_identity.

Qed.

(*   intros. *)
(*   subst. *)
(*   symmetry. *)
(*   erewrite SCAM_update_overwrite. *)
(*   eauto using SCAM_update_identity. *)

(* Grab Existential Variables. *)

(* introv hpts. *)
(* forwards: oreq' hpts. *)
(* unfold SCAM_update in H4. *)
(* rewrite pi_inverse in H4. *)
(* came. *)
(* rewrite update_overwrite in H4. *)
(* eauto. *)

(* unfold SCAM_update in *. *)
(* erewrite (@pi_inverse _ _ _ _ project inject _ _ _) in dreq'. *)
(* erewrite (@pi_inverse _ _ _ _ project inject _ _ _) in oreq'. *)
(* came. *)
(* destruct (project R). *)
(* came. *)

(* destruct dreq'; [ tauto | ]. *)
(* destruct dreq; [ | tauto ]. *)


(* ---------------------------------------------------------------------------- *)

End Operations.

Implicit Arguments CAM_empty [ MA H H0 ].

(* ---------------------------------------------------------------------------- *)

(* Agreement. *)

Section Agreement.

Context `{A: Type}.
Context `{AS: Type}.
Context `{projectA: A -> aMap AS}.
Context `{injectA: aMap AS -> A -> A}.
Context `{subA: @SubStructure (aMap AS) A projectA injectA}.

Context `{B: Type, MB: MSA B}.
Context `{BS: Type, MBS: MSA BS, PRB: @Present BS MBS, PTB: @Pointsto BS MBS}.
Context `{projectB: B -> ClosedAMap BS}.
Context `{MpB: @MSAMorphism B MB (ClosedAMap BS) (msa_ClosedAMap BS) projectB}.
Context `{injectB: ClosedAMap BS -> B -> B}.
Context `{MiB: @MSABiMorphism (ClosedAMap BS) (msa_ClosedAMap BS) B MB injectB}.
Context `{subB: @SubStructure (ClosedAMap BS) B projectB injectB}.

(* Context `{A: Type, MA: MSA A}. *)
(* Context `{AS: Type, MAS: MSA AS, PRA: @Present AS MAS, PTA: @Pointsto AS MAS}. *)
(* Context `{projectA: A -> ClosedAMap AS}. *)
(* Context `{MpA: @MSAMorphism A MA (ClosedAMap AS) (msa_ClosedAMap AS) projectA}.  *)
(* Context `{injectA: ClosedAMap AS -> A -> A}. *)
(* Context `{MiA: @MSABiMorphism (ClosedAMap AS) (msa_ClosedAMap AS) A MA injectA}. *)
(* (* Context `{subA: @MSASub (ClosedAMap AS) (msa_ClosedAMap AS) A MA projectA injectA MpA MiA}. *) *)

(* Context `{B: Type, MB: MSA B}. *)
(* Context `{BS: Type, MBS: MSA BS, PRB: @Present BS MBS, PTB: @Pointsto BS MBS}. *)
(* Context `{projectB: B -> ClosedAMap BS}.  *)
(* Context `{MpB: @MSAMorphism B MB (ClosedAMap BS) (msa_ClosedAMap BS) projectB}.  *)
(* Context `{injectB: ClosedAMap BS -> B -> B}. *)
(* Context `{MiB: @MSABiMorphism (ClosedAMap BS) (msa_ClosedAMap BS) B MB injectB}. *)
(* (* Context `{subB: @MSASub (ClosedAMap BS) (msa_ClosedAMap BS) B MB projectB injectB MpB MiB}. *) *)

Context `{agrees: AS -> BS -> Prop}.

Definition SCAM_agrees (RA: A) (RB: B) :=
  let RA := projectA RA in
  let (RB, _) := projectB RB in
  aMap_agrees agrees RA RB.

(* Definition CAM_agrees (R1 : aMap A) (R2 : ClosedAMap B) := *)
(*   let (R2, _) := R2 in *)
(*   aMap_agrees agrees R1 R2. *)

(* Ltac scame := *)
(*   unfold SCAM_empty, SCAM_maps, SCAM_allocate, SCAM_limit, SCAM_update; *)
(*   intros; *)
(*   repeat match goal with *)
(*         | hpi: context[project (inject _ _)] |- _ => *)
(*             rewrite pi_inverse in hpi *)
(*         | |- context[project (inject _ _)] => *)
(*             rewrite pi_inverse *)
(*       end; *)
(*   try match goal with *)
(*            | hc: consistent _ |- _ => *)
(*                forwards: (@morph_consistent B _ _ _ project _) hc *)
(*          end; *)
(*   try match goal with *)
(*            | hr: rio _ _ |- _ => *)
(*                forwards: morph_rio hr *)
(*          end; *)
(*   repeat match goal with *)
(*         | hc: context[project (core _)] |- _ => *)
(*             rewrite (@morph_core _ _ _ _ project _) in hc *)
(*         | |- context[project (core _)] => *)
(*             rewrite (@morph_core _ _ _ _ project _) *)
(*       end; *)
(*   repeat match goal with *)
(*         | hs: context[project (star _ _)] |- _ => *)
(*             rewrite (@morph_star _ _ _ _ project _) in hs *)
(*         | |- context[project (star _ _)] => *)
(*             rewrite (@morph_star _ _ _ _ project _) *)
(*       end; *)
(*   repeat match goal with *)
(*            | h: context[project ?r] |- _ => *)
(*                destruct (project r) *)
(*            | |- context[project ?r] => *)
(*                destruct (project r) *)
(*          end; *)
(*   try match goal with *)
(*         | |- consistent (project _) => eapply (@morph_consistent _ _ _ _ project _); eauto *)
(*         | |- consistent (inject _ _) => eapply (@bimorph_consistent _ _ _ _ inject _); eauto *)
(*       end; *)
(*   came. *)

Lemma SCAM_agrees_empty:
  forall (RA: A) (RB: B),
  SCAM_agrees (SAM_empty injectA RA) (SCAM_empty injectB RB).
Proof.
  unfold SCAM_empty, SAM_empty.
  unfold SCAM_agrees.
  intros.
  do 2 rewrite pi_inverse.
  came.
  repeat econstructor.
Qed.

Lemma SCAM_agrees_consistent:
  (forall a b, agrees a b -> consistent b) ->
  forall R1 R2,
  SCAM_agrees R1 R2 ->
  consistent (projectB R2).
Proof.
  unfold SCAM_agrees.
  intros.
  destruct (projectA R1); destruct (projectB R2).
  inversion H0; subst.
  came.
  inversion H0; subst.
  eapply Forall2_Forall_modus_ponens; eauto.
Qed.

Lemma SCAM_agrees_maps_left_to_right:
  forall { R1 R2 l a },
  SCAM_agrees R1 R2 ->
  SAM_maps projectA R1 l a ->
  exists b,
  SCAM_maps projectB R2 l b /\
  agrees a b.
Proof.
  unfold SCAM_agrees, SCAM_maps, SAM_maps.
  intros.
  destruct (projectA R1); destruct (projectB R2); came.
  dependent destruction H.
  dependent destruction H0.
  inversion H; subst.
  forwards: Forall2_is_nth_left_to_right; [ eauto | eauto | unpack ].
  xplit. econstructor; eauto. eauto.
Qed.

Lemma SCAM_agrees_maps_right_to_left:
  forall { R1 R2 l b },
  SCAM_agrees R1 R2 ->
  SCAM_maps projectB R2 l b ->
  exists a,
  SAM_maps projectA R1 l a /\ agrees a b.
Proof.
  unfold SCAM_agrees, SCAM_maps, SAM_maps.
  intros.
  destruct (projectA R1); destruct (projectB R2); came.
  dependent destruction H.
  dependent destruction H0.
  inversion H; subst.
  forwards: Forall2_is_nth_right_to_left; [ eauto | eauto | unpack ].
  xplit. econstructor; eauto. eauto.
Qed.

Lemma SCAM_agrees_maps_maps:
  forall { R1 R2 l a b },
  SCAM_agrees R1 R2 ->
  SAM_maps projectA R1 l a ->
  SCAM_maps projectB R2 l b ->
  agrees a b.
Proof.
  unfold SCAM_agrees, SCAM_maps, SAM_maps.
  intros.
  destruct (projectA R1); destruct (projectB R2); came.
  dependent destruction H.
  dependent destruction H.
  dependent destruction H0.
  dependent destruction H1.
  eapply Forall2_is_nth; eauto.
Qed.

Lemma SCAM_agrees_allocate:
  forall R1 R2 a b reqB,
  SCAM_agrees R1 R2 ->
  agrees a b ->
  SCAM_agrees (SAM_allocate projectA injectA R1 a)
              (SCAM_allocate projectB injectB R2 b reqB).
Proof.
  unfold SCAM_agrees, SCAM_allocate, SAM_allocate.
  intros.
  do 2 rewrite pi_inverse.
  destruct (projectA R1); destruct (projectB R2); came; dependent destruction H.
  econstructor.
  eapply Forall2_app; eauto.
Qed.

Lemma SCAM_agrees_update:
  forall R1 R2 l a b dreqB oreqB,
  SCAM_agrees R1 R2 ->
  agrees a b ->
  SCAM_agrees (SAM_update projectA injectA R1 l a)
              (SCAM_update BS B projectB injectB R2 l b dreqB oreqB).
Proof.
  unfold SCAM_agrees, SCAM_update, SAM_update.
  intros.
  do 2 rewrite pi_inverse.
  destruct (projectA R1); destruct (projectB R2); came; dependent destruction H.
  econstructor.
  eauto using Forall2_set_nth.
Qed.

Lemma SCAM_agrees_limit:
  forall R1 R2,
  SCAM_agrees R1 R2 ->
  SAM_limit projectA R1 = SCAM_limit projectB R2.
Proof.
  unfold SCAM_agrees, SCAM_limit, SAM_limit.
  intros.
  destruct (projectA R1); destruct (projectB R2); came; dependent destruction H.
  came.
  eauto using Forall2_length.
Qed.

End Agreement.

(* Implicit Arguments CAM_agrees [ A B M H H0 ]. *)

(* ---------------------------------------------------------------------------- *)

(* This tactic is copied and adapted from [MSAAllocationMapOperations]. *)

Ltac SCAM_maps_maximal msa_morph_proj :=
  instantiate;
  match goal with
  | h: SCAM_maps ?proj ?R ?l ?a |- SCAM_maps ?proj (star ?R1 ?R2) ?l ?b =>
    (* Force [a] and [b] to be the same element. If this fails, perhaps
       another hypothesis [h] will be tried; there might be several [CAM_maps]
       facts at hand. *)
    assert (a = b); [ solve [ eauto with maps_maximal_reflexivity_hints ] |
      (* Accumulate consistency hypotheses as we go down, in order to avoid
         a quadratic overhead. *)
      assert (consistent (star R1 R2)); [ solve [ eauto with consistent_split ] | ];
      (* Determine whether we should go left or right. *)
      match R1 with
      | context[R] =>
          (* [R] occurs inside [R1]: go left. *)
          eapply (SCAM_maps_maximal_left msa_morph_proj); [ | eassumption | eauto with maps_maximal_hints ];
          SCAM_maps_maximal msa_morph_proj
      | _ =>
          match R2 with
          | context[R] =>
              (* [R] occurs inside [R2]: go right. *)
              eapply (SCAM_maps_maximal_right msa_morph_proj); [ | eassumption | eauto with maps_maximal_hints ];
              SCAM_maps_maximal msa_morph_proj
          | _ =>
              idtac "SCAM_maps_maximal: I don't know whether to go left or right."
          end
      end
    ]
  | _ =>
    eassumption (* we are done going down and can hopefully solve the goal *)
  end.

(* Hint Extern 1 (SCAM_maps ?proj _ _ _) => (SCAM_maps_maximal proj) : SCAM_maps_maximal. *)
 
(* ---------------------------------------------------------------------------- *)

(* Tactics to help apply certain lemmas in forward mode. *)

Ltac SCAM_maps_injective :=
  repeat match goal with h1: SCAM_maps ?proj ?R ?x _, h2: SCAM_maps ?proj ?R ?x _ |- _ =>
    unpack (SCAM_maps_injective proj h1 h2); clear h1
  end;
  injections.

(* ---------------------------------------------------------------------------- *)

(* This tactic helps solve goals of the form [CAM_limit R1 = CAM_limit R2], where
   [R2] occurs within [R1]. *)

Ltac SCAM_limit_split msa_morph_proj :=
  match goal with
  | |- SCAM_limit ?proj (star ?R1 ?R2) = SCAM_limit ?proj ?R =>
    match R1 with
    | context[R] =>
        rewrite (SCAM_limit_split_left msa_morph_proj) by eauto with consistent_split;
        SCAM_limit_split msa_morph_proj
    | _ =>
        match R2 with
        | context[R] =>
            rewrite (SCAM_limit_split_right msa_morph_proj) by eauto with consistent_split;
            SCAM_limit_split msa_morph_proj
        | _ =>
            idtac "SCAM_limit_split: I don't know whether to go left or right"
        end
    end
  | _ =>
      reflexivity
  end.

(* Hint Extern 1 (_ = _) => *)
(*   (instantiate; SCAM_limit_split) : SCAM_limit_split. *)

