Set Implicit Arguments.
Require Import MyTactics.
Require Import MyLogic.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Signature.
Require Import Dup.
Require Import Exclusive.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JDrop.
Require Import JDuplication.
Require Import JClassification.
Require Import JDecomposition.
Require Import JInstantiation.
Require Import DecomposeRecompose.

(* ------------------------------------------------------------------------- *)

(* The following two lemmas help prove [JpsDynamicAppears] correct. *)

(* Only a valid memory location can receive an exclusive type. *)

Lemma exclusive_implies_location:
  forall R K v T,
  exclusive T ->
  jctop R K v T ->
  consistent R ->
  exists l, v = TLoc l /\ l < heap_limit R.
Proof.
  inversion 1; subst; intros.
  (* Case: TyCon. *)
  classification. decomposition_TyCon. dca.
  eexists. split. eauto. eapply (SCAM_maps_limit heap_project).
    SCAM_maps_maximal heap_project_morphism.
  (* Case: TyBlock. *)
  classification. dca.
  eexists. split. eauto. eapply (SCAM_maps_limit heap_project).
    SCAM_maps_maximal heap_project_morphism.
Qed.

(* If [v] is a valid memory location, then it is safe to let the permission
   [v @ TyDynamic] appear. *)

Lemma dynamic_appears:
  forall R K v T,
  jctop R K v T ->
  consistent R ->
  (exists l, v = TLoc l /\ l < heap_limit R) ->
  jpi R K (TyStar (TyAnchored v T) (TyAnchored v TyDynamic)).
Proof.
  intros. unpack. subst.
  econstructor; [ eauto | | eapply core_right_unit ].
  eapply JcTopDynamic.
  rewrite (SCAM_limit_core heap_project_morphism).
  assumption.
Qed.

(* ------------------------------------------------------------------------- *)

(* This preliminary lemma states that if the subtyping fact [T <= U] is
   semantically sound, then the permission subsumption fact [ys @ T <= ys @ U]
   is semantically sound as well. This lemma is later used to apply the fact
   [T <= U] to a list of adoptees, so as to prove that it is sound for
   [adopts] clauses to be covariant. *)

Lemma sub_zap:
  forall K T U,
  (forall R v , jctop R K       v T  -> consistent R -> jctop R K       v U) ->
  (forall ys R, jpi    R K (zap ys T) -> consistent R -> jpi    R K (zap ys U)).
Proof.
  induction ys; simpl; intros; decomposition; econstructor; eauto with consistent_split.
Qed.

(* ------------------------------------------------------------------------- *)

(* Subsumption is sound with respect to the natural notion of implication that
   arises out of the interpretation of permissions. *)

(* This key lemma is the reason why we do not care how complicated the
   definition of [jps] might be. *)

(* The proof of this lemma is carried out by well-founded induction over the
   index [n], and is the reason why we introduce this index in the first
   place. Ordinary structural induction would not work here because there is
   at least one case where a substitution must be applied to a sub-term before
   exploiting the induction hypothesis that comes with it. *)

Lemma jps_sub_sound:
(
  forall n K P Q,
  jps n K P Q ->
  forall R,
  jpi R K P -> 
  consistent R ->
  jpi R K Q
) /\ (
  forall n K T U,
  sub n K T U ->
  forall R v,
  jctop R K v T ->
  consistent R ->
  jctop R K v U
) /\ (
  forall n K Ts Us,
  row_sub n K Ts Us ->
  forall R vs,
  jpi R K (zip vs Ts) ->
  consistent R ->
  jpi R K (zip vs Us)
).
Proof.
  eapply simultaneous_well_founded_induction_3; intros n ih_jps ih_sub ih_row_sub;
  introv h; dependent destruction h; intros; subst;
  decomposition;
  try solve [
    eauto using @star_commutative, @star_associative_symmetric,
                @core_left_unit, @core_right_unit,
                jpi_drop_right, j_duplicable,
                JcTopFrame, JcTopSingleton
           with jpi wk
  ].

  (* JpsHideDuplicablePrecondition *)
  (* Remember some well-kindedness facts, which are lost by [decomposition]. *)
  forwards: jctop_wk. eauto. wk_inversion.
  (* A value that admits a function type must be a function. *)
  classification.
  decomposition.
  (* Build a new proof. *)
  econstructor; eauto with wk.
  eapply JFunction; [ eauto | eauto | ].
  (* Upon entry to the function body, move the permission [P], which is
     ambient, to the argument, where the function expects it. *)
  eapply (@JSubLeft 1); [ eauto | ].
  jps_left JpsStarCommutative.
  eapply JpsMixStarIntro.

  (* JpsExistsHoistLeft *)
  econstructor; eauto using wk_weakening with wk.
  simpl_subst_goal. rewrite subst_lift.
  solve [ eauto with j ].

  (* JpsForallType2Perm *)
  (* The goal boils down to proving that [v] has type [T] in the presence
     of a fresh variable [x]. *)
  econstructor.
    eauto with substitutable.
  (* We know that [v] has type [forall x.T]. We use weakening to introduce
     a fresh unused [x], then instantiate the universally quantified [x]
     with the new [x] -- a pun. *)
  forwards h: (last6 j_weakening); eauto. simpl_lift_all. instantiate (1 := 0) in h. rewrite plus_0_r in h.
  forwards i: instantiation_TyForall_TYPE; eauto using WkTVar, lookup_insert_bingo with harmless. (* instantiate with TVar 0 *)
  match type of i with context[subst (TVar 0) 0 (shift 1 ?T)] => forwards eq: (pun_2 T 0); rewrite eq in i; clear eq end.
  assumption.

  (* JpsForallPerm2Type *)
  (* Derive some well-kindedness facts. *)
  forwards: (last6 j_wk). eauto. wk_inversion.
  (* Again, the goal boils down to proving that [v] has type [T] in the
     presence of a fresh [x]. We have a universal permission stating
     that for every [x], [v] has type [T], so it is again a pun. *)
  eapply JcTopForallIntro; [ assumption | | assumption ].
  forwards h: (last6 j_weakening); eauto.
    simpl_lift_all. instantiate (1 := 0) in h. rewrite plus_0_r in h.
  forwards i: instantiation_TyForall_PERM; eauto using WkTVar, lookup_insert_bingo. (* instantiate with TVar 0 *)
  simpl_subst_all.
  do 2
  match type of i with context[subst (TVar 0) 0 (shift 1 ?T)] => forwards eq: (pun_2 T 0); rewrite eq in i; clear eq end.
  assumption.

  (* JpsForallElimType *)
  solve [ eauto using instantiation_TyForall_TYPE ].

  (* JpsForallElimPerm *)
  solve [ eauto using instantiation_TyForall_PERM ].

  (* JpsVacuousForallIntro *)
  econstructor. eauto.
  solve [ eapply j_weakening; eauto ].

  (* JpsDecomposeBlock *)
  classification.
  decomposition.
  (* Apply the fundamental lemma that explains why this rule is sound. *)
  match goal with h3: jpi _ _ (zip _ _) |- _ =>
    forwards: (>> JpsDecomposeBlockSound h3); [ eauto | eauto | unpack; subst ]
  end.
  (* Build a proof. *)
  eapply JpiExists.
  simpl_subst_goal.
  rewrite subst_subst by omega.
  simpl_subst_goal.
  repeat rewrite subst_lift.
  (* The focused value. *)
  eapply JpiStar; [ | eauto | eapply reconfigure05 ].
  (* The block. *)
  solve [ eapply JcTopLoc; eauto ].
  (* Two administrative premises of [JpiExists]. *)
  solve [ eauto using wk_empty with wk ].
  solve [ do 3 (econstructor; eauto using wk_weakening, wk_substitution with wk insert_insert) ].

  (* JpsRecomposeBlock *)
  classification.
  decomposition.
  eapply JcTopLoc;
    eauto using reconfigure05,
          JpsRecomposeBlockSound, consistent_12p3_13.

  (* JpsUnifyBlock *)
  econstructor; eauto using @core_right_unit with j.
  eapply JcTopSingletonEq.
    solve [ eauto using JpsUnifyBlockSound ].
    solve [ eauto with wk wk_inversion ].

  (* JpsFold *)
  solve [ eauto using JcTopFold ].

  (* JpsUnfold *)
  classification.
  decomposition_TyCon.
  forwards: tc_is_record_and_tc_tag; [ eauto | eauto | unpack; subst ].
  assumption.

  (* JpsDirectedUnfold *)
  econstructor; eauto using JcTopLoc.
  classification.
  decomposition_TyCon.
  (* We have two [TyBlock] permissions for the same value. The tags
     must coincide. *)
  match goal with
  h1: jctop _ _ _ (TyBlock _ ?tag1 _ _),
  h2: jctop _ _ _ (TyBlock _ ?tag2 _ _) |- _ =>
    assert (tag1 = tag2); [ eapply JpsDirectedUnfoldSound; eauto | subst ]
  end.
  (* The tag determines the manner in which an algebraic data type
     is unfolded. *)
  match goal with
  h1: tc_tag ?tc ?tag ?Ts1,
  h2: tc_tag ?tc ?tag ?Ts2 |- _ =>
    assert (Ts1 = Ts2); [ eapply tc_tag_injective; eauto | subst ]
  end.
  assumption.

  (* JpsDynamicAppears *)
  (* The goal boils down to proving that [l] must be a valid memory location.
     This follows from the fact that [T] is exclusive. *)
  eapply dynamic_appears; eauto.
  eapply exclusive_implies_location; eauto.

  (* JpsBottom *)
  solve [ classification ].

  (* JpsCoArrow *)
  (* A value that admits a function type must be a lambda-abstraction. *)
  classification.
  (* The function body must be well-typed. *)
  decomposition.
  (* The subtyping judgement for the domain is expressed using [sub].
     Unfold the definition of [sub] so as to express it in terms of
     [jps]. *)
  sub_inversion.
  (* We now build a new type derivation by pushing the subtyping rule
     under the lambda-abstraction. We use [JSubLeft] for the domain 
     and [JSubRight] for the codomain. The resulting type derivation
     is canonical because the use of subtyping under a lambda-abstraction
     is allowed: it takes place one step ``later''. Thus, the induction
     hypotheses are not needed. *)
  eapply JpiAtomicEmpty; [ eauto with wk | ].
  eapply JFunction; eauto with wk dup.
  eapply JSubLeft; [ | jps_left JpsDisappearsLeft; eassumption ].
  eapply JSubRight; [ | eapply j_weakening; eauto ].
  eassumption.

  (* JpsCoBlock *)
  (* A value that admits a memory block type must be a memory location. *)
  classification.
  (* The values held at this location must have types [Ts], and every
     adoptee must have type [T]. *)
  decomposition.
  (* By the induction hypothesis, these values also have types [Us]. By
     the lemma [sub_zap] and by the induction hypothesis, every adoptee
     must have type [U]. There remains to build a new type derivation. *)
  solve [
    eapply JcTopLoc;
      eauto 7 using sub_zap, consistent_abandon_all with consistent_split wk
  ].

  (* JpsCoStar *)
  solve [ eauto 8 with j consistent_split ].

  (* JpsCoForall *)
  jpi_inversion. econstructor; solve [ eauto ].

  (* Sub *)
  (* Because [jctop] is defined in terms of [jpi], we are able to
     directly apply the induction hypothesis [ih_jps]. *)
  eapply ih_jps; eauto.
  (* The judgement that justifies this use of the induction hypothesis
     is obtained by substituting [v] for the name [x] that is
     universally quantified in the definition of [sub]. *)
  solve [
    eapply j_substitution; eauto using jctop_value with substitutable;
    simpl_subst_goal; f_equal; eapply subst_lift
  ].

  (* RowSubCons *)
  (* If the row of types has at least one component, then the row of values
     has at least one component. *)
  match goal with h: jpi _ _ (zip ?vs (RowCons _ _)) |- _ =>
    destruct vs; simpl in h; revert h
  end; intro hjpi; dependent destruction hjpi; subst; simpl.
  (* Build a new type derivation. *)
  econstructor; eauto with consistent_split.

Qed.

Definition jps_sound := get1 jps_sub_sound.
Definition sub_sound := get2 jps_sub_sound.

